using System;
using System.Xml;
using System.Collections;
using System.Collections.Specialized;

namespace Xtatic {

  using Xtatic.Value;

  namespace Value {
    public class Attributes : HybridDictionary {}

    public abstract class Seq {
      public enum Kind {Element, Object, Bool, Int, Char, String};
      public Kind kind;
      public Seq next;
    }

    public abstract class SeqTag : Seq {
      public Attributes attributes;
      public Seq contents;
    }

    public class SeqElement : SeqTag {
      public Sym.Tag tag;
      public SeqElement(Sym.Tag tag, Seq contents, Seq next) { 
        kind = Kind.Element; 
        this.tag = tag;
        this.contents = contents;
        this.next = next;
      }
      public SeqElement(string tag, Seq contents, Seq next) { 
        kind = Kind.Element; 
        this.tag = Sym.Intern(tag); 
        this.contents = contents; 
        this.next = next; 
      }
    }

    public class SeqObject : SeqTag {
      public object tag;
      public SeqObject(object tag, Seq contents, Seq next) { 
        kind = Kind.Object; 
        this.tag = tag; 
        this.contents = contents; 
        this.next = next; 
      }
    }

    public class SeqBool : Seq {
      public bool val;
      public SeqBool(bool val, Seq next) { 
        kind = Kind.Bool; 
        this.val = val;
        this.next = next;
      }
    }

    public class SeqInt : Seq {
      public int val;
      public SeqInt(int val, Seq next) { 
        kind = Kind.Int; 
        this.val = val;
        this.next = next;
      }
    }

    public class SeqChar : Seq {
      public char val;
      public SeqChar(char val, Seq next) { 
        kind = Kind.Char; 
        this.val = val;
        this.next = next;
      }
    }

    public class SeqString : Seq {
      public string val;
      public SeqString(string val, Seq next) { 
        kind = Kind.String; 
        this.val = val;
        this.next = next;
      }
    }
  }

  public class Util {

    private static Seq convText(string txt, int i, Seq next) {
      if (i >= txt.Length) return next;
      else return new SeqChar(txt[i],convText(txt,i+1,next));
    }

    private static Seq convSeq(XmlNodeList nodes, int i) {
      if (i >= nodes.Count) return null;
      else return convNode(nodes[i],convSeq(nodes,i+1));
    }

    private static Seq convNode(XmlNode n, Seq next) {
      switch (n.NodeType) {
        case XmlNodeType.Element:
          SeqElement el = new SeqElement(n.Name, convSeq(n.ChildNodes,0), next);
          XmlAttributeCollection atts = n.Attributes;
          if (atts.Count != 0) {
            Attributes d = new Attributes();
            foreach (XmlAttribute a in atts) d[a.Name] = a.Value;
            el.attributes = d;
          }
          return el;
        case XmlNodeType.Text:
          return convText(n.InnerText, 0, next);
        case XmlNodeType.Comment: return next;
        case XmlNodeType.EntityReference: 
          if (n.Name == "nbsp") return next;
          else throw new SystemException("convNode: unexpected entity reference " + n.Name);
        default:
          throw new SystemException("convNode: unexpected node type" + n);
      }
    }

    public static Seq LoadXml(string fname) {
      XmlDocument doc = new XmlDocument();
      doc.Load(fname);
      return convNode(doc.DocumentElement,null);
    }

    public static void PrintVal(Seq s) {
      if (s == null) return;
      switch (s.kind) {
        case Seq.Kind.Element: {
          Console.Write("<");
          SeqElement st = (SeqElement) s;
          Console.Write(Sym.Name(st.tag));
          if (st.attributes != null)
            foreach (DictionaryEntry de in st.attributes) Console.Write(" " + de.Key + "=\"" + de.Value + "\"");
          if (st.contents == null) {
            Console.Write("/>");
          } else {
            Console.Write(">");
            PrintVal(st.contents);
            Console.Write("</" + st.tag + ">");
          }
          break;
        } case Seq.Kind.Object: {
          Console.Write("<");
          SeqObject st = (SeqObject) s;
          Console.Write(st.tag);
          if (st.attributes != null)
            foreach (DictionaryEntry de in st.attributes) Console.Write(" " + de.Key + "=\"" + de.Value + "\"");
          Console.Write(">");
          PrintVal(st.contents);
          Console.Write("</" + st.tag + ">");
          break;
        } case Seq.Kind.Bool:
          SeqBool sb = (SeqBool) s;
          Console.Write(sb.val);
          if (s.next != null) Console.Write(" ");
          break;
        case Seq.Kind.Int:
          SeqInt si = (SeqInt) s;
          Console.Write(si.val);
          if (s.next != null) Console.Write(" ");
          break;
        case Seq.Kind.Char:
          SeqChar sc = (SeqChar) s;
          Console.Write(sc.val);
          if (s.next != null) Console.Write(" ");
          break;
        case Seq.Kind.String:
          SeqString ss = (SeqString) s;
          Console.Write("\"" + ss.val + "\"");
          if (s.next != null) Console.Write(" ");
          break;
      };
      PrintVal(s.next);
    }

    public static string ToString(Seq s) {
      if (s == null) return "";
      string res = "";
      switch (s.kind) {
        case Seq.Kind.Element: {
          SeqElement st = (SeqElement) s;
          res = "<" + Sym.Name(st.tag);
          if (st.attributes != null)
            foreach (DictionaryEntry de in st.attributes) res += (" " + de.Key + "=\"" + de.Value + "\"");
          if (st.contents == null) {
            res += "/>";
          } else {
            res += ">";
            res += ToString(st.contents);
            res += ("</" + st.tag + ">");
          }
          break;
        } case Seq.Kind.Object: {
          SeqObject st = (SeqObject) s;
          res = "<" + st.tag;
          if (st.attributes != null)
            foreach (DictionaryEntry de in st.attributes) res += (" " + de.Key + "=\"" + de.Value + "\"");
          res += ">";
          res += ToString(st.contents);
          res += ("</" + st.tag + ">");
          break;
        } case Seq.Kind.Bool:
          SeqBool sb = (SeqBool) s;
          res = "" + sb.val;
          if (s.next != null) res += " ";
          break;
        case Seq.Kind.Int:
          SeqInt si = (SeqInt) s;
          res = "" + si.val;
          if (s.next != null) res += " ";
          break;
        case Seq.Kind.Char:
          SeqChar sc = (SeqChar) s;
          res = "" + sc.val;
          if (s.next != null) res += " ";
          break;
        case Seq.Kind.String:
          SeqString ss = (SeqString) s;
          res = "\"" + ss.val + "\"";
          if (s.next != null) res += " ";
          break;
      };
      return res + ToString(s.next);
    }

    public static Seq Append(Seq v1, Seq v2) {
      if (v1 == null) return v2;
      else {
        Seq tmp = Append(v1.next, v2);
        switch (v1.kind) {
          case Seq.Kind.Element: {
            SeqElement x1 = (SeqElement) v1;
            SeqElement res = new SeqElement(x1.tag, x1.contents, tmp);
            res.attributes = x1.attributes;
            return res;
          } case Seq.Kind.Object: {
            SeqObject x1 = (SeqObject) v1;
            SeqObject res = new SeqObject(x1.tag, x1.contents, tmp);
            res.attributes = x1.attributes;
            return res;
          } case Seq.Kind.Bool:
            SeqBool x2 = (SeqBool) v1;
            return new SeqBool(x2.val, tmp);
          case Seq.Kind.Int:
            SeqInt x3 = (SeqInt) v1;
            return new SeqInt(x3.val, tmp);
          case Seq.Kind.Char:
            SeqChar x4 = (SeqChar) v1;
            return new SeqChar(x4.val, tmp);
          case Seq.Kind.String:
            SeqString x5 = (SeqString) v1;
            return new SeqString(x5.val, tmp);
          default:
            throw new SystemException("Append: should never be reached");
        }
      }
    }

    public static Seq Prefix(Seq v1, Seq v2) {
      if ((object)v1 == (object)v2) return null;
      else if (v1 == null) throw new SystemException("Prefix: unexpected empty sequence: ");
      else {
        Seq tmp = Prefix(v1.next, v2);
        switch (v1.kind) {
          case Seq.Kind.Element: {
            SeqElement x1 = (SeqElement) v1;
            SeqElement res = new SeqElement(x1.tag, x1.contents, tmp);
            res.attributes = x1.attributes;
            return res;
          } case Seq.Kind.Object: {
            SeqObject x1 = (SeqObject) v1;
            SeqObject res = new SeqObject(x1.tag, x1.contents, tmp);
            res.attributes = x1.attributes;
            return res;
          } case Seq.Kind.Bool:
            SeqBool x2 = (SeqBool) v1;
            return new SeqBool(x2.val, tmp);
          case Seq.Kind.Int:
            SeqInt x3 = (SeqInt) v1;
            return new SeqInt(x3.val, tmp);
          case Seq.Kind.Char:
            SeqChar x4 = (SeqChar) v1;
            return new SeqChar(x4.val, tmp);
          case Seq.Kind.String:
            SeqString x5 = (SeqString) v1;
            return new SeqString(x5.val, tmp);
          default:
            throw new SystemException("Prefix: should never be reached");
        }
      }
    }

    public static Seq FixedPrefix(Seq v1, int cnt) {
      if (cnt == 0) return null;
      else if (v1 == null) throw new SystemException("FixedPrefix: unexpected empty sequence: " + cnt);
      else {
        Seq tmp = FixedPrefix(v1.next, cnt-1);
        switch (v1.kind) {
          case Seq.Kind.Element: {
            SeqElement x1 = (SeqElement) v1;
            SeqElement res = new SeqElement(x1.tag, x1.contents, tmp);
            res.attributes = x1.attributes;
            return res;
          } case Seq.Kind.Object: {
            SeqObject x1 = (SeqObject) v1;
            SeqObject res = new SeqObject(x1.tag, x1.contents, tmp);
            res.attributes = x1.attributes;
            return res;
          } case Seq.Kind.Bool:
            SeqBool x2 = (SeqBool) v1;
            return new SeqBool(x2.val, tmp);
          case Seq.Kind.Int:
            SeqInt x3 = (SeqInt) v1;
            return new SeqInt(x3.val, tmp);
          case Seq.Kind.Char:
            SeqChar x4 = (SeqChar) v1;
            return new SeqChar(x4.val, tmp);
          case Seq.Kind.String:
            SeqString x5 = (SeqString) v1;
            return new SeqString(x5.val, tmp);
          default:
            throw new SystemException("FixedPrefix: should never be reached");
        }
      }
    }

    public static Attributes CreateAttributes(object key, object value) {
      Attributes res = new Attributes();
      res.Add(key,value);
      return res;
    }

    public static Attributes ProcessAtts(Seq atts) {
      Attributes res = new Attributes();
      while (atts != null) {
        switch (atts.kind) {
          case Seq.Kind.Element: {
            SeqElement s = (SeqElement) atts;
            res.Add(s.tag, s.contents);
            atts = s.next;
            break;
          } default: {
            throw new SystemException("ProcessAtts: only element nodes allowed in attributes");
          }
        }
      }
      return res;
    }

    public static Seq CreateTree(Sym.Tag tag, Seq atts, Seq contents) {
      SeqElement el = new SeqElement(tag, contents, null);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq CreateTree(object tag, Seq atts, Seq contents) {
      SeqObject el = new SeqObject(tag, contents, null);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq CreateTree(bool tag, Seq atts, Seq contents) {
      SeqBool el = new SeqBool(tag, null);
      return el;
    }

    public static Seq CreateTree(int tag, Seq atts, Seq contents) {
      SeqInt el = new SeqInt(tag, null);
      return el;
    }

    public static Seq CreateTree(char tag, Seq atts, Seq contents) {
      SeqChar el = new SeqChar(tag, null);
      return el;
    }

    public static Seq CreateTree(string tag, Seq atts, Seq contents) {
      SeqString el = new SeqString(tag, null);
      return el;
    }

    public static Seq StringToPcdata(string s) {
      return convText(s, 0, null);
    }

  }
}
