type name = string 
type path = name list

exception NameAlreadyUsed of name
exception AmbiguousName of name


(*************** Namespaces ***********************************************)

(** A value of space type contains, in a hierarchical form,
    information about names of members declared in the namespaces of a
    program.  (This includes namespace, class and regtype
    declarations, but not 'using' directives.) For a namespace that
    has multiple declaration bodies, the names declared in all the
    bodies are collected together.
*)

type leafkind = LClas | LIntf
type leaf = Ident.id * leafkind

type space = 
    (name * member)list
    (** The members list is sorted by names. *)

and member = 
  | Leaf of leaf
  | Space of space

let spaceEmpty = []

let spaceFindChildMem space name : member = 
  let rec finder = function 
    | [] -> raise Not_found
    | (n, mem)::rest -> 
	(match compare name n with
	| 0 -> mem
	| c when c<0 -> raise Not_found
	| c when c>0 -> finder rest
	| _ -> assert false )
  in finder space


(** NEVER raises Not_found, but rather returns back the result which
    is semantically equivalent to the input *)
let rec spaceFindDeepMem space path : member * path = 
  match path with
  | [] -> (Space space, [])
  | ph::pt -> 
      try
	(match spaceFindChildMem space ph with
	| Leaf lf -> (Leaf lf, pt)
	| Space ns -> spaceFindDeepMem ns pt )
      with Not_found -> (Space space, path)


(** Types of functions performing a change to a namespace member.  On
    input, 'Some m' provides the member m that is to be changed,
    'None' informs the function that a new member is to be created
    from scratch.
*)
type altleaf = leaf option -> leaf
type altspace = space option -> space
type altmem = member option -> member

let spaceAlterChildMem space name (alt: altmem) : space = 
  (* Find the spot in the sorted list of members, and perform 
     the alteration there *)
  let rec alterator = function 
    | [] -> [(name, alt None)] 
    | ((n, mem) as p)::rest -> 
	(match compare name n with
	| 0 -> (n, alt (Some mem)) :: rest
	| c when c<0 -> (name, alt None) :: p :: rest
	| c when c>0 -> p :: (alterator rest)
	| _ -> assert false  )
  in alterator space
   

let spaceAlterChildLeaf space name (alt: altleaf) : space = 
  let altmem = function
    | None -> Leaf(alt None)
    | Some(Leaf lf) -> Leaf(alt(Some lf))
    | Some(Space sp) -> raise (NameAlreadyUsed name)
  in spaceAlterChildMem space name altmem

let spaceAlterChildSpace space name (alt: altspace) : space = 
   let altmem = function 
     | None -> Space(alt None)
     | Some(Leaf _) -> raise (NameAlreadyUsed name)
     | Some(Space s) -> Space(alt (Some s))
   in spaceAlterChildMem space name altmem

let spaceAddChildSpace (parent: space) child_name (child: space) : space =
  spaceAlterChildSpace parent child_name 
    (function
       | None -> child
       | Some _ -> raise (NameAlreadyUsed child_name) )

let rec spaceAlterDeepLeaf space (path,name) (alt : altleaf) 
  : space = 
  match path with 
    | [] -> spaceAlterChildLeaf space name alt
    | p::ps -> 
	spaceAlterChildSpace space p 
	(function 
	   | Some s -> spaceAlterDeepLeaf s (ps,name) alt
	   | None -> spaceAlterDeepLeaf spaceEmpty (ps,name) alt
	)

let spaceAddChildLeaf space lf : space =
  let id, _ = lf in
  let name =  Ident.name id in 
    spaceAlterChildLeaf space name 
      (function 
       | None -> lf
       | Some _ -> raise (NameAlreadyUsed name) )


(************* Imports ***************)

type import = 
  | AliasMember of name * member
  | OpenSpace of space



(**************  Name contexts ******************************************)

module AliasTable = Hashtbl.Make(struct 
				   type t = name   
				   let equal = (=)
				   let hash = Hashtbl.hash
				 end)
type ctx = 
    layer list

and layer = 
  | LSpace of space
  | LAliases of member AliasTable.t 

let ctxEmpty = []

let ctxApplyImports ctx imports = 
  let atbl = AliasTable.create (List.length imports) in
  let adder = function 
    | AliasMember (name, member) -> AliasTable.add atbl name member
    | OpenSpace space -> List.iter (fun (n,m) -> AliasTable.add atbl n m) space
  in
    List.iter adder imports;
    (LAliases atbl) :: ctx


let rec ctxFindChildMem (ctx: ctx) name = 
  match  ctx with
  | (LSpace ns) :: rest -> 
      (try spaceFindChildMem ns name
      with Not_found -> ctxFindChildMem rest name )
  | (LAliases atbl) :: rest -> 
      (match AliasTable.find_all atbl name with 
	 | [] -> ctxFindChildMem rest name
	 | [mem] -> mem
	 | _::_ -> raise(AmbiguousName name)  )
  | [] -> raise Not_found
      

(** raises Invalid_argument when path is empty, Not_found when
    unsuccessful, AmbiguousName when conflicting 'uses open'
    directives are encountered. *)
let ctxFindDeepMem (ctx: ctx) path = 
  let (phead, ptail) = match path with
  | hd :: tl -> hd, tl
  | [] -> raise (Invalid_argument "Naming.ctxFindLeaf")
  in
  let child_mem = ctxFindChildMem ctx phead in
  match child_mem with
  | Leaf a -> (Leaf a, ptail)
  | Space ns -> spaceFindDeepMem ns ptail


