(** Datastructures for handling C# namespaces and name visibility.  *)

(**
   There are three data structures provided by this module, each with
   its particular intended usage: 

  - space, a namespace, -- a hierarchical structure for named
  namespace members (which can be leafs (containing class ids) or
  other namespaces -- similar to a directory structure)

  - imports, for using directives, -- to represent the effect of
  'using directives' in the beginning of a namespace body (a
  directive either declares an alias for a long name, or imports
  names of all the members of a namespace referenced by a long name)

  - ctx, a naming context, -- represents the names visible at a
  particular point in a program, which includes names at the top level
  in the global namespace, and names introduced by the using
  directives of the current and enclosing bodies; a naming context
  provides an ability to resolve a long name into the corresponding
  member entry.
*)

type name = string 
type path = name list

exception NameAlreadyUsed of name
exception AmbiguousName of name

type space 

type leafkind = LClas | LIntf
type leaf = Ident.id * leafkind

type member = 
  | Leaf of leaf
  | Space of space


(*************** Namespaces ***********************************************)

val spaceEmpty: space

val spaceFindChildMem:  space -> name -> member

val spaceFindDeepMem: space -> path -> member * path
    (** NEVER raises Not_found, but rather returns back the result which
    is semantically equivalent to the input *)

(** Types of "alteration" functions that can perform an "alteration"
    to a namespace member.  (They are to be supplied by a client of
    this module.) The alteration is either to create a new entry, or to
    modify an existing entry, whatever is requested according to the
    following contract:
    - 'None' requests the alteration to return a fresh entry;
    - 'Some m' requests the alteration to modify the entry m.
*)
type altleaf = leaf option -> leaf
type altspace = space option -> space
type altmem = member option -> member

val spaceAlterChildLeaf: space -> name -> altleaf -> space 
val spaceAddChildLeaf: space -> leaf -> space

val spaceAlterChildSpace: space ->  name -> altspace -> space 
val spaceAddChildSpace: space -> name -> space -> space

(************** Imports  ************************************************)
type import = 
  | AliasMember of name * member
  | OpenSpace of space


(************** Name contexts ********************************************)

type ctx 

val ctxEmpty: ctx

val ctxApplyImports: ctx -> import list -> ctx 

val ctxFindDeepMem: ctx -> path -> member * path
  (** raises 
     Invalid_argument when path is empty, 
     Not_found        when unsuccessful, 
     AmbiguousName    when encountering conflicting 'uses open' directives *)

