(*-- open Il2cs *)
open Cs
module SO = SynOp
let sprintf = Printf.sprintf

let zchar x = sprintf "'%s'" (Char.escaped x)
let zfloat x = sprintf "'%f'" x
let zint x = sprintf "%d" x
let zstring x = sprintf "\"%s\"" (String.escaped x)

let zoption f z x = match x with
| Some y -> sprintf f (z y)
| None -> ""

let rec zlist z l sep = match l with
| a::b::x -> (z a) ^ sep ^ (zlist z (b::x) sep)
| [a] -> z a
| [] -> ""

let zoption_list f z l sep = match l with
| [] -> ""
| _ -> sprintf f (zlist z l sep)

(* for zynOp.ml *)

(* binary operator *)
let zbinaryop = function
| SO.BiPlus -> 8, "+"
| SO.BiMinus -> 8, "-"
| SO.BiMult -> 6, "*"
| SO.BiDiv -> 6, "/"
| SO.BiMod -> 6, "%"
| SO.BiLt -> 12, "<"
| SO.BiGt -> 12, ">"
| SO.BiLeq -> 12, "<="
| SO.BiGeq -> 12, ">="
| SO.BiEq -> 14, "=="
| SO.BiNeq -> 14, "!="
| SO.BiAnd -> 22, "&&"
| SO.BiOr -> 24, "||"

(* method modifier *)
let zmethod_modifier = function
| SO.Virtual -> "virtual "
| SO.Override -> "override "
| SO.New -> "new "
| SO.Normal -> ""

(* for compile.ml *)

let zlocvarsym x = x
let zmembvarsym x = x
let zmethsym x = x
let ztypsym x = x
let znssym x = x
let zlabsym x = x

(* boolean *)
let zbool = function
| true -> "true"
| fals -> "false"

(* constant value *)
let zconstval = function
| TCNull -> "null"
| TCBool x -> zbool x
| TCInt x -> zint x
| TCChar x -> zchar x
| TCString x -> zstring x
| TCSymbol x -> "Tags." ^ Compile.escape_tag x

(* base type *)
let zbasetyp = function
| TBTBool -> "Seq.Kind.Bool"
| TBTInt -> "Seq.Kind.Int"
| TBTChar -> "Seq.Kind.Char"
| TBTObject -> "Seq.Kind.Object"
| TBTElement -> "Seq.Kind.Element"
| TBTEmpty -> "Seq.Kind.Empty"
| TBTPcdata ->
    let pcenc = !Pref.pcdataencoding in
    if pcenc = "string" then "Seq.Kind.String"
    else if pcenc = "substring" then "Seq.Kind.Substring"
    else if pcenc = "hybrid" then
      "Seq.Kind.String: case Seq.Kind.Substring: case Seq.Kind.Char"
    else assert false

(* base value *)
let zbaseval = function
| TBVBool x -> zbool x
| TBVInt x -> zint x
| TBVChar x -> zchar x
| TBVSymbol x -> assert false

(* pattern *)
let zpat = function 
| TPType x -> sprintf "case %s:" (zbasetyp x)
| TPVal x -> sprintf "case %s:" (zbaseval x)
| TPDefault -> "default:"
| TPLabel x -> sprintf "%s:" x

(* predefined type *)
let rec zpredefty = function
| TPTyVoid -> "void"
| TPTyBool -> "bool"
| TPTyInt -> "int"
| TPTyChar -> "char"
| TPTyString -> "string"
| TPTyArray t -> (ztyp t) ^ "[]"

(* type *)
and ztyp = function
| TTYPredef x -> zpredefty x
| TTYNamed x -> ztypsym x

(* actual parameter *)
let rec zact_param (x,y) = sprintf "%s%s"
  (if x then "ref " else "") (zexp max_int y)

(* expression *)
and zexp p = function 
| TEConst x -> zconstval x
| TENew (x,y) -> sprintf "new %s(%s)" (ztyp x) (zlist (zexp max_int) y ", ")
| TENewArray (x,y,z) ->
    sprintf "new %s[%s]%s" (ztyp x) (zexp max_int y)
      (let rec loop = function
        | 0 -> ""
        | n -> "[]"^(loop (n-1))
      in loop z)
| TEArrayAccess (x,y) -> sprintf "%s[%s]" (zexp 3 x) (zexp max_int y)
| TELocal x -> zlocvarsym x
| TEField (x,y) -> sprintf "%s.%s" (zexp 3 x) (zmembvarsym y)
| TEStaticField (x,y) -> sprintf "%s.%s" (ztypsym x) (zmembvarsym y)
| TECall (x,y,s) -> sprintf "%s.%s(%s)" 
  (zexp 3 x) (zmethsym y) (zlist zact_param s ", ")
| TEStaticCall (x,y,s) -> sprintf "%s.%s(%s)"
  (ztypsym x) (zmethsym y) (zlist zact_param s ", ")
| TEAssign (x,y) ->
    if p > 28 then
      sprintf "%s = %s" (zexp 28 x) (zexp 29 y)
    else
      sprintf "(%s = %s)" (zexp 28 x) (zexp 29 y)
| TEUnOp (SO.UnNot,y) ->
    if p > 4 then
      sprintf "!%s" (zexp 5 y)
    else
      sprintf "(!%s)" (zexp 5 y)
| TEUnOp (SO.UnPlus,y) -> sprintf "%s ++" (zexp 3 y)
| TEUnOp (SO.UnMinus,y) -> sprintf "%s --" (zexp 3 y)
| TEBinOp (x,y,s) ->
    let (p', r) = zbinaryop x in
    if p > p' then
      sprintf "%s %s %s" (zexp (p' + 1) y) r (zexp p' s)
    else
      sprintf "(%s %s %s)" (zexp (p' + 1) y) r (zexp p' s)
| TECast (x,y) ->
    if p > 4 then
      sprintf "(%s)%s" (ztyp x) (zexp 5 y)
    else
      sprintf "((%s)%s)" (ztyp x) (zexp 5 y)
| TEIs (x,y) ->
    if p > 12 then
      sprintf "%s is %s" (zexp 13 x) (ztyp y)
    else
      sprintf "(%s is %s)" (zexp 13 x) (ztyp y)
| TEType t -> sprintf "typeof(%s)" (ztyp t)

(* local variable *)
(*--VG??  is the next line deprecated? *)
let zlocals (x,y) = sprintf "%s%s" x (zoption "= %s" (zexp max_int) y)
let zlocals (x,y) = match x with
| [] -> ""
| _ -> 
  let z (u,v) = match v with
  | Some w -> sprintf "%s = %s" (zlocvarsym u) (zexp max_int w)
  | None -> sprintf "%s" (zlocvarsym u) in
  sprintf "%s %s;\n" (ztyp y) (zlist z x ", ")

(* block *)
let rec zblock x = zlist zstmt x ""

(* statement *)
and zstmt = function 
| TSBlock x -> sprintf "{\n%s}\n" (zblock x)
| TSExp x -> sprintf "%s;\n" (zexp max_int x)
| TSLocals (x,y) -> zlocals (x,y)
| TSReturn x -> sprintf "return%s;\n" (zoption " %s" (zexp max_int) x)
| TSIf (x,y,u) -> sprintf "if (%s) {\n%s}%s\n"
  (zexp max_int x) (zblock y) (zoption_list "else {\n%s}" zstmt u "")
| TSWhile (x,y) -> sprintf "while (%s) {\n%s}\n"
  (zexp max_int x) (zblock y)
| TSSwitch (x,y) -> sprintf "switch (%s) {\n%s}\n"
  (zexp max_int x) (zlist zclause y "")
| TSBreak -> "break;\n"
| TSEmpty -> ";\n"
| TSError x -> sprintf "throw new SystemException(%s);\n" (zstring x)
| TSLabel x -> sprintf "\n%s:\n" (zlabsym x)
| TSGoto x -> sprintf "goto %s;\n" (zlabsym x)

(* switch clause *)
and zclause (x, y) =
  sprintf "%s{\n%s}\n" (zpat x) (zblock y)

(* formal parameter *)
let zform_param (x,y,s) = sprintf "%s%s %s" 
  (if x then "ref " else "")
  (ztyp s) (zlocvarsym y)

let zstatus = function
  | SO.Static -> "static "
  | SO.Instance -> ""

(* special function *)
let zinvoke = function 
| TINVNone -> ""
| TINVThis x -> sprintf "this(%s);" (zlist zact_param x ", ")
| TINVSuper x -> sprintf "base(%s);" (zlist zact_param x ", ")

(* class member *)
let zclmember_invoke = function
| TINVNone -> ""
| TINVThis x -> sprintf " : this(%s)" (zlist zact_param x ", ")
| TINVSuper x -> if x=[] then "" else
  sprintf " : base(%s)" (zlist zact_param x ", ")

let zclmember c = function
| TCMConstr (x,y,s) -> sprintf "public %s (%s)%s {\n%s}\n" 
  c (zlist zform_param y ", ") (zclmember_invoke x) (zblock s)
| TCMStaticConstr(body) ->
    sprintf "static %s() {\n%s}\n"  c (zblock body)
| TCMFld (x, s, y, e) -> 
    sprintf 
    "public %s%s %s %s;" 
    (zstatus s)
    (ztyp y)
    (zmembvarsym x)
    (match e with
       | None -> ""
       | Some e -> (" = " ^ (zexp 0 e)))
| TCMMeth (x,stat,m,y,u,v) -> sprintf "public %s%s%s %s (%s) {\n%s}\n"
  (zstatus stat)
  (zmethod_modifier m) (ztyp u) (zmethsym x) 
  (zlist zform_param y ", ") (zblock v)

let zmeth_decl (x,y,u) =
  sprintf "%s %s (%s);\n" (ztyp u) (zmethsym x) (zlist zform_param y ", ")

(* class *)
let zclasext_and_implements ext impl =
  match ext, impl with
  | "object", [] -> ""
  | x, [] -> sprintf " : %s" (ztypsym x)
  | "object", l -> zoption_list " : %s" ztypsym l ", "
  | x, l -> sprintf " : %s%s" (ztypsym x) (zoption_list ", %s" ztypsym l ", ")

let zclas x = 
  let clas_code = 
    sprintf "public class %s%s {\n%s}\n"
      (ztypsym x.tCLName)
      (zclasext_and_implements x.tCLExt x.tCLImpls)
      (zlist (zclmember x.tCLName) x.tCLMems "\n")
  in 
    if x.tCLHomeNS = "" then clas_code
    else sprintf "namespace %s {\n%s}\n" x.tCLHomeNS clas_code

let zintf x = 
  let intf_code = 
    sprintf "public interface %s%s {\n%s}\n"
      (ztypsym x.tINName)
      (zoption_list " : %s" ztypsym x.tINImpls ", ")  
      (zlist zmeth_decl x.tINMems "\n")
  in 
    if x.tINHomeNS = "" then intf_code
    else sprintf "namespace %s {\n%s}\n" x.tINHomeNS intf_code


(*--VG-- not used!
(* interface member *) 
let zifmember = function
| TIMFld (x,y) -> sprintf "%s %s;" (ztyp y) (zmembvarsym x)
| TIMMeth (x,y,u) -> sprintf "%s %s (%s);"
  (ztyp u) (zmethsym x) (zlist zform_param y "\n")

(* interface *)
let ziface x = sprintf "interface %s : %s {\n%s}\n"
  (ztypsym x.tIFName)
  (zlist ztypsym x.tIFExts ", ")
  (zlist zifmember x.tIFMems "")
*)

(* using directive *)
let zusedir = function
| TUSAlias (x,y) -> sprintf "alias %s = %s;\n" (znssym x) (znssym y)
| TUSNSpace x -> sprintf "using %s;\n" (znssym x)

(*--VG--
(* namespace member *)
let rec znsmember = function
| TNMNspace x -> znspace x
| TNMClas x -> zclas x
| TNMIf x -> ziface x

(* namespace *)
and znspace x = sprintf "namespace %s {\nusing %s;\n%s}\n"
  (znssym x.tNSName) (zlist zusedir x.tNSUsing ", ")
  (zlist znsmember x.tNSMems "\n")
*)

(* tag sym *)

let zprog (uses,intfs,classes) = 
  sprintf "%s\n%s\n%s\n"
    (zlist zusedir uses "")
    (zlist zintf intfs "\n")
    (zlist zclas classes "\n")

(* indentation *)
let ncol = 65
let ntab = 2
let blank = String.make ncol ' '
let space n = String.sub blank 0 (min ncol n)

let name x =
  let c = Char.code in
  let y = c x in
  x == '_' || x == '.' || x == '=' || x == '&' || x == '|'
  || (y >= (c 'A') && y<= (c 'Z'))
  || (y >= (c 'a') && y<= (c 'z'))
  || (y >= (c '0') && y<= (c '9'))
  
let almost x = x=' ' || x='\n' || x=';' || x=','

let rec delimit s n i x =
  if i>=n then n else
  let c = String.get s i in
  if c=x then i else
  if c='\\' then
  delimit s n (i+2) x else
  delimit s n (i+1) x


(* indent string length index prefix output offset *)
let rec indent s n i o j p =
  if i>=n then () else (
  let c = String.get s i in
  let d = if i=n-1 then '\000' else String.get s (i+1) in
  match c with
  | '\n' when d='}' ->
    output_string o ("\n" ^ (space (p-ntab)) ^ "}");
    indent s n (i+2) o (p-ntab+1) (p-ntab)
  | '\n' ->
    output_string o ("\n" ^ (space p));
    indent s n (i+1) o p p
  | '{' ->
    output_char o c; 
    indent s n (i+1) o (j+1) (p+ntab)
  | '\'' | '"' ->
    let k = delimit s n (i+1) c in
    let m = k - i + 1 in
    output o s i m;
    indent s n (k+1) o (j+m) p 
  | _ when j>=ncol && not (name c) && not (almost d) ->
    output_char o c;
    output_string o ("\n" ^ (space (p+ntab)));
    indent s n (i+1) o (p+ntab) p
  | _ -> 
    output_char o c; 
    indent s n (i+1) o (j+1) p
  )

let toplevel c o =
  output_string o (sprintf
    "class __%s { public static void Main(string[] x) { new %s().Main(x); }}\n\n"
    c c)

(* [gen compiled_prog start_class] *)
let gen x c =
  let f = !Pref.outputfile in
  let f' = if f="" then (c ^ ".cs") else f in
  let o = open_out f' in
  let s = zprog x in
  let n = String.length s in
  indent s n 0 o 0 0;
(*    toplevel c o; *)
  close_out o
