module ST = SynTyped
module SD = SynDenot
module SO = SynOp
module ID = Ident
module SL = Sortedlist
module AR = Array
module UT = Utils
module INT = Intersection
open List
open Il


(******************************)
(******* Compilation **********)
(******************************)

(****** Symbol generators *)

let _generators_ = ref []

let create_generator (prefix : string) : string * int ref =
  let gens = remove_assoc prefix !_generators_ in
  let gen = prefix, ref 0 in
  _generators_ := gen :: gens;
  gen

let gensym gen : string =
  let r = snd gen in
  incr r;
  fst gen ^ "_" ^ string_of_int (!r)

let get_gensym gen : int = !(snd gen)

let set_gensym gen i : unit = (snd gen) := i

let z_gen = create_generator "z"
let y_gen = create_generator "y"

let reset_generators () = iter (fun (_,r) -> r := 0) !_generators_


(****** Tag list *)

let acompare x y = compare y x

let (uni : int list -> int list -> int list) = SL.merge compare

let _tags_ = ref []

let add_tag x = _tags_ := SL.merge compare [x] !_tags_

let reset_tags () = _tags_ := []

let escape_tag x =
  let x = Str.global_replace (Str.regexp "-") "__" x in 
  let x = Str.global_replace (Str.regexp ":") "_c_" x in 
    x
  

let escape_tag_add_prefix x =
  escape_tag (!Pref.tagprefix ^ x)

let tag_decls () =
  map (fun str ->
         let e = IEStaticCall("Tag","Intern",[false, IEConst (ICString str)]) in
         let id = escape_tag_add_prefix str in
         ICMFld(id,SO.Static,ty_tag,Some e))
    !_tags_


(******* Finding loop breakers among tree automata states *)

let stsdebug = map (fun s -> s.Ta.num)

type state =
  { num : int;    (*VG was: state.id *)
    mutable final : bool;
    mutable finpre : Ta.pre;
    mutable next : trans list;
    desc : SD.seqty option;
    mutable empty : bool option; 
    mutable binders : Ta.binder list option;
    mutable preclos : Ta.pre;
  }
and trans = (SD.labty * SD.attsty * int * int * Ta.pre) 

let statedebug (s:Ta.state) : state = {
  num = s.Ta.num;
  final = s.Ta.final;
  finpre = s.Ta.finpre;
  next = map (fun (l,a,s1,s2,p) -> l,a,s1.Ta.num,s2.Ta.num,p) s.Ta.next;
  empty = s.Ta.empty;
  binders = s.Ta.binders;
  desc = s.Ta.desc;
  preclos = s.Ta.preclos;
}

let _loop_breakers_ = ref []

let reset_loop_breakers () = _loop_breakers_ := []

let is_disj_labtys (l1 : SD.labty) (l2 : SD.labty) : bool =
   match Sub.minimum_labty [l1; l2] with
     | None -> true
     | Some _ -> false

let is_multiple_disj_trans (i : int) : bool =
  match (Ta.index_to_state i).Ta.next with
    | [] | [_] -> false
    | ((l : SD.labty), _, _, _, _) :: rest ->
        UT.ormap (fun ((l2 : SD.labty), _, _, _, _) -> is_disj_labtys l l2) rest

let is_more_than_2_disj_trans (i : int) : bool =
  let rec loop (trs : Ta.trans list) (seen : SD.labty list) (cnt : int) =
    cnt == 0 ||
      (match trs with
        | [] -> false
        | ((l1 : SD.labty), _, _, _, _) :: trs ->
            let isdiff =
              UT.ormap (fun ((l2 : SD.labty), _, _, _, _) -> is_disj_labtys l1 l2) trs in
            let isnew = for_all (fun (l2) -> is_disj_labtys l1 l2) seen in
            if isdiff && isnew then loop trs (l1::seen) (cnt - 1)
            else loop trs seen cnt)
  in loop ((Ta.index_to_state i).Ta.next) [] 2

let substates (s : int) : int list =
  let add_states s1 s2 acc =
    uni [s1.Ta.num] (uni [s2.Ta.num] acc) in
  fold_right
    (fun (_,_,s1,s2,_) acc -> add_states s1 s2 acc)
    (Ta.index_to_state s).Ta.next
    []

let find_loop_breakers (ss : Ta.state list) : int list =
  let _ss_ = stsdebug ss in
  let n = Ta.number_of_states () in
  let pred = AR.make n [] in
  let succ = AR.make n [] in
  let outdeg = AR.make n 0 in
  let indeg = AR.make n 0 in
  let init = AR.make n false in
  let smalls = [!INT.ta_any.Ta.num] in
  let is_empty s = (s.Ta.next == []) && (s.Ta.final) in
  let is_small (i : int) : bool =
    memq i smalls or is_empty (Ta.index_to_state i) in
  let diff_smalls = filter (fun x -> not (is_small x)) in
  let get_indeg (i : int) : int =
    if is_small i then 0
    else indeg.(i)
  in let set_indeg (i : int) (v : int) : unit =
    if not (is_small i) then indeg.(i) <- v
  in let get_outdeg (i : int) : int =
    if is_small i then 0
    else outdeg.(i)
  in let set_outdeg (i : int) (v : int) : unit =
    if not (is_small i) then outdeg.(i) <- v
  in let get_succ (i : int) : int list =
    if is_small i then []
    else succ.(i)
  in let set_succ (i : int) (v : int list) : unit =
    if not (is_small i) then succ.(i) <- v
  in let get_pred (i : int) : int list =
    if is_small i then []
    else pred.(i)
  in let set_pred (i : int) (v : int list) : unit =
    if not (is_small i) then pred.(i) <- v
  in let is_entry_set (i : int) : bool =
    is_small i || init.(i)
  in let set_entry (i : int) : unit =
    if not (is_small i) then (
      init.(i) <- true;
      set_pred i []
    )
  in let filt (f : int * bool * int * int -> bool) : int list =
    let res = ref [] in
    for i = 0 to n - 1 do
      if (not (is_small i)) && f (i,init.(i),indeg.(i),outdeg.(i)) then res := i :: !res
    done;
    rev !res
  in let find_max_deg ss =
    let rec loop ss max res =
      match ss with
        | [] -> res
        | i :: ss ->
            let ind = get_indeg i in
            let outd = get_outdeg i in
            let deg = ind + outd in
            if memq i (get_succ i) then i
            else if ind == 0 || outd == 0 then i
            else if deg > max then loop ss deg i
            else loop ss max res
    in let max = get_outdeg (hd ss) + get_indeg (hd ss) in
    let res = hd ss in
    loop (tl ss) max res
  in let ss =
    ref (diff_smalls (Sort.list (<) (map (fun x -> x.Ta.num) ss)))
  in while !ss != [] do
    let s = hd !ss in
    ss := tl !ss;
    if not (is_entry_set s) then set_entry s;
    let succs = diff_smalls (substates s) in
    set_succ s succs;
    set_outdeg s (length succs);
    iter
      (fun s2 ->
         if not (is_entry_set s2) then (
           ss := uni [s2] !ss;
           set_entry s2;
         );
         set_indeg s2 (1 + (get_indeg s2));
         set_pred s2 (uni [s] (get_pred s2)))
      succs;
  done;
  let res1 = filt (fun (_,flag,ind,_) -> flag && ind > 1) in
  let res2 = filt (fun (i,flag,_,_) -> flag && is_multiple_disj_trans i) in
  let res3 = [] in
  let res4 = filt (fun (i,flag,ind,_) -> flag && ind > 1 && is_multiple_disj_trans i) in
  let res5 = filt (fun (_,flag,_,_) -> flag ) in
  let lb = !Pref.lb in
  let res = match !Pref.lb with
    | "lb-min" -> ref res3
    | "lb-max" -> ref res5
    | "lb-indeg" -> ref res1
    | "lb-indeg-multtrans" -> ref res4
    | "lb-multtrans" -> ref res2
    | _ -> raise (Bug.Bug("Compile.find_loop_breakers", "non-existant loop breaker method"))
  in 
  ss := filt (fun (_,init,_,outd) -> init && outd == 0);
  while !ss != [] do
    let s = hd !ss in
    ss := tl !ss;
    iter
      (fun s2 ->
         set_outdeg s2 (get_outdeg s2 - 1);
         set_succ s2 (SL.diff compare (get_succ s2) [s]);
         if get_outdeg s2 == 0 then ss := uni [s2] !ss)
      (get_pred s)
  done;
  ss := filt (fun (_,init,ind,outd) -> init && ind > 0 && outd > 0);
  while !ss != [] do
    let s = find_max_deg !ss in
    ss := SL.diff compare !ss [s];
    if not (get_indeg s == 0 || get_outdeg s == 0) then
      res := uni [s] !res;
    iter
      (fun s2 ->
         set_indeg s2 (get_indeg s2 - 1);
         set_pred s2 (SL.diff compare (get_pred s2) [s]))
      (get_succ s);
    iter
      (fun s2 ->
         set_outdeg s2 (get_outdeg s2 - 1);
         set_succ s2 (SL.diff compare (get_succ s2) [s]))
      (get_pred s)
  done;
  !res

let add_loop_breakers (ss : Ta.state list) : unit =
  let _ss1_ = map statedebug ss in
  let ss = find_loop_breakers ss in
  let _ss2_ = map (fun s -> statedebug (Ta.index_to_state s)) ss in
  _loop_breakers_ := uni !_loop_breakers_ ss

let is_loop_breaker (s : Ta.state) : bool =
  memq s.Ta.num !_loop_breakers_


(******* Dealing with Ta and Ta.pre *)

let ty_to_state (s0 : Ta.state) (t : SD.pa) : Ta.state =
  let s1 =
      match (SD.ty' t) with
        | SD.Thost(_) -> assert false
        | SD.Tseq(sty) -> Buildta.build_ta sty
        | SD.PTyBind(id,t) -> assert false
  in if !Pref.tb then (
    if !Pref.debug_configs then (
      fpr outf "--------@.";
      Ta.print outf [s1]; 
      fpr outf "\n-----@.";
      Ta.print outf [s0]
    );
    let s = INT.inter s1 s0 in
    if !Pref.debug_configs then (
      fpr outf "\n-----@.";
      Ta.print outf [s];
      fpr outf "@."
    );
    add_loop_breakers [s];
    s
  ) else s1

(** [[T]] x => [[T x]] *)
let ty_rem_binds (t : SD.pa) : SD.ty =
  let rec loop t =
    match (SD.ty' t) with
      | SD.Thost(_) ->
          fpr outf "%a@." Pretty.pr_ty t;
          assert false
      | SD.Tseq(sty) -> sty
      | SD.PTyBind(id,t) -> SD.seqbind id (loop t)
  in  SD.ty (SD.Tseq (loop t))

let rec merge_var_ty_pairs (ls1 : (locvarsym * SD.hostty) list) ls2 = 
  match (ls1,ls2) with
    | [], _ -> ls2
    | _, [] -> ls1
    | ((x1,t1) as p1)::rest1, ((x2,t2) as p2)::rest2 ->
        let c = compare x1 x2 in
        if c < 0 then p1 :: merge_var_ty_pairs rest1 ls2
        else if c > 0 then p2 :: merge_var_ty_pairs ls1 rest2
        else (x1, Sub.join_hostty t1 t2) :: (merge_var_ty_pairs rest1 rest2)

let labty_to_var_types (t : SD.labty) : (locvarsym * SD.hostty) list =
  let rec loop names = function
    | SD.TLab(t : SD.hostty) ->
      fold_right (fun x acc -> merge_var_ty_pairs [x,t] acc) names []
    | SD.PLabBind((x : ID.id), (lty : SD.labty)) ->
      loop (ID.name x :: names) (SD.labty' lty)
  in loop [] (SD.labty' t)

let attsty_to_var_types (t : SD.attsty) : (locvarsym * SD.hostty) list =
  match (SD.attsty' t) with
    | SD.TAttAny -> []
    | SD.PAtt((atts : SD.attpat list), (rest : (ID.id * SD.seqty) option)) ->
        let acc = UT.ifopt (fun (x,_) -> [ID.name x, SD.hty_Seq]) [] rest in
        fold_right (fun (_,x,_) acc -> merge_var_ty_pairs
	  [ID.name x, SD.hty_Seq] acc) atts acc

let rec seqty_to_var_types (t : SD.seqty) : (locvarsym * SD.hostty) list =
  match (SD.seqty' t) with 
    | SD.TTree((lty : SD.labty), (atts : SD.attsty), (t : SD.seqty)) ->
        merge_var_ty_pairs (labty_to_var_types lty)
          (merge_var_ty_pairs (attsty_to_var_types atts) (seqty_to_var_types t))
    | SD.TCat(x : SD.seqty list) ->
        fold_right (fun t acc -> merge_var_ty_pairs (seqty_to_var_types t) acc) x []
    | SD.TAlt(x : SD.seqty list) ->
        fold_right (fun t acc -> merge_var_ty_pairs (seqty_to_var_types t) acc) x []
    | SD.TRep(x : SD.seqty) -> []
    | SD.TSeqRef(x : ID.id) -> []
    | SD.PSeqBind((x : ID.id), (t : SD.seqty)) ->
        merge_var_ty_pairs [ID.name x, SD.hty_Seq] (seqty_to_var_types t)
    | SD.PTailBind((x : ID.id), (t : SD.seqty)) ->
        merge_var_ty_pairs [ID.name x, SD.hty_Seq] (seqty_to_var_types t)

let ty_to_var_types (t : SD.ty) : (locvarsym * SD.hostty) list =
  match (SD.ty' t) with
    | SD.Thost(hty) -> assert false
    | SD.Tseq(sty) -> seqty_to_var_types sty
    | SD.PTyBind(id,t) -> assert false


(******* Compiling types *)

let rec compile_hostty (t : SD.hostty) : typ =
  match (SD.hostty' t) with
    | SD.TClasRef(x : ID.id) 
    | SD.TIntfRef(x : ID.id) -> ITYNamed (ID.longname x)
    | SD.TObject -> ty_object
    | SD.TNull -> assert false
    | SD.TBool -> ITYPredef IPTyBool
    | SD.Tbool(x : bool) -> ITYPredef IPTyBool
    | SD.TInt -> ITYPredef IPTyInt
    | SD.Tint(x : int) -> ITYPredef IPTyInt
    | SD.TChar -> ITYPredef IPTyChar
    | SD.Tchar(x : char) -> ITYPredef IPTyChar
    | SD.TString -> ITYPredef IPTyString
    | SD.Tstring(x : string) -> ITYPredef IPTyString
    | SD.TTag  -> ty_tag
    | SD.Ttag(x : string) -> ty_tag
    | SD.TSeq -> ty_seq
    | SD.TArray t -> ITYPredef (IPTyArray (compile_type t))

and compile_type (t : SD.ty) : typ =
  match (SD.ty' t) with
    | SD.Thost(hty) -> compile_hostty hty
    | SD.Tseq(sty) -> ty_seq
    | SD.PTyBind(id,t) -> assert false

let compile_rettype (t : SD.ty option) : typ =
  match t with
    | None -> ITYPredef IPTyVoid
    | Some(t) -> compile_type t

let bottom_value (t : typ) : exp = match t with
  | ITYNamed _ -> IEConst ICNull
  | ITYPredef u -> match u with
    | IPTyVoid -> assert false
    | IPTyBool -> IEConst (ICBool false)
    | IPTyInt -> IEConst (ICInt 0)
    | IPTyChar -> IEConst (ICChar ' ')
    | IPTyString -> IEConst (ICString "")
    | IPTyArray t -> IEConst ICNull


(******* Various side-effect constructs for the compiler *)

type lab_var_frame = {
  mutable iLVLab: int;
  mutable iLVVar: int;
}

let (_lab_vars_ : lab_var_frame Stack.t) = Stack.create ()

let inc_lab () =
  let frame = Stack.top _lab_vars_ in
  let lab = frame.iLVLab in
  frame.iLVLab <- lab + 1;
  lab

let next_var n =
  let frame = Stack.top _lab_vars_ in
  let v = frame.iLVVar in
  frame.iLVVar <- max v n;
  v

let new_lab_var_frame () = Stack.push { iLVLab = 0; iLVVar = 0 } _lab_vars_

let pop_lab_var_frame () = ignore (Stack.pop _lab_vars_)

let reset_lab_vars () = Stack.clear _lab_vars_


(******* Code generation utilities *)

let gen_numbered_named_id x n = x ^ "_" ^ (string_of_int n)
let gen_numbered2_named_id x n k = x ^ "_" ^ (string_of_int n) ^ "_" ^ (string_of_int k)
let gen_numbered_x_id n = gen_numbered_named_id "x" n
let gen_numbered_y_id n = gen_numbered_named_id "y" n
let gen_numbered_z_id n = gen_numbered_named_id "z" n
let gen_numbered_u_id n = gen_numbered_named_id "u" n
let gen_numbered_k_id n = gen_numbered_named_id "k" n
let gen_numbered_l_id n = gen_numbered_named_id "l" n
let gen_numbered_a_id n = gen_numbered_named_id "a" n
let gen_numbered_w_id n = gen_numbered_named_id "w" n
let gen_numbered_lp_id n = gen_numbered_named_id "lp" n

let gen_id_based_on_typ t i = match t with
  | ITYPredef IPTyBool -> gen_numbered_named_id "lb" i
  | ITYPredef IPTyInt -> gen_numbered_named_id "li" i
  | ITYPredef IPTyChar -> gen_numbered_named_id "lc" i
  | ITYPredef IPTyVoid -> assert false
  | _ ->
      if t == ty_seq then gen_numbered_x_id i
      else if t == ty_tag then gen_numbered_named_id "ll" i
      else if t == ty_string_pcdata then gen_numbered_lp_id i
      else gen_numbered_named_id "lo" i

let rec make_seq (e1 : exp) (e2 : exp) : exp = match (e1,e2) with
  | IEEmpty, _ -> e2
  | _, IEEmpty -> e1
  | IESeq(e11,e12),_ -> IESeq(e11, make_seq e12 e2)
  | _ -> IESeq(e1,e2)

let make_local x = IELocal (UT.intern x)

let make_ipvar x t = IPVar(UT.intern x,t)

let make_prefix exp1 exp2 exp3 =
  let args = [false,exp1; false,exp2; false,exp3] in
  if exp1=exp2 then exp3
  else IEStaticCall("Utils", "Prefix", args)

let make_append e1 e2 =
  if e1 == the_empty_exp then e2
  else if e2 == the_empty_exp then e1
  else IEStaticCall("Utils", "Append", [false, e1; false, e2])

let make_literal_string_seq str =
  let pcenc = !Pref.pcdataencoding in
  if pcenc = "substring" then
    IENew(ty_seqsubstring,[IEConst (ICString str); the_empty_exp])
  else if pcenc = "string" then
    IENew(ty_seqstring,[IEConst (ICString str); the_empty_exp])
  else if pcenc = "hybrid" then
    let k = String.length str in
    if k == 1 then
      IENew(ty_seqchar,[IEConst (ICChar str.[0]); the_empty_exp])
    else
      IENew(ty_seqsubstring,[IEConst (ICString str); the_empty_exp])
  else
    assert false

let compact_switch e : exp =
  let compact e cls f = match cls with
    | [IPDefault, e] -> e
    | _ -> f e cls
  in match e with
    | IESwitchK(e,cls) -> compact e cls (fun e cls -> IESwitchK(e,cls))
    | IESwitchV(e,cls) -> compact e cls (fun e cls -> IESwitchV(e,cls))
    | IESwitchO(e,cls) -> compact e cls (fun e cls -> IESwitchO(e,cls))
    | IESwitchA(e,cls) -> compact e cls (fun e cls -> IESwitchA(e,cls))
    | IESwitchP(e,cls) -> compact e cls (fun e cls -> IESwitchP(e,cls))
    | _ -> e


(******* Dealing with catch blocks *)

type idvarmap = (ID.id * locvarsym) list

type comp_params =
    int list * Ta.state list * int list * typ list * 
    Ta.state list list * exp list * int list * idvarmap list

type lblock_param_map = (comp_params, int) Hashtbl.t

type lblock_frame = {
  iLBParamMap: lblock_param_map;
  mutable iLBLabMap: (int * int list * typ list * exp) list;
}

let (_lblocks_ : lblock_frame Stack.t) = Stack.create ()

let reset_lblocks () = Stack.clear _lblocks_

let new_lblock_frame () =
  Stack.push { iLBParamMap = Hashtbl.create 13; iLBLabMap = [] } _lblocks_

let pop_lblock_frame () = (Stack.pop _lblocks_).iLBLabMap

let lookup_lblock params : exp option =
  let ks, _, ms, mtys, _, _, _, _ = params in
  let frame = Stack.top _lblocks_ in
  try let l = Hashtbl.find frame.iLBParamMap params in
      let ktys = map (fun _ -> ty_seq) ks in
      let xs, tys = SL.merge2 acompare ks ms ktys mtys in
      let ts = map2 (fun t x -> make_local (gen_id_based_on_typ t x)) tys xs in
      Some(IEExit(UT.intern (gen_numbered_k_id l), ts))
  with Not_found -> None

let set_lblock params (e:exp) =
  let l = inc_lab () in
  let ks, _, ms, tys, _, _, _, _ = params in
  let ktys = map (fun _ -> ty_seq) ks in
  let xs, tys = SL.merge2 acompare ks ms ktys tys in
  let frame = Stack.top _lblocks_ in
  Hashtbl.add frame.iLBParamMap params l;
  frame.iLBLabMap <- (l,xs,tys,e)::frame.iLBLabMap;
  let ts = map2 (fun t x -> make_local (gen_id_based_on_typ t x)) tys xs in
  IEExit(UT.intern (gen_numbered_k_id l), ts)


(******* Generating matching procedures *)

let (_matching_proc_states_ : (Ta.state * Ta.state * bool) list ref) = ref []
let (_matching_proc_decls_ : meth_decl list ref) = ref []

let reset_matching_procs () =
  _matching_proc_states_ := [];
  _matching_proc_decls_ := []

let matching_proc_decls () = map (fun x -> ICMMeth x) !_matching_proc_decls_

let is_matching_proc (name:methsym) : bool =
  exists (fun (n,_,_,_,_,_) -> n == name) !_matching_proc_decls_

let number_of_matching_procs () = length !_matching_proc_decls_


(******* Generating regex declarations *)

type regex =
  | IRChar of string * Ta.pre
  | IRAlt of regex * regex
  | IRCat of regex * regex
  | IRStar of regex
  | IREmpty

type pcdatavar =
  | IVPCBeg of ID.id
  | IVPCEnd of ID.id
  | IVPCAll of ID.id
  | IVPCRest of ID.id

type simplicity_specification =
  | ISPStr
  | ISPAny
  | ISPComplex
  | ISPFixed of int

type regex_decl = simplicity_specification * string

let (_regex_decls_ : (regexsym * regex_decl) list ref) = ref []
let (_regex_names_ : (regex * regexsym) list ref) = ref []

let lookup_regexsym r =
  assoc r !_regex_decls_

let reset_regexs () =
  _regex_decls_ := []


(****** Type stamp list *)

let (_typestamps_ : (string * typestampsym) list ref) = ref []
let _number_of_typestamps_ = ref 0

let add_type_stamp ts_str =
  try assoc ts_str !_typestamps_
  with Not_found ->
    let x = gen_numbered_named_id "typestamp" !_number_of_typestamps_ in
    incr _number_of_typestamps_;
    _typestamps_ := (ts_str,x) :: !_typestamps_;
    x

let typestamp_decls () =
  map (fun (str,x) ->
         let e = IEStaticCall("Typestamp","Intern",[false, IEConst (ICString str)]) in
         ICMFld(x,SO.Static,ty_object,Some e))
    !_typestamps_


(****** Pcdata compilation *)


let (uni_b : Ta.pre -> Ta.pre -> Ta.pre ) = SL.merge compare

let binders_to_pcdata_vars bnds restbnds =
  let (uni : pcdatavar list -> pcdatavar list -> pcdatavar list) = SL.merge compare in
  let exists_beg_id id =
    exists (function | Ta.Begin(id2) -> (id == id2) | _ -> false) bnds in
  let endids, bnds, restbnds =
    fold_right (fun b (ids,bnds1,bnds2) -> match b with
                 | Ta.End id ->
                     if exists_beg_id id then (id::ids, (uni_b [b] bnds1), bnds2)
                     else ids, bnds1, uni_b [b] bnds2
                 | _ -> ids, bnds1, uni_b [b] bnds2)
      restbnds
      ([],bnds,[]) in
  let rec loop bnds ids res = match bnds with
    | [] -> fold_right (fun x -> uni [IVPCBeg x]) ids res
    | b::bnds ->
        (match b with
          | Ta.Begin(id) -> loop bnds (id::ids) res
          | Ta.Rest(id) -> loop bnds ids (uni [IVPCRest id] res)
          | Ta.End(id) ->
              if memq id ids then
                loop bnds (filter (fun x -> x!=id) ids) (uni [IVPCAll id] res)
              else
                loop bnds ids (uni [IVPCEnd id] res)
          | Ta.Fixed(_,_) -> assert false) 
  in loop bnds [] [], bnds, restbnds, endids

let regex_binders r : Ta.pre =
  let rec loop r = match r with
    | IREmpty -> []
    | IRChar(_,bnds) -> fold_right (fun b -> uni_b [b]) bnds []
    | IRAlt(r1,r2)
    | IRCat(r1,r2) -> uni_b (loop r1) (loop r2)
    | IRStar(r) -> loop r
  in loop r

let regex_leading_bnds r : Ta.pre * regex * bool =
  let rec loop r = match r with
    | IREmpty
    | IRStar(_) ->
        [], r, false
    | IRChar(str,bnds) ->
        bnds, IRChar(str,[]), true
    | IRAlt(r1,r2) ->
        let bnds1,r1,f1 = loop r1 in
        let _,r2,_ = loop r2 in
        bnds1, IRAlt(r1,r2), f1
    | IRCat(r1,r2) ->
        let bnds1,r1,f1 = loop r1 in
        if f1 then bnds1, IRCat(r1,r2), true
        else let bnds2,r2,f2 = loop r2 in
        bnds2, IRCat(r1,r2), f2
  in loop r

let regex_to_string r pcdatavars endids : string =
  let closed_id id : bool =
    let rec loop = function
      | [] -> false
      | IVPCAll id2::xs when id=id2 -> true
      | _::xs -> loop xs
    in loop pcdatavars in
  let rec loop p = function
    | IREmpty -> ""
    | IRChar(str,bnds) ->
        fold_right
          (fun b acc -> match b with
            | Ta.Begin(id) ->
                if closed_id id then "(?<" ^ ID.name id ^ ">" ^ acc
                else "(?<_beg_" ^ ID.name id ^ ">" ^ acc
            | Ta.End(id) -> ")" ^ acc
            | Ta.Rest(id) -> "(?<_rest_" ^ ID.name id ^ ">" ^ acc
            | Ta.Fixed(id,_) -> assert false)
          (sort acompare bnds)
          str
    | IRAlt(r1,r2) ->
        if p > 4 then loop 4 r1 ^ "|" ^ loop 5 r2
        else "(" ^ loop 4 r1 ^ "|" ^ loop 5 r2 ^ ")"
    | IRCat(r1,r2) ->
        if p > 2 then loop 2 r1 ^ loop 3 r2
        else "(" ^ loop 2 r1 ^ loop 3 r2 ^ ")"
    | IRStar(r) ->
        loop 1 r ^ "*" in
  let prefix =
    fold_right
      (fun x acc -> match x with
        | IVPCEnd id -> "(?<_end_" ^ ID.name id ^ ">" ^ acc     (* ) balancing paren *)
        | _ -> acc)
      pcdatavars
      "" in
  let suffix =
    fold_right
      (fun x acc -> match x with
        | IVPCRest _                          (* ( *)
        | IVPCBeg _ -> ")" ^ acc
        | _ -> acc)
      pcdatavars
      (String.make (length endids) ')')
  in prefix ^ loop 100 r ^ suffix

let register_regex s1 s2 (r:regex) pcdatavars issimple endids isrest : regexsym =
  if r == IREmpty && isrest then
    let rid = "r_" ^ (string_of_int s1.Ta.num) ^ "_" ^ (string_of_int s2.Ta.num) in
    _regex_decls_ := (rid,(ISPAny,""))::!_regex_decls_;
    rid
  else if mem_assoc r !_regex_names_ then
    assoc r !_regex_names_
  else
    let rid = "r_" ^ (string_of_int s1.Ta.num) ^ "_" ^ (string_of_int s2.Ta.num) in
    let str = regex_to_string r pcdatavars endids in
    let issimple = if str = wildcard_pcdata_regex then ISPAny else issimple in
    _regex_names_ := (r,rid)::!_regex_names_;
    _regex_decls_ := (rid,(issimple,str))::!_regex_decls_;
    rid

let regex_decls () =
  let escape ch = match ch with
    | '"' -> "\\\""
    | '{' -> "\\{"
    | '}' -> "\\}"
    | '$' -> "\\$"
    | '.' -> "\\."
    | '?' -> "\\?"
    | _ -> String.make 1 ch in
  let escape str =
    if str = wildcard_char_regex then str
    else if str = wildcard_pcdata_regex then str
    else
      let res = ref "" in
      let n = String.length str in
      for i = 0 to n - 1 do
        let ch = String.get str i in
        res := !res ^ escape ch
      done;
      !res in
  fold_right (fun (id,(issimple,r)) acc ->
         if issimple != ISPComplex then acc
         else
           let e = IENew(ty_regex,[IEConst (ICString ("\\G" ^ escape r)); opt_compiled]) in
           ICMFld(id,SO.Static,ty_regex,Some e) :: acc)
    !_regex_decls_
    []


(******* Converting states to regular expressions *)

let uni_s = SL.merge Ta.compare_state
let diff_s = SL.diff Ta.compare_state

(* This predicate determines whether a given tree automaton
   transition corresponds to a Any pattern in the
   intermediate language *)
let is_trans_any hty atrs s1 s2 = 
  let any_id = (!INT.ta_any).Ta.num in
  match SD.hostty' hty with
    | SD.TObject ->
        (match SD.attsty' atrs with
          | SD.TAttAny -> (s1.Ta.num == any_id && s2.Ta.num == any_id)
          | _ -> false)
    | _ -> false

(* This predicate determines whether a given tree automaton
   transition corresponds to a AnyAtom pattern in the
   intermediate language *)
let is_trans_anyatom hty atrs s1 =
  let any_id = (!INT.ta_any).Ta.num in
  match SD.hostty' hty with
    | SD.TObject ->
        (match SD.attsty' atrs with
          | SD.TAttAny -> (s1.Ta.num == any_id)
          | _ -> false)
    | _ -> false

let is_real_char_trans (l,atrs,s1,s2,_) =
  let (h,_) = SD.crack_labty l in
  match SD.hostty' h with
    | SD.TChar | SD.Tchar _ -> true
    | _ -> false

let is_char_trans (l,atrs,s1,s2,_) =
  let (h,_) = SD.crack_labty l in
  match SD.hostty' h with
    | SD.TChar | SD.Tchar _ -> true
    | SD.TObject -> is_trans_anyatom h atrs s1
    | _ -> false

let is_nonchar_trans (l,atrs,s1,s2,_) =
  let (h,_) = SD.crack_labty l in
  match SD.hostty' h with
    | SD.TChar | SD.Tchar _ -> false
    | _ -> true

let find_reachable_char_states s0 =
  let rec loop sts seen res = match sts with
    | [] -> res
    | s::sts ->
        let final = s.Ta.final || exists is_nonchar_trans s.Ta.next in
        let succs =
          fold_right
            (fun ((_,_,_,s2,_) as r) acc ->
               if is_char_trans r then uni_s [s2] acc else acc)
            s.Ta.next
            [] in
        let seen = uni_s [s] seen in
        loop (uni_s (diff_s succs seen) sts) seen ((s,final,false,[],false)::res)
  in loop [s0] [] []

let is_pcdata_state s =
  s.Ta.final &&
  s.Ta.finpre == [] &&
  match s.Ta.next with
    | [l,_,_,s2,p] ->
        let (h,_) = SD.crack_labty l in
        (match SD.hostty' h with
          | SD.TChar -> (p == [] && s2.Ta.num == s.Ta.num)
          | _ -> false)
    | _ -> false

let is_any_state s =
  s.Ta.num == (!INT.ta_any).Ta.num

let states_to_regex (s:Ta.state) :
  (pcdatavar list * regexsym * Ta.state * Ta.pre * bool * Ta.pre) list =
  let states = AR.of_list (find_reachable_char_states s) in
  let n = AR.length states in
  let pred = AR.make n [] in
  let succ = AR.make n [] in
  let get_state_id s =
    let res = ref (-1) in
    for i = 0 to n - 1 do
      let (s',_,_,_,_) = states.(i) in
      if s==s' then res := i
    done;
    !res in
  let rec make_alt r1 r2 = match (r1,r2) with
    | IRAlt(r1',r1''),IRAlt(r2',r2'') ->
        let c = compare r1' r2' in
        if c==0 then IRAlt(r1',make_alt r1'' r2'')
        else if c<0 then IRAlt(r1',make_alt r1'' r2)
        else IRAlt(r2',make_alt r1 r2'')
    | _,IRAlt(r2',r2'') ->
        let c = compare r1 r2' in
        if c==0 then r2
        else if c<0 then IRAlt(r1,r2)
        else IRAlt(r2',make_alt r1 r2'')
    | IRAlt(_,_),_ ->
        make_alt r2 r1
    | _ ->
        let c = compare r1 r2 in
        if c==0 then r2
        else if c<0 then IRAlt(r1,r2)
        else IRAlt(r2,r1) in
  let rec make_cat r1 r2 = match r1 with
    | IRCat(r1',r1'') -> IRCat(r1',make_cat r1'' r2)
    | IREmpty -> r2
    | _ -> if r2==IREmpty then r1 else IRCat(r1,r2) in
  let make_star r = match r with
    | IRStar(_)
    | IREmpty -> r
    | _ -> IRStar(r) in
  let is_simple_regex =
    let rec loop r = match r with
      | IRChar(ch,bnds) ->
          if ch <> wildcard_char_regex && bnds == [] then ISPStr
          else ISPComplex
      | IRAlt(_,_) -> ISPComplex
      | IRCat(r1,r2) ->
          let x = loop r1 in
          let y = loop r2 in
          if x == ISPStr && y = ISPStr then ISPStr
          else ISPComplex
      | IRStar r ->
          (match r with
            | IRChar(ch,bnds) when (ch = wildcard_char_regex && bnds == []) -> ISPAny
            | _ -> ISPComplex)
      | IREmpty -> ISPStr
  in loop in
  let is_fixed_size =
    let rec loop r = match r with
      | IRChar(ch,_) ->
          let k = if ch = wildcard_char_regex then 1 else String.length ch in
          ISPFixed k
      | IRAlt(r1,r2) ->
          (match (loop r1, loop r2) with
            | ISPFixed k1, (ISPFixed k2 as x) when k1 == k2 -> x
            | _ -> ISPComplex)
      | IRCat(r1,r2) ->
          (match (loop r1, loop r2) with
            | ISPFixed k1, ISPFixed k2 -> ISPFixed (k1+k2)
            | _ -> ISPComplex)
      | IRStar r -> ISPComplex
      | IREmpty -> ISPFixed 0
  in loop in
  let set_succ s1 s2 (r:regex) =
    let tbl = succ.(s1) in
    if mem_assq s2 tbl then
      let r' = assq s2 tbl in
      succ.(s1) <- (s2,make_alt r' r)::(remove_assq s2 tbl)
    else succ.(s1) <- (s2,r)::tbl in
  let set_pred s1 s2 (r:regex) =
    let tbl = pred.(s1) in
    if mem_assq s2 tbl then
      let r' = assq s2 tbl in
      pred.(s1) <- (s2,make_alt r' r)::(remove_assq s2 tbl)
    else pred.(s1) <- (s2,r)::tbl in
  let rem_succ s1 s2 =
    let tbl = succ.(s1) in
    succ.(s1) <- remove_assq s2 tbl in
  let rem_pred s1 s2 =
    let tbl = pred.(s1) in
    pred.(s1) <- remove_assq s2 tbl in
  let has_succ s1 s2 =
    let tbl = succ.(s1) in
    mem_assq s2 tbl in
  let filt p =
    let res = ref [] in
    for i = 0 to n - 1 do
      let (s,final,isrest,restpre,finalany) = states.(i) in
      if p i s final isrest restpre finalany then res := (i::!res);
    done;
    rev !res in
  let elim_state s =
    let succtbl = succ.(s) in
    let predtbl = pred.(s) in
    if mem_assq s succtbl then
      let r = assq s succtbl in
      let succtbl = remove_assq s succtbl in
      let predtbl = remove_assq s predtbl in
      iter (fun (s1,r1) ->
              iter (fun (s2,r2) ->
                      let r = make_cat r1 (make_cat (make_star r) r2) in
                      set_succ s1 s2 r;
                      rem_succ s1 s;
                      set_pred s2 s1 r;
                      rem_pred s2 s)
                succ.(s))
        pred.(s);
    else 
      iter (fun (s1,r1) ->
              iter (fun (s2,r2) ->
                      let r = make_cat r1 r2 in
                      set_succ s1 s2 r;
                      rem_succ s1 s;
                      set_pred s2 s1 r;
                      rem_pred s2 s)
                succ.(s))
        pred.(s) in
  let char_transitions trans =
    fold_right
      (fun ((l,atrs,s1,s2,p) as r) acc ->
         let (h,ids) = SD.crack_labty l in
         match SD.hostty' h with
           | SD.TChar ->
               if ids!=[] then
                 raise
                   (Bug.Bug("Compile.char_transitions", "cannot handle binders in chars"));
               (wildcard_char_regex, p, get_state_id s2) :: acc
           | SD.Tchar ch ->
               if ids!=[] then
                 raise
                   (Bug.Bug("Compile.char_transitions", "cannot handle binders in chars"));
               (String.make 1 ch, p, get_state_id s2) :: acc
           | SD.TObject when is_trans_anyatom h atrs s1 ->
               if ids!=[] then
                 raise
                   (Bug.Bug("Compile.char_transitions", "cannot handle binders in chars"));
               (wildcard_char_regex, p, get_state_id s2) :: acc
           | _ -> acc)
      trans
      [] in
  let mark_final_any_states () =
    for i = 0 to n - 1 do
      let (s,final,_,_,_) = states.(i) in
      if is_pcdata_state s || is_any_state s then states.(i) <- (s,final,false,[],true)
    done in
  let is_finalany i =
    let (_,_,_,_,f) = states.(i) in
    f
  in mark_final_any_states ();
  for i = 0 to n - 1 do
    let (s,final,_,_,finalany) = states.(i) in
    if not finalany then
      let prs = char_transitions s.Ta.next in
      iter
        (fun (str,bnds,s2) ->
           if is_finalany s2 && s.Ta.final && s.Ta.finpre = bnds then
             states.(i) <- (s,final,true,bnds,false);
           let r = IRChar(str,bnds) in
           let s = get_state_id s in
           set_succ s s2 r;
           set_pred s2 s r)
        prs;
  done;
  let ss = filt (fun _ s' final _ _ any -> s != s' && not any && not final) in
  iter elim_state ss;
  let ss = filt (fun _ s' final _ _ any -> s != s' && not any && final) in
  let s' = get_state_id s in
  let prs =
    let do_pair s s2 =
      if s==s2 then
        if has_succ s s then [IRStar(assq s succ.(s)), s]
        else [IREmpty, s]
      else if has_succ s s2 then
        let r = assq s2 succ.(s) in
        let r2 =
          if has_succ s2 s2 then make_cat r (make_star (assq s2 succ.(s2))) else r in
        if has_succ s s then [make_cat (make_star (assq s succ.(s))) r2, s2]
        else [r2, s2]
      else [] in
    let rec calc s stbl ptbl ss stbls ptbls = match ss with
      | [] -> do_pair s s
      | [s2] -> do_pair s s2
      | s2::ss2 ->
          elim_state s2;
          let stbls2 = map (fun i -> succ.(i)) ss2 in
          let ptbls2 = map (fun i -> pred.(i)) ss2 in
          let prs = calc s succ.(s) pred.(s) ss2 stbls2 ptbls2 in
          succ.(s) <- stbl;
          pred.(s) <- ptbl;
          iter2 (fun s stbl -> succ.(s) <- stbl) ss stbls;
          iter2 (fun s ptbl -> pred.(s) <- ptbl) ss ptbls;
          iter elim_state ss2;
          do_pair s s2 @ prs
    in let succs = map (fun i -> succ.(i)) ss in
       let preds = map (fun i -> pred.(i)) ss in
       calc s' succ.(s') pred.(s') ss succs preds
  in map
    (fun (r,s2) ->
       let (s2,_,isrest,restbnds,_) = states.(s2) in
       let s2' = if isrest then !INT.ta_any else s2 in
       let begbnds, r, _ = regex_leading_bnds r in
       let bnds = regex_binders r in
       let xs, bnds, restbnds, endids = binders_to_pcdata_vars bnds restbnds in
       let issimple = is_simple_regex r in
       let issimple = if issimple == ISPComplex then is_fixed_size r else issimple in
       let rid = register_regex s s2 r xs issimple endids isrest in 
       xs, rid, s2', begbnds, isrest, restbnds)
    prs


(******* Pattern representation in various compiler functions *)

type patkind =
  | IPKEmpty
  | IPKLab
  | IPKObj
  | IPKPcdata
  | IPKBase of basetyp
  | IPKAtom
  | IPKDefault

type patrep =
  | IPREmpty
  | IPRLab of SD.hostty * locvarsym list * SD.attsty * Ta.state * Ta.state
  | IPRObj of SD.hostty * locvarsym list * SD.attsty * Ta.state * Ta.state
  | IPRPcdata of regexsym * pcdatavar list * Ta.state * bool * Ta.pre
  | IPRBase of SD.hostty * locvarsym list * SD.attsty * Ta.state * Ta.state
  | IPRAtom of Ta.state
  | IPRDefault

type patrepdebug =
  | IPREmpty'
  | IPRLab' of SD.hostty * locvarsym list * SD.attsty * int * int
  | IPRObj' of SD.hostty * locvarsym list * SD.attsty * int * int
  | IPRPcdata' of regexsym * pcdatavar list * int * bool * Ta.pre
  | IPRBase' of SD.hostty * locvarsym list * SD.attsty * int * int
  | IPRAtom' of int
  | IPRDefault'

let is_object = function
  | IPRObj(h,_,_,_,_) ->
      (match SD.hostty' h with
        | SD.TObject -> true
        | _ -> false)
  | _ -> false

let patrepdebug rep = match rep with
  | IPREmpty -> IPREmpty'
  | IPRDefault -> IPRDefault'
  | IPRLab(x,y,z,s1,s2) -> IPRLab'(x,y,z,s1.Ta.num,s2.Ta.num)
  | IPRObj(x,y,z,s1,s2) -> IPRObj'(x,y,z,s1.Ta.num,s2.Ta.num)
  | IPRBase(x,y,z,s1,s2) -> IPRBase'(x,y,z,s1.Ta.num,s2.Ta.num)
  | IPRPcdata(x,y,s,b,p) -> IPRPcdata'(x,y,s.Ta.num,b,p)
  | IPRAtom(s) -> IPRAtom'(s.Ta.num)

let tplsdebug tpls =
  map (fun (k, rep, sts, t, b, f, map, ms, tys) ->
         (k, patrepdebug rep, stsdebug sts, t, b, f, map, ms, tys))
    tpls

(* This function turns a transition into a pattern representation
   value. The goal of the latter is to identify a kind of pattern
   that will be used in the intermediate language pattern
   matching clause *)
let trans_to_patrep l atrs s1 s2 =
  let (hty,ids) = SD.crack_labty l in
  let xs = map ID.name ids in
  let hty' = SD.hostty' hty in
  match hty' with
    | SD.TInt | SD.Tint _ -> [IPKBase IBTInt, IPRBase(hty,xs,atrs,s1,s2)]
    | SD.TBool | SD.Tbool _ -> [IPKBase IBTBool, IPRBase(hty,xs,atrs,s1,s2)]
    | SD.TChar | SD.Tchar _ ->
        if !Pref.pcdataencoding <> "char" then []
        else [IPKBase IBTChar, IPRBase(hty,xs,atrs,s1,s2)]
    | SD.TTag | SD.Ttag _ -> [IPKLab, IPRLab(hty,xs,atrs,s1,s2)]
    | SD.TObject ->
        if is_trans_any hty atrs s1 s2 then [IPKDefault, IPRDefault]
        else if is_trans_anyatom hty atrs s1 then [IPKAtom, IPRAtom s2]
        else [IPKObj, IPRObj(hty,xs,atrs,s1,s2)]
    | SD.TIntfRef _ | SD.TClasRef _ | SD.TString | SD.Tstring _ ->
        [IPKObj, IPRObj(hty,xs,atrs,s1,s2)]
    | _ -> assert false

let substates0 _ sts = sts

let substates1 rep sts = match rep with
  | IPREmpty -> assert false
  | IPRDefault -> !INT.ta_any::sts
  | IPRLab(_,_,_,_,s)
  | IPRObj(_,_,_,_,s)
  | IPRBase(_,_,_,_,s)
  | IPRPcdata(_,_,s,_,_) 
  | IPRAtom(s) -> s::sts

let substates2 rep sts = match rep with
  | IPREmpty
  | IPRPcdata(_,_,_,_,_) -> assert false
  | IPRDefault -> !INT.ta_any::!INT.ta_any::sts
  | IPRAtom(s) -> !INT.ta_any::s::sts
  | IPRLab(_,_,_,s1,s2)
  | IPRObj(_,_,_,s1,s2)
  | IPRBase(_,_,_,s1,s2) -> s1::s2::sts


(******* Compiling match statements *)

let _conf_id_ = ref 0

let next_conf_id id =
  if !Pref.debug_configs then
    let next_id = !_conf_id_ in
    incr _conf_id_;
    fpr outf "config transition: %d -> %d@." id next_id;
    next_id
  else 0

let print_config id params = 
  let ks, in_sts, ms, _, stss, ts, bs, _ = params in
  let sts = fold_right (fun x acc -> SL.merge Ta.compare_state [x] acc) in_sts [] in
  let sts =
    fold_right (fun sts acc ->
                  fold_right (fun x acc -> SL.merge Ta.compare_state [x] acc) sts acc)
      stss sts
  in Ta.print outf sts;
  fpr outf "Configuration %d:@." id;
  fpr outf "   ks: ";
  iter (fun x -> fpr outf "%d " x) ks;
  fpr outf "@.";
  fpr outf "   in_sts: ";
  iter (fun x -> fpr outf "%d " x.Ta.num) in_sts;
  fpr outf "@.";
  fpr outf "   ms: ";
  iter (fun x -> fpr outf "%d " x) ms;
  fpr outf "@.";
  fpr outf "   bs: ";
  iter (fun x -> fpr outf "%d " x) bs;
  fpr outf "@.";
  Printf.printf "   ts: ";
  iter (fun x -> Printil.exp stdout x; Printf.printf " ") ts;
  fpr outf "@.";
  iter (fun sts ->
          fpr outf "   ";
          iter (fun x -> fpr outf "%d " x.Ta.num) sts;
          fpr outf "@.")
    stss

let is_wildcard_state (s : Ta.state) : bool = (s == !INT.ta_any)

let squish_params (xs : int list) in_sts (stss : Ta.state list list)
    (cmds : exp list) (bs : int list) =
  let same (sts : Ta.state list) : bool =
    match sts with
      | s :: sts ->
          let rec loop = function
            | [] -> true
            | [s2] -> s.Ta.num == s2.Ta.num || (is_wildcard_state s
	      && not (Ta.has_binders s2))
            | s2::sts -> s.Ta.num == s2.Ta.num && loop sts
          in loop sts
      | _ -> assert false
  in let compress stss cmds bs =
    UT.split3 (UT.setify (UT.map3 (fun sts c b -> sts,c,b) stss cmds bs))
  in let transp (stss : Ta.state list list) (ls : 'a list) : Ta.state list list =
    if stss == [] then map (fun _ -> []) ls
    else let rec loop = function
      | []::_ -> []
      | ls -> map hd ls :: loop (map tl ls)
    in loop stss
  in let prune_all_any_rows unit =
    let stss2 = ref [] in
    let ts2 = ref [] in
    let bs2 = ref [] in
    let (clump : (Ta.state list * exp * int) list ref) = ref [] in
    let clump_compare (sts1,_,_) (sts2,_,_) : int =
      let w1 = for_all is_wildcard_state sts1 in
      let w2 = for_all is_wildcard_state sts2 in
      if w1 && w2 then 0
      else if w1 then (-1)
      else if w2 then 1
      else 0
    in let rec loop stss ts bs i =
      if stss == [] || i != hd bs then (
        clump := sort clump_compare !clump;
        stss2 := (map (fun (x,_,_) -> x) !clump) @ !stss2;
        ts2 := (map (fun (_,x,_) -> x) !clump) @ !ts2;
        bs2 := (map (fun (_,_,x) -> x) !clump) @ !bs2;
      );
      match (stss,ts,bs) with
        | [],[],[] ->
            rev !stss2, rev !ts2, rev !bs2
        | sts::stss, t::ts, b::bs ->
            if i==b then (
              clump := (sts,t,b) :: !clump;
              loop stss ts bs i
            ) else (
              clump := [sts,t,b];
              loop stss ts bs b
            )
        | _ -> assert false
    in loop stss cmds bs (if bs==[] then 0 else hd bs)
  in let combine_almost_equal_rows ((stss, cmds, bs) as params) =
    params
  in let check_any_of_case ((stss,cmds,bs) as params) =
     if hd stss == [] then [hd stss], [hd cmds], [hd bs]
     else params
  in let sts_eq sts1 sts2 =
    let rec loop = function
      | [],[] -> true
      | s1::sts1, s2::sts2 -> (s1.Ta.num=s2.Ta.num) && loop (sts1,sts2)
      | _ -> false
    in loop (sts1,sts2)
  in let remove_last_equal_rows ((stss,cmds,bs) as params) =
(*  also checks for identical index afterwards *)
    let (stss,cmds,bs) as params =
      match rev stss with
        | sts1::sts2::stss when sts_eq sts1 sts2 ->
            let rec loop stss cmds bs cmd b = match stss with
              | [] -> [sts1], [cmd], [b]
              | sts::stss2 ->
                  if sts=sts1 then loop stss2 (tl cmds) (tl bs) (hd cmds) (hd bs)
                  else rev (sts1::stss), rev (cmd::cmds), rev (b::bs)
            in
            loop stss (tl (tl (rev cmds))) (tl (tl (rev bs))) (hd (tl (rev cmds)))
              (hd (tl (rev bs)))
        | _ -> params
    in match bs with
      | [] -> params
      | b::bs ->
          if for_all (fun x -> x == b) bs then
            if for_all (for_all (fun s -> not (Ta.has_binders s))) stss then
              [[]], [hd cmds], [b]
            else params
          else params
(*    in let remove_last_equal_rows params = params  (* ,,, *) *)
  in let params = prune_all_any_rows ()
  in let params = check_any_of_case params
  in let ((stss, cmds, bs) as params) = combine_almost_equal_rows params
  in let rec loop xs in_sts stss ys in_sts2 res =
    let _stss_ = map stsdebug stss in
    match (xs,in_sts,stss) with
      | [], _, _ -> rev ys, rev in_sts2, compress (transp (rev res) cmds) cmds bs
      | x::xs, s0::in_sts, sts::stss ->
          let _sts_ = stsdebug sts in
          let f = not (Ta.has_binders (hd sts)) in
          let issame = same sts in
          if f && issame then loop xs in_sts stss ys in_sts2 res
          else loop xs in_sts stss (x :: ys) (s0 :: in_sts2) (sts :: res)
      | _ -> assert false
  in let ks,in_sts,params = loop xs in_sts (transp stss cmds) [] [] []
  in let params = remove_last_equal_rows params
  in ks,in_sts,params

let branching_factor =
  let rec loop cur rest = match cur with
    | [] ->
        (match rest with
          | [] -> 0
          | s::rest -> 1 + loop [s] rest)
    | s::cur ->
        let rest,same = partition (INT.disjoint s) rest in
        loop (same @ cur) rest
  in loop []

let rec exists_loop_breaker = function
  | [s] -> Ta.has_binders s && is_loop_breaker s
  | s::sts -> is_loop_breaker s || exists_loop_breaker sts
  | _ -> assert false

let exists_nonrec_col params =
  let _, _, _, _, stss, _, _, _ = params in
  let rec loop stss i best_i max =
    if hd stss == [] then best_i
    else if exists_loop_breaker (map hd stss) then loop (map tl stss) (i+1) best_i max
    else if not !Pref.tb then i
    else let bf = branching_factor (map hd stss) in
         if bf > max then loop (map tl stss) (i+1) i bf
         else loop (map tl stss) (i+1) best_i max
  in loop stss 0 (-1) 0

let gen_any_of_branch params =
  let _, _, _, _, _, ts, _, _ = params in
  match ts with
    | e::_ -> set_lblock params e
    | _ -> assert false

let split_rec_state params =
  let ks, in_sts, _, _, stss, ts, bs, maps = params in
  let sts = map hd stss in
  let s = find is_loop_breaker sts in
  let sts2 = UT.extract_after (fun x -> s == x) sts sts in
  let stss2 = UT.extract_after (fun x -> s == x) sts stss in
  let ts2 = UT.extract_after (fun x -> s == x) sts ts in
  let bs2 = UT.extract_after (fun x -> s == x) sts bs in
  let maps2 = UT.extract_after (fun x -> s == x) sts maps in
  (hd ks, hd in_sts, tl ks, tl in_sts, s,
   UT.extract_before (fun x -> s == x) sts stss,
   [tl (hd (UT.extract_where (fun x -> s == x) sts stss))],
   tl stss2,
   UT.extract_before (fun x -> s == x) sts ts,
   [hd (UT.extract_where (fun x -> s == x) sts ts)],
   tl ts2,
   UT.extract_before (fun x -> s == x) sts bs,
   [hd (UT.extract_where (fun x -> s == x) sts bs)],
   tl bs2,
   UT.extract_before (fun x -> s == x) sts maps,
   [hd (UT.extract_where (fun x -> s == x) sts maps)],
   tl maps2)

let add_any_row stss in_sts ts bs maps t (b:int) =
  stss @ [in_sts],
  ts @ [t],
  bs @ [b],
  maps @ [[]]

let gen_case_opt_vars bs k :
  (ID.id * locvarsym * typ) list * int list * typ list * int list * ID.id list =
  let rec loop bs k prs ms ms2 tys ids eids = match bs with
    | [] -> prs, ms, tys, ms2, eids
    | x :: bs ->
        match x with
          | Ta.BNSeq(Ta.Begin(id)) | Ta.BNSeq(Ta.Fixed(id,_))
          | Ta.BNSeq(Ta.Rest(id)) | Ta.BNAttr(id) ->
              let y = gen_numbered_x_id k in
              if memq id ids then loop bs k prs ms ms2 tys ids eids
              else
                loop bs (k+1) ((id,y,ty_seq)::prs) (k::ms) ms2 (ty_seq::tys)
                  (id::ids) eids
          | Ta.BNSeq(Ta.End(id)) ->
              if memq id ids then loop bs k prs ms ms2 tys ids eids
              else loop bs (k+1) prs ms (k::ms2) tys (id::ids) (id::eids)
          | Ta.BNLabel(id,h) ->
              let t = compile_hostty h in
              let y = gen_id_based_on_typ t k in
              loop bs (k+1) ((id,y,t)::prs) (k::ms) ms2 (t::tys) (id::ids) eids
  in loop bs k [] [] [] [] [] []

let gen_submaps ids maps : locvarsym list * locvarsym list * idvarmap =
  let rec loop i xs zs submaps = function
    | [] -> xs, zs, submaps
    | id::ids ->
        let x = assq id maps in
        let z = gen_numbered_z_id i in
        loop (i+1) (x::xs) (z::zs) ((id,z)::submaps) ids
  in loop 0 [] [] [] ids

let handle_acts (v:locvarsym) (e:exp) (map:idvarmap) (p:Ta.pre) : exp * idvarmap = 
  fold_right
    (fun a (e,map) -> match a with
       | Ta.Begin(id) ->
           e, (id, v) :: map
       | Ta.End(id) ->
           (try let x0 = assq id map in
                let exp1 = make_local x0 in
                let exp2 = make_local v in
                let exp3 = the_empty_exp in
                let t = make_prefix exp1 exp2 exp3 in
                make_seq (IEAssign(make_local (ID.name id), t)) e, map
            with Not_found ->
              let x0 = "_end_" ^ ID.name id in
              make_seq (IEAssign(make_local x0, make_local v)) e, map)
       | Ta.Rest(id) ->
           make_seq (IEAssign(make_local (ID.name id), make_local v)) e, map
       | Ta.Fixed(id,i) ->
           let args = [false, make_local v; false, IEConst (ICInt i)] in
           let t = IEStaticCall("Utils", "FixedPrefix", args) in
           make_seq (IEAssign(make_local (ID.name id), t)) e, map)
    p
    (e,map)

let unroll_altss k x s alts finalopt sts e b mp ms tys exists_char_trans =
  let is_beg_bind = function
    | IVPCBeg _ -> true
    | _ -> false in
  let do_binds p isfree =
    let p = sort acompare p in
    let e, mp = handle_acts x e mp p in
    if isfree then
      e,ms,tys,mp
    else
      let ms,tys = SL.merge2 acompare [k] ms [ty_seq] tys in
      e,ms,tys,mp in
  let rest =
    if !Pref.pcdataencoding <> "char" && exists_char_trans then
      let tuples = states_to_regex s in
      let _tuples_ = map (fun (x,r,s,p1,isrest,p2) -> x,r,s.Ta.num,p1,isrest,p2) tuples in
      let _s_ = statedebug s in
      map (fun (xs,r,s,begbnds,isrest,restbnds) ->
             let isfree = begbnds == [] && not (exists is_beg_bind xs) in
             let e,ms,tys,mp = do_binds begbnds isfree in
             IPKPcdata,IPRPcdata(r,xs,s,isrest,restbnds),sts,e,b,isfree,mp,ms,tys)
        tuples
    else [] in
  let rest = 
    fold_right
      (fun (l,atrs,s1,s2,p) acc ->
         let prs = trans_to_patrep l atrs s1 s2 in
         let isfree = p==[] in
         let e,ms,tys,mp = do_binds p isfree in
         (map (fun (patk,patr) -> patk,patr,sts,e,b,isfree,mp,ms,tys) prs @ acc))
      alts
      rest
  in match finalopt with
    | None -> rest
    | Some(p) -> 
        let p = sort acompare p in
        let e, mp = handle_acts x e mp p in
        let f = p==[] in
        let ms, tys = if f then (ms, tys) else SL.merge2 acompare [k] ms [ty_seq] tys in
        (IPKEmpty, IPREmpty, sts, e, b, f, mp, ms, tys) :: rest

let extract_col (params:comp_params) (k:int) (x:locvarsym) i =
  let _, _, ms, tys, stss, ts, bs, maps = params in
  let _stss_ = map stsdebug stss in
  let f =
    exists
      (fun sts ->
         let s = fst (UT.rem_inner sts i) in exists is_real_char_trans s.Ta.next)
      stss in
  let rec loop stss ts bs maps = match stss with
    | [] -> []
    | sts :: stss ->
        let s, sts = UT.rem_inner sts i in
        let finalopt = if s.Ta.final then Some s.Ta.finpre else None in
        let alts = filter (fun (_,_,s1,s2,_) -> not (Ta.is_empty s1 || Ta.is_empty s2))
                     s.Ta.next
        in unroll_altss k x s alts finalopt sts (hd ts) (hd bs) (hd maps) ms tys f @
        loop stss (tl ts) (tl bs) (tl maps)
  in loop stss ts bs maps

let gen_lab_id pat i = match pat with
  | IPKLab -> gen_numbered_named_id "ll" i
  | IPKObj -> gen_numbered_named_id "lo" i
  | IPKBase(b) ->
      (match b with
        | IBTBool -> gen_numbered_named_id "lb" i
        | IBTInt -> gen_numbered_named_id "li" i
        | IBTChar -> gen_numbered_named_id "lc" i
        | IBTElement | IBTObject | IBTEmpty -> assert false)
  | IPKEmpty | IPKPcdata | IPKAtom | IPKDefault -> assert false

let compare_patkind p1 p2 =
  if p1 = p2 then 0
  else match (p1,p2) with
    | _, IPKDefault -> (-1)
    | IPKDefault, _ -> 1
    | _, IPKAtom -> (-1)
    | IPKAtom, _ -> 1
    | _, IPKObj -> (-1)
    | IPKObj, _ -> 1
    | _, IPKBase _ -> (-1)
    | IPKBase _, _ -> 1
    | _, IPKLab -> (-1)
    | IPKLab, _ -> 1
    | _, IPKPcdata -> (-1)
    | IPKPcdata, _ -> 1
    | _ -> 0

let basetyp_to_typ (bt : basetyp) : typ = match bt with
  | IBTBool -> ITYPredef IPTyBool
  | IBTInt -> ITYPredef IPTyInt
  | IBTChar -> ITYPredef IPTyChar
  | _ -> assert false

let get_next_var k ks rest_ks ms f =
  let v = 0 in
  let hd' ls = if ls == [] then 0 else hd ls in
  if f then max k (1 + max (hd' rest_ks) (max (hd' ms) v))
  else max k (1 + max (hd' ks) (max (hd' ms) v))

let named_to_object (t : typ) : typ =
  if t == ty_seq || t == ty_tag || t == ty_string_pcdata then t
  else match t with
    | ITYNamed(x) -> ty_object
    | ITYPredef _ -> t

let handle_const_pat h xs k_lab v_lab e ms tys = 
  let f = xs == [] in
  let t = compile_hostty h in
  let ms, tys = if f then (ms, tys) else SL.merge2 acompare ms [k_lab] tys [t] in
  let x = make_local v_lab in
  let e2 = if t==ty_seq then x else IECast(t,x) in
  let e = fold_right (fun x e -> make_seq (IEAssign (make_local x, e2)) e) xs e
  in e, ms, tys

let needs_rest = exists (function | IVPCBeg _ | IVPCRest _ -> true | _ -> false)

let insert_assign x0 y prefix e =
  let rec loop e = match e with
    | IESeq((IEAssign(_,_) as e1), e2) -> IESeq(e1,loop e2)
    | IELet((IPVar(x,t) as pat),bnd,e) ->
        if x=x0 then
          IELet(pat,bnd,make_seq (IEAssign(make_local y, prefix)) e)
        else
          IELet(pat,bnd,loop e)
    | e -> make_seq (IEAssign(make_local y, prefix)) e
  in loop e

let handle_rest_of_pcdata v1 v2 strv r (e:exp) map (restbnds:Ta.pre) : exp * idvarmap =
  let get_empty_ids bnds =
    let rec loop bnds ids = match bnds with
      | [] -> []
      | b::bnds ->
          (match b with
            | Ta.Begin id -> loop bnds (id::ids)
            | Ta.End id ->
                if memq id ids then id :: loop bnds ids
                else loop bnds ids
            | _ -> loop bnds ids)
    in loop (sort compare bnds) [] in
  let rem_id_bnds id bnds =
    fold_right (fun b acc -> match b with
                  | Ta.Begin id2
                  | Ta.End id2 -> if (id == id2) then acc else b :: acc
                  | _ -> b :: acc)
      bnds
      [] in
  let empty_ids = get_empty_ids restbnds in
  let restbnds = fold_right (fun id bnds -> rem_id_bnds id bnds) empty_ids restbnds in
  let e = fold_right
            (fun id e -> make_seq (IEAssign(make_local (ID.name id), the_empty_exp)) e)
            empty_ids
            e
  in fold_right
    (fun a (e,map) -> match a with
       | Ta.Begin(id) ->
           e, (id, v2) :: map
       | Ta.End(id) ->
           (try let x0 = assq id map in
                let exp1 = make_local x0 in
                let exp2 = make_local v1 in
                let str = make_local strv in
                let exp3 = IERegexField(r,"__match__", str, the_empty_exp) in
                let t = make_append (make_prefix exp1 exp2 the_empty_exp) exp3 in
                insert_assign x0 (ID.name id) t e, map
            with Not_found ->
              let x0 = "_end_" ^ ID.name id in
              insert_assign v2 x0 (make_local v2) e, map)
       | Ta.Rest(id) ->
           insert_assign v2 (ID.name id) (make_local v2) e, map
       | Ta.Fixed(id,i) ->
           let args = [false, make_local v2; false, IEConst (ICInt i)] in
           let t = IEStaticCall("Utils", "FixedPrefix", args) in
           insert_assign v2 (ID.name id) t e, map)
    restbnds
    (e,map)

let handle_pcdata_pat
    xs e r str_k strv allv rest_k restv' ks ms tys map isfree isrest restbnds =
  let ms, tys = SL.merge2 acompare [str_k] ms [ty_string_pcdata] tys in
  let k = get_next_var 0 [] ks ms isfree in
  let flag = needs_rest xs || (isrest && restbnds != []) in
  let ms, tys = if flag then SL.merge2 acompare [k] ms [ty_seq] tys else ms, tys in
  let newk = if flag then k+1 else k in
  let restv = if flag then gen_numbered_x_id k else restv' in
  let rec loop xs k e ms tys map = match xs with
    | [] ->  e, ms, tys, map
    | x::xs ->
        (match x with
          | IVPCBeg id -> 
              let ms, tys = SL.merge2 acompare [k] ms [ty_seq] tys in
              let v = gen_numbered_x_id k in
              let map = (id,v)::map in
              let beg_x = "_beg_" ^ (ID.name id) in
              let rest = make_local restv in
              let str = make_local strv in
              let field = IERegexField(r,beg_x,str,rest) in
              let e = IELet(make_ipvar v ty_seq, Some(field), e) in
              loop xs (k+1) e ms tys map
          | IVPCEnd id ->
              (try let x0 = assq id map in
                   let end_x = "_end_" ^ (ID.name id) in
                   let all = make_local allv in
                   let rest = the_empty_exp in
                   let str = make_local strv in
                   let tail = IERegexField(r,end_x,str,rest) in
                   let prefix = make_prefix (make_local x0) all tail in
                   let e = insert_assign x0 (ID.name id) prefix e in
                   let ms,tys = SL.merge2 acompare ms [rest_k] tys [ty_seq] in
                   loop xs k e ms tys map
               with Not_found ->
                 let x0 = "_end_" ^ ID.name id in
                 let e = make_seq (IEAssign(make_local x0, make_local allv)) e in
                 loop xs k e ms tys map)
          | IVPCAll id ->
              let x = ID.name id in
              let rest = the_empty_exp in
              let str = make_local strv in
              let field = IERegexField(r,x,str,rest) in
              let e = make_seq (IEAssign (make_local x, field)) e in
              loop xs k e ms tys map
          | IVPCRest id ->
              let x = ID.name id in
              let ms, tys = SL.merge2 acompare [k] ms [ty_seq] tys in
              let v = gen_numbered_x_id k in
              let rest_x = "_rest_" ^ (ID.name id) in
              let rest = make_local restv in
              let str = make_local strv in
              let field = IERegexField(r,rest_x,str,rest) in
              let e = make_seq (IEAssign (make_local x, make_local v)) e in
              let e = IELet(make_ipvar v ty_seq, Some(field), e) in
              loop xs (k+1) e ms tys map) in
  let (e,ms,tys,map) = loop xs newk e ms tys map in
  let e =
    if flag then
      let rest = make_local restv' in
      let str = make_local strv in
      let field = IERegexField(r,"__rest__",str,rest) in
      IELet(make_ipvar restv ty_seq, Some(field), e)
    else e in
  let e, map =
    if flag then handle_rest_of_pcdata allv restv strv r e map restbnds else e, map
  in e, ms, tys, map

let is_wild_const h =
  let h = SD.hostty' h in
  match h with
    | SD.TObject | SD.TBool | SD.TInt | SD.TChar | SD.TTag -> true
    | SD.TClasRef _ | SD.TIntfRef _ | SD.Tbool _ | SD.Tint _ | SD.Tchar _ | SD.Ttag _
    | SD.TString | SD.Tstring _ -> false
    | _ -> assert false

let trans_hostty_pat h : baseval = match (SD.hostty' h) with
  | SD.TClasRef(x) | SD.TIntfRef (x) -> IBVObj (ID.longname x)
  | SD.TString -> IBVObj ("string")
  | SD.Tbool x -> IBVBool x
  | SD.Tint x -> IBVInt x
  | SD.Tchar x -> IBVChar x
  | SD.Ttag x ->
      add_tag x;
      IBVSymbol x
  | x -> assert false

(* This function processes an attribute part of a transition,
   replaces variables mentioned in the pattern by the variables
   with internal names, and introduces an assignment expression
   to copy the contents of an internal variable into the
   corresponding source program variable after pattern matching
   is finished *)
let handle_att_pat atrs e ks ms tys : pat * exp * int list * typ list =
  match SD.attsty' atrs with
    | SD.TAttAny -> IPDefault, e, ms, tys
    | SD.PAtt(atts, resto) ->
        let k = get_next_var 0 [] ks ms true in
        let rec loop atts k prs e ms tys = match atts with
          | [] ->
              (match resto with
                | None -> IPAtt(rev prs, None), e, ms, tys
                | Some(id,_) ->
                    let ms, tys = SL.merge2 acompare [k] ms [ty_seq] tys in
                    let x = ID.name id in
                    let v = gen_numbered_x_id k in
                    let e = make_seq (IEAssign (make_local x, make_local v)) e in
                    IPAtt(rev prs, Some(make_ipvar v ty_seq)), e, ms, tys)
          | (a,id,_)::atts ->
              add_tag a;
              let ms, tys = SL.merge2 acompare [k] ms [ty_seq] tys in
              let x = ID.name id in
              let v = gen_numbered_x_id k in
              let e = make_seq (IEAssign (make_local x, make_local v)) e in
              loop atts (k+1) ((a,make_ipvar v ty_seq)::prs) e ms tys
        in loop atts k [] e ms tys

let gen_let_ids prs ms endids (e:exp) (map:idvarmap) =
  fold_right2 (fun m id e ->
                let x = ID.name id in
                let x1 = gen_numbered_x_id m in
                try
                  let beg_x = assq id map in
                  let exp1 = make_local beg_x in
                  let exp2 = make_local x1 in
                  let exp3 = the_empty_exp in
                  let t = make_prefix exp1 exp2 exp3 in
                  make_seq (IEAssign (make_local x, t)) e
                with Not_found ->
                  let end_x = "_end_" ^ x in
                  make_seq (IEAssign(make_local end_x, make_local x1)) e)
    ms
    endids
    (fold_right (fun (id,x,t) e ->
                   let v = make_local x in
                   let e2 = if t==ty_seq then v else IECast(t,v) in
                   make_seq (IEAssign (make_local (ID.name id), e2)) e) prs e)

let rec gen_match_proc_call (s:Ta.state) (s0 : Ta.state) x xeids endids prs e expo =
  let s0 = if !Pref.tb && !Pref.tbprocs then s0 else !INT.ta_any in   (* ,,, *)
  let exh = match expo with None -> true | _ -> false in
  let name = gen_numbered2_named_id (if exh then "me" else "mne") s.Ta.num s0.Ta.num in
  let xeids = map gen_numbered_x_id xeids in
  let pats = map (fun (_,x,t) -> (x, named_to_object t)) prs in
  let pats = map (fun x -> UT.intern x, ty_seq) xeids @ pats in
  if (not (mem (s,s0,exh) !_matching_proc_states_)) then (
    _matching_proc_states_ := (s,s0,exh) :: !_matching_proc_states_;
    let x0 = gen_numbered_x_id 0 in
    let formvars1 = map (fun (id,_,_) -> ID.name id) prs in
    let formvars = map (fun id -> "_end_" ^ (ID.name id)) endids @ formvars1 in
    let formparams = map2 (fun u (_,t) -> true, u, t) formvars pats in
    let formparams = (false, x0, ty_seq) :: formparams in
    let body = gen_match_proc_body s s0 formvars1 exh in
    let retty = if exh then IPTyVoid else IPTyBool in
    let decl = name, SO.Static, SO.Normal, formparams, ITYPredef retty, body in
    _matching_proc_decls_ := decl :: !_matching_proc_decls_
  );
  IEProcCall(name,UT.intern x,pats,e,expo)

and gen_match_proc_body s s0 (formvars:locvarsym list) exh =
  let stss = [[s]] in
  let retexp = IEReturn (Some (IEConst (ICBool true))) in
  let e0 = if exh then IEReturn None else retexp in
  new_lab_var_frame ();
  let res =
    if exh then compile_init_conf s0 stss [e0] [0] false
    else let retexp = IEReturn (Some (IEConst (ICBool false))) in
         let (stss, ts, bs, _) = add_any_row stss [s0] [e0] [0] [] retexp 1 in
         compile_init_conf s0 stss ts bs false
  in pop_lab_var_frame ();
  res

and gen_rec_state_query id params k s0 rest_ks rest_ins s stss ts bs maps stss2
    ts2 bs2 maps2 =
  let bnds = Ta.get_binders s in
  let ks, in_sts, ms, tys, _, _, _, _ = params in
  let prs, ms', tys', ms'', endids =
    gen_case_opt_vars bnds (get_next_var 0 [] rest_ks ms true) in
  let ms3, tys3 = SL.merge2 acompare ms' ms'' tys' (map (fun _ -> ty_seq) ms'') in
  let ms2 = ms3 @ ms in
  let bnds = map (fun (id,x,t) -> make_ipvar (ID.name id) t, Some (make_local x)) prs in
  let ts = map2 (gen_let_ids prs ms'' endids) ts maps in
  let x = gen_numbered_x_id k in
  let tys1 = tys3 @ tys in
  if stss2 == [] then
    let cmd1 =
      comp (next_conf_id id) (rest_ks, rest_ins, ms2, tys1, stss, ts, bs, maps) true
    in if ms' == [] && ms'' == [] then cmd1
    else gen_match_proc_call s s0 x ms'' endids prs cmd1 None
  else 
    let cmd2 = comp (next_conf_id id) (ks, in_sts, ms, tys, stss2, ts2, bs2, maps2) true
    in let stss, ts, bs, maps =
      if !Pref.tb && !Pref.tbprocs then
        add_any_row stss rest_ins ts bs maps cmd2 (-1)
      else add_any_row stss (map (fun _ -> !INT.ta_any) (hd stss)) ts bs maps cmd2 (-1)
    in let ktys = map (fun _ -> ty_seq) ks in
    let ms2, tys2 = SL.merge2 acompare ks ms2 ktys tys1 in
    let cmd1 =
      comp (next_conf_id id) (rest_ks, rest_ins, ms2, tys2, stss, ts, bs, maps) true 
    in let x = gen_numbered_x_id k in
    gen_match_proc_call s s0 x ms'' endids prs cmd1 (Some cmd2)

and gen_rec_state id params
      (k, s0, rest_ks, rest_ins, s, stss1, stss, stss2, ts1, ts, ts2,
       bs1, bs, bs2, maps1, maps, maps2) =
  if !Pref.debug_configs then (
    fpr outf "rec state config: %d %d@." id s.Ta.num;
  );
  let cmd =
    gen_rec_state_query id params k s0 rest_ks rest_ins s stss ts bs maps stss2
      ts2 bs2 maps2
  in if stss1 == [] then cmd
  else let ks, in_sts, ms, tys, _, _, _, maps = params in
       let stss1,ts1,bs1,maps1 =
         if !Pref.tb && !Pref.tbprocs then
           add_any_row stss1 in_sts ts1 bs1 maps1 cmd (-1)
         else add_any_row stss1 (map (fun _ -> !INT.ta_any) (hd stss1)) ts1
           bs1 maps1 cmd (-1)
       in let ktys = map (fun _ -> ty_seq) ks in
       let ms, tys = SL.merge2 acompare ks ms ktys tys in
       comp (next_conf_id id) (ks, in_sts, ms, tys, stss1, ts1, bs1, maps1) true

and gen_col_case id (params:comp_params) i =
  let ks, in_sts, ms, tys, stss, ts, bs, maps = params in
  let x,rest_ks = UT.rem_inner ks i in
  let s0,rest_ins = UT.rem_inner in_sts i in
  let v = gen_numbered_x_id x in
  let tpls = extract_col params x v i in
  let _tpls_ = tplsdebug tpls in
  let cls = gen_val_branches id s0 x ks rest_ks rest_ins ms tpls in
  set_lblock params (compact_switch (IESwitchK (make_local v, cls)))

and gen_val_branches id s0 curk ks rest_ks rest_ins ms tpls : clause list =
  let patreps' = map (fun (k, _, _, _, _, _, _, _, _) -> k) tpls in
  let patreps' = fold_right (fun x acc -> (SL.merge compare_patkind [x] acc)) patreps' []
  in let rec loop tpls pat res other f patreps = match tpls with
    | [] ->
        let k = get_next_var curk ks rest_ks ms f in
        let cl = gen_val_branch id s0 curk k rest_ks rest_ins pat (rev res) in
        if other==[] || pat == IPKDefault || pat == IPKAtom then [cl]
        else cl :: loop (rev other) (hd patreps) [] [] true (tl patreps)
    | ((pat2, patr, _, _, _, isfree, _, _, _) as tpl) :: tpls ->
        let new_f = isfree && f in
        if pat2 == IPKDefault && pat == IPKPcdata then
          loop tpls pat res (tpl::other) new_f patreps
        else if pat2 == IPKDefault && pat2 != pat then
          loop tpls pat (tpl::res) (tpl::other) new_f patreps
        else if pat != IPKObj && pat2 == IPKObj && (is_object patr) then
          loop tpls pat (tpl::res) (tpl::other) new_f patreps
        else if pat2 == IPKAtom then
          match pat with
            | IPKEmpty | IPKPcdata -> loop tpls pat res (tpl::other) f patreps
            | _ -> loop tpls pat (tpl::res) (tpl::other) new_f patreps
        else if pat ==  IPKDefault || pat == IPKAtom || pat = pat2 then
          loop tpls pat2 (tpl::res) other new_f patreps
        else loop tpls pat res (tpl::other) f patreps
  in loop tpls (hd patreps') [] [] true (tl patreps')

and gen_val_branch id s0 curk k rest_ks rest_ins patrep tpls : clause =
  let _tpls_ = tplsdebug tpls in
  match patrep with
    | IPKEmpty ->
        if !Pref.debug_configs then (
          fpr outf "empty config: %d@." id;
        );
        IPEmpty, recurse_m id s0 rest_ks rest_ins tpls substates0
    | IPKAtom ->
        let rest = gen_numbered_x_id k in
        let ks = k::rest_ks in
        let pat = IPAnyAtom (make_ipvar rest ty_seq) in
        gen_anyatom_branch id s0 pat ks rest_ins tpls
    | IPKPcdata ->
        let ks = k+1::rest_ks in
        let contv = gen_numbered_lp_id k in
        let restv = gen_numbered_x_id (k+1) in
        let allv = gen_numbered_x_id curk in
        let cls = gen_pcdata_branches id s0 k contv allv (k+1) restv ks rest_ins tpls
        in IPPcdata {iTag = IPDontCare; iAtt = IPDontCare;
                     iCont = make_ipvar contv ty_string_pcdata;
                     iRest = make_ipvar restv ty_seq},
        compact_switch (IESwitchP(make_local contv, cls))
    | IPKDefault ->
        if !Pref.debug_configs then (
          fpr outf "default config: %d@." id;
        );
        let cmd = recurse_m id s0 rest_ks rest_ins tpls substates0 in
        IPDefault, cmd
    | _ -> 
        let ks = k+3::k+2::rest_ks in
        let contv = gen_numbered_x_id (k+3) in
        let restv = gen_numbered_x_id (k+2) in
        let attv = gen_numbered_a_id (k+1) in
        let labv = gen_lab_id patrep k in
        let cls = gen_const_branches id s0 k labv attv ks rest_ins tpls in
        (match patrep with
          | IPKLab ->
              IPElement {iTag = make_ipvar labv ty_tag; iAtt = make_ipvar attv ty_att;
                      iCont = make_ipvar contv ty_seq; iRest = make_ipvar restv ty_seq},
              compact_switch (IESwitchV(make_local labv, cls))
          | IPKObj ->
              IPObj {iTag = make_ipvar labv ty_object; iAtt = make_ipvar attv ty_att;
                      iCont = make_ipvar contv ty_seq; iRest = make_ipvar restv ty_seq},
              compact_switch (IESwitchO(make_local labv, cls))
          | IPKBase bt ->
              IPBaseType(bt,
                 {iTag = make_ipvar labv (basetyp_to_typ bt);
                  iAtt = make_ipvar attv ty_att;
                  iCont = make_ipvar contv ty_seq; iRest = make_ipvar restv ty_seq}),
              compact_switch (IESwitchV(make_local labv, cls))
          | IPKDefault | IPKPcdata | IPKAtom | IPKEmpty -> assert false)

and gen_pcdata_branches id s0 str_k strv allv rest_k restv ks rest_ins tpls :
    clause list =
  let _tpls_ = tplsdebug tpls in
  let rec loop tpls pat res other = match tpls with
    | [] ->
        let cl = gen_pcdata_branch id s0 ks rest_ins pat (rev res) in
        if other==[] or pat == IPDefault then [cl]
        else cl :: loop (rev other) IPDefault [] []
    | ((patk,patrep,sts,e,b,f,map,ms,tys) as tpl) :: tpls ->
        (match patrep with
          | IPRPcdata(r,xs,s,isr,bnds) ->
              let (e,ms,tys,map) =
                handle_pcdata_pat xs e r str_k strv allv rest_k restv ks ms tys map f isr
                  bnds in
              let f = bnds == [] in
              let ms,tys =
                if f then ms,tys else SL.merge2 acompare [rest_k] ms [ty_seq] tys in
              let tpl2 = patk,patrep,sts,e,b,f,map,ms,tys in
              let pat2 = IPRegex(r, xs == [] && bnds == []) in
              if pat == IPDefault or pat = pat2 then loop tpls pat2 (tpl2::res) other
              else loop tpls pat res (tpl::other)
          | _ -> assert false)
  in loop tpls IPDefault [] []

and gen_pcdata_branch id s0 ks rest_ins pat tpls : clause =
  pat, recurse_m id s0 ks rest_ins tpls substates1

and gen_const_branches id s0 k_lab v_lab v_att ks rest_ins tpls : clause list =
  let rec loop tpls pat res other = match tpls with
    | [] ->
        let cl = gen_const_branch id s0 v_att ks rest_ins pat (rev res) in
        if other==[] or pat == IPDefault then [cl]
        else cl :: loop (rev other) IPDefault [] []
    | ((patk, patrep, sts, e, b, f, map, ms, tys) as tpl) :: tpls ->
        (match patrep with
          | IPRDefault
          | IPRAtom _ ->
              loop tpls pat (tpl::res) (tpl::other)
          | IPRLab(h,xs,_,_,_)
          | IPRObj(h,xs,_,_,_)
          | IPRBase(h,xs,_,_,_) ->
              let (e,ms,tys) = handle_const_pat h xs k_lab v_lab e ms tys in
              let tpl = patk, patrep, sts, e, b, f, map, ms, tys in
              if is_wild_const h then loop tpls pat (tpl::res) (tpl::other)
              else let pat2 = IPBaseVal (trans_hostty_pat h) in
              if pat == IPDefault or pat = pat2 then loop tpls pat2 (tpl::res) other
              else loop tpls pat res (tpl::other)
          | _ -> assert false)
  in loop tpls IPDefault [] []

and pr_pat outf = function
  | IPRLab(hty,_,_,_,_)
  | IPRObj(hty,_,_,_,_)
  | IPRBase(hty,_,_,_,_) ->
      Pretty.pr_hostty outf hty
  | _ -> assert false

and gen_const_branch id s0 v_att ks rest_ins pat tpls : clause =
  if !Pref.debug_configs then (
    fpr outf "const config: %d @?" id;
    Printf.fprintf stdout "%a" Printil.pat pat;
    fpr outf "@.";
  );
  let any_att_pat (_,patr,_,_,_,_,_,_,_) = match patr with
    | IPRDefault
    | IPRAtom _ -> true
    | IPRLab(_,_,a,_,_)
    | IPRObj(_,_,a,_,_)
    | IPRBase(_,_,a,_,_) -> (SD.attsty' a == SD.TAttAny)
    | _ -> assert false
  in let _tpls_ = tplsdebug tpls in
  match pat with
    | IPDefault when for_all any_att_pat tpls ->
        pat, recurse_m id s0 ks rest_ins tpls substates2
    | _ -> 
        let cls = gen_att_branches id s0 ks rest_ins tpls in
        pat, compact_switch (IESwitchA(make_local v_att, cls))

and gen_att_branches id s0 ks rest_ins tpls : clause list =
  let rec loop tpls pat res other = match tpls with
    | [] ->
        let cl = gen_att_branch id s0 pat ks rest_ins (rev res) in
        if other==[] or pat == IPDefault then [cl]
        else cl :: loop (rev other) IPDefault [] []
    | ((patk,patr,sts,e,b,f,map,ms,tys) as tpl) :: tpls ->
        (match patr with
          | IPRDefault
          | IPRAtom _ ->
              loop tpls pat (tpl::res) (tpl::other)
          | IPRLab(_,_,atrs,_,_)
          | IPRObj(_,_,atrs,_,_)
          | IPRBase(_,_,atrs,_,_) ->
              let pat2, e, ms, tys = handle_att_pat atrs e ks ms tys in
              let tpl2 = patk, patr, sts, e, b, f, map, ms, tys in
              if pat2 == IPDefault then loop tpls pat (tpl2::res) (tpl::other)
              else if pat == IPDefault or pat = pat2 then loop tpls pat2 (tpl2::res) other
              else  loop tpls pat res (tpl::other)
          | _ -> assert false)
  in loop tpls IPDefault [] []

and gen_att_branch id s0 pat ks rest_ins tpls : clause =
  pat, recurse_m id s0 ks rest_ins tpls substates2

and gen_anyatom_branch id s0 pat ks rest_ins tpls : clause =
  pat, recurse_m id s0 ks rest_ins tpls substates1

and union2 s1 s2 =
  let s = INT.union2 s1 s2 in
  if s == s1 || s == s2 then s
  else (
    add_loop_breakers [s];
    s
  )

and match_trans_patrep (l:SD.labty) patr =
  let (hty,_) = SD.crack_labty l in
  match patr with
    | IPREmpty -> false
    | IPRLab(hty2,_,_,_,_)
    | IPRObj(hty2,_,_,_,_)
    | IPRBase(hty2,_,_,_,_) ->
        (match Sub.minimum_hostty [hty;hty2] with
          | None -> false
          | _ -> true)
    | IPRPcdata(_,_,_,_,_) ->
        (match Sub.minimum_hostty [hty; SD.hostty SD.TChar] with
          | None -> false
          | _ -> true)
    | IPRAtom _
    | IPRDefault -> true

and filter_trans_by_patrep patr trs =
  sort Ta.compare_trans
    (filter (fun (l,_,_,_,_) -> match_trans_patrep l patr) trs)

and compare_patr patr1 patr2 = match (patr1,patr2) with
  | IPREmpty, IPREmpty -> 0
  | _, IPREmpty -> 1
  | IPREmpty, _ -> (-1)
  | IPRLab(_,_,_,_,_), IPRLab(_,_,_,_,_) -> 0
  | _, IPRLab(_,_,_,_,_) -> 1
  | IPRLab(_,_,_,_,_), _ -> (-1)
  | IPRObj(_,_,_,_,_), IPRObj(_,_,_,_,_) -> 0
  | _, IPRObj(_,_,_,_,_) -> 1
  | IPRObj(_,_,_,_,_), _ -> (-1)
  | IPRBase(_,_,_,_,_), IPRBase(_,_,_,_,_) -> 0
  | _, IPRBase(_,_,_,_,_) -> 1
  | IPRBase(_,_,_,_,_), _ -> (-1)
  | IPRPcdata(_,_,_,_,_), IPRPcdata(_,_,_,_,_) -> 0
  | _, IPRPcdata(_,_,_,_,_) -> 1
  | IPRPcdata(_,_,_,_,_), _ -> (-1)
  | IPRAtom _, IPRAtom _ -> 0
  | _, IPRAtom _ -> 1
  | IPRAtom _, _ -> (-1)
  | _ -> 0

and principle_patr tpls =
  hd (sort compare_patr (map (fun (_,patr,_,_,_,_,_,_,_) -> patr) tpls))

and recurse_m id s0 ks rest_ins tpls substates =
  let _tpls_ = tplsdebug tpls in
  let patr = principle_patr tpls in
  let alts = filter_trans_by_patrep patr s0.Ta.next in
  let (s1,s2) =
    match alts with
      | [] -> Ta.empty, Ta.empty
      | (_,_,s1,s2,_)::alts ->
          fold_right (fun (_,_,q1,q2,_) (s1,s2) ->
                        union2 s1 q1,
                        union2 s2 q2)
            alts (s1, s2) in
  let in_sts =
    if substates == substates0 then []
    else if substates == substates1 then [s2]
    else if substates == substates2 then [s1; s2]
    else assert false in
  let rec loop tpls stss ts bs maps ms tys = match tpls with
    | [] ->
        comp (next_conf_id id)
          (ks, in_sts @ rest_ins, ms, tys, rev stss, rev ts, rev bs, rev maps) true
    | (_,patrep,sts,t,b,_,map,ms2,tys2) :: tpls ->
        let sts = substates patrep sts in
        let ms, tys = SL.merge2 acompare ms ms2 tys tys2 in
        loop tpls (sts::stss) (t::ts) (b::bs) (map::maps) ms tys
  in loop tpls [] [] [] [] [] []

and print_int_list ks =
  iter (fpr outf "%d\t") ks;
  fpr outf "\n"

(* assume stss (ts bs) != [] *)
and comp (id : int) (params : comp_params) recur =
(*    let control = Gc.get () in *)
(*    fpr outf "Stack limit: %d\n" control.Gc.stack_limit; *)
(*    Gc.print_stat stdout; *)
(*   if mem id [500;510;520;530;540;550;560;570] then *)
(*     fpr outf ""; *)
  let ks, in_sts, ms, tys, stss, ts, bs, maps = params in
  ignore (next_var (get_next_var 0 [] ks ms true));
  if !Pref.debug_configs then print_config id params;
  let ks,in_sts,(stss,ts,bs) = squish_params ks in_sts stss ts bs in
  let params = ks, in_sts, ms, tys, stss, ts, bs, maps in
  if !Pref.debug_configs then print_config id params;
  match lookup_lblock params with
    | Some(code) -> code
    | None ->
        if hd stss == [] then gen_any_of_branch params
        else let col = exists_nonrec_col params in
        if col>=0 or not recur then gen_col_case id params (if col<0 then 0 else col)
        else set_lblock params (gen_rec_state id params (split_rec_state params))

and compile_init_conf s0 (stss : Ta.state list list) ts bs (recur : bool) : exp =
  let k = next_var 0 in
  new_lblock_frame ();
  let maps = map (fun _ -> []) bs in
  let body = comp (next_conf_id (-1)) ([k], [s0], [], [], stss, ts, bs, maps) recur in
  let rec loop = function
    | [] -> body
    | [k,ks,tys,cmd] -> cmd
    | (k,ks,tys,cmd)::rest ->
        let catch_pats =
          map2 (fun k t ->  (UT.intern (gen_id_based_on_typ t k), named_to_object t)) ks tys
        in IECatch(loop rest, UT.intern (gen_numbered_k_id k), catch_pats, cmd)
  in loop (rev (pop_lblock_frame ()))

let compile_match (test : exp) (in_type : SD.ty)
                  (pats : SD.pa list) (ts0 : exp list) : exp =
  let v = gen_numbered_x_id (next_var 0) in
  let pats = map ty_rem_binds pats in
  let s0 =
    match (SD.ty' in_type) with
      | SD.Thost(h) ->
          Pretty.pr_hostty outf h;
          assert false
      | SD.Tseq(sty) ->
          Buildta.build_ta sty
      | SD.PTyBind(id,t) ->
          assert false
  in let sts = map (ty_to_state s0) pats in
  add_loop_breakers sts;
  if !Pref.debug_configs then (
    fpr outf "Loop breakers: ";
    iter (fun x -> fpr outf "%d " x) !_loop_breakers_;
    fpr outf "\n"
  );
  Ta.print outf (map Ta.index_to_state !Pref.print_states);
  let var_tys_list = map ty_to_var_types pats in
  let var_tys = fold_right merge_var_ty_pairs var_tys_list [] in
  fold_right
    (fun (x,t) e -> IELet (make_ipvar x (compile_hostty t),None,e))
    var_tys
    (let rec loop ts0 var_tys_list ts bs = match ts0 with
       | [] ->
           let stss = map (fun s -> [s]) sts in
           compile_init_conf s0 stss (rev ts) (rev bs) true
       | t::ts0 ->
           let var_tys = hd var_tys_list in
           let catch_pats = map (fun (x,t) -> (UT.intern x, compile_hostty t)) var_tys in
           let vs = map fst var_tys in
           let l = inc_lab () in
           let lab = UT.intern (gen_numbered_k_id l) in
           let ex = IEExit(lab, map (fun x -> make_local x) vs) in
           IECatch (loop ts0 (tl var_tys_list) (ex::ts) (l::bs), lab, catch_pats, t)
     in IELet(make_ipvar v ty_seq, Some(test), loop ts0 var_tys_list [] []))


(******* Compiling expressions and statements *)

let rec compile_actparam (e : ST.exp) : act_param = false, compile_exp e

and compile_exp {ST.exp_desc=e;ST.exp_type=_} : exp = match e with
  | ST.EClass cid ->
      failwith "compilation of static class access not implemented in compile.ml"
  | ST.Ebool(x : bool) -> IEConst (ICBool x)
  | ST.Eint(x : int) -> IEConst (ICInt x)
  | ST.Echar(x : char) -> IEConst (ICChar x)
  | ST.Estring(x : string) -> IEConst (ICString x)
  | ST.Etag(x : string) ->
      add_tag x;
      IEConst (ICSymbol x)
  | ST.Enull -> IEConst ICNull
  | ST.EThis -> make_local "this"
  | ST.EVar(x : ID.id) -> make_local (ID.name x)
  | ST.ENew((id : ID.id), (exps : ST.exp list)) ->
      IENew (ITYNamed (ID.longname id), map compile_exp exps)
  | ST.ENewArray (baset, exp, dims) ->
      IENewArray (compile_type baset, compile_exp exp, dims)
  | ST.EFieldAccess(({ST.exp_desc=base_desc;ST.exp_type=_} as base), (fname : ID.id)) ->
      (match base_desc with
        | ST.EClass cid ->
            IEStaticField (ID.longname cid, ID.name fname)
        | _ -> IEField (compile_exp base, ID.name fname))
  | ST.EArrayAccess (ar, ind) ->
      IEArrayAccess (compile_exp ar, compile_exp ind)
  | ST.EMethInvk(({ST.exp_desc=base_desc;ST.exp_type=_} as base), _,
                 (mname : ID.id), mangled_name, (args : ST.exp list)) ->
      (match base_desc with
        | ST.EClass cid ->
            IEStaticCall (ID.longname cid, mangled_name, map compile_actparam args)
        | _ -> IECall (compile_exp base, mangled_name, map compile_actparam args))
  | ST.EProcMethInvk(({ST.exp_desc=base_desc;ST.exp_type=_} as base), _,
                     (mname : ID.id), mangled_name, (args : ST.exp list)) ->
      (match base_desc with
        | ST.EClass cid ->
            IEStaticCall (ID.longname cid, mangled_name, map compile_actparam args)
        | _ -> IECall (compile_exp base, mangled_name, map compile_actparam args))
  | ST.EAssign((left : ST.exp), (right : ST.exp)) ->
      IEAssign(compile_exp left, compile_exp right)
  | ST.ECast((ty : SD.ty), (e : ST.exp), issafe) ->
      let {ST.exp_desc=e_desc;ST.exp_type=e_type} = e in
      let exp = compile_exp e in
      (match SD.ty' ty with
        | SD.Thost(hty) -> IECast(compile_type ty, exp)
        | SD.Tseq(sty) ->
            let xid = gensym z_gen in
            let x = make_local xid in
            let str = SD.toTypeStampString_seqty sty in
            let y = add_type_stamp str in
            let field1 = IEField(x,"kind") in
            let field2 = IEStaticField("Seq","Kind.Typestamp") in
            let test1 = IEBinOp(SO.BiEq, field1, field2) in
            let cast = IECast(ty_seqtypestamp,x) in
            let field1 = IEField(cast,"typestamp") in
            let field2 = IEStaticField("Typestamps",y) in
            let test2 = IEBinOp(SO.BiEq, field1, field2) in
            let test = IEBinOp(SO.BiAnd, test1, test2) in
            let next = IEField(x,"next") in
            let ty2 = compile_type ty in
            (match e_type with
              | None -> assert false
              | Some e_type ->
                  let e_type =
                    (match SD.ty' e_type with
                      | SD.Thost(_) -> SD.ty_any
                      | _ -> e_type)
                  in IELet(make_ipvar xid ty2,
                           Some(IECast(ty_seq,exp)),
                           make_seq
                             (IEIf(test,
                                   IEAssign(x,next),
                                   (if not issafe then IEEmpty
                                    else
                                      (compile_match x e_type [ty; e_type]
                                         [IEEmpty; IEError "cast failure"]))))
                          x))
        | SD.PTyBind(id,t) -> assert false)
  | ST.ETypeStamp(sty, e) ->
      let str = SD.toTypeStampString_seqty sty in
      let x = add_type_stamp str in
      let e2 = compile_exp e in
      IETypeStamp(x,e2)
  | ST.EUnary((op : SO.unaryop), (e : ST.exp)) -> IEUnOp(op, compile_exp e)
  | ST.EBinaryArith((e1 : ST.exp), (op : SO.binaryop), (e2 : ST.exp)) ->
      IEBinOp(op, compile_exp e1, compile_exp e2)
  | ST.EBinaryBool((e1 : ST.exp), (op : SO.binaryop), (e2 : ST.exp)) -> 
      IEBinOp(op, compile_exp e1, compile_exp e2)
  | ST.EConcatString((e1 : ST.exp), (e2 : ST.exp)) ->
      IEBinOp (SO.BiPlus, compile_exp e1, compile_exp e2)
  | ST.EBinaryEq((e1 : ST.exp), (eq : bool), (e2 : ST.exp)) ->
      if eq then IEBinOp(SO.BiEq, compile_exp e1, compile_exp e2)
      else IEBinOp(SO.BiNeq, compile_exp e1, compile_exp e2)
  | ST.ESeq(exps : ST.exp list) -> compile_catexp exps
  | ST.ETree((tag : ST.exp), (atts : ST.exp list), (cont : ST.exp list)) ->
      IEStaticCall("Utils", "CreateTree",
        [compile_actparam tag; (false, compile_catexp atts); (false, compile_catexp cont)])
  | ST.EPcdata tag ->
      make_literal_string_seq tag
  | ST.EString2Pcdata(x : ST.exp) ->
      IEStaticCall("Utils", "StringToPcdata", [compile_actparam x])

and compile_catexp (exps : ST.exp list) : exp = match exps with
  | [] -> the_empty_exp
  | [e] -> compile_exp e
  | e::rest -> make_append (compile_exp e) (compile_catexp rest)

let rec compile_rest (e : exp) (rest : ST.block) : exp = match rest with
  | [] -> e
  | _ -> make_seq e (compile_block rest)

and compile_stmt (s : ST.stmt) (rest : ST.block) : exp = match s with
  | ST.SBreak -> IEBreak
  | ST.SBlock (x : ST.block) -> compile_rest (compile_block x) rest
  | ST.SExp (x : ST.exp) -> compile_rest (compile_exp x) rest
  | ST.SLocVars ((decls : (ID.id * ST.exp option) list), (ty : SD.ty), (flag : bool)) ->
      let ty2 = compile_type ty in
      if not flag then
        fold_right
          (fun (id, expo) body ->
             IELet(make_ipvar (ID.name id) ty2, UT.doopt compile_exp expo, body))
          decls
          (compile_block rest)
      else
        (match decls with
          | [id, Some(e)]  ->
              if ID.is_bogus id then
                compile_match (compile_exp e) (ST.get_real_type e) [ty]
                  [compile_block rest]
              else
                let x = ID.name id in
                IELet(make_ipvar x ty2,
                      Some(compile_exp e), 
                      compile_match (make_local x) (ST.get_real_type e) [ty]
                        [compile_block rest])
          | _ -> assert false)
  | ST.SReturn (x : ST.exp option) ->
      (match x with
        | None -> IEReturn None
        | Some(e) -> IEReturn (Some (compile_exp e)))
  | ST.SIf ((test : ST.exp), (b1 : ST.block), (b2 : ST.block)) ->
      compile_rest (IEIf (compile_exp test, compile_block b1, compile_block b2)) rest
  | ST.SWhile ((test : ST.exp), (body : ST.block)) ->
      compile_rest (IEWhile (compile_exp test, compile_block body)) rest
  | ST.SForeachRE ((pat : SD.seqpa), (e : ST.exp), (body : ST.block)) ->
      (*          foreach (p in e) { s1; s2; ... }   *)
      (*            =>                               *)
      (*          [[any]] x = e;                     *)
      (*          while (x != null) {                *)
      (*            match (x) {                      *)
      (*              case [[p, any y]]:             *)
      (*                x = y;                       *)
      (*                s1;                          *)
      (*                s2; ...                      *)
      (*            }                                *)
      (*          }                                  *)
      let x = gensym z_gen in
      let x_id = ID.create x in
      let y = gensym y_gen in
      let y_id = ID.create y in
      let test = IEBinOp(SO.BiNeq, make_local x, the_empty_exp) in
      let sty = SD.cat [pat; SD.seqbind y_id SD.sty_any] in
      let pat = SD.ty (SD.Tseq sty) in
      let asgn_type = SD.ty_any in
      let lhs_desc = ST.EVar x_id in
      let rhs_desc = ST.EVar y_id in
      let lhs = ST.make lhs_desc asgn_type in
      let rhs = ST.make rhs_desc asgn_type in
      let asgn_desc = ST.EAssign (lhs,rhs) in
      let asgn = ST.make asgn_desc asgn_type in
      let cl = (ST.SExp asgn) :: body in
      compile_rest
        (IELet(make_ipvar x ty_seq,
               Some(compile_exp e), 
               IEWhile (test, compile_block [ST.SMatch (lhs, [pat,cl])])))
        rest
  | ST.SMatch ((test : ST.exp), (cls : ST.clause list)) ->
      let (pats, blocks) = split cls in
      compile_rest (compile_match (compile_exp test) (ST.get_real_type test) pats
                      (map compile_block blocks)) rest
  | ST.SEmpty ->
      compile_block rest

and compile_block (b : ST.block) : exp = match b with
  | [] -> IEEmpty
  | s::rest -> compile_stmt s rest


(******* Compiling methods, classes, namespaces, and programs *)

let compile_invoke (inv : ST.invoke) : invoke =
  match inv with
    | ST.InvkNone -> IINVNone
    | ST.InvkThis(exps) -> IINVThis (map compile_actparam exps)
    | ST.InvkSuper(exps) -> IINVSuper (map compile_actparam exps)

let compile_formparam ((x : ID.id), (t : SD.ty)) : form_param =
  false, UT.intern (ID.name x), compile_type t

let compile_constr ((inv : ST.invoke), (xs : ST.form_param list), (b : ST.block)) : clmember =
  reset_generators ();
  new_lab_var_frame ();
  let res = ICMConstr(compile_invoke inv, map compile_formparam xs, compile_block b) in
  pop_lab_var_frame ();
  res

let compile_field ((id : ID.id), (status, (ty : SD.ty), init)) : clmember =
  ICMFld(ID.name id, status,  compile_type ty,
         match init with
         | None -> None
         | Some e -> Some (compile_exp e))

let compile_meth (cname : typsym)
                 ((id : ID.id),
                  (mods, stat, (xs : ST.form_param list), (resty : SD.ty option),
                   (b : ST.block)),
                  (mangled_name : string))
  : clmember =
  reset_generators ();
  new_lab_var_frame ();
  let old_configs = !Pref.debug_configs in
(*    let _tmp_ = ID.name id in *)
(*    fpr outf "%s@." _tmp_; *)
(*    if _tmp_ = "empty" then *)
(*      (); (* ,,, *) *)
  if mem (ID.name id) !Pref.debug_method_configs then Pref.debug_configs := true;
  let blk = compile_block b in
  if mem (ID.name id) !Pref.debug_method_configs then (
    Pref.debug_configs := old_configs;
  );
  let res =
    ICMMeth(mangled_name, stat, mods, map compile_formparam xs, compile_rettype resty,
            blk)
  in pop_lab_var_frame ();
  res

let compile_meth_decl (iname : typsym)
                 ((id : ID.id),
                  ((xs : ST.form_param list), (resty : SD.ty option)),
                  (mangled_name : string))
  : intf_meth_decl =
    (mangled_name, map compile_formparam xs, compile_rettype resty)

let compile_class (id : ID.id) (c : ST.clas) startclass : clas =
  let init = ICMStaticConstr(
    let enc = IEConst(ICString !Pref.pcdataencoding) in
    let setenc = IEStaticCall("Utils", "SetPcdataEncoding", [false, enc]) in
    let app = IEConst(ICString !Pref.appendpolicy) in
    let setapp = IEStaticCall("Utils", "SetAppendPolicy", [false, app]) in
    make_seq setenc setapp) in
  let cname = ID.name id in {
    iCLHomeNS = ID.location_longname id;
    iCLName = ID.name id;
    iCLExt = ID.longname c.ST.clas_parent;
    iCLImpls = List.map ID.name c.ST.clas_implements;
    iCLMems = (compile_constr c.ST.clas_constr) ::
              (if startclass == id then [init] else []) @
              (map compile_field c.ST.clas_fields) @
              (map (compile_meth cname) c.ST.clas_meths)
}
  
let compile_intf (id : ID.id) (i : ST.intf) : intf =
  let iname = ID.name id in {
    iINHomeNS = ID.location_longname id;
    iINName = ID.name id;
    iINImpls = List.map ID.name i.ST.intf_implements;
    iINMems = map (compile_meth_decl iname) i.ST.intf_meths_decls;
}

let reset_compiler () =
  INT.reset ();
  reset_tags ();
  reset_generators ();
  reset_loop_breakers ();
  reset_lab_vars ();
  reset_lblocks ();
  reset_matching_procs ();
  reset_regexs ()

let compile_prog (prog : ST.prog) startclass : prog =
  reset_compiler ();
  let uses_decls = [IUSNSpace "System";
		    IUSNSpace "Xtatic";
		    IUSNSpace "System.Text.RegularExpressions";
		    IUSNSpace "System.Collections";
		   ] in
  let user_classes = 
    ID.fold
      (fun (cclasses : clas list) ((id : ID.id), (c : ST.clas)) ->
         (compile_class id c startclass) :: cclasses)
      [] prog.ST.classes in
  let user_intfs = 
    ID.fold
      (fun (cintfs : intf list) ((id : ID.id), (i : ST.intf)) ->
         (compile_intf id i) :: cintfs)
      [] prog.ST.intfs in
  let regexs_class = 
    {iCLHomeNS = "";
     iCLName = "Regexs";
     iCLExt = "object";
     iCLImpls = [];
     iCLMems = regex_decls ()} in
  let tags_class =
    {iCLHomeNS = "";
     iCLName = "Tags";
     iCLExt = "object";
     iCLImpls = [];
     iCLMems = tag_decls ()} in
  let typestamps_class =
    {iCLHomeNS = "";
     iCLName = "Typestamps";
     iCLExt = "object";
     iCLImpls = [];
     iCLMems = typestamp_decls ()} in
  let matching_class =  
    { iCLHomeNS = "";
      iCLName = "Matching";
      iCLExt = "object";
      iCLImpls = [];
      iCLMems = matching_proc_decls ()} 
  in 
    (uses_decls, user_intfs,
     regexs_class :: tags_class :: typestamps_class :: matching_class :: user_classes)

