module C = Contexts
module SO = SynOp
module SS = SynSource
module SD = SynDenot
module ST = SynTyped
module R = Result

exception TypeCheck of Finfo.t * string

let fpr = Format.fprintf
let outf = !Pref.output_formatter

(******************** Extending the context **************************)

(** Datastructures to store classing context information
  from the user's program. *)

type cls = {
  cls_extends: Ident.id;
  cls_implements: Ident.id list;
  cls_sealed: bool;
  cls_constr: constr;
  (* the bool says whether the field is static *)
  cls_fields: (Ident.id * (SO.status * SD.ty)) list;
  cls_meths_ret: (Ident.id * (SO.status * SO.method_modifier * SD.ty option)) list; 
  cls_meths: (string * meth) list;
  (* the following contain the names (mangled or not) to use *)
  cls_meths_names: (Ident.id * string) list;
}

and constr = SD.ty list
and meth = Ident.id * SO.status * SO.method_modifier * SD.ty option * (SD.ty list)
and intf_meth_decl = Ident.id * SD.ty option * (SD.ty list)

type intf = {
  intf_implements: Ident.id list;
  intf_meths: (string * intf_meth_decl) list;
  (* the following contain the names (mangled or not) to use *)
  intf_meths_names: (Ident.id * string) list;
}
  
let extract_param nctx (_, (ty: SS.ty)) = Src2denot.ty nctx ty

let extract_constr nctx (inf, (_, params, _, _)) = 
  List.map (extract_param nctx) params

let extract_field nctx (inf, (dn, mangopt, expopt, status, (ty: SS.ty))) = 
  (Ident.create (Finfo.drop dn),
   (status, Src2denot.ty nctx ty))

let extract_meth nctx (inf, (mname, mangopt, status, modifier, retty, params, _)) =
  let paramty = List.map (extract_param nctx) params in
  let name = Finfo.drop mname in
  let id = MethIdent.getid name paramty in
  let ret_ty = match retty with
     | (inf, None) -> None
     | (inf, Some ty) -> Some (Src2denot.ty nctx (inf, ty))
  in
  (* first meth_ret *)
  (id, (status, Finfo.drop modifier, ret_ty)),
  (* then meth *)
  (name, (
     id,
     status,
     Finfo.drop modifier,
     ret_ty,
     paramty))

let extract_imeth nctx (mname, (inf, (retty, params))) = 
  let paramty = List.map (extract_param nctx) params in
  let name = Finfo.drop mname in
  let id = MethIdent.getid name paramty in
  let ret_ty = match retty with
     | (inf, None) -> None
     | (inf, Some ty) -> Some (Src2denot.ty nctx (inf, ty))
  in
  (name, (
     id,
     ret_ty,
     paramty))

let extract_clas ((nctx: Naming.ctx), ((clas_inf, clas'): SS.clas))
  : cls = 
   let cid = Naming.resolve_class nctx [clas'.SS.clas_name] in 
   let mret,m = List.split (List.map (extract_meth nctx) clas'.SS.clas_meths) in
  { cls_extends = Semant.parent cid;
    cls_implements = Semant.implements cid;
    cls_sealed = clas'.SS.clas_sealed;
    cls_constr = extract_constr nctx clas'.SS.clas_constr;
    cls_fields = List.map (extract_field nctx) clas'.SS.clas_fields;
    cls_meths_ret = mret;
    cls_meths = m;
    cls_meths_names = [];
  }

let extract_intf ((nctx: Naming.ctx), ((intf_inf, intf'): SS.intf))
  : intf = 
  { intf_implements = List.map (Naming.resolve_interface nctx) (intf'.SS.intf_implements);
    intf_meths = List.map (extract_imeth nctx) intf'.SS.intf_meths;
    intf_meths_names = [];
  }

let get_cls_entry clstbl cid = 
  try Some (Ident.find_same clstbl cid)
  with Not_found -> None

let get_intf_entry intftbl iid =
  try Some (Ident.find_same intftbl iid)
  with Not_found -> None

let find_sealed (lib,usr) cid =
  match get_cls_entry usr cid with
  | None -> lib.C.cc_sealed cid
  | Some cls -> cls.cls_sealed
  
let rec field (lib, usr) cid fid = 
  match get_cls_entry usr cid with
  | None -> lib.C.cc_field cid fid  (* search in the library *)
  | Some cls -> 
      try List.assq fid cls.cls_fields
      with Not_found -> (* no fid in this class, search in the parent *)
	field (lib, usr) cls.cls_extends fid

let rec find_field (lib, usr) cid fname = 
  match get_cls_entry usr cid with
  | None -> lib.C.cc_find_field cid fname  (* search in the library *)
  | Some cls -> 
      try List.find (fun (id, _) -> fname = (Ident.name id)) cls.cls_fields
      with Not_found -> (* no fid in this class, search in the parent *)
	find_field (lib, usr) cls.cls_extends fname

let rec fields_all (lib, usr) cid = 
  match get_cls_entry usr cid with
  | None -> lib.C.cc_fields cid
  | Some cls -> 
      let this_fields = cls.cls_fields in 
      let inherited_fields = fields_all (lib, usr) cls.cls_extends in 
	inherited_fields @ this_fields

let constr_type (lib,usr) cid = 
   match get_cls_entry usr cid with
  | None -> lib.C.cc_constrtype cid
  | Some cls -> cls.cls_constr

let rec meth_ret_type (lib,usr) cid methid = 
   match get_cls_entry usr cid with
  | None -> lib.C.cc_methrettype cid methid
  | Some cls -> 
      try List.assoc methid cls.cls_meths_ret
      with Not_found -> meth_ret_type (lib,usr) cls.cls_extends methid
      
let meth_this_type (lib,usr) cid name =
  match get_cls_entry usr cid with
  | None -> lib.C.cc_meththistype cid name 
  | Some cls -> 
      List.fold_left (fun acc (n,m) -> if n = name then m :: acc else acc) []
      cls.cls_meths

let rec intf_cls_all (lib, usr, intftbl) cid =
  let x = Ident.name cid in
(*  fpr outf "interfaces of %s@." x; *)
(*  fpr outf "+++@."; *)
(*  Ident.iter (fun id _ -> fpr outf "   %s@." (Ident.name id)) usr; *)
(*  fpr outf "++++@."; *)
(*  Ident.iter (fun id _ -> fpr outf "   %s@." (Ident.name id)) intftbl; *)
(*  fpr outf "+++++@."; *)
  match get_cls_entry usr cid with
  | None -> lib.C.cc_implements cid
  | Some cls ->
      let this_intf = List.fold_left 
                        (fun acc intfid -> (intf_intf_all lib intftbl  intfid) @ acc)
                        []
                        cls.cls_implements
      in
      let inherited_intf = intf_cls_all (lib, usr, intftbl) cls.cls_extends in
      inherited_intf @ this_intf

and intf_intf_all lib intftbl iid =
  let x = Ident.name iid in
(*  fpr outf "%s@." x; *)
(*  fpr outf "---@."; *)
(*  Ident.iter (fun id _ -> fpr outf "   %s@." (Ident.name id)) intftbl; *)
(*  fpr outf "----@."; *)

  match get_intf_entry intftbl iid with
  | None -> lib.C.cc_implements iid
  | Some intf ->
      let inherited_intf = List.fold_left 
                        (fun acc intfid -> (intf_intf_all lib intftbl  intfid) @ acc)
                        []
                        intf.intf_implements
      in
      intf.intf_meths @ inherited_intf
      
let intf_direct (lib, usr, intftbal) cid =
  match get_cls_entry usr cid with
  | None -> (
    match get_intf_entry intftbal cid with
    | None -> lib.C.cc_intf_direct cid
    | Some intf -> intf.intf_implements)
  | Some cls -> cls.cls_implements
      
let mangle_meths_names (lib, usr) =
  Ident.fold (
    fun acc (_, cls) ->
      Contexts.extend_mangled_names acc cls.cls_meths_names)
  lib.C.cc_mangled_name
  usr 
      
let extend (clsLib: C.clsctx) (srcUsr: SynSource.prog) : C.clsctx = 
(*  fpr outf "---@."; *)
(*  Ident.iter (fun id _ -> fpr outf "   %s@." (Ident.name id)) srcUsr.SS.cls; *)
(*  fpr outf "----@."; *)

(*  fpr outf "Classes@."; *)
  let clsUsr = Ident.map 
		 (fun cid clasentry ->
(*  fpr outf "   %s@." (Ident.name cid); *)
                    extract_clas clasentry)
		 srcUsr.SS.cls
  in
(*  fpr outf "Interfaces@."; *)
  let intfUsr = Ident.map 
		 (fun iid intfentry ->
(*  fpr outf "   %s@." (Ident.name iid); *)
                    extract_intf intfentry)
		 srcUsr.SS.intfs
  in
  let should_mangle args =
    (* mangle if one arg is a seqtype that is not any *)
    List.exists (fun ty -> Semant.bsub ty SD.ty_any && not (Semant.bsub SD.ty_any ty))
      args
  in
  let mangle_name name types =
    Printf.sprintf "%s_%s" name (Digest.to_hex (Digest.string (SD.toString_list_ty types)))
  in
  let clsUsr = 
    Ident.map (
      fun cid cls ->
        { cls with cls_meths_names = (
          List.fold_left (
            fun acc (name, (mid, stat, modi, retty, args)) ->
              if List.mem_assoc mid acc then acc else
              let mn = 
                if should_mangle args then begin
                  let mn = mangle_name name args in
                  Debug.trace (Printf.sprintf "\n Mangling %s at types %s: %s" 
                    name 
                    (SD.toString_list_ty args)
                    mn);
                  mn
                end else
                  name
              in
              (mid, mn) :: acc)
            []
            cls.cls_meths
        )
        }
    )
    clsUsr
  in
  let intfUsr = 
    Ident.map (
      fun cid intf ->
        { intf with intf_meths_names = (
          List.fold_left (
            fun acc (name, (mid, retty, args)) ->
              if List.mem_assoc mid acc then acc else
              let mn = 
                if should_mangle args then begin
                  let mn = mangle_name name args in
                  Debug.trace (Printf.sprintf "\n Mangling %s at types %s: %s" 
                    name 
                    (SD.toString_list_ty args)
                    mn);
                  mn
                end else
                  name
              in
              (mid, mn) :: acc)
            []
            intf.intf_meths
        )
        }
    )
    intfUsr
  in
(*  fpr outf "---@."; *)
(*  Ident.iter (fun id _ -> fpr outf "   %s@." (Ident.name id)) clsUsr; *)
(*  fpr outf "----@."; *)
(*  Ident.iter (fun id _ -> fpr outf "   %s@." (Ident.name id)) intfUsr; *)
(*  fpr outf "-----@."; *)
  { C.cc_sealed = find_sealed (clsLib, clsUsr);
    C.cc_fields = fields_all (clsLib, clsUsr);
    C.cc_field = field (clsLib, clsUsr);
    C.cc_find_field = find_field (clsLib, clsUsr);
    C.cc_constrtype = constr_type (clsLib, clsUsr);
    C.cc_methrettype = meth_ret_type (clsLib, clsUsr); 
    C.cc_meththistype = meth_this_type (clsLib,clsUsr);
    C.cc_mangled_name = mangle_meths_names (clsLib, clsUsr);
    C.cc_implements = intf_cls_all (clsLib, clsUsr, intfUsr); 
    C.cc_intf_direct = intf_direct (clsLib, clsUsr, intfUsr);
   }



(******************** Typechecking **********************************)

(******
let check_meth_main (ctx,loc) meth_name meth_type = 
  let main_type = (SY.TItem(J.Tvoid), 
		   [SY.TItem(J.TArray(J.TClas StdIds.string_id))]) in
  if not ("main" = meth_name) then ()
  else if (meth_type = main_type) then ()
  else raise (WrongTypeOfMainMethInClass (I.name loc.classID))
*******)


(*TODO: this function is very inefficient due to using find_field ... *)
let check_field (nctx, cctx, env) (fld : SS.field) = 
  let (decname, mangopr, expopt, stat, ty)  =  Finfo.drop fld in
  let (inf_name, fname) = decname in 
  try
    let _ = cctx.C.cc_find_field (TyEnv.get_super env) fname
    in raise (TypeCheck(inf_name, "Field name already used in ancestor " ^
			"(NYI: field hiding)."))
  with Not_found -> 
    let fid, (stat, ty) =
      try cctx.C.cc_find_field (TyEnv.get_this env) fname
      with Not_found -> assert false 
	(* fname must be within the fields of the current class, 
	   since it was put there by extract_field *)
    in
    match expopt with
    | None -> fid, (stat, ty, None)
    | Some exp ->
        let ({ST.exp_desc=_;ST.exp_type=t} as e) =
          Tycheck.typ_exp nctx cctx env exp in
        let rett = match t with
            | Some t -> t
            | None -> raise 
              (Tycheck.TyCheck(Finfo.get exp, "Cannot initialize to a value whose type is VOID"))
        in
        if not (Result.boolOf(Semant.sub rett ty)) then
          raise (Tycheck.TyCheck(Finfo.get exp, "initialization not subtype of field type"));
        fid, (stat, ty, Some e)
  
let check_params (nctx, cctx, env) params = 
  List.map 
    (fun (dname_opt, s_ty) -> 
       let dname = 
	 Bug.from_opt "CtxCls.check_params" "missing param name" dname_opt in
	 (Ident.create (Finfo.drop dname), Src2denot.ty nctx s_ty))
    params

let check_invoke (nctx, cctx, env) (inf_invk, invk') = 
  match invk' with 
  | SS.InvkNone -> ST.InvkNone
  | SS.InvkThis _ -> 
      raise (TypeCheck (inf_invk, "Call to another constructor " ^ 
			"(NYI: multiple constructors)." ))
  | SS.InvkSuper s_args -> 
      let paramtypes = cctx.C.cc_constrtype (TyEnv.get_super env) in 
      let typedargs = List.map 
	  (fun s_arg -> 
	    let ({ST.exp_desc=_;ST.exp_type=t} as e) =
              Tycheck.typ_exp nctx cctx env s_arg in 
            let rett = match t with
            | Some t -> t
            | None -> raise (Tycheck.TyCheck(Finfo.get s_arg, "Cannot cast something whose type is VOID")) in
            (e, rett))
	  s_args  in
      let (t_args, typ_args) = List.split typedargs in 
      (try 
	let num = ref 0 in
	List.iter2 
	  (fun arg_ty par_ty -> 
	    (if not (Result.boolOf(Semant.sub  arg_ty  par_ty)) then 
	      let inf = Finfo.get(List.nth s_args !num) in
	      raise (TypeCheck(inf, 
			   "Argument is not subtype of formal parameter")));
	    incr num
	  )
	  typ_args  paramtypes
      with Invalid_argument _ -> 
	raise (TypeCheck(inf_invk, 
	 "Wrong number of actual arguments in constructor invokation")));
      ST.InvkSuper t_args
	

let check_constr (nctx, cctx, env) (constr : SS.constr) = 
  let (inf_constr, constr') = constr in
  let (clsname, s_params, s_invk, s_body_opt) = constr' in
  let (inf_clsname, str_clsname) = clsname in
  if str_clsname <> (Ident.name (TyEnv.get_this env)) 
  then raise (TypeCheck(inf_clsname, 
			"Type of constructor is not the current class"));
  let t_params = check_params (nctx, cctx, env) s_params in 
  let env1 = List.fold_left TyEnv.addVar env t_params in 
  let t_invk = check_invoke (nctx, cctx, env1) s_invk in 
  let s_body = Bug.from_opt "CtxCls.check_constr" "missing body" s_body_opt in
  let (t_body, rty_body) = Tycheck.typ_block nctx cctx env1 s_body in 
  (* Check there was no return in the body: *)
  match rty_body with 
  | Tycheck.NoRet -> (t_invk, t_params, t_body)
  | _     -> raise (TypeCheck(Finfo.get s_body, 
			       "Return from constructor body."))


let check_param_bounds infs tys bounds msg = 
  match infs, tys, bounds with 
  |  [], [], [] -> ()
  | f::fs, t::ts, b::bs -> 
      if not (Result.boolOf(Semant.sub t b))
      then raise (TypeCheck(f, msg))
  | _ -> assert false (* list legths must be the same *)

let check_equal_param_bounds clas_inf tys bounds msg =
  match tys, bounds with 
  | [], [] -> ()
  | t::ts, b::bs -> 
      if not (Result.boolOf(Semant.sub t b) &&
              Result.boolOf(Semant.sub b t))
      then raise (TypeCheck(clas_inf, msg))
  | _ -> assert false (* list legths must be the same *)

let check_meth_common nctx cctx env mname s_retty s_params =
  let ty_retty = 
    match  s_retty with
    | (_, None) -> None
    | (inf, Some t) -> Some (Src2denot.ty nctx (inf,t)) in 
  let ty_params  = 
    List.map (fun(_, ty) -> Src2denot.ty nctx ty) s_params in 
  Debug.trace ("Typechecking method "^mname^": "^(SD.toString_list_ty ty_params)^" -> "^(SD.toString_ty_option ty_retty));
  let meth_id = MethIdent.getid mname ty_params in
  let mangled_name = Ident.find_same cctx.C.cc_mangled_name meth_id in

  (* Check method overriding *)
  ty_params, ty_retty, meth_id, mangled_name
  
let check_meth (nctx, cctx, env) (meth : SynSource.meth) =
  let m_decname, mangopt, is_static,(info_mod, mods), s_retty, s_params, s_body_opt = Finfo.drop meth in 
  let mname = Finfo.drop m_decname in
  let (inf_retty, _) = s_retty in 
  let id_params  = 
    List.map (fun(dnopt, _) -> 
		let dn = 
		  Bug.from_opt "CtxCls.check_meth" "missing paramer id" dnopt
		in Ident.create (Finfo.drop dn))
      s_params in
  let ty_params, ty_retty, meth_id, mangled_name =
    check_meth_common nctx cctx env mname s_retty s_params in
  (try
    (* should we do something special for static methods ? *)
    (let _, modifier, anc_retty = cctx.C.cc_methrettype (TyEnv.get_super env) meth_id in 
  (*TODO check 'Main' constraints *)
     (match modifier, mods with
     | (SO.Normal | SO.New), SO.Normal -> Tycheck.warning info_mod "hiding method without specifying New"
     | (SO.Virtual | SO.Override), SO.Virtual -> raise (TypeCheck(info_mod, 
     "method cannot be virtual since it overrides another method"))
     | (SO.Virtual | SO.Override), (SO.Normal | SO.New) -> raise
     (Tycheck.TyCheck(info_mod,
     "NOT IMPLEMENTED: cannot hide a virtual method, use override"))
     | (SO.Normal | SO.New), (SO.Virtual | SO.Override) -> raise
     (Tycheck.TyCheck(info_mod,
     "Method already implemented, use new to hide it"))
     | _ -> ());
     match anc_retty, ty_retty with
     | None, None -> ()
     | Some t1, Some t2 -> 
         if not (Result.boolOf(Semant.sub t1 t2) && 
                 Result.boolOf(Semant.sub t2 t1))
         then raise (TypeCheck(inf_retty, "Method return type is not equal to " ^ 
	  "that of the method being overriden."))
     | _ -> raise (TypeCheck(inf_retty, "Method return type is not equal to " ^ 
			  "that of the method being overriden."))
                          (* no need to check for argument types, they determine the method*)
    )
  with Not_found -> (
    match mods with
    | SO.New -> Tycheck.warning info_mod 
    "Method specifies New but hides no previous method"
    | SO.Override -> raise (Tycheck.TyCheck(info_mod, "no virtual method to override"))
    | _ -> ()
  )
  );
  (* Check method body *)
    let s_body = Bug.from_opt "CtxCls.check_meth" "missing method body" s_body_opt in
  let env1 = List.fold_left2 
      (fun e id ty -> TyEnv.addVar e (id, ty))
      env id_params ty_params in
  let (t_body, rty_body) = Tycheck.typ_block nctx cctx env1 s_body in 
  Debug.trace "Method body checked, checking return type";
  (match rty_body, ty_retty with
    | (Tycheck.NoRet | Tycheck.SureRet None | Tycheck.MaybeRet None), None -> ()
    | (Tycheck.NoRet | Tycheck.SureRet None | Tycheck.MaybeRet None), _ -> 
        raise (TypeCheck(Finfo.get s_body, "Return from method expected."))
    | Tycheck.MaybeRet _, _ -> raise (TypeCheck(Finfo.get s_body, "Return from method not certain."))
    | _, None -> raise (TypeCheck(Finfo.get s_body, "Return with a value not allowed here."))
    | Tycheck.SureRet (Some rty), Some t -> 
        (*--
        Debug.trace ("Body type: " ^ (SD.toString_ty rty) ^ "\n" ^
		     "Decl return: " ^ (SD.toString_ty t) ^ "\n" );*)
        (*--
	if not (Result.boolOf(Semant.sub rty t)) then
	  raise (TypeCheck(Finfo.get s_body, 
			   "Actual return not subtype of declared. \n" ^
			   "Body type: " ^ (SD.toString_ty rty) ^ "\n" ^
			   "Decl return: " ^ (SD.toString_ty t) ^ "\n" )) *)
        match Semant.sub rty t with 
        | R.Succ -> ()
        | R.Fail f -> 
  	    raise (TypeCheck(Finfo.get s_body, 
			   "Actual return not subtype of declared. \n" ^
			   "Body type: " ^ (SD.toString_ty rty) ^ "\n" ^
			   "Decl return: " ^ (SD.toString_ty t) ^ "\n" ^
                           "At path: " ^ (Typath.toString f) ^ "\n" ))
  );
  (meth_id, 
   (mods, is_static, List.map2 (fun x y -> x,y) id_params ty_params, ty_retty, t_body),
   mangled_name)

let check_meth_decl (nctx, cctx, env) (dname_meth, (meth_decl : SynSource.imeth)) =
  let (_, mname) = dname_meth in
  let (_, (s_retty, s_params)) = meth_decl in 
  let id_params  = 
    List.map (fun(dnopt, _) -> 
		let dn = Bug.from_opt "CtxCls.check_meth_decl" 
			   "missing parameter name" dnopt in
		  Ident.create (Finfo.drop dn)) s_params in
  let ty_params, ty_retty, meth_id, mangled_name =
    check_meth_common nctx cctx env mname s_retty s_params in
  (meth_id, (List.map2 (fun x y -> x,y) id_params ty_params, ty_retty), mangled_name)
  
let check_for_intf_implementation (nctx, cctx, inf_clas, cid) (iname, sig_intf) =
  try
    let (iid, s_retty, s_params) = sig_intf in 
    let meth_stat, modifier, ret_ty = cctx.C.cc_methrettype cid iid in
    begin
      match ret_ty, s_retty with
      | None, None -> ()
      | Some t1, Some t2 -> 
          if not (Result.boolOf(Semant.sub t1 t2) && 
                  Result.boolOf(Semant.sub t2 t1))
       then raise (TypeCheck(inf_clas, "Method return type of "^iname^" is not equal to " ^ 
			  "that of the interface being implemented."))
      | _ -> raise (TypeCheck(inf_clas, "Method return type of "^iname^" is not equal to " ^ 
			  "that of the interface being implemented."))
    end
  with Not_found ->
    raise (TypeCheck(inf_clas,"method "^iname^
    " is not implemented or with incorrect argument types"))

let check_clas (cctx) (nctx, (inf_clas, clas')) : ST.clas= 
  let cid = Naming.resolve_class nctx [clas'.SS.clas_name] in
  Debug.trace ("Typechecking "^(Finfo.drop clas'.SS.clas_name));
  let pid = Semant.parent cid in
  if (cctx.C.cc_sealed pid) 
  then raise (TypeCheck(inf_clas, "class "^(Finfo.drop clas'.SS.clas_name)^" extends a sealed class"));
  let env = TyEnv.empty_instance cid pid in 
  let env = List.fold_left  TyEnv.addFld  env  (cctx.C.cc_fields cid)
  in
  let clas_fields = 
    List.map  (check_field (nctx, cctx, env)) clas'.SS.clas_fields  in
  let clas_constr =  check_constr (nctx, cctx, env) clas'.SS.clas_constr  in
  let clas_meths = List.map (check_meth (nctx, cctx, env)) clas'.SS.clas_meths in
  let impl = cctx.C.cc_implements cid in
  List.iter (check_for_intf_implementation (nctx, cctx, inf_clas, cid)) impl;
  {ST.clas_id = cid;
   ST.clas_parent = pid;
   ST.clas_implements = cctx.C.cc_intf_direct cid;
   ST.clas_fields = clas_fields; 
   ST.clas_constr = clas_constr;
   ST.clas_meths = clas_meths;
 } 

let check_intf (cctx) (nctx, (inf_intf, intf')) : ST.intf =
  let iid = Naming.resolve_interface nctx [intf'.SS.intf_name] in
  Debug.trace ("Typechecking "^(Finfo.drop intf'.SS.intf_name));
  let pid = Semant.parent iid in
  let env = TyEnv.empty_instance iid pid in 
  let intf_meths_decls = List.map (check_meth_decl (nctx, cctx, env)) intf'.SS.intf_meths in
  {ST.intf_id = iid;
   ST.intf_implements = cctx.C.cc_intf_direct iid;
   ST.intf_meths_decls = intf_meths_decls;
 } 
    

let typecheck (clsFull: C.clsctx) (srcUsr: SS.prog) : ST.prog = 
  { ST.classes = 
    Ident.fold (
      fun tbl (cid, sclas_entry) -> 
        Ident.add tbl (cid, check_clas clsFull sclas_entry) )
    Ident.empty srcUsr.SS.cls;
    ST.intfs =
    Ident.fold (
      fun tbl (iid, sintf_entry) -> 
        Ident.add tbl (iid, check_intf clsFull sintf_entry) )
    Ident.empty srcUsr.SS.intfs;
  }

