module SS = SynSource
module M = Semant

exception InheritanceLoop of string
exception MultipleInheritance of Finfo.t * string
exception InterfaceExtendsClass of Finfo.t * string
      
let check_for_inheritance_loops datactx prog=
  let tbl = Hashtbl.create 101 in
  let msg_to_string endid seen =
    let rec loop str = function
    | id :: rest -> let str2 = " -> " ^ (Ident.name id) in
      if id = endid then str ^ str2 else loop (str ^ str2) rest
    | [] -> assert false
    in
    "Circular inheritance: " ^ (Ident.name endid) ^ (loop "" seen)
  in
  let rec iter_intf seen intfid =
    if Hashtbl.mem tbl intfid then ()
    else if List.mem intfid seen then
      raise (InheritanceLoop (msg_to_string intfid seen))
    (* the top of the interface tree is the object class *)
    else if intfid == datactx.M.dc_topclass then () else
    List.iter (iter_intf (intfid::seen)) (Ident.find_same datactx.M.dc_implements intfid);
    Hashtbl.add tbl intfid intfid
  in
  let rec iter_cls seen clsid =
    if Hashtbl.mem tbl clsid then ()
    else if List.mem clsid seen then raise (InheritanceLoop (msg_to_string clsid seen))
    else if clsid == datactx.M.dc_topclass then () else
      begin
        List.iter (iter_intf (clsid::seen)) (Ident.find_same datactx.M.dc_implements clsid);
        iter_cls (clsid::seen) (Ident.find_same datactx.M.dc_parent clsid);
        Hashtbl.add tbl clsid clsid
      end
  in
  Ident.iter (fun intfid _ -> iter_intf [] intfid) prog.SS.intfs;
  Ident.iter (fun clsid _ -> iter_cls [] clsid) prog.SS.cls


let extend (origctx: M.datctx) (prog: SS.prog) : M.datctx = 
  (* for each function, first check the origctx, then the user program *)

  (* Parent table  *)
  let parenttbl = origctx.M.dc_parent in  (* parents from original context *)
  let parenttbl =                         (* add parent classes *)
    Ident.fold        
      (fun  ptbl (cid, (nctx, clas)) -> 
	 let extended_classes = (* filter out interfaces *)
	   List.fold_left
	     (fun classes n -> match Naming.resolve_leaf nctx n with 
		| (eid, Cnaming.LClas) -> eid::classes
		| (_, _) -> classes)
	     [] (Finfo.drop clas).SS.clas_extends
	 in match extended_classes with 
	   | [eid] -> Ident.add ptbl (cid, eid)
	   | [] -> Ident.add ptbl (cid, origctx.M.dc_topclass)
	   | _ ->  raise (MultipleInheritance(Finfo.get clas, ("Class "^(Finfo.drop (Finfo.drop clas).SS.clas_name)^" extends several classes"))) )
      parenttbl prog.SS.cls  in
  let parenttbl = Ident.fold  (* add Object as the parent of every interface *)
	       (fun ptbl (iid, _) -> 
		  Ident.add ptbl (iid, origctx.M.dc_topclass)) 
               parenttbl prog.SS.intfs  in

  (* Implements table *)
  let implemtbl = origctx.M.dc_implements in
  let implemtbl =     (* add when a class implements interfaces  *)
    Ident.fold
      (fun itbl (cid, (nctx, clas)) -> 
	 let implemented_intfs =  (* filter out classes *)
	   List.fold_left (fun intfs n -> 
			     match Naming.resolve_leaf nctx n with 
			       | (eid, Cnaming.LIntf) -> eid::intfs
			       | _ -> intfs)
	     [] (Finfo.drop clas).SS.clas_extends   
	 in 
	   Ident.add itbl (cid, implemented_intfs)
      )
      implemtbl prog.SS.cls in
  let implemtbl =     (*  add when one interface implements another *)
    Ident.fold 
      (fun itbl (iid, (nctx, intf)) -> 
	 let inherited_intfs =  (* check there are no classes *)
	   List.fold_left (fun intfs n -> 
			     match Naming.resolve_leaf nctx n with
			       | (eid, Cnaming.LIntf) -> eid::intfs
			       | (_, Cnaming.LClas) ->  raise 
             (InterfaceExtendsClass(Finfo.get intf, ("Interface "^(Finfo.drop (Finfo.drop intf).SS.intf_name)^" extends a class")))
		     )
	     [] (Finfo.drop intf).SS.intf_implements 
	 in
	   Ident.add itbl (iid, inherited_intfs)
      )
      implemtbl prog.SS.intfs 
  in
    
  let res =
    { M.dc_topclass = origctx.M.dc_topclass;
      M.dc_any_id = origctx.M.dc_any_id;
      M.dc_parent = parenttbl;
      M.dc_implements = implemtbl;
      M.dc_seqdef = origctx.M.dc_seqdef;
      M.dc_seqdef_untranslated = prog.SS.seq;
    } 
  in
  (* we now check if there are no inheritance loop*)
    check_for_inheritance_loops res prog;
    res

   
