#define _GNU_SOURCE
#include <assert.h>
#include <pnet/il_dumpasm.h>
#include <pnet/il_program.h>
#include <pnet/il_system.h>


// gcc -std=c99 -o dasm dasm.c -lILImage -lILEngine -lILSupport -lpthread

char *type_prim_str (int t) {
  switch (t) {
  case IL_META_ELEMTYPE_VOID: return "void";
  case IL_META_ELEMTYPE_BOOLEAN: return "bool";
  case IL_META_ELEMTYPE_CHAR: return "char";
  case IL_META_ELEMTYPE_I1: return "sbyte";
  case IL_META_ELEMTYPE_U1: return "byte";
  case IL_META_ELEMTYPE_I2: return "short";
  case IL_META_ELEMTYPE_U2: return "ushort";
  case IL_META_ELEMTYPE_I4: return "int";
  case IL_META_ELEMTYPE_I: return "int";
  case IL_META_ELEMTYPE_U4: return "uint";
  case IL_META_ELEMTYPE_U: return "uint";
  case IL_META_ELEMTYPE_I8: return "long";
  case IL_META_ELEMTYPE_U8: return "ulong";
  case IL_META_ELEMTYPE_R4: return "float";
  case IL_META_ELEMTYPE_R8: return "double";
  case IL_META_ELEMTYPE_R: return "double";
  case IL_META_ELEMTYPE_TYPEDBYREF: return "typedbyref";
  default: exit(1);
  }
}


char *type_class_str (ILClass *c) {
  const char *n1 = ILClassGetNamespace(c);
  const char *n2 = ILClassGetName(c);

  // pnet bug? null for System.Runtime.InteropServices.ELEMDESC.DESCUNION
  if (!n1) return "dasm.c.bug";

  char *s;
  asprintf(&s, "%s.%s", n1, n2);
  return s;
}


char *type_str (ILType *t) {
  if (ILType_IsPrimitive(t)) return type_prim_str(ILType_ToElement(t));
  else if (ILType_IsClass(t)) return type_class_str(ILType_ToClass(t));
  else if (ILType_IsValueType(t)) return type_class_str(ILType_ToClass(t));
  else if (ILType_IsComplex(t)) return "complex";
  else assert(0);    
}


void do_field (ILField *f) {
  if (!ILField_IsPublic(f)) return;
  const char *n = ILField_Name(f);
  ILType *t = ILField_Type(f);
  char *ts = type_str(t);
  printf("\n  (\"%s\", \"%s\");", ts, n);
}


void do_param (ILType* t) {
  int pn = ILTypeNumWithParams(t);  
  for (int i=1; i<=pn; i++) {
    ILType *pt = ILTypeGetParam(t, i);
    char *pts = type_str(pt);
    printf("\"%s\"; ", pts);
  }
}

void do_ctor (ILMethod *m) {
  if (!ILMethod_IsPublic(m)) return;
  ILType *t = ILMethod_Signature(m);
  printf("\n  [");
  do_param(t);
  printf("];");
}


void do_method (ILMethod *m) {
  if (!ILMethod_IsPublic(m)) return;
  ILType *t = ILMethod_Signature(m);
  const char *n = ILMethod_Name(m);
  ILType *rt = ILTypeGetReturn(t);
  char *rts = type_str(rt);
  printf("\n  (\"%s\", \"%s\", [", n, rts);
  do_param(t);
  printf("]);");
}


void do_iment (ILClass* c) {
  ILImplements* i = 0;
  while (i = ILClassNextImplements(c, i)) {
    ILClass *f = ILImplementsGetInterface(i);
    const char *n1 = ILClassGetNamespace(f);
    const char *n2 = ILClassGetName(f);
    printf("\n\"%s.%s\";", n1, n2);
  }
  printf("], [");
}


void do_class (ILClass* c) {
  const char* n1 = ILClassGetNamespace(c);
  const char* n2 = ILClassGetName(c);
  ILClass *p = ILClassGetParentRef(c);
  if (!p) p = c; // System.Object
  const char* pn1 = ILClassGetNamespace(p);
  const char* pn2 = ILClassGetName(p);
  printf("\nClass (\"%s.%s\", \"%s.%s\", [", n1, n2, pn1, pn2);
  do_iment(c);

  ILMember* m = 0; 
  int k = IL_META_MEMBERKIND_FIELD;
  while (m = ILClassNextMemberByKind(c, m, k))
    do_field((ILField *)m);
  printf("], [");

  m = 0; k = IL_META_MEMBERKIND_METHOD;
  while (m = ILClassNextMemberByKind(c, m, k))
    if (ILMethod_IsConstructor((ILMethod *)m)) 
      do_ctor((ILMethod *)m);
  printf("], [");

  m = 0; k = IL_META_MEMBERKIND_METHOD;
  while (m = ILClassNextMemberByKind(c, m, k))
    if (!ILMethod_IsConstructor((ILMethod *)m)) 
      do_method((ILMethod *)m);
  printf("]);\n");
}


void do_iface (ILClass* c) {
  const char* n1 = ILClassGetNamespace(c);
  const char* n2 = ILClassGetName(c);
  printf("\nIface (\"%s.%s\", [", n1, n2);
  do_iment(c);

  ILMember* m = 0;
  int k = IL_META_MEMBERKIND_METHOD;
  while (m = ILClassNextMemberByKind(c, m, k))
    do_method((ILMethod *)m);
  printf("]);\n");
}


int main () {
  printf("open Asm\nlet x = [\n");
  ILContext *context = ILContextCreate(); 
  assert(context);

  int mask = IL_LOADFLAG_FORCE_32BIT 
    | IL_LOADFLAG_PRE_VALIDATE
    | IL_LOADFLAG_NO_RESOLVE;
  ILImage *image = 0;
  int error = ILImageLoad(stdin, "stdin", context, &image, mask);
  assert(!error);  

  int kind = IL_META_TOKEN_TYPE_DEF;
  int n = ILImageNumTokens(image, kind);
  for (int i=1; i<=n; ++i) {
    ILClass* c = ILImageTokenInfo(image, kind | i);
    if (!ILClass_IsPublic(c)) continue;
    if (ILClass_IsRegularClass(c)) do_class(c);
    else if (ILClass_IsInterface(c)) do_iface(c);
    else assert(0);
  }
  printf("]\n");
}
