module SD = SynDenot
module ID = Ident
module SO = SynOp

(************************************************)
(******* Intermediate language data types *******)
(************************************************)

type tagsym = string
type locvarsym = string
type membvarsym = string
type methsym = string
type typsym = string
type nssym = string
type labsym = string
type regexsym = string
type typestampsym = string
type regexfield = string

type constval =
  | ICNull
  | ICBool of bool
  | ICInt of int
  | ICChar of char
  | ICString of string
  | ICSymbol of string

type basetyp =
  | IBTEmpty
  | IBTElement
  | IBTObject
  | IBTBool
  | IBTInt
  | IBTChar

type baseval =
  | IBVBool of bool
  | IBVInt of int
  | IBVChar of char
  | IBVSymbol of tagsym
  | IBVObj of typsym

type predefty =
  | IPTyVoid
  | IPTyBool
  | IPTyInt
  | IPTyChar
  | IPTyString
  | IPTyArray of typ

and typ =
  | ITYPredef of predefty
  | ITYNamed of typsym

type patcont =
    { iTag : atompat; iAtt : atompat; iCont : atompat; iRest : atompat }

and atompat =
  | IPVar of locvarsym * typ
  | IPDontCare

and pat =
  | IPEmpty
  | IPDefault
  | IPAnyAtom of atompat
  | IPElement of patcont
  | IPPcdata of patcont
  | IPObj of patcont
  | IPBaseType of basetyp * patcont
  | IPAtt of (tagsym * atompat) list * atompat option
  | IPBaseVal of baseval
  | IPRegex of regexsym * bool

let ty_seq = ITYNamed "Seq"
let ty_seq_source = SD.ty (SD.Thost (SD.hostty (SD.TClasRef (ID.create "Seq"))))

let ty_object = ITYNamed "object"
let ty_tag = ITYNamed "Tag"
let ty_att = ITYNamed "Attributes"
let ty_regex = ITYNamed "Regex"
let ty_string = ITYNamed "string"
let ty_string_pcdata = ITYNamed "__string_pcdata"
let ty_seqstring = ITYNamed "SeqString"
let ty_seqsubstring = ITYNamed "SeqSubstring"
let ty_seqchar = ITYNamed "SeqChar"
let ty_seqtypestamp = ITYNamed "SeqTypestamp"

let fpr = Format.fprintf
let outf = !Pref.output_formatter

type act_param = bool * exp

and exp = 
  | IEConst of constval
  | IENew of typ * exp list
  | IENewArray of typ * exp * int
  | IEArrayAccess of exp * exp
  | IELocal of locvarsym
  | IEField of exp * membvarsym
  | IEStaticField of typsym * membvarsym
  | IECall of exp * methsym * act_param list
  | IEStaticCall of typsym * methsym * act_param list
  | IEAssign of exp * exp
  | IEUnOp of SO.unaryop * exp
  | IEBinOp of SO.binaryop * exp * exp
  | IECast of typ * exp
  | IETypeStamp of typestampsym * exp
  | IELet of atompat * exp option * exp
  | IESeq of exp * exp
  | IEReturn of exp option
  | IEIf of exp * exp * exp
  | IEProcCall of methsym * locvarsym * (locvarsym * typ) list * exp * exp option
      (* method of class "Matching" *) (* return values (call by reference) *)
  | IEWhile of exp * exp
  | IESwitchK of exp * clause list
  | IESwitchV of exp * clause list
  | IESwitchO of exp * clause list
  | IESwitchA of exp * clause list
  | IESwitchP of exp * clause list
  | IEBreak
  | IECatch of exp * labsym * (locvarsym * typ) list * exp
  | IEExit of labsym * exp list
  | IEEmpty
  | IEError of string
  | IERegexField of regexsym * regexfield * exp * exp

and clause = pat * exp

let opt_compiled = IEField(IELocal "RegexOptions", "Compiled")
let the_empty_exp = IEStaticField("Utils", "the_empty_seq")

let wildcard_char_regex = "(.|\n)"
let wildcard_pcdata_regex = "(.|\n)(.|\n)*"

type form_param = bool * locvarsym * typ

type invoke = 
  | IINVNone
  | IINVThis of act_param list
  | IINVSuper of act_param list

type meth_decl = methsym * SO.status * SO.method_modifier * form_param list * typ * exp

type clmember =
  | ICMConstr of invoke * form_param list * exp
  | ICMStaticConstr of exp
  | ICMFld of membvarsym * SO.status * typ * exp option
  | ICMMeth of meth_decl

type clas = {
  iCLHomeNS: string;   (*--VG full longname of the parent namespace *)
  iCLName: typsym;
  iCLExt: typsym;
  iCLImpls: typsym list;
  iCLMems: clmember list;
}

type intf_meth_decl = methsym * form_param list * typ

type intf = {
  iINHomeNS: string;   (*--VG full longname of the parent namespace *)
  iINName: typsym;
  iINImpls: typsym list;
  iINMems: intf_meth_decl list;
}

type usedir =
  | IUSAlias of nssym * nssym
  | IUSNSpace of nssym

type prog = usedir list * intf list * clas list


