open Il
module L = List

let _cnt_ = ref 0

let bump k = _cnt_ := !_cnt_ + k

let fpr = Format.fprintf
let outf = !Pref.output_formatter

let exp =
  let rec do_opt e = match e with
    | None -> 0
    | Some e -> loop e
  and do_list exps = L.fold_left (fun sz e -> sz + loop e) 0 exps
  and do_act_params params = L.fold_left (fun sz (_,e) -> sz + loop e) 0 params
  and do_clauses cls = L.fold_left (fun sz (p,e) -> sz + pat p + loop e) 0 cls
  and patcont { iTag=p1; iAtt=p2; iCont=p3; iRest=p4 } =
    atompat p1 + atompat p2 + atompat p3 + atompat p4
  and atompat = function
    | IPVar(_,_) ->
        bump 2;
        2
    | IPDontCare ->
        bump 1;
        1
  and pat = function
    | IPEmpty ->
        bump 1;
        1
    | IPDefault ->
        bump 1;
        1
    | IPAnyAtom a ->
        bump 1;
        1 + atompat a
    | IPElement pc ->
        bump 1;
        1 + patcont pc
    | IPPcdata pc ->
        bump 1;
        1 + patcont pc
    | IPObj pc ->
        bump 1;
        1 + patcont pc
    | IPBaseType(_,pc) ->
        bump 2;
        2 + patcont pc
    | IPAtt(pairs,ao) ->
        (L.fold_left (fun sz (_,a) -> sz + atompat a) 0 pairs) +
        (match ao with
          | None -> 0
          | Some a -> atompat a)
    | IPBaseVal _ ->
        bump 1;
        1
    | IPRegex(_,_) ->
        bump 2;
        2
  and loop e = match e with
    | IEConst(_) ->
        bump 1;
        1
    | IERegexField(_,_,e1,e2) ->
        bump 3;
        3 + loop e1 + loop e2
    | IENew(_,exps) ->
        bump 2;
        2 + do_list exps
    | IENewArray(_,e,_) ->
        bump 3;
        3 + loop e
    | IEArrayAccess(e1,e2) ->
        bump 1;
        1 + loop e1 + loop e2
    | IELocal(_) ->
        bump 1;
        1
    | IEField(e,_) ->
        bump 1;
        1 + loop e
    | IEStaticField(_,_) ->
        bump 2;
        2
    | IECall(e,_,params) ->
        bump 2;
        2 + do_act_params params + loop e
    | IEStaticCall(_,_,params) ->
        bump 3;
        3 + do_act_params params
    | IEAssign(e1,e2) ->
        bump 1;
        1 + loop e1 + loop e2
    | IEUnOp(_,e) ->
        bump 2;
        2 + loop e
    | IEBinOp(_,e1,e2) ->
        bump 2;
        2 + loop e1 + loop e2
    | IECast(_,e) ->
        bump 2;
        2 + loop e
    | IETypeStamp(_,e) ->
        bump 2;
        2 + loop e
    | IELet(a,e1,e2) ->
        bump 1;
        1 + atompat a + loop e2 + do_opt e1
    | IESeq(e1,e2) ->
        bump 1;
        1 + loop e1 + loop e2
    | IEReturn(e) ->
        bump 1;
        1 + do_opt e
    | IEIf(e1,e2,e3) ->
        bump 1;
        1 + loop e1 + loop e2 + loop e3
    | IEProcCall(_,_,params,e1,e2) ->
        bump 3;
        3 + L.length params + loop e1 + do_opt e2
    | IEWhile(e1,e2) ->
        bump 1;
        1 + loop e1 + loop e2
    | IESwitchK(e,cls)
    | IESwitchV(e,cls)
    | IESwitchO(e,cls)
    | IESwitchP(e,cls)
    | IESwitchA(e,cls) ->
        bump 1;
        1 + loop e + do_clauses cls
    | IEBreak ->
        bump 1;
        1
    | IECatch(e1,_,params,e2) ->
        bump 2;
        2 + loop e1 + loop e2
    | IEExit(l,exps) ->
        bump 2;
        2 + do_list exps
    | IEEmpty ->
        bump 1;
        1
    | IEError _ ->
        bump 2;
        2
  in loop

let meth_decl (m,status,md,params,t,e) =
  fpr outf "method %s: %d@." m (exp e)
    
let clmember = function
  | ICMConstr(i,params,e) ->
      fpr outf "constructor: %d@." (exp e)
  | ICMStaticConstr(e) ->
      fpr outf "static_constructor: %d@." (exp e)
  | ICMFld(n,s,t,e) ->
      (match e with
        | None -> ()
        | Some e -> fpr outf "field %s: %d@." n (exp e))
  | ICMMeth(decl) ->
      meth_decl decl

let clas {iCLName=n; iCLExt=n2; iCLImpls=ns; iCLMems=mems} =
  fpr outf "Class %s:@." n;
  L.iter clmember mems

let prog (udl,intfs,classes) =
  _cnt_ := 0;
  L.iter clas classes;
  !_cnt_
