open Format
open Printing
module SD = SynDenot

exception NoLabsMaximum of (Ident.id * SynDenot.hostty list)

let fpr = Format.fprintf
let outf = !Pref.output_formatter

let ta_any = ref Ta.empty   (** this duplicates Store.ta_any *)

let reset () =
  let s = Buildta.build_ta !Store.any_seqty_def in
  match s.Ta.next with
    | [_,_,s1,_,_] ->
        ta_any := s1
    | _ -> assert false

module StateTbl = 
  Hashtbl.Make
    (struct 
      type t = Ta.state
      let equal = (==)
      let hash x = x.Ta.num
    end)

 
  
(**VG(comment) Computes y.Ta.preclos, which is the "closure of finpre sets",
   consisting of the union of Ta.finpre sets of y and all its
   descendants via Ta.next. *)
let rec collect_preactions tbl dirty y =
  if StateTbl.mem tbl y
  then y.Ta.preclos
  else begin
    StateTbl.add tbl y ();
    let c = 
      List.fold_left (fun c (_,_,y1,y2,pre) ->
	Sortedlist.merge compare pre
	  (Sortedlist.merge compare c
	     (Sortedlist.merge compare (collect_preactions tbl dirty y1)
		(collect_preactions tbl dirty y2))))
	(*init*) y.Ta.finpre
	(*list*) y.Ta.next in
    if Sortedlist.compare compare y.Ta.preclos c == 0 then c
    else begin
      dirty := true;
      y.Ta.preclos <- c;
      c
    end
  end
     
let rec find_preactions x =
  let tbl = StateTbl.create 121 in
  let dirty = ref false in
  ignore(collect_preactions tbl dirty x);
  if !dirty then find_preactions x


(* Specialize a given pattern y with respect to a given type x:
   the resulting pattern exactly matches values in the type x *)

let rec specialize_states asm x y =
  let any_id = (!ta_any).Ta.num in
  if any_id == x.Ta.num then y
  else if any_id == y.Ta.num || x.Ta.num == y.Ta.num then x
  else 
    try Ta.StStMap.find (x, y) !asm
    with Not_found ->
      begin
	let fin = x.Ta.final && y.Ta.final in
        let z = Ta.create_state fin [] None in
        if fin then z.Ta.finpre <- y.Ta.finpre;
        z.Ta.preclos <- y.Ta.preclos;
        asm := Ta.StStMap.add (x, y) z !asm;
        z.Ta.next <- specialize_pats asm x.Ta.next y.Ta.next;
        z
      end

and specialize_pats asm xl yl =
  Listutil.mapcat
    (fun (c, a, x1, x2, _) ->
      Listutil.mapcat
	(fun (d, b, y1, y2, pre) ->
	  let (ch, c_vars) = SD.crack_labty c in
	  let (dh, d_vars) = SD.crack_labty d in
	  match Sub.minimum_hostty [ch; dh] with 
	  | None -> []  (* labels are disjoint *)
	  | Some cdh -> 
(*VG?? just an optimization?
	      if ( Isect.disjoint [x1] [y1] 
		|| Isect.disjoint [x2] [y2]) then [] 
	      else 
*)
	      let cd = 
		SD.slap_labty (Sortedlist.merge compare c_vars d_vars) cdh
	      in [(cd, b,
		   specialize_states asm x1 y1,
		   specialize_states asm x2 y2,
		   pre)]
	) yl)
    xl

(*old -- relies on ld_inter   and Isect.disjoint 
and specialize_pats asm xl yl =
  Listutil.mapcat
    (fun (c, a, x1, x2, _) ->
      Listutil.mapcat
	(fun (d, b, y1, y2, pre) ->
	  if (boot.sb_labdenot.ld_disjoint c d 
	    || Isect.disjoint [x1] [y1] 
	    || Isect.disjoint [x2] [y2]) then [] 
	  else
	    [(boot.sb_labdenot.ld_inter c d,
	      specialize_states asm x1 y1,
	      specialize_states asm x2 y2,
	      pre)]
	)
      yl)
    xl
*)


(* We obtain a type from a pattern by replacing each closing variable
   with the empty sequence. *)

let contains_closing var pre =
  List.exists (function
      Ta.End var' when var=var' -> true
    |	_ -> false) 
    pre

let rec trim_closing asm z var =
  if not (contains_closing var z.Ta.preclos) then
    z 
  else
    try Ta.StMap.find z !asm
    with Not_found ->
      begin
	let fin = 
	  z.Ta.final || 
	  List.exists
	    (fun (_,_,_,_,pre) -> contains_closing var pre) z.Ta.next in
	let zt =
	  List.filter
	    (fun (_,_,_,_,pre) -> not(contains_closing var pre)) z.Ta.next in
	let w = Ta.create_state fin [] None in
	asm := Ta.StMap.add z w !asm;
	w.Ta.next <- 
	  List.map
	    (fun (l,a,z1,z2,_) -> (l, a, z1, trim_closing asm z2 var, []))
	    zt;
	w
      end
      
(* We obtain a type from a pattern by trimming off the type to the
   given fixed length. *)

let rec trim_fixed asm z len =
  try Ta.StMap.find z !asm
  with Not_found ->
    if len == 0 then begin
      let w = Ta.create_state true [] None in
      asm := Ta.StMap.add z w !asm;
      w
    end
    else begin
      let w = Ta.create_state z.Ta.final [] None in
      asm := Ta.StMap.add z w !asm;
      w.Ta.next <- 
	List.map
	  (fun (l,a,z1,z2,_) -> (l, a, z1, trim_fixed asm z2 (len-1), []))
	  z.Ta.next;
      w
    end

(* We retrieve a type for a variable from the specialized pattern. *)

let ensure_var var m = 
  try Hashtbl.find m var
  with Not_found -> 
    let x = Ta.create_state false [] None in
    Hashtbl.add m var x;  
    x

let add_final var m = 
  (ensure_var var m).Ta.final <- true

let add_next var t m = 
  let ts = Ta.elim_empty_l [t] in
  let x = ensure_var var m in
  x.Ta.next <- ts @ x.Ta.next


let contains_any_opening pre =
  List.exists (function
      Ta.Begin _ | Ta.Rest _ | Ta.Fixed _ -> true
    |	_ -> false) 
    pre


let rec retrieve_seq_bindings asm stTbl z =
  if contains_any_opening z.Ta.preclos && not(Ta.StSet.mem z !asm) then begin
    asm := Ta.StSet.add z !asm;
    List.iter (function
	Ta.Rest var -> add_final var stTbl
      | Ta.Begin var -> add_final var stTbl
      | Ta.Fixed (var, _) -> add_final var stTbl
      | _ -> ())
      z.Ta.finpre;
    List.iter (fun (l, a, z1, z2, pre) ->
      retrieve_seq_bindings asm stTbl z1;
      retrieve_seq_bindings asm stTbl z2;
      List.iter (function
	  Ta.Rest var -> 
	    add_next var (l, a, z1, z2, []) stTbl
	| Ta.Begin var -> 
	    if contains_closing var pre then add_final var stTbl else
	    add_next var 
	      (l, a, z1, trim_closing (ref Ta.StMap.empty) z2 var, []) stTbl
	| Ta.Fixed (var, len) ->
	    if len=0 then add_final var stTbl else
	    add_next var 
	      (l, a, z1, trim_fixed (ref Ta.StMap.empty) z2 (len-1), []) stTbl
	| _ -> ())
	pre)
      z.Ta.next
  end



let retrieve_lab_att_bindings  zz = 
  let visited = ref Ta.StSet.empty in 
    (* states (z and its descendants) already visited in traversal  *) 
  let (styTbl: (Ident.id, SD.seqty list) Hashtbl.t) = Hashtbl.create 13 in
    (* maps an attr var to list of all seqtypes associated to it under z *)
  let (htyTbl: (Ident.id, SD.hostty list) Hashtbl.t) = Hashtbl.create 13 in
    (* maps a label var to the list of types ever associated to it under z  *)
  (* Update functions for the above tables: *)
  let sty_for_var s v = 
    let curr = try Hashtbl.find styTbl v 
	with Not_found -> [] in
    Hashtbl.replace styTbl v (s::curr)
  in
  let hty_for_var h v = 
    let curr = try Hashtbl.find htyTbl v 
	with Not_found -> [] in
    Hashtbl.replace htyTbl v (h::curr)
  in
  let rec retrieve z = 
    if not(Ta.StSet.mem z !visited) then 
      begin
	visited := Ta.StSet.add z !visited;
	List.iter 
	  (fun (l, a, z1, z2, pre) ->
	    retrieve z1;   retrieve z2;
	    let (hty, vars) = SD.crack_labty l in 
	    List.iter (hty_for_var hty) vars;
	    match SD.attsty' a with
	    | SD.TAttAny -> ()
	    | SD.PAtt(attpairs, catchopt) ->
		List.iter (fun (_, aid, sty) -> sty_for_var sty aid) attpairs;
		match catchopt with
		| Some(vid, sty) -> sty_for_var sty vid
		| None -> ()
	  ) z.Ta.next  
      end
  in 
  retrieve zz;
  let bind = ref IdBind.empty in 
  begin
    (* Add bindings for label variables *)
    Hashtbl.iter 
      (fun v htys -> 
	let jointy = List.fold_left Sub.join_hostty SD.hty_Null htys in
	bind := IdBind.add !bind (v, (SD.ty (SD.Thost jointy)))
      ) htyTbl;
    (* Add bindings for attribute variables *)
    Hashtbl.iter
      (fun v stys -> 
	let unionty = SD.alt stys in
	bind := IdBind.add !bind (v, (SD.ty (SD.Tseq unionty)))
      ) styTbl
  end;
  !bind



(************* Inference entry points ***********************)


let tyMatch (x: Ta.state) (y: Ta.state) 
    : Ta.state * (SynDenot.ty IdBind.t) * (Ident.id, Ta.state) Hashtbl.t = 
  find_preactions y;
  let z = specialize_states Store.inf_asm x y in
  let stTbl = Hashtbl.create 121 in
  retrieve_seq_bindings (ref Ta.StSet.empty) stTbl z;
  let labattBind = retrieve_lab_att_bindings z in
  (z, labattBind, stTbl)
  (* By returning z we rely on the fact that specialize_states 
     returns a pattern which is actually the intersection of the original 
     and the input type.  *)
 
  

