let set_print_states str =
  Pref.print_states := List.map int_of_string (Str.split (Str.regexp " ") str)

let set_meth_configs str =
  Pref.debug_method_configs := Str.split (Str.regexp " ") str

let usage_msg = 
  Format.sprintf "Usage: %s [options] classname file1 ... [-- arg1 ...]" Sys.argv.(0)
 
let process_args() =
  let anonyms = ref [] in  (* for  'classname file1 ... filen' portion *)
  let args = ref [] in     (* for  '-- arg1 ... argk' portion *)
  Arg.parse
    [ 
      "-version", Arg.Unit (fun () -> Format.printf "FX version %s@." Pref.version; exit 0), "show version";
      "-showphase", Arg.Unit (fun() -> Pref.showphase := true), "show phases";
      "-noeval", Arg.Unit (fun() -> ()), "deprecated, no-op";
      "-debug_configs", Arg.Unit (fun() -> Pref.debug_configs := true),
        "debug congigurations";
      "-debug_intersections", Arg.Unit (fun() -> Pref.debug_intersections := true),
        "debug intersection operations";
      "-dumpil", Arg.Unit (fun() -> Pref.dumpil := true), "dump IL";
      "-dumpoptil", Arg.Unit (fun() -> Pref.dumpoptil := true), "dump optimized IL";
      "-dumpbreakil", Arg.Unit (fun() -> Pref.dumpbreakil := true), "dump IL with breaks";
      "-dumpilsize", Arg.Unit (fun() -> Pref.dumpilsize := true),
        "dump initial IL size";
      "-dumpoptilsize", Arg.Unit (fun() -> Pref.dumpoptilsize := true),
        "dump optimized IL size before breaks are added";
      "-dumpbreakilsize", Arg.Unit (fun() -> Pref.dumpbreakilsize := true),
        "dump final IL size";
      "-skipopt", Arg.Unit (fun() -> Pref.skipopt := true), "skip optimization";
      "-skiptb", Arg.Unit (fun() -> Pref.tb := false), "skip type-based optimization";
      "-skiptbprocs", Arg.Unit (fun() -> Pref.tbprocs := false),
        "avoid type-based helper procedures";
      "-o", Arg.String (fun s -> Pref.outputfile := s), "c sharp output filename";
      "-tempfile", Arg.String (fun s -> Pref.tempfile := s), "c sharp temp filename";
      "-tagprefix", Arg.String (fun s -> Pref.tagprefix := s), "tag field prefix";
      "-print_states", Arg.String set_print_states, "ids of TA states to be printed";
      "-debug_method_configs", Arg.String set_meth_configs,
        "method names for which to turn configuration debugging";
      "-outputxducesrc", Arg.Unit (fun() -> Pref.outputxducesrc := true),
        "output xduce sources";
      "-outputxduceval", Arg.Unit (fun() -> Pref.outputxduceval := true),
        "output xduce values";
      "-trace", Arg.Unit (fun() -> Pref.trace := true), "trace compilation";
      "-tracepat", Arg.Unit (fun() -> Pref.tracepat := true), "trace pat compilation";
      "-loopbreakers", Arg.String (fun s -> Pref.lb := s), "specifies a method for determining loop breakers [lb-min|lb-max|lb-indeg|lb-indeg-multtrans|lb-multtrans]";
      "-pcdataencoding", Arg.String (fun s -> Pref.pcdataencoding := s), "specifies encoding of pcdata [char|string|substring|hybrid]";
      "-appendpolicy", Arg.String (fun s -> Pref.appendpolicy := s), "specifies sequence concatenation algorithm [eager|lazynormalize|eagernormalize]";
      "-runtimefile", Arg.String (fun s -> let new_alt = 
				    match !Pref.altern_runtime with 
				      | None -> [s] 
				      | Some lst -> lst @ [s] 
				  in Pref.altern_runtime := Some new_alt), 
      "xti file with specs for an alternative runtime; the corresponding cs file(s) should be supplied to the csharp compiler instead of xtatic.cs";
(*??
      "-path", Arg.String (fun s -> Pref.searchpaths := s::!Pref.searchpaths), "add a string to the search paths";
*)

      "--", Arg.Rest (fun arg -> args := !args @ [arg]), "arguments to pass to Xtatic program";
    ] 
    (fun anon -> anonyms := !anonyms @ [anon])
    usage_msg;
  Pref.args := !args;
  match !anonyms with
  | [] -> failwith "Start class and files are not specified"
  | classname :: filenames -> (classname, filenames)



open Format
(* open Print *)

let rec print_exn  =  
  function
    | XtLexer.Lexing(finfo, msg) ->
	Finfo.print finfo;
	eprintf "@[%s@]@." msg
    | Parse.Parse(finfo, msg) ->
	Finfo.print finfo;
	eprintf "@[%s@]@." msg
    | Source.ElaborationError(finfo, msg) -> 
	eprintf "@[%s: %s@]@." (Finfo.finfo_to_string finfo) msg
    | Naming.ElaborationError(finfo, msg) -> 
	eprintf "@[%s: %s@]@." (Finfo.finfo_to_string finfo) msg
    | Src2denot.ElaborationError(finfo, msg) -> 
	eprintf "@[%s: %s@]@." (Finfo.finfo_to_string finfo) msg
    | CtxDat.MultipleInheritance(finfo,msg) ->
	eprintf "@[%s: %s@]@." (Finfo.finfo_to_string finfo) msg
    | CtxDat.InterfaceExtendsClass(finfo,msg) ->
	eprintf "@[%s: %s@]@." (Finfo.finfo_to_string finfo) msg
    | CtxDat.InheritanceLoop(msg) ->
	eprintf "@[%s@]@."  msg
    | CtxCls.TypeCheck(finfo, msg) -> 
	eprintf "@[%s: %s@]@." (Finfo.finfo_to_string finfo) 
	  ("From CtxCls: " ^ msg)
    | Tycheck.TyCheck(finfo, msg) -> 
	eprintf "@[%s: @[%s@]@]@." (Finfo.finfo_to_string finfo) 
	  ("From Tycheck: " ^ msg)
    | e -> raise e 

 
let fpr fmtstr = Format.fprintf !Pref.output_formatter fmtstr

let _ = Format.pp_set_max_boxes !Pref.output_formatter 1000
let _ = Format.pp_set_margin !Pref.output_formatter 75

let pr_startPhase phname = 
  if !Pref.showphase then
    fpr "@[BEGIN %s ... @]@."  phname

let pr_finishPhase phname = 
  if !Pref.showphase then
    fpr "@[END %s ... @]@." phname

      
let tycheck_timer = Timer.create "tycheck_timer"

(** The revised structure of main() to support use of xti files for
  standard libraries, to be implemented next.  Note that the stages
  have repeating structure, which will be factored out into
  load_typlib function. *)
(*
let main () = 
  (**** Stage 0,  empty contexts  *)
  (*  nspace, datctx, clsctx  -- empty, will grow by stages *)
  (*  nctx, prog  -- empty starters that will be re-created at each stage *)

  (****  Stage 1, .NET core type library  ****)
  (*   Load dotnet_core.xti as tunits, *)
  (*   Extend nspace with tunits  *)
  (*   Make nctx from the nspace and  core aliases *)
  (*   Extend empty SS.prog with the tunits (under the nctx)  *)
  (*   Extend datctx from prog, Semant.setup, extend clsctx from prog *)

  (****  Stage 2, standard .NET type libraries ****)
  (*   Load dotnet_standard.xti as tunits *)
  (*   Extend nspace with tunits *)
  (*   Make nctx from the nspace and  core aliases *)
  (*   Extend empty SS.prog with the tunits (under the nctx)  *)
  (*   Extend datctx from prog, Semant.setup, extend clsctx from prog *)

  (****  Stage 3, other libraries ****)
  (*   Load ???.xti as tunits *)
  (*   Proceed as in Stages 1, 2 ! *)

  (****  Stage 4, user program ****)
  (*   Load user's xt files as tunits *)
  (*   Proceed as in Stages 1, 2 ! *)
  (*   Now typecheck and compile user's program *)
  ()
*)

let load_typlib (nspace,datctx,clsctx) tunits : 
  (SynSource.prog * (Naming.space * Semant.datctx * Contexts.clsctx)) = 
  let new_nspace = 
    List.fold_left Source.extend_top_space nspace tunits in
  let curr_nctx = Naming.ctxInit new_nspace in
  let curr_nctx = Naming.ctxApplyImports curr_nctx Kerlib.nimports in
  let empty_prog = 
    { SynSource.cls = Ident.empty;
      SynSource.intfs = Ident.empty;
      SynSource.seq = Ident.empty; } in
  let curr_prog = 
    List.fold_left (Source.add_top_unit curr_nctx new_nspace.Naming.cs)
      empty_prog tunits in
   let new_datctx = CtxDat.extend datctx curr_prog in
   let () = Semant.setup new_datctx in
   let new_clsctx = CtxCls.extend clsctx curr_prog 
   in
     (curr_prog, (new_nspace, new_datctx, new_clsctx))


let main () = 

 (**** Stage 1, loading .NET core type libraries ****)

  let nspace = Kerlib.nspace in
  let datctx = Kerlib.datctx in
  let clsctx = Kerlib.clsctx in

  (**** Stage 2, loading .NET standard libraries ****)
  let xt_runtime = match !Pref.altern_runtime with 
    | None -> [ Pref.homedir ^ "/../lib/poorman.xti";
		Pref.homedir ^ "/../lib/xtatic.xti"; ]
    | Some files -> files
  in 
  let std_tunits = List.map Parse.fromFile 
		     ([Pref.homedir ^ "/../lib/dotnet_std.xti";
(*--		      Pref.homedir ^ "/../lib/poorman.xti";
		      Pref.homedir ^ "/../lib/xtatic.xti"; *)
		     ] @ xt_runtime) in
  let (_, (nspace, datctx, clsctx)) = 
    load_typlib (nspace, datctx, clsctx) std_tunits in

  (**** Stage 3, loading type libraries referenced by the user program *****)

  (**** Stage 4, reading off type libraries defined by the user program  *****)
  let (startclassname, inputfiles)  = process_args() in
  let user_tunits = List.map Parse.fromFile inputfiles in
  let (user_prog, (nspace, _, clsctx)) = 
    load_typlib (nspace, datctx, clsctx) user_tunits in
  let startclassid = Naming.findClass nspace startclassname in

  if !Pref.outputxducesrc then Xdprint.pr_prog !Pref.output_formatter user_prog;
  let _ = pr_startPhase "typechecking" in

  Infer.reset ();

  let typed_prog = 
    Timer.wrap tycheck_timer (fun () -> CtxCls.typecheck clsctx user_prog) in
  let _ = pr_startPhase "compiling" in
  let startclassname = Ident.name startclassid in
  SynDenot.set_module_name startclassname;
  let compiled_prog = Compile.compile_prog typed_prog startclassid in
  if !Pref.dumpilsize then (
    fpr "IL size before opt@.";
    let sz = Ilsize.prog compiled_prog in
    fpr "Total size = %d@." sz
  );
  if !Pref.dumpil then Printil.dump compiled_prog;
  let compiled_prog =
    if not !Pref.skipopt then
      (let res = Simp.prog compiled_prog in
       if !Pref.dumpoptilsize then (
         fpr "IL size after opt before breaks are added@.";
         let sz = Ilsize.prog compiled_prog in
         fpr "Total size = %d@." sz
       );
       if !Pref.dumpoptil then Printil.dump res;
       res)
    else compiled_prog
  in let compiled_prog = Addbreaks.prog compiled_prog in
  if !Pref.dumpbreakilsize then (
    fpr "final IL size@.";
    let sz = Ilsize.prog compiled_prog in
    fpr "Total size = %d@." sz
  );
  if !Pref.dumpbreakil then Printil.dump compiled_prog;
  let compiled_prog = Il2cs.translate_prog compiled_prog in
  let _ = pr_finishPhase "compiling" in
  let _ = Codegen.gen compiled_prog startclassname in
  ()

let exec func arg =
  try
    func arg
(*      Timer.dump(); *)
(*      Counter.dump() *)
  with
    e -> 
      print_exn e;
      Timer.dump();
      Counter.dump();
      exit 1

let _ = exec main ()


