module CN = Cnaming
module XN = Xnaming

type space = {
  cs: CN.space;
  xs: XN.spaces; }

type imports = {
  ci: CN.import list;
  xi: XN.import list; }
 
type ctx = {
  cc: CN.ctx;
  xc: XN.ctx; }

exception StartClassNotFound
exception ElaborationError of Finfo.t * string

(************** Names (for types and variables) ************************)

type name = string Finfo.finfo
type longname = name list 

type qualname = name option * name     (* (prefix opt, local) *) 

(*## These types have only mnemonic value now; perhaps can be in-lined... *)
type decname = name      (** for declaring a new name *)
type refname = longname  (** for referencing a type through its long name *)
type onename = name      (** for referencing a type through its short name *)


let eq_name (_, n1) (_, n2) = (n1 = n2)

let eq_longname ln1 ln2 = 
  try 
    List.fold_left2 (fun prev n1 n2 -> prev && (eq_name n1 n2)) 
      true ln1 ln2 
  with Invalid_argument _ -> false

let info_of_longname lname = 
  assert ((List.length lname) > 0);
  let (ifirst, _) = List.nth lname 0 in
  let (ilast, _) = List.nth lname (List.length lname - 1) in 
  {Finfo.file_name = ifirst.Finfo.file_name;
   Finfo.start_pos = ifirst.Finfo.start_pos;
   Finfo.end_pos = ilast.Finfo.end_pos; }

let path_of_longname ln : string list = 
  List.map (fun nm -> Finfo.drop nm) ln

let dirty_string_of_longname ln = 
  let rec dotter = function 
     | [n] -> n
     | n::ns -> n ^ "." ^ (dotter ns)
     | [] -> "<empty longname>"
  in dotter (path_of_longname ln)


let info_of_qualname qualname = 
  match qualname with 
    | (None, loc) -> Finfo.get loc
    | (Some pref, loc) -> 
	let ipref = Finfo.get pref in 
	let iloc = Finfo.get loc in 
	  {Finfo.file_name = ipref.Finfo.file_name;
	   Finfo.start_pos = ipref.Finfo.start_pos;
	   Finfo.end_pos = iloc.Finfo.end_pos}

let qname_of_qualname = function
  | (None, loc) -> (XN.NoPref, Finfo.drop loc)
  | (Some pref, loc) -> (XN.Pref (Finfo.drop pref), Finfo.drop loc)
      

let dirty_string_of_qualname qualname = 
  match qname_of_qualname qualname with 
    | (XN.NoPref, loc) -> loc
    | (XN.Pref pref, loc) -> pref ^ ":" ^ loc

(*****************  ****************)

let ctxInit space = 
  { cc = CN.ctxApplyImports CN.ctxEmpty [CN.OpenSpace space.cs];
    xc = XN.ctxInit space.xs; }

let ctxApplyImports ctx imports = 
  { cc = CN.ctxApplyImports ctx.cc imports.ci;
    xc = XN.ctxApplyImports ctx.xc imports.xi }



(***************** long name resolvers **********************)

(** raises Not_found, does NOT raise N.AmbiguousName *)
let resolve_as_memtail (ctx: ctx) refname : CN.member * CN.path =
  try
    CN.ctxFindDeepMem ctx.cc (path_of_longname refname) 
  with 
  | CN.AmbiguousName n -> 
      let rn_info = info_of_longname refname in
      let rn_string = dirty_string_of_longname refname in
      raise 
	(ElaborationError
	   (rn_info, "The component '" ^ n ^ "' of reference name '" ^ 
	    rn_string ^ "' is ambiguous in this context."))


let resolve_member ctx refname =
  let rn_info = info_of_longname refname in
  let rn_string = dirty_string_of_longname refname in
  try 
    match resolve_as_memtail ctx refname with
    | (referent, []) -> referent
    | (referent, x::xs) -> raise 
	  (ElaborationError
	     (rn_info, "The suffix " ^ 
	      (List.fold_left (fun nn n -> nn ^ "." ^ n) x xs) ^
	      " of the reference name does not " ^
	      "resolve into a namespace or a type."))
  with
  | Not_found -> raise 
	(ElaborationError(rn_info, "The name '" ^ rn_string ^ 
			  "' is undefined in this context."))

let resolve_leaf ctx refname = 
  match resolve_member ctx refname with
  | CN.Leaf a -> a
  | CN.Space _ -> raise
	(ElaborationError
	   (info_of_longname refname, "The name '" ^ 
	      (dirty_string_of_longname refname) ^ 
	      "' is expected to refer to a leaf, not a namespace")) 


let resolve_class ctx refname = 
  match resolve_leaf ctx refname with 
    | (id, CN.LClas) -> id
    | _ -> raise (ElaborationError
		    (info_of_longname refname, 
		     "The name '" ^ (dirty_string_of_longname refname) ^ 
		     "' is expected to refer to a class"))

let resolve_interface ctx refname = 
  match resolve_leaf ctx refname with 
    | (id, CN.LIntf) -> id
    | _ -> raise (ElaborationError
		    (info_of_longname refname, 
		     "The name '" ^ (dirty_string_of_longname refname) ^ 
		     "' is expected to refer to an interface"))

let chop_leaf nctx refname : (CN.leaf * CN.path) option = 
  try 
    match resolve_as_memtail nctx refname with
      | (CN.Leaf lf, path) -> Some (lf, path)
      | (CN.Space _, _) -> raise
	  (ElaborationError
	     (info_of_longname refname, 
	      "The name '" ^ (dirty_string_of_longname refname) ^ 
	      "' is expected to refer to a leaf, not a namespace")) 
    with Not_found -> None


let resolve_seqty ctx qualname = 
  let qname = qname_of_qualname qualname in 
    try 
      XN.ctxResolveQname ctx.xc qname 
    with Not_found -> raise 
      (ElaborationError(info_of_qualname qualname, 
			"This regtype name could not be resolved: " ^ 
			(dirty_string_of_qualname qualname)))



(****************  Class finder **************************)


(** Uses dots in the input string to split it into a list of short
  names.  Mutliple dots in a row and dots trailing in the beginning
  and in the end are ignored. *)
let parse_longname strlongname = 
  let rec find_nondot pos = (* looking from the right end *)
    if pos < 0 then raise Not_found
    else if ('.' = String.get strlongname pos) then find_nondot (pos-1)
    else pos
  in
  let rec parse lname pos = 
    let rbound = 
      try find_nondot pos
      with Not_found -> (-1) in
    let lbound = 
      try String.rindex_from strlongname rbound '.' 
      with Not_found -> (-1) 
    in 
      if (rbound<0) then lname
      else 
	let name = String.sub strlongname (lbound+1) (rbound-lbound)
	in parse (name::lname) lbound
  in
  let lname = parse [] ((String.length strlongname) -1)
  in lname


let findClass  nspace lname : Ident.id = 
  let startclassid = 
    try 
      match CN.spaceFindDeepMem nspace.cs (parse_longname lname) with
      | (CN.Leaf (id, CN.LClas), []) -> id
      | (CN.Leaf (id, _), []) -> 
	  raise (ElaborationError
		   (Finfo.bogus(), 
		    "The namespace leaf with name " ^
		    lname ^ " is not a class"))
      | _ -> raise Not_found
    with Not_found -> raise StartClassNotFound
  in startclassid
