open Format
open DenBoot

(* PATTERN OPTIMIZATION transforms a given pattern (which may be the
   union of patterns of several match clauses) to another more
   efficient pattern, by using the type information for input values
   given from the context.  

   PATTERN OPTIMIZATION has two phases:

   (i) Propagate the type information through the structure of the
   given pattern and, for each subpattern, compute a type of values
   that may be matched against the subpattern.  That is, once such a
   type is computed, we can guarantee that any values NOT in this type
   will never be matched against the subpattern.  On the other hand,
   this analysis may be conservative in that the type may contain
   extra values that will actually never be matched against.

   (ii) Reconstruct the given pattern by using replacing each
   subpattern by a new one in the following way:

     + If the computed type is included by the subpattern and the
     subpattern does not contain pattern varialbles, then we replace
     the subpattern with the Any pattern since matching aginst this
     subpattern will always succeed.

     + If the computed type and the subpattern are disjoint, then we
     replace the subpattern with the failure pattern since matching
     aginst this subpattern will always fail.

     + If the computed type is the union of types with all the same
     label L and the subpattern is the union of patterns also with all
     the same label L, then we replace each label in this union of
     patterns with the ~ label (i.e., any labels) since the label
     checking will always succeed.

*)

(* Type propagation *)

module TypePatternSet = Ta.StStSet

module PatternTable = 
  Map.Make
    (struct
      type t = Ta.state
      let compare = Ta.compare_state
    end)

let rec propagate_type_to_pattern_rec boot set typ pat =
  if not(TypePatternSet.mem (typ,pat) !set) then begin
    set := TypePatternSet.add (typ,pat) !set;
    List.iter 
      (fun (k,_,typ1,typ2,_) ->
	List.iter
	  (fun (l,_,pat1,pat2,pre) ->
	    if boot.sb_labdenot.ld_sub k l then begin
	      propagate_type_to_pattern_rec boot set typ1 pat1;
	      propagate_type_to_pattern_rec boot set typ2 pat2
	    end)
	  pat.Ta.next)
      typ.Ta.next
  end
 
let convert_type_pattern_set_to_map set =
  let map = ref PatternTable.empty in
  let merge (typ,pat) =
    try 
      let typs = PatternTable.find pat !map in 
      typs := Sortedlist.merge Ta.compare_state !typs [typ]
    with
      Not_found -> 
	map := PatternTable.add (*key*)pat (*data*)(ref [typ]) !map in
  TypePatternSet.iter merge !set;
  !map

let propagate_type_to_pattern boot ty pat =
  let set = ref TypePatternSet.empty in
  propagate_type_to_pattern_rec boot set ty pat;
  convert_type_pattern_set_to_map set

(* Pattern reconstruction *)

let rec reconstruct_pattern_rec boot hyp map pat =
  try PatternTable.find pat !hyp with 
    Not_found ->
      let typs = try !(PatternTable.find pat map) with Not_found -> [] in
      if (List.for_all Ta.is_empty typs ||
	Sub.subtype boot typs [pat] == Result.Succ) && 
	pat.Ta.preclos = [] then begin
	  let any = boot.sb_engine.Eng.any_ta  in
	  hyp := PatternTable.add pat any !hyp;
	  any
      end (*else 
      if Isect.disjoint typs [pat.Ta.typ]
      then begin
	hyp := PatternTable.add pat Ta.empty !hyp;
	Ta.empty
      end*) else begin
	let pat1 = Ta.create_state false [] None in
	hyp := PatternTable.add pat pat1 !hyp;
	if Ta.state_set_final typs && pat.Ta.final then begin
	  pat1.Ta.final <- true;
	  pat1.Ta.finpre <- pat.Ta.finpre
	end;
	pat1.Ta.next <- 
	  reconstruct_pattern_trans boot hyp map (Ta.state_set_next typs) pat.Ta.next;
	pat1
      end

and reconstruct_pattern_trans boot hyp map tytrans pattrans =
  List.map
    (fun (l,a,pat2,pat3,pre) -> 
      (l, a,
       reconstruct_pattern_rec boot hyp map pat2,
       reconstruct_pattern_rec boot hyp map pat3,
       pre))
    pattrans


let reconstruct_pattern boot map pat =
  let hyp = ref PatternTable.empty in
  reconstruct_pattern_rec boot hyp map pat

(* Pattern optimization entry point *)

let optimize_pattern_by_type boot ty pat =
  let map = propagate_type_to_pattern boot ty pat in
  Infer.find_preactions pat;
  reconstruct_pattern boot map pat


