module L = List
open Il

let fpr = Printf.fprintf

let list print_item =
  let rec loop ff = function
    | [] -> fpr ff ""
    | [x] -> print_item ff x
    | x::xs -> fpr ff "%a %a" print_item x loop xs
  in loop

let tagsym ff x = fpr ff "%s" x
let locvarsym ff x = fpr ff "%s" x
let membvarsym ff x = fpr ff "%s" x
let methsym ff x = fpr ff "%s" x
let typsym ff x = fpr ff "%s" x
let nssym ff x = fpr ff "%s" x
let labsym ff x = fpr ff "%s" x
let regexsym ff x = fpr ff "%s" x
let typestampsym ff x = fpr ff "%s" x
let regexfield ff x = fpr ff "%s" x

let constval ff = function
  | ICNull -> fpr ff "(null)"
  | ICBool(b) -> fpr ff "%b" b
  | ICInt(i) -> fpr ff "%d" i
  | ICChar(c) -> fpr ff "#\\%c" c
  | ICString(s) -> fpr ff "\"%s\"" s
  | ICSymbol(s) -> fpr ff "'%s" s

let rec predefty ff = function
  | IPTyVoid -> fpr ff "void"
  | IPTyBool -> fpr ff "bool"
  | IPTyInt -> fpr ff "int"
  | IPTyChar -> fpr ff "char"
  | IPTyString -> fpr ff "string"
  | IPTyArray(t) -> fpr ff "(array %a)" typ t

and typ ff = function
  | ITYPredef(pt) -> predefty ff pt
  | ITYNamed(n) -> typsym ff n

let basetyp ff = function
  | IBTEmpty -> fpr ff "()"
  | IBTElement -> fpr ff "elem"
  | IBTObject -> fpr ff "obj"
  | IBTBool -> fpr ff "bool"
  | IBTInt -> fpr ff "int"
  | IBTChar -> fpr ff "char"

let baseval ff = function
  | IBVBool(x) -> fpr ff "%b" x
  | IBVInt(x) -> fpr ff "%d" x
  | IBVChar(x) -> fpr ff "#\\%c" x
  | IBVSymbol(x) -> fpr ff "'%s" (Str.global_replace (Str.regexp "Sym.Tag.") "" x)
  | IBVObj(x) -> fpr ff "(class %a)" typsym x

let rec patcont ff { iTag=p1; iAtt=p2; iCont=p3; iRest=p4 } =
  fpr ff "%a %a %a %a" atompat p1 atompat p2 atompat p3 atompat p4 

and atompat ff = function
  | IPVar(v,t) -> fpr ff "%a:%a" locvarsym v typ t
  | IPDontCare -> fpr ff "_"

and attpat ff (a,p) = fpr ff "(%a %a)" tagsym a atompat p

and patopt ff = function
  | None -> fpr ff ""
  | Some(p) -> fpr ff " %a" atompat p

and pat ff = function
  | IPEmpty -> fpr ff "()"
  | IPDefault -> fpr ff "_"
  | IPAnyAtom(p) -> fpr ff "(anyatom %a)" atompat p
  | IPElement(pc) -> fpr ff "(elem %a)" patcont pc
  | IPPcdata(pc) -> fpr ff "(pcdata %a)" patcont pc
  | IPObj(pc) -> fpr ff "(obj %a)" patcont pc
  | IPBaseType(t,pc) -> fpr ff "(%a %a)" basetyp t patcont pc
  | IPAtt(pats,p) -> fpr ff "(att (%a)%a)" (list attpat) pats patopt p
  | IPBaseVal(bv) -> baseval ff bv
  | IPRegex(r,_) -> fpr ff "(regex %a)" regexsym r

let unaryop ff = function 
  | SO.UnNot -> fpr ff "not "
  | SO.UnPlus -> fpr ff "+"
  | SO.UnMinus -> fpr ff "-"

let binaryop ff = function 
  | SO.BiPlus  -> fpr ff "+"
  | SO.BiMinus   -> fpr ff "-"
  | SO.BiMult   -> fpr ff "*"
  | SO.BiDiv   -> fpr ff "/"
  | SO.BiMod   -> fpr ff "%%"
  | SO.BiLt   -> fpr ff "<"
  | SO.BiGt  -> fpr ff ">"
  | SO.BiLeq  -> fpr ff "<="
  | SO.BiGeq  -> fpr ff ">="
  | SO.BiEq  -> fpr ff "=="
  | SO.BiNeq  -> fpr ff "!="
  | SO.BiAnd  -> fpr ff "&&"
  | SO.BiOr  -> fpr ff "||"

let rec act_param ff (flag,e) =
  if flag then fpr ff "(byref %a)" exp e
  else fpr ff "%a" exp e

and expopt ff = function
  | None -> fpr ff "(void)"
  | Some(e) -> exp ff e

and expopt2 ff = function
  | None -> fpr ff ""
  | Some(e) -> fpr ff " %a" exp e

and expseq ff = function
  | IESeq(e1,e2) -> fpr ff "%a %a" exp e1 expseq e2
  | e -> exp ff e

and bnd ff (v,t) = fpr ff "%a:%a" locvarsym v typ t

and handler ff (l,bnds,e2) =
  fpr ff "(%a (%a) %a)" labsym l (list bnd) bnds expseq e2

and regex ff s =
  fpr ff "\"%s\"" s

and exp ff = function
  | IEConst(c) -> constval ff c
  | IENew(t,es) -> fpr ff "(new %a (%a))" typ t (list exp) es
  | IENewArray(t,e,i) -> fpr ff "(new %a [%a] [%d])" typ t exp e i
  | IEArrayAccess(e1,e2) -> fpr ff "(getarray %a %a)" exp e1 exp e2
  | IELocal(v) -> locvarsym ff v
  | IEField(e,f) -> fpr ff "(dot %a %a)" exp e membvarsym f
  | IEStaticField(t,f) -> fpr ff "%a.%a" typsym t membvarsym f
  | IECall(e,m,params) -> fpr ff "(%a %a %a)" methsym m exp e (list act_param) params
  | IEStaticCall(t,m,params) ->
      fpr ff "(%a.%a %a)" typsym t methsym m (list act_param) params
  | IEAssign(e1,e2) -> fpr ff "(set! %a %a)" exp e1 exp e2
  | IEUnOp(o,e) -> fpr ff "(%a %a)" unaryop o exp e
  | IEBinOp(o,e1,e2) -> fpr ff "(%a %a %a)" binaryop o exp e1 exp e2
  | IECast(t,e) -> fpr ff "(cast %a %a)" typ t exp e
  | IETypeStamp(x,e2) -> fpr ff "(typestamp %a %a)" typestampsym x exp e2
  | IELet(p,e1,e2) -> fpr ff "(let ((%a %a)) %a)" atompat p expopt e1 expseq e2
  | IESeq(e1,e2) -> fpr ff "(begin %a %a)" exp e1 expseq e2
  | IEReturn(e) -> fpr ff "(return%a)" expopt2 e
  | IEIf(e1,e2,e3) -> fpr ff "(if %a %a %a)" exp e1 exp e2 exp e3
  | IEProcCall(m,v,bnds,e1,e2) ->
      fpr ff "(if (%a %a %a) %a%a)" methsym m locvarsym v (list bnd) bnds exp e1 expopt2
        e2
  | IEWhile(e1,e2) -> fpr ff "(while %a %a)" exp e1 expseq e2
  | IESwitchK(e,cls) -> fpr ff "(case %a %a)" exp e (list clause) cls
  | IESwitchV(e,cls) -> fpr ff "(case %a %a)" exp e (list clause) cls
  | IESwitchO(e,cls) -> fpr ff "(case %a %a)" exp e (list clause) cls
  | IESwitchA(e,cls) -> fpr ff "(case %a %a)" exp e (list clause) cls
  | IESwitchP(e,cls) -> fpr ff "(case %a %a)" exp e (list clause) cls
  | IEBreak -> fpr ff "(break)"
  | IECatch(e1,l,bnds,e2) ->
      let rec loop e1 res = match e1 with
        | IECatch(e1,l,bnds,e2) -> loop e1 ((l,bnds,e2)::res)
        | _ -> fpr ff "(catch %a %a)" exp e1 (list handler) res
      in loop e1 [l,bnds,e2]
  | IEExit(l,exps) -> fpr ff "(exit %a %a)" labsym l (list exp) exps
  | IEEmpty -> fpr ff "()"
  | IEError(s) -> fpr ff "(error %s)" s
  | IERegexField(r,f,e1,e2) ->
      fpr ff "(regex_field %a %a %a %a)" regexsym r regexfield f exp e1 exp e2

and clause ff (p,e) = fpr ff "(%a %a)" pat p expseq e

let form_param ff (flag,v,t) =
  if flag then fpr ff "%a" locvarsym v
  else fpr ff "%a" locvarsym v

let invoke ff = function
  | IINVNone -> fpr ff ""
  | IINVThis(params) -> fpr ff "(this %a) " (list act_param) params
  | IINVSuper(params) -> fpr ff "(super %a) " (list act_param) params

let mmod ff = function
  | SO.Virtual -> fpr ff "virtual "
  | SO.Override -> fpr ff "override "
  | SO.New -> fpr ff "new "
  | SO.Normal -> fpr ff ""

let meth_decl ff (m,status,md,params,t,e) =
  match status with 
  (*--VG?? the clauses are identical; is this a bug? *)
    | SO.Static -> 
	fpr ff "(define (%a %a) %a %a)" methsym m (list form_param) params typ t exp e
    | SO.Instance -> 
	fpr ff "(define (%a %a) %a %a)" methsym m (list form_param) params typ t exp e

let intf_meth_decl ff (m,params,t) =
    fpr ff "(define (%a %a) %a)" methsym m (list form_param) params typ t

let exp_init ff = function
  | None -> ()
  | Some e -> fpr ff " = %a" exp e
    
let clmember ff = function
  | ICMConstr(i,params,e) ->
      fpr ff "(constructor %a(%a) %a)" invoke i (list form_param) params exp e
  | ICMStaticConstr(e) ->
      fpr ff "(static_constructor %a)" exp e
  | ICMFld(n,s,t,e) -> fpr ff "(field %a %s%a %a)" 
        membvarsym n 
        (match s with 
	   | SynOp.Static -> "static "
	   | SynOp.Instance -> "")
        typ t
        exp_init e
  | ICMMeth(decl) -> fpr ff "%a" meth_decl decl

let clas ff {iCLName=n; iCLExt=n2; iCLImpls=ns; iCLMems=mems} =
  fpr ff "(class %a (%a) (%a) %a)" typsym n typsym n2 (list typsym) ns (list clmember) mems

let intf ff {iINName=n; iINImpls=ns; iINMems=mems} =
  fpr ff "(interface %a (%a) %a)" typsym n (list typsym) ns (list intf_meth_decl) mems

(*--VG-- not used! 
let ifmember ff = function
  | IIMFld(n,t) -> fpr ff "(field %a %a)" membvarsym n typ t
  | IIMMeth(m,params,t) ->
      fpr ff "(method %a (%a) %a)" methsym m (list form_param) params typ t

let iface ff {iIFName=n; iIFExts=ns; iIFMems=mems} =
  fpr ff "(interface %a (%a) %a)" typsym n (list typsym) ns (list ifmember) mems
*)

let usedir ff = function
  | IUSAlias(s1,s2) -> fpr ff "(useas %a %a)" nssym s1 nssym s2
  | IUSNSpace(s) -> fpr ff "(use %a)" nssym s

(*--VG-- not used!
let rec nsmember ff = function
  | INMNspace(s) -> nspace ff s
  | INMClas(c) -> clas ff c
  | INMIf(i) -> iface ff i

and nspace ff {iNSName=n; iNSUsing=udl; iNSMems=mems} =
  fpr ff "(namespace %a (%a) %a)" nssym n (list usedir) udl (list nsmember) mems
*)

(*--VG
let cunit ff (tl,udl,mems) =
  fpr ff "(cunit (%a) (%a) %a)" (list tagsym) tl (list usedir) udl (list nsmember) mems
*)

let prog ff (udl,intfs,classes) =
  fpr ff "(prog (%a) (%a) %a)"      (list usedir) udl 
                                    (list intf) intfs 
                                    (list clas) classes


let dump u =
  let f = !Pref.tempfile in
  let f = if f="" then "/tmp/temp.ss" else f in
  let o = open_out f in
  prog o u;
  close_out o;
  ignore (Sys.command (Pref.homedir ^ "/pp " ^ Pref.homedir ^ " " ^ f));
  Format.fprintf !Pref.output_formatter "%s\n" (Utils.read_file f)
