module SO = SynOp
module ID = Ident
module L = List
module SL = Sortedlist
module C = Compile
open Il


(********** handling patterns *)

type env = (string * string) list

let cmp_env (x,_) (y,_) = compare x y

let merge_env = SL.merge cmp_env

let extend_env x y env = merge_env [x,y] env

let (uni : string list -> string list -> string list) = SL.merge compare

let (diff : string list -> string list -> string list) = SL.diff compare

let (inter : string list -> string list -> string list) = SL.isect compare

let pvars_atom = function
  | IPVar(x,_) -> [x]
  | IPDontCare -> []

let pvars p : string list =
  let rec patcontvars {iTag=p1; iAtt=p2; iCont=p3; iRest=p4} =
    uni (pvars_atom p1) (uni (pvars_atom p2) (uni (pvars_atom p3) (pvars_atom p4)))
  and attpatsvars pats =
    L.fold_right (fun (_,p) acc -> uni (pvars_atom p) acc) pats []
  and do_opt = function
    | None -> []
    | Some p -> pvars_atom p
  and loop = function
    | IPAnyAtom p -> pvars_atom p
    | IPPcdata pc
    | IPElement pc
    | IPObj pc
    | IPBaseType(_,pc) -> patcontvars pc
    | IPAtt(attpats,po) -> uni (attpatsvars attpats) (do_opt po)
    | IPRegex(_,_)
    | IPEmpty
    | IPDefault
    | IPBaseVal _ -> []
  in loop p

let rem_useless_var_atom p vs = match p with
  | IPVar(x,_) -> if L.memq x vs then p else IPDontCare
  | IPDontCare -> p

let rem_useless_var p vs =
  let rec patcont {iTag=p1; iAtt=p2; iCont=p3; iRest=p4} =
    {iTag=rem_useless_var_atom p1 vs; iAtt=rem_useless_var_atom p2 vs;
     iCont=rem_useless_var_atom p3 vs; iRest=rem_useless_var_atom p4 vs}
  and rem_attpats pats =
    L.map (fun (l,p) -> l, rem_useless_var_atom p vs) pats
  and do_opt = function
    | None -> None
    | Some p -> Some (rem_useless_var_atom p vs)
  and loop p = match p with
    | IPEmpty
    | IPBaseVal _
    | IPRegex(_,_)
    | IPDefault -> p
    | IPAnyAtom p -> IPAnyAtom (rem_useless_var_atom p vs)
    | IPPcdata pc -> IPPcdata (patcont pc)
    | IPElement pc -> IPElement (patcont pc)
    | IPObj pc -> IPObj (patcont pc)
    | IPBaseType(t,pc) -> IPBaseType(t,patcont pc)
    | IPAtt(attpats,po) -> IPAtt(rem_attpats attpats, do_opt po)
  in loop p

let is_non_binding_pat =
  let rec patcontvars {iTag=p1; iAtt=p2; iCont=p3; iRest=p4} =
    atompat p1 && atompat p2 && atompat p3 && atompat p4
  and atompat = function
    | IPVar(x,_) -> false
    | IPDontCare -> true
  and attpats pats =
    L.for_all (fun (_,p) -> atompat p) pats
  and opt = function
    | None -> true
    | Some p -> atompat p
  and loop = function
    | IPEmpty
    | IPBaseVal _
    | IPDefault -> true
    | IPRegex(_,isfree) -> isfree
    | IPAnyAtom p -> atompat p
    | IPPcdata pc -> patcontvars pc
    | IPElement pc -> patcontvars pc
    | IPObj pc -> patcontvars pc
    | IPBaseType(_,pc) -> patcontvars pc
    | IPAtt(pats,po) -> attpats pats && opt po
  in loop

let is_var_pat p : string = match p with
  | IPVar(x,_) -> x
  | IPDontCare -> ""

let gen_fresh_names vars env_vars env =
  let rec loop vars pairs env i = match vars with
    | [] -> pairs,env
    | x :: vars ->
        let rec loop2 i =
          let y = C.gen_numbered_y_id i in
          if L.memq y env_vars then loop2 (i+1)
          else loop vars ((x,y)::pairs) (extend_env x y env) (i+1)
        in loop2 i
  in loop vars [] env 1

let rename_env vars env =
  let env = L.fold_right L.remove_assq vars env in
  let dom_vars = L.fold_right (fun (x,_) vs -> uni [x] vs) env [] in
  let codom_vars = L.fold_right (fun (_,x) vs -> uni [x] vs) env [] in
  let env_vars = uni dom_vars codom_vars in
  let ren_vars = inter vars codom_vars in
  gen_fresh_names ren_vars env_vars env

let rename_pat_atom p env =
  let vars = pvars_atom p in
  let prs,env = rename_env vars env in
  (match p with
    | IPVar(x,t) ->
        (try let y = L.assq x prs in IPVar(y,t)
         with Not_found -> p)
    | IPDontCare -> p),
  (L.fold_right (fun x vs -> uni [try L.assq x prs with Not_found -> x] vs) vars []),
  env

let rename_pat p env =
  let vars = pvars p in
  let prs,env = rename_env vars env in
  (let rec patcont {iTag=p1; iAtt=p2; iCont=p3; iRest=p4} =
     {iTag=atompat p1; iAtt=atompat p2;
      iCont=atompat p3; iRest=atompat p4}
   and attpats pats =
     L.map (fun (l,p) -> l, atompat p) pats
   and opt = function
     | None -> None
     | Some p -> Some (atompat p)
   and atompat p = match p with
     | IPVar(x,t) ->
         (try let y = L.assq x prs in IPVar(y,t)
          with Not_found -> p)
     | IPDontCare -> p
   and loop p = match p with
     | IPEmpty
     | IPBaseVal _
     | IPRegex(_,_)
     | IPDefault -> p
     | IPAnyAtom p -> IPAnyAtom (atompat p)
     | IPPcdata pc -> IPPcdata (patcont pc)
     | IPElement pc -> IPElement (patcont pc)
     | IPObj pc -> IPObj (patcont pc)
     | IPBaseType(t,pc) -> IPBaseType(t,patcont pc)
     | IPAtt(pats,po) -> IPAtt(attpats pats, opt po)
   in loop p),
  (L.fold_right (fun x vs -> uni [try L.assq x prs with Not_found -> x] vs) vars []),
  env

let rename_vars xs env =
  let vars = L.sort compare xs in
  let prs,env = rename_env vars env in
  (L.map (fun x -> try L.assq x prs with Not_found -> x) xs),
  env


(********** inlining table and their operations *)

type handler_info = (string * typ) list * exp * env

let (_catch_inline_tbl_ : (labsym * handler_info option ref) list ref) = ref []

let (_catch_use_tbl_ : (labsym * int ref) list ref) = ref []

let _is_second_phase_ = ref false

let reset_catch_tables () =
  _is_second_phase_ := false;
  _catch_use_tbl_ := [];
  _catch_inline_tbl_ := []

let add_inlined_label l =
  _catch_inline_tbl_ := (l,ref None)::!_catch_inline_tbl_

let calc_inlinable_catches () =
  _is_second_phase_ := true;
  _catch_inline_tbl_ := [];
  L.iter (fun (l,cnt) -> if !cnt==1 then add_inlined_label l) !_catch_use_tbl_

let set_inlined_label l xs t env =
  _catch_inline_tbl_ := (l, ref (Some(xs,t,env))) :: !_catch_inline_tbl_

let reset_inlined_label l params t env =
  try let box = L.assq l !_catch_inline_tbl_ in
  box := Some(params,t,env)
  with Not_found -> ()

let is_inlined_catch l =
  try let _ = L.assq l !_catch_inline_tbl_ in true
  with Not_found -> false

let inc_use_counter l delta =
  try let box = L.assq l !_catch_use_tbl_ in
  box := !box + delta
  with Not_found -> assert false

let reset_use_counter l : unit =
  _catch_use_tbl_ := (l, ref 0)::!_catch_use_tbl_

let is_unused_catch l : bool =
  try let box = L.assq l !_catch_use_tbl_ in
  !box == 0
  with Not_found -> assert false

let count_catch_uses t delta =
  let rec do_opt e = match e with
    | None -> ()
    | Some e -> loop e
  and do_act_params params = L.iter (fun (_,e) -> loop e) params
  and do_clauses cls = L.iter (fun (_,e) -> loop e) cls
  and loop e = match e with
    | IEConst(_) -> ()
    | IERegexField(_,_,_,_) -> ()
    | IENew(_,exps) -> L.iter loop exps
    | IENewArray(_,e,_) -> loop e
    | IEArrayAccess(e1,e2) -> loop e1; loop e2
    | IELocal(_) -> ()
    | IEField(e,_) -> loop e
    | IEStaticField(_,_) -> ()
    | IECall(e,_,params) -> do_act_params params; loop e
    | IEStaticCall(_,_,params) -> do_act_params params
    | IEAssign(e1,e2) -> loop e1; loop e2
    | IEUnOp(_,e) -> loop e
    | IEBinOp(_,e1,e2) -> loop e1; loop e2
    | IECast(_,e) -> loop e
    | IETypeStamp(_,e) -> loop e
    | IELet(_,e1,e2) -> loop e2; do_opt e1
    | IESeq(e1,e2) -> loop e1; loop e2
    | IEReturn(e) -> do_opt e
    | IEIf(e1,e2,e3) -> loop e1; loop e2; loop e3
    | IEProcCall(_,_,params,e1,e2) -> loop e1; do_opt e2
    | IEWhile(e1,e2) -> loop e1; loop e2
    | IESwitchK(e,cls)
    | IESwitchV(e,cls)
    | IESwitchO(e,cls)
    | IESwitchP(e,cls)
    | IESwitchA(e,cls) -> loop e; do_clauses cls
    | IEBreak -> ()
    | IECatch(e1,_,params,e2) -> loop e1; loop e2
    | IEExit(l,exps) ->
        inc_use_counter l delta;
        L.iter loop exps
    | IEEmpty -> ()
    | IEError _ -> ()
  in loop t

let is_second_phase () = !_is_second_phase_


(********** simplifying terms *)

type switch_kind =
  | IESwKK
  | IESwKV
  | IESwKO
  | IESwKA
  | IESwKP

let bounded_equal_exp e1 e2 n =
  let rec do_list exps1 exps2 n = match exps1,exps2 with
    | [],[] -> n
    | e::exps,x::xps -> loop e x (do_list exps xps n)
    | _ -> raise Not_found
  and do_opt e1 e2 n = match e1,e2 with
    | None,None -> n
    | Some e,Some x -> loop e x n
    | _ -> raise Not_found
  and do_act_params params1 params2 n = match params1,params2 with
    | [],[] -> n
    | (_,e)::params1,(_,x)::params2 -> loop e x (do_act_params params1 params2 n)
    | _ -> raise Not_found
  and loop e1 e2 n =
    if n <= 0 then raise Not_found
    else match e1,e2 with
      | IEConst(x),IEConst(y) when x=y ->
          n-1
      | IENew(x,exps1),IENew(y,exps2) when x=y ->
          do_list exps1 exps2 (n-1)
      | IENewArray(x1,e1,x2),IENewArray(y1,e2,y2) when x1=y1 && x2=y2 ->
          loop e1 e2 (n-1)
      | IEArrayAccess(e1,e2),IEArrayAccess(x1,x2) ->
          do_list [e1;e2] [x1;x2] (n-1)
      | IELocal(x),IELocal(y) when x=y ->
          n-1
      | IEField(e1,f),IEField(e2,g) when f=g ->
          loop e1 e2 (n-1)
      | IEStaticField(x1,x2),IEStaticField(y1,y2) when x1=y1 && x2=y2 ->
          n-1
      | IECall(e1,x,params1),IECall(e2,y,params2) when x=y ->
          do_act_params params1 params2 (loop e1 e2 (n-1))
      | IEStaticCall(x1,x2,params1),IEStaticCall(y1,y2,params2) when x1=y1 && x2=y2 ->
          do_act_params params1 params2 (n-1)
      | IEAssign(e1,e2),IEAssign(x1,x2) ->
          do_list [e1;e2] [x1;x2] (n-1)
      | IEUnOp(x,e1),IEUnOp(y,e2) when x=y ->
          loop e1 e2 (n-1)
      | IEBinOp(x,e1,e2),IEBinOp(y,x1,x2) when x=y ->
          do_list [e1;e2] [x1;x2] (n-1)
      | IECast(x,e1),IECast(y,e2) when x=y ->
          loop e1 e2 (n-1)
      | IETypeStamp(x,e1),IETypeStamp(y,e2) when x = y ->
          loop e1 e2 (n-1)
      | IELet(x,e1,e2),IELet(y,x1,x2) when x=y ->
          loop e2 x2 (do_opt e1 x1 (n-1))
      | IESeq(e1,e2),IESeq(x1,x2) ->
          do_list [e1;e2] [x1;x2] (n-1)
      | IEReturn(e1),IEReturn(e2) ->
          do_opt e1 e2 (n-1)
      | IEIf(e1,e2,e3),IEIf(x1,x2,x3) ->
          loop e1 x1 (loop e2 x2 (loop e3 x3 (n-1)))
      | IEProcCall(x,y,pms1,e1,e2),IEProcCall(z,w,pms2,x1,x2)
        when x=z && y=w && pms1=pms2 ->
          loop e1 x1 (do_opt e2 x2 (n - L.length pms1 - 2))
      | IEWhile(e1,e2),IEWhile(x1,x2) ->
          do_list [e1;e2] [x1;x2] (n-1)
      | IESwitchK(e1,cls1),IESwitchK(e2,cls2)
      | IESwitchV(e1,cls1),IESwitchV(e2,cls2)
      | IESwitchO(e1,cls1),IESwitchO(e2,cls2)
      | IESwitchP(e1,cls1),IESwitchP(e2,cls2)
      | IESwitchA(e1,cls1),IESwitchA(e2,cls2) when L.map fst cls1 = L.map fst cls2 ->
          do_list (e1 :: L.map snd cls1) (e2 :: L.map snd cls2) (n-1)
      | IEBreak,IEBreak ->
          n-1
      | IECatch(e1,x,pms1,e2),IECatch(x1,y,pms2,x2) when x=y && pms1=pms2 ->
          loop e1 x1 (loop e2 x2 (n - L.length pms1 - 1))
      | IEExit(x,exps1),IEExit(y,exps2) when x=y ->
          do_list exps1 exps2 (n-1)
      | IEEmpty,IEEmpty ->
          n-1
      | IEError x,IEError y when x=y ->
          n-1
      | _ -> raise Not_found
  in try ignore (loop e1 e2 n); true
  with Not_found -> false

let is_small_size e n =
  let rec do_list exps n = match exps with
    | [] -> n
    | e::exps -> loop e (do_list exps n)
  and do_opt e n = match e with
    | None -> n
    | Some e -> loop e n
  and do_act_params params n = match params with
    | [] -> n
    | (_,e)::params -> loop e (do_act_params params n)
  and loop e n =
    if n < 0 then raise Not_found
    else match e with
      | IEConst(_) -> n-1
      | IERegexField(_,_,e1,e2) -> do_list [e1;e2] (n-1)
      | IENew(_,exps) -> do_list exps (n-1)
      | IENewArray(_,e,_) -> loop e (n-1)
      | IEArrayAccess(e1,e2) -> do_list [e1;e2] (n-1)
      | IELocal(_) -> n-1
      | IEField(e,_) -> loop e (n-1)
      | IEStaticField(_,_) -> n-1
      | IECall(e,_,params) -> do_act_params params (loop e (n-1))
      | IEStaticCall(_,_,params) -> do_act_params params (n-1)
      | IEAssign(e1,e2) -> do_list [e1;e2] (n-1)
      | IEUnOp(_,e) -> loop e (n-1)
      | IEBinOp(_,e1,e2) -> do_list [e1;e2] (n-1)
      | IECast(_,e) -> loop e (n-1)
      | IETypeStamp(_,e) -> loop e (n-1)
      | IELet(_,e1,e2) -> loop e2 (do_opt e1 (n-1))
      | IESeq(e1,e2) -> do_list [e1;e2] (n-1)
      | IEReturn(e) -> do_opt e (n-1)
      | IEIf(e1,e2,e3) -> loop e1 (loop e2 (loop e3 (n-1)))
      | IEProcCall(_,_,params,e1,e2) -> loop e1 (do_opt e2 (n - L.length params - 2))
      | IEWhile(e1,e2) -> do_list [e1;e2] (n-1)
      | IESwitchK(e,cls)
      | IESwitchV(e,cls)
      | IESwitchO(e,cls)
      | IESwitchP(e,cls)
      | IESwitchA(e,cls) -> do_list (e :: L.map snd cls) (n-1)
      | IEBreak -> n-1
      | IECatch(e1,_,params,e2) -> loop e1 (loop e2 (n - L.length params - 1))
      | IEExit(_,exps) -> do_list exps (n-1)
      | IEEmpty -> n-1
      | IEError _ -> n-1
  in try ignore (loop e n); true
  with Not_found -> false

let is_var_exp (e : exp option) : string = match e with
  | Some (IELocal x) -> x
  | _ -> ""

let is_same_pat_exp (p:atompat) (e:exp) = match (p,e) with
  | IPVar(x1,_), IELocal(x2) -> (x1 == x2)
  | _ -> false

let rec do_opt t env = match t with
  | None -> None, [] 
  | Some t ->
      let t,vs = simp t env in
      Some t, vs

and do_let bnds body env bodyenv =
  let scan_small_bnds bnds =
    let rec loop bnds env2 bs = match bnds with
      | [] -> bs,env2
      | (p,t)::rest ->
          let t,tvars = do_opt t env in
          let x = is_var_pat p in
          let y = is_var_exp t in
          if (x<>"") && (y<>"") then
            if x==y then loop rest (L.remove_assq x env2) bs
            else loop rest (extend_env x y (L.remove_assq x env2)) bs
          else loop rest env2 ((p,t,[],tvars)::bs)
    in loop bnds [] [] in
  let scan_big_binds bnds env =
    let rec loop bnds env2 vars bs = match bnds with
      | [] -> bs,vars,env2
      | (p,t,_,tvars) :: bnds ->
          let p,pvars,newenv = rename_pat_atom p env in
          loop bnds (merge_env newenv env2) (uni pvars vars) ((p,t,pvars,tvars)::bs)
    in loop bnds [] [] [] in
  let is_small_binding = function
    | None -> true
    | Some(t) ->
        (match t with
          | IEConst(_)
          | IELocal(_)
          | IEStaticField(_,_)
          | IERegexField(_,_,_,_) -> true
          | _ -> false) in
  let recons_let bs0 body vars =
    let rec loop bs pvars tvars bnds = match bs with
      | [] -> (match bnds with
          | [] -> body,vars
          | _ -> (L.fold_right (fun (p,t) e -> IELet(p,t,e)) bnds body,
                  uni tvars (diff vars pvars)))
      | (p,t,pvs,tvs) :: bs ->
          let p = rem_useless_var_atom p vars in
          if p==IPDontCare && is_small_binding t then loop bs pvars tvars bnds
          else loop bs (uni pvars pvs) (uni tvars tvs) ((p,t)::bnds)
    in loop bs0 [] [] [] in
  let bs,env = scan_small_bnds bnds in
     let bs,vars,newenv = scan_big_binds bs bodyenv in
     let bodyenv =
       merge_env newenv (merge_env env (L.fold_right L.remove_assq vars bodyenv)) in
     let body,xs = simp body bodyenv in
     recons_let bs body xs

and do_list ts env =
  let rec loop = function
    | [] -> [], []
    | t::ts ->
        let t, vs1 = simp t env in
        let ts, vs2 = loop ts in
        t::ts, uni vs1 vs2
  in loop ts

and do_proc_call m x params e eo env =
  let x = try L.assq x env with Not_found -> x in
  let vars,tys = L.split params in
  let vars,env = rename_vars vars env in
  let eo,xs1 = do_opt eo env in
  let e,xs2 = simp e env in
  match (e,eo,params) with
    | IEReturn (Some (IEConst (ICBool true))),
      Some (IEReturn (Some (IEConst (ICBool false)))),
      [] ->
        IEReturn (Some (IEStaticCall("Matching", m, [false, IELocal x]))), [x]
    | _ -> 
        let params = L.combine vars tys in
        IEProcCall(m,x,params,e,eo), uni [x] (diff (uni xs1 xs2) vars)

and do_clauses cls env = match cls with
  | [] -> assert false
  | _ ->
      let rec loop cls res vars = match cls with
        | [] -> L.rev res, vars
        | (p,t)::cls ->
            (match p with
              | IPAnyAtom(_) ->
                  ()
              | _ -> ());
            let p,pvars,env = rename_pat p env in
            let t,tvars = simp t env in
            let p = rem_useless_var p tvars in
            loop cls ((p,t)::res) (uni (diff tvars pvars) vars)
      in loop cls [] []

and do_switch sw term clauses env =
  let exp,vs1 = simp term env in
  let cls,vs2 = do_clauses clauses env in
  let is_one_non_bind cls = match cls with
    | [p,_] -> is_non_binding_pat p
    | _ -> false in
  let all_non_bind_same cls =
    let pats,exps = L.split cls in
    let e = L.hd exps in
    L.for_all is_non_binding_pat pats &&
    L.for_all (fun x -> bounded_equal_exp x e !Pref.traverse_exp_depth) (L.tl exps) in
  let last_clause (cls : 'a list) : 'a list * 'a =
    let rec loop cls pre = match cls with
      | [] -> assert false
      | [cl] -> L.rev pre, cl
      | cl::cls -> loop cls (cl::pre)
    in loop cls [] in
  let like_any_atom_pat p : atompat option = match p with
    | IPPcdata {iTag=IPDontCare; iAtt=IPDontCare; iCont=IPDontCare; iRest=p} -> Some p
    | IPElement {iTag=IPDontCare; iAtt=IPDontCare; iCont=IPDontCare; iRest=p} -> Some p
    | IPObj {iTag=IPDontCare; iAtt=IPDontCare; iCont=IPDontCare; iRest=p} -> Some p
    | IPBaseType(_,{iTag=IPDontCare; iAtt=IPDontCare; iCont=IPDontCare; iRest=p}) -> Some p
    | _ -> None in
  let drop_after_any_clause cls =
    let rec loop = function
      | [] -> []
      | ((IPAtt([],Some _), e) as cl) :: _ -> [cl]
      | cl::cls -> cl :: loop cls
    in loop cls
  in if is_one_non_bind cls then
    snd (L.hd cls), vs2
  else if all_non_bind_same cls then
    let t=snd (L.hd cls) in
    count_catch_uses t (1 - L.length cls);
    t,vs2
  else
    let cls = drop_after_any_clause cls in
    let precls,cl = last_clause cls in
    let vs = uni vs1 vs2 in
    let cls =
      if is_non_binding_pat (fst cl) then precls @ [IPDefault,snd cl]
      else match like_any_atom_pat (fst cl) with
        | None -> cls
        | Some p -> precls @ [IPAnyAtom p, snd cl]
    in match sw with
      | IESwKK -> IESwitchK(exp, cls), vs
      | IESwKV -> IESwitchV(exp, cls), vs
      | IESwKO -> IESwitchO(exp, cls), vs
      | IESwKA -> IESwitchA(exp, cls), vs
      | IESwKP -> IESwitchP(exp, cls), vs

and do_catch t1 l params t2 env =
  let xs,tys = L.split params in
  let xs,env2 = rename_vars xs env in
  let params = L.combine xs tys in
  let helper t2 vs2 =
    reset_use_counter l;
    reset_inlined_label l params t2 env2;
    let t1,vs1 = simp t1 env in
    if is_unused_catch l then
      (if vs2 != None then count_catch_uses t2 (-1);
       t1,vs1)
    else
      let xs = L.sort compare xs in
      match vs2 with
        | None -> assert false (* impossible since inlined_catch implies unused_catch *)
        | Some vs2 -> IECatch(t1,l,params,t2), uni vs1 (diff vs2 xs)
  in if is_second_phase () && is_inlined_catch l then
    helper t2 None
  else
    let t2,vs2 = simp t2 env2 in
    if is_small_size t2 (1 + L.length params) then
      (count_catch_uses t2 (-1);
       set_inlined_label l params t2 env2;
       simp t1 env)
    else
      helper t2 (Some vs2)

and inline_label l ts env1 =
  try let box = L.assq l !_catch_inline_tbl_ in
      match !box with
        | None -> assert false
        | Some(params,hndl,env2) ->
            Some (do_let (L.map2 (fun (x,ty) t -> IPVar(x,ty), Some t) params ts) hndl
                    env1 env2)
  with Not_found -> None

and do_exit l ts env =
  if l="k_23" then
    (); (* ,,, *)
  match inline_label l ts env with
    | None ->
        let ts,vs = do_list ts env in
        inc_use_counter l 1;
        IEExit(l,ts), vs
    | Some(e,xs) -> e,xs

and do_act_params params env =
  let rec loop = function
    | [] -> [], []
    | (flag,t)::ts ->
        let t, vs1 = simp t env in
        let ts, vs2 = loop ts in
        (flag,t)::ts, uni vs1 vs2
  in loop params

and do_seq e1 e2 env =
  let e1,xs1 = simp e1 env in
  match e1 with
    | IEReturn(_) -> e1,xs1
    | _ ->
        let e2,xs2 = simp e2 env in
        IESeq(e1,e2), uni xs1 xs2

and do_if e1 e2 e3 env =
  let e1,xs1 = simp e1 env in
  let e2,xs2 = simp e2 env in
  let e3,xs3 = simp e3 env in
  match (e2,e3) with
    | IEReturn (Some (IEConst (ICBool true))),
      IEReturn (Some (IEConst (ICBool false))) ->
        e1, xs1
    | _ -> IEIf(e1,e2,e3), uni xs1 (uni xs2 xs3)

and simp t env = match t with
  | IEConst(_) -> t, []
  | IENew(t,es) ->
      let es,xs = do_list es env in
      IENew(t,es), xs
  | IENewArray(t,e,i) ->
      let e,xs = simp e env in
      IENewArray(t,e,i), xs
  | IEArrayAccess(e1,e2) ->
      let e1,xs1 = simp e1 env in
      let e2,xs2 = simp e2 env in
      IEArrayAccess(e1,e2), uni xs1 xs2
  | IELocal(x) ->
      (try
         let y = L.assq x env in
         IELocal(y), [y]
       with Not_found ->
         t,[x])
  | IEField(e,f) ->
      let e,xs = simp e env in
      IEField(e,f), xs
  | IEStaticField(_,_) -> t, []
  | IECall(e,m,params) ->
      let e,xs1 = simp e env in
      let params,xs2 = do_act_params params env in
      IECall(e,m,params), uni xs1 xs2
  | IEStaticCall(t,m,params) ->
      let params,xs = do_act_params params env in
      IEStaticCall(t,m,params), xs
  | IEAssign(e1,e2) ->
      let e1,xs1 = simp e1 env in
      let e2,xs2 = simp e2 env in
      IEAssign(e1,e2), uni xs1 xs2
  | IEUnOp(op,e) ->
      let e,xs = simp e env in
      IEUnOp(op,e), xs
  | IEBinOp(op,e1,e2) ->
      let e1,xs1 = simp e1 env in
      let e2,xs2 = simp e2 env in
      IEBinOp(op,e1,e2), uni xs1 xs2
  | IECast(t,e) ->
      let e,xs = simp e env in
      IECast(t,e), xs
  | IETypeStamp(x,e) ->
      let e,xs = simp e env in
      IETypeStamp(x,e), xs
  | IELet(p,eo,e) ->
      do_let [p,eo] e env env
  | IESeq(e1,e2) ->
      do_seq e1 e2 env
  | IEReturn(e) ->
      let e,xs = do_opt e env in
      IEReturn(e),xs
  | IEIf(e1,e2,e3) ->
      do_if e1 e2 e3 env
  | IEProcCall(m,x,params,e,eo) ->
      do_proc_call m x params e eo env
  | IEWhile(e1,e2) ->
      let e1,xs1 = simp e1 env in
      let e2,xs2 = simp e2 env in
      IEWhile(e1,e2), uni xs1 xs2
  | IESwitchK(e,cls) -> do_switch IESwKK e cls env
  | IESwitchV(e,cls) -> do_switch IESwKV e cls env
  | IESwitchO(e,cls) -> do_switch IESwKO e cls env
  | IESwitchA(e,cls) -> do_switch IESwKA e cls env
  | IESwitchP(e,cls) -> do_switch IESwKP e cls env
  | IEBreak -> t, []
  | IECatch(e,k,params,e2) ->
      do_catch e k params e2 env
  | IEExit(k,es) ->
      do_exit k es env
  | IEEmpty -> t, []
  | IEError(_) -> t, []
  | IERegexField(r,f,e1,e2) ->
      let e1,xs1 = simp e1 env in
      let e2,xs2 = simp e2 env in
      IERegexField(r,f,e1,e2), uni xs1 xs2


(********** simplifying modules *)

let catches e =
  reset_catch_tables ();
  let (e,_) = simp e [] in
  calc_inlinable_catches ();
  simp e []

let meth_decl ((n:string),stat,mods,params,retty,e) =
  if n="do_techreport" then
    (); (* ,,, *)
  (n,stat,mods,params,retty,fst (catches e))

let clmember = function
  | ICMConstr(i,params,e) -> ICMConstr(i,params,fst (catches e))
  | ICMStaticConstr(e) -> ICMStaticConstr(fst (catches e))
  (* maybe optimize initialization of field ... *)
  | ICMFld(_,_,_,_) as f -> f
  | ICMMeth(d) -> ICMMeth(meth_decl d)

let clas cl = 
  { cl with iCLMems = L.map clmember cl.iCLMems }
 
let prog (usedirs,intfs,classes) = (usedirs, intfs, L.map clas classes)
