module SS = SynSource
module SD = SynDenot

(*##ATT  Otherwise, this module is independent of Kerids *)
let ref_pcdata = SD.seqref Kerids.pcdata_id
(*-- let ref_anyseq = SD.seqref Kerids.any_id *)
let ref_attseq = SD.seqref Kerids.attseq_id

exception ElaborationError of Finfo.t * string


(***************** linearity checkers ************************) 

(* A linearity checker collects a list of bound variables of a pattern
(as a list of strings), and checks for the linearity conditions along
the way. *)
(*  lin_XXXty: SS.XXXty -> string list *)

(*## cf. Xtatic/xduce/src/wf.ml *)

let str_of_vars (vars: string list) = 
  List.fold_left (fun str v -> str ^ " " ^ v) "" vars
    

let lin_attpat (att: SS.attpat) : string= 
  match Finfo.drop att with 
  | (aname, (_, v)) -> v

let lin_attsty atts : string list = 
  match atts with 
  | (inf, SS.TAttAny) -> []
  | (inf, SS.PAtt(attpats, catchopt)) ->
      let pairvars = 
	List.fold_left (fun vars attpat -> 
	  let v = lin_attpat attpat in
	  if Sortedlist.mem compare v vars
	  then raise (ElaborationError(inf, "Variable '" ^ v ^ 
          		    "' is used multiple times in attribute patern."))
	  else Sortedlist.merge compare [v] vars)
	  [] attpats
      in match catchopt with 
      | None -> pairvars
      | Some (_, v) -> 
	  if Sortedlist.mem compare v pairvars
	  then raise (ElaborationError(inf, "Catch-all variable '" ^ v ^ 
          	      "' is already used in attribute patern."))
	  else Sortedlist.merge compare [v] pairvars

let lin_hostty tt = []  

let rec lin_labty ll = 
  let (inf, ll') = ll in 
  match ll' with 
  | SS.TLab h -> lin_hostty h
  | SS.PLabBind((_, v), lty) -> 
      let vars = lin_labty lty in 
      if Sortedlist.mem compare v vars
      then raise (ElaborationError(inf, "Variable '" ^ v ^ 
          		"' is used multiple times in label patern."))
      else Sortedlist.merge compare [v] vars		    

let rec lin_seqty sty = 
  let (inf, sty') = sty in 
  match sty' with 
  | SS.TTree(l, a, s) -> 
      let lvars = lin_labty l in 
      let avars = lin_attsty a in 
      let svars = lin_seqty s in 
      let isect_la = Sortedlist.isect compare lvars avars in 
      let isect_ls = Sortedlist.isect compare lvars svars in 
      let isect_as = Sortedlist.isect compare avars svars in 
      if (isect_la, isect_ls, isect_as) = ([], [], [])
      then Sortedlist.merge compare 
	  lvars (Sortedlist.merge compare avars svars)
      else
	let dupls = Sortedlist.merge compare 
	  isect_la (Sortedlist.merge compare isect_ls isect_as) in 
	raise (ElaborationError(inf, 
	    "The following variables are bound multiple times in pattern: " ^ 
		(str_of_vars dupls)))
  | SS.TCat ss -> 
      List.fold_left (fun vars s -> 
	let vs = lin_seqty s in 
	let dupls = Sortedlist.isect compare vs vars in 
	if dupls = [] 
	then Sortedlist.merge compare vs vars
	else raise (ElaborationError
		      (inf, "The following variables are bound " ^ 
		       "multiple times in pattern: " ^ (str_of_vars dupls)))
		     ) [] ss
  | SS.TAlt [] -> []
  | SS.TAlt (alt1::alts) -> 
      let expected_vars = lin_seqty alt1 in 
      List.iter (fun alt -> 
	let vs = lin_seqty alt in 
	if Sortedlist.compare compare vs expected_vars <> 0
	then raise (ElaborationError
		      (inf, "Not all branches of pattern bind the same vars"))
		) alts;
      expected_vars
  | SS.TRep s -> 
      let vars = lin_seqty s in 
      if vars = [] 
      then [] 
      else raise (ElaborationError
		    (inf, "Bound variables in a repetition (under *): " ^ 
		     (str_of_vars vars) ))
  | SS.TSeqRef _ -> []
  | SS.PSeqBind((_, v), s) -> 
      let vars = lin_seqty s in 
      if Sortedlist.mem compare v vars
      then raise (ElaborationError(inf, "Variable '" ^ v ^ 
          			   "' is used multiple times in patern."))
      else Sortedlist.merge compare [v] vars

let rec lin_ty ty = 
  let (inf, ty') = ty in 
  match ty' with 
  | SS.Thost h -> lin_hostty h 
  | SS.Tseq s -> lin_seqty s
  | SS.PTyBind((_, v), t) -> 
      let vars = lin_ty t in 
      if Sortedlist.mem compare v vars
      then raise (ElaborationError(inf, "Variable '" ^ v ^ 
          		"' is used multiple times in patern."))
      else Sortedlist.merge compare [v] vars		    
     

(********************* SS -> SD transformations ***************)

let s2d_attpat nctx vctx (att: SS.attpat) = 
  match Finfo.drop att with
  | (aname, dn) -> 
      (Finfo.drop aname, 
       List.assoc (Finfo.drop dn) vctx,
       ref_pcdata)

let s2d_attsty nctx vctx (atts: SS.attsty) : SD.attsty = 
  match Finfo.drop atts with 
  | SS.TAttAny -> SD.attsty_any
  | SS.PAtt(s_attpairs, s_catchopt) ->  
      let d_attpairs = List.map (s2d_attpat nctx vctx) s_attpairs in
      let d_catchopt = match s_catchopt with
      | Some dn -> Some (List.assoc (Finfo.drop dn) vctx,  ref_attseq)
      | None -> None
      in SD.attsty(SD.PAtt(d_attpairs, d_catchopt))

let rec s2d_hostty nctx vctx tt = 
  SD.hostty (match Finfo.drop tt with
  | SS.TCSharp rn -> 
     (match Naming.resolve_leaf nctx rn with 
	| (id, Cnaming.LClas) -> SD.TClasRef id
	| (id, Cnaming.LIntf) -> SD.TIntfRef id  )
  | SS.TNull -> SD.TNull
  | SS.TBool -> SD.TBool
  | SS.Tbool b -> SD.Tbool b
  | SS.TInt -> SD.TInt
  | SS.Tint n -> SD.Tint n
  | SS.TString -> SD.TString
  | SS.Tstring s -> SD.Tstring s
  | SS.TChar -> SD.TChar
  | SS.Tchar c -> SD.Tchar c 
  | SS.TTag -> SD.TTag
  | SS.Ttag t -> SD.Ttag t
  | SS.TSeq -> SD.TSeq
  | SS.TArray t -> SD.TArray (s2d_ty nctx vctx t)
	    )

and s2d_labty nctx vctx (ll: SS.labty) = 
  SD.labty (match Finfo.drop ll with 
  | SS.TLab hty -> SD.TLab (s2d_hostty nctx vctx hty)
  | SS.PLabBind (dname, lty) -> 
      let vid =  List.assoc (Finfo.drop dname) vctx in 
      SD.PLabBind(vid, s2d_labty nctx vctx lty)
	   )

and s2d_seqty nctx vctx (ss: SS.seqty) = 
  (match Finfo.drop ss with
  | SS.TTree(l, ats, s) -> 
      SD.tree (s2d_labty nctx vctx l) (s2d_attsty nctx vctx ats) (s2d_seqty nctx vctx s)
  | SS.TCat sls -> SD.cat (List.map (s2d_seqty nctx vctx) sls)
  | SS.TAlt sls -> SD.alt (List.map (s2d_seqty nctx vctx) sls)
  | SS.TRep s -> SD.rep (s2d_seqty nctx vctx s)
  | SS.TSeqRef qn -> SD.seqref (Naming.resolve_seqty nctx qn)
  | SS.PSeqBind(dname, s) -> 
      let vid = List.assoc (Finfo.drop dname) vctx in 
      SD.seqbind vid (s2d_seqty nctx vctx s)
  )

and s2d_ty nctx vctx (tt: SS.ty) : SD.ty = 
  SD.ty (match Finfo.drop tt with 
  | SS.Thost hty ->  SD.Thost (s2d_hostty nctx vctx hty)
  | SS.Tseq s -> SD.Tseq (s2d_seqty nctx vctx s)
  | SS.PTyBind (dname, t) -> 
      let vid = List.assoc (Finfo.drop dname) vctx in 
      SD.PTyBind (vid,  s2d_ty nctx vctx t)
	)


(******************* exported functions *************************)

let seqty nctx (ss: SS.seqty) : SD.seqty = 
  let vars = lin_seqty ss in
  let vctx = List.map (fun v -> (v, Ident.create v)) vars in
  s2d_seqty nctx vctx ss

let ty nctx (ss: SS.ty) : SD.ty = 
  let vars = lin_ty ss in
  let vctx = List.map (fun v -> (v, Ident.create v)) vars in
  s2d_ty nctx vctx ss


