
(*************************************************************************
                          Host types 1/2
**************************************************************************)
module C = Contexts 
module SD = SynDenot
module R = Result



let rec sub_class clsid1 clsid2 = 
(*
  Printf.printf "testing if class %s is sub of class %s\n" 
    (Ident.toString clsid1) (Ident.toString clsid2);
    flush stdout;
*)
  (* can be augmented by result caching *)
  if clsid1 == clsid2 then true
  else if clsid1 == !Store.topclass then false
  else sub_class (Ident.find_same !Store.parent clsid1) clsid2

let rec sub_intf intfid1 intfid2 =
  (* we definitely want to cache this info*)
  (* BEWARE: MAY LOOP if we do not check before that there is no circular
     inheritance among interfaces
     This is now done when building the data context
     *)
  if intfid1 == intfid2 then true
  else 
    let implements = (Ident.find_same !Store.implements intfid1) in
    (* this works since List.exists return false for the empty list *)
    List.exists (fun intf -> sub_intf intf intfid2) implements

let rec sub_clas_intf clsid intfid =
(*
  Printf.printf "testing if class %s is sub of interface %s\n" 
    (Ident.toString clsid) (Ident.toString intfid);
  flush stdout;
*)
  (* we definitely want to cache this info*)
  let implements = (Ident.find_same !Store.implements clsid) in
  if List.exists (fun intf -> sub_intf intf intfid) implements then
    true
  else
  sub_clas_intf (Ident.find_same !Store.parent clsid) intfid


let rec join_class cid1 cid2 = 
  let topclass = !Store.topclass in
  let parent id = Ident.find_same !Store.parent id in 
  if (cid1 == topclass || cid2 == topclass) then topclass
  else if cid1 == cid2 then cid1 
  else if sub_class cid1 cid2 then cid2 
  else if sub_class cid2 cid1 then cid1
  else join_class (parent cid1) (parent cid2)


let join_hostty hty1 hty2 = 
  let t1 = SD.hostty' hty1 in
  let t2 = SD.hostty' hty2 in
  SD.hostty (match (t1, t2) with
  | _, SD.TObject  |  SD.TObject, _ -> SD.TObject

  | SD.TNull, _ -> t2
  | _, SD.TNull -> t1

  | SD.TClasRef c1, SD.TClasRef c2 -> SD.TClasRef(join_class c1 c2)
  | SD.TClasRef _, _  |  _, SD.TClasRef _ -> SD.TObject

  | SD.TSeq, SD.TSeq -> SD.TSeq

  | SD.Tbool x1, SD.Tbool x2 -> if x1=x2 then SD.Tbool x1 else SD.TBool
  | SD.TBool, SD.TBool  
  | SD.Tbool _, SD.TBool |  SD.TBool, SD.Tbool _ -> SD.TBool
 
  | SD.Tint x1, SD.Tint x2 -> if x1=x2 then SD.Tint x1 else SD.TInt
  | SD.TInt, SD.TInt 
  | SD.Tint _, SD.TInt |  SD.TInt, SD.Tint _ -> SD.TInt
 
  | SD.Tchar x1, SD.Tchar x2 -> if x1=x2 then SD.Tchar x1 else SD.TChar
  | SD.TChar, SD.TChar
  | SD.Tchar _, SD.TChar |  SD.TChar, SD.Tchar _ -> SD.TChar
 
  | SD.Tstring x1, SD.Tstring x2 -> if x1=x2 then SD.Tstring x1 else SD.TString
  | SD.TString, SD.TString 
  | SD.Tstring _, SD.TString |  SD.TString, SD.Tstring _ -> SD.TString
 
  | SD.Ttag x1, SD.Ttag x2 -> if x1=x2 then SD.Ttag x1 else SD.TTag
  | SD.TTag, SD.TTag 
  | SD.Ttag _, SD.TTag |  SD.TTag, SD.Ttag _ -> SD.TTag

  | _, _ -> SD.TObject
	    )


(**************************************************************************
                           Automata 1/2
**************************************************************************)

open Format


let rec hashl xs =
  match xs with
    [] -> 0
  | x :: xs' -> x.Ta.num + 239 * hashl xs'

let hash_hyp xs1 xs2 = (hashl xs1 + 17 * hashl xs2) land 0x3FFFFFFF


(*
   coalesce_label: f(A, B) | g(A, B) = (f | g) (A, B)
   coalesce_first: f(A, B) | f(A', B) = f(A | A', B)
   coalesce_second: f(A, B) | f(A, B') = f(A, B | B')
*)


(*VG ld_union below cannot be simply changed to ld_maximum; the whole
  algorithm needs to be reconsidered into something like this:

   - sort the enties as now, but also split the whole list into groups
     with equal state pairs;

   - for labels of each group, find maximal labels (there can be more
     than one "local maximum" for the group), and use them as the
     coalesced labels.

   For the latter, NEED TO GENERALIZE ld_maximum to become ld_maximums.
*)

let coalesce_first xsl =
  let xsl =
    Sort.list 
      (fun (c, _, xs2) (d, _, ys2) -> 
	SD.compare_labty c d < 0 &&
	Ta.compare_state_sets xs2 ys2 < 0) xsl in
  List.fold_right
    (fun (c, xs1, xs2) rem ->
      match rem with
      |	(d, ys1, ys2) :: rem' 
	when SD.compare_labty c d = 0 && 
	  Ta.compare_state_sets xs2 ys2 = 0 ->
	    (c, Ta.union_state_sets xs1 ys1, xs2) :: rem'
      | _ ->
          (c, xs1, xs2) :: rem)
    xsl []

let coalesce_second xsl =
  let xsl =
    Sort.list 
      (fun (c, xs1, _) (d, ys1, _) -> 
	SD.compare_labty c d < 0 &&
	Ta.compare_state_sets xs1 ys1 < 0) xsl in
  List.fold_right
    (fun (c, xs1, xs2) rem ->
      match rem with
      |	(d, ys1, ys2) :: rem' 
	when SD.compare_labty c d = 0 && 
	  Ta.compare_state_sets xs1 ys1 = 0 ->
	    (c, xs1, Ta.union_state_sets xs2 ys2) :: rem'
      | _ ->
          (c, xs1, xs2) :: rem)
    xsl []




let rec for_all f = function
    [] -> Result.Succ
  | e :: l -> (match f e with 
    | Result.Succ -> for_all f l 
    | Result.Fail w -> Result.Fail w)



(*************************************************************************
                          Host types 2/2
**************************************************************************)

let rec sub_hostty (hty1: SD.hostty) (hty2 : SD.hostty) 
  : Typath.t R.t = 
  let t1 = SD.hostty' hty1 in
  let t2 = SD.hostty' hty2 in
  match (t1, t2) with
  | _, SD.TObject -> R.Succ
  | (SD.TObject | SD.TBool | SD.Tbool _ | SD.TInt | SD.Tint _ | SD.TString |
     SD.Tstring _ | SD.TChar | SD.Tchar _ | SD.TTag | SD.Ttag _ | SD.TSeq),
     SD.TIntfRef _ -> R.Fail(`Host(hty1, hty2)) 
  | SD.TIntfRef id1, SD.TIntfRef id2 -> if sub_intf id1 id2 then R.Succ else R.Fail(`Host(hty1, hty2))
(*--VG subsumed by above  | SD.TIntfRef _, SD.TObject -> R.Succ  *)
  | SD.TIntfRef _, _ -> R.Fail(`Host(hty1,hty2))
  | SD.TClasRef id1, SD.TIntfRef id2 -> if sub_clas_intf id1 id2 then R.Succ else R.Fail(`Host(hty1, hty2))
  | SD.TArray a1, SD.TArray a2 -> sub_ty a1 a2
  | SD.TArray _, _ -> R.Fail(`Host(hty1, hty2))
  | _, SD.TArray _ -> R.Fail(`Host(hty1, hty2))
  | SD.TObject, _ -> R.Fail(`Host(hty1, hty2))

  | SD.TNull, _ -> R.Succ
  | _, SD.TNull -> R.Fail(`Host(hty1, hty2))

  | SD.TClasRef id1, SD.TClasRef id2 -> 
      if sub_class id1 id2 then R.Succ else R.Fail(`Host(hty1,hty2))
  | SD.TClasRef _, _  -> R.Fail(`Host(hty1, hty2))
  | _, SD.TClasRef _  -> R.Fail(`Host(hty1, hty2)) 

  | SD.Tbool x1, SD.Tbool x2 -> 
      if x1=x2 then R.Succ else R.Fail(`Host(hty1, hty2))
  | SD.Tbool _, SD.TBool -> R.Succ
  | SD.TBool, SD.TBool -> R.Succ
  | SD.Tbool _, _ |  _, SD.Tbool _ -> R.Fail(`Host(hty1, hty2))
  | SD.TBool, _    |  _, SD.TBool -> R.Fail(`Host(hty1, hty2))

  | SD.Tint x1, SD.Tint x2 -> 
      if x1=x2 then R.Succ else R.Fail(`Host(hty1, hty2))
  | SD.Tint _, SD.TInt -> R.Succ
  | SD.TInt, SD.TInt -> R.Succ
  | SD.Tint _, _ |  _, SD.Tint _ -> R.Fail(`Host(hty1, hty2))
  | SD.TInt, _    |  _, SD.TInt -> R.Fail(`Host(hty1, hty2))

  | SD.Tstring x1, SD.Tstring x2 -> 
      if x1=x2 then R.Succ else R.Fail(`Host(hty1, hty2))
  | SD.Tstring _, SD.TString -> R.Succ
  | SD.TString, SD.TString -> R.Succ
  | SD.Tstring _, _ |  _, SD.Tstring _ -> R.Fail(`Host(hty1, hty2))
  | SD.TString, _    |  _, SD.TString -> R.Fail(`Host(hty1, hty2))

  | SD.Tchar x1, SD.Tchar x2 -> 
      if x1=x2 then R.Succ else R.Fail(`Host(hty1, hty2))
  | SD.Tchar _, SD.TChar -> R.Succ
  | SD.TChar, SD.TChar -> R.Succ
  | SD.Tchar _, _ |  _, SD.Tchar _ -> R.Fail(`Host(hty1, hty2))
  | SD.TChar, _    |  _, SD.TChar -> R.Fail(`Host(hty1, hty2))

  | SD.Ttag x1, SD.Ttag x2 -> 
      if x1=x2 then R.Succ else R.Fail(`Host(hty1, hty2))  
  | SD.Ttag _, SD.TTag -> R.Succ
  | SD.TTag, SD.TTag -> R.Succ
  | SD.Ttag _, _ |  _, SD.Ttag _ -> R.Fail(`Host(hty1, hty2))
  | SD.TTag, _    |  _, SD.TTag -> R.Fail(`Host(hty1, hty2))

  | SD.TSeq, SD.TSeq -> R.Succ
(*  | SD.TSeq, _   | _, SD.TSeq -> R.Fail(`Host(hty1, hty2)) *) 
             (* subsumed by earlier cases *)


and bsub_hostty t1 t2 = R.boolOf(sub_hostty t1 t2)


and maximums_hostty htyps = 
  let ord =  bsub_hostty in
  Extrem.extremums ord htyps

and maximum_hostty htyps = 
  match maximums_hostty htyps with 
  | [m] -> Some m
  | _ -> None


and minimums_hostty htyps = 
  let ord h1 h2 = bsub_hostty h2 h1 
                             (* note exchange of places for arguments *)
  in Extrem.extremums ord htyps

and minimum_hostty htyps = 
  match minimums_hostty htyps with 
  | [m] -> Some m
  | _ -> None

(*************************************************************************
                       Label types
**************************************************************************)


and sub_labty lty1 lty2 = 
  match SD.labty' lty1, SD.labty' lty2 with
  | SD.TLab hty1, SD.TLab hty2 -> 
      sub_hostty hty1 hty2
  | SD.PLabBind(_, l1), _ -> sub_labty l1 lty2
  | _, SD.PLabBind(_, l2) -> sub_labty lty1 l2


(** Returns all the "local" minimums w.r.t. subtyping present in the
    argument set.  Any lab binders are lost. *)
and minimums_labty ltys : SD.labty list = 
   let (htys, _) = List.split (List.map SD.crack_labty ltys) in
   let min_htys = minimums_hostty htys in
   List.map (SD.slap_labty []) min_htys


(** Returns the smallest of the arguments if there is such w.r.t.
    subtyping, and None otherwise.  Any lab binders are lost. *)
and minimum_labty ltys : SD.labty option = 
  match minimums_labty ltys with
  | [] -> None
  | [lty] -> Some lty
  | _::_ -> None
   

(** Returns all the "local" maximums w.r.t. subtyping present in the
    argument set.  Any lab binders are lost. *)
and maximums_labty ltys : SD.labty list = 
   let (htys, _) = List.split (List.map SD.crack_labty ltys) in
   let max_htys = maximums_hostty htys in
   List.map (SD.slap_labty []) max_htys


(** Returns the largest of the arguments if there is such w.r.t.
    subtyping, and None otherwise.  Any lab binders are lost. *)
and maximum_labty ltys : SD.labty option = 
  match maximums_labty ltys with
  | [] -> None
  | [lty] -> Some lty
  | _::_ -> None


(*??
let disjoint_labty lty1 lty2 = 
   match minimum_labty [lty1; lty2] with
   | None -> true
   | Some _ -> false
(*## Usage of 'disjoint' should be phased out in the cases where it is
  used in conjunction with minimum, since this duplicates the work. *)
ML: NB! I put a copy of this definition in compile.ml
??*)

(**************************************************************************
                           Automata 2/2
**************************************************************************)

and coalesce_label xsl =
  let xslc = Listutil.split_into_classes 
      (fun (_, xs1, xs2) (_, ys1, ys2) -> 
	match 
	  Ta.compare_state_sets xs1 ys1,  Ta.compare_state_sets xs2 ys2 
	with
	| -1, _ -> -1
	| 0, -1 -> -1
	| 0, 0 -> 0
	| _ -> 1 )   xsl
  in
  Listutil.mapcat 
    (fun cls -> 
      let (x1, x2) = (* take first, since all are equal *)
	match cls with (_,x1,x2)::_ -> (x1,x2) | _ -> assert false in
      let labs = List.map (fun (c,_,_) -> c) cls in
      let maxlabs = maximums_labty labs in
      List.map (fun c -> (c,x1,x2)) maxlabs
     ) xslc
	

(*old
let coalesce_label xsl =
  let xsl =
    Sort.list
      (fun (_, xs1, xs2) (_, ys1, ys2) -> 
	Ta.compare_state_sets xs1 ys1 < 0 &&
	Ta.compare_state_sets xs2 ys2 < 0) xsl in
  List.fold_right
    (fun (c, xs1, xs2) rem ->
      match rem with
        (d, ys1, ys2) :: rem' when
	  Ta.compare_state_sets xs1 ys1 = 0 && 
	  Ta.compare_state_sets xs2 ys2 = 0 ->
          (boot.sb_labdenot.ld_union c d, xs1, xs2) :: rem'
      | _ ->
          (c, xs1, xs2) :: rem)
    xsl []
*)

and coalesce xsl =
  coalesce_first
    (coalesce_second
       (coalesce_label xsl))

(*VG Restricts ysl to contain only transitions with labels that are in
     some sense "interesting" w.r.t. the label c. *)
(* 
   + c(x1,x2) <: d(y1,y2)|T iff c(x1,x2) <: (c&d)(x1,x2)|T
   + c(x1,x2) <: d(y1,y2)|T and d = 0 iff c(x1,x2) <: T
   + c(x1,x2) <: d(y1,y2)|...|e(z1,z2)|... 
     and c !<: d ... and c <: e ... and c !<: (d|...)
     iff c(x1,x2) <: e(z1,z2)|...
*)
(*old
let tidy_labels c ysl =
   (*VG First, restrict labels in ysl to their parts common with c *)
  let labden = boot.sb_labdenot in
  let ysl = List.map 
      (fun (d, ys1, ys2) -> (labden.ld_inter c d, ys1, ys2)) ysl in
  let ysl = List.filter 
      (fun (d, _, _) -> not (SD.is_emptylab d)) ysl in
  (*VG Compute union of those labels that do not cover c *)
  let non_covers =  
    List.fold_right
      (fun (d, _, _) d' -> 
	if labden.ld_sub c d then d' else labden.ld_union d d')
      ysl SD.emptylab in
 (*VG If c is not covered even by the union of non-covers, then throw
    out all the non-covers and leave covers only [totally unclear what
    is the logic behind this, but correctness of the call from
    subtype_many_many to subtype_one_one seems to depend on this.]  *)
 if not(labden.ld_sub c non_covers) 
  then List.filter (fun (d, _, _) -> labden.ld_sub c d) ysl
  else ysl
*)


and tidy_labels c ysl =
  let ysl = Listutil.filtermap_opt 
      (fun (d, ys1, ys2) -> match minimum_labty [c; d] with
      |	Some cd -> Some (cd, ys1, ys2)
      |	None -> None)
      ysl   in 
  (*VG We need an analogue of the last sentence of the original
  tidy_labels...  For host types (taking into account the "open world
  assumption" for subclassing), a host type cannot be assumed to be
  covered by any set of its proper subtypes.  Therefore, to do the
  job, we just trow away all the non-covers of c. *)
   (* Note that all the 1st part of tidy_labels is now redundant... *)
   List.filter (fun(d,_,_) -> Result.boolOf(sub_labty c d)) ysl

and subtype_main hyps xs ys =
  let state_set_is_obviously_any x = 
    List.exists (fun s -> s == !Store.any_ta) x
  in
  (* First, attempt to solve without traversing the successors... *)
  let ds = Ta.diff_state_sets xs ys in
  if ds = [] then Result.Succ else
  if ys = [] then Result.Fail `Any else
  if state_set_is_obviously_any ys then Result.Succ else
  if Ta.state_set_final ds && not (Ta.state_set_final ys) 
  then Result.Fail `Empty 
  (* ... if not succeeding, consider the successors *)
  else subtype_handle_recursion hyps ds ys

and subtype_handle_recursion hyps xs ys =
  let t = (hash_hyp xs ys, xs, ys) in
  if Ta.IntLsLsMap.mem t !Store.failure_cache then 
    begin
      Result.Fail (Ta.IntLsLsMap.find t !Store.failure_cache)
    end else
  if Ta.IntLsLsSet.mem t !hyps then Result.Succ else 
  let hyps' = ref (Ta.IntLsLsSet.add t !hyps) in
  match 
    subtype_many_many hyps' (Ta.state_set_next xs) (Ta.state_set_next ys) 
  with
  | Result.Fail w -> 
      begin 
	Store.failure_cache := Ta.IntLsLsMap.add t w !Store.failure_cache; 
	Result.Fail w 
      end
  | Result.Succ -> 
      begin hyps := !hyps'; Result.Succ end

and subtype_many_many hyps xl yl =
  for_all
    (fun (c,_, x1, x2, _) -> 
      let ysl = List.map (fun (d,_, y1, y2, _) -> (d, [y1], [y2])) yl in
      let ysl = tidy_labels c ysl in
      let ysl = coalesce ysl in
      match ysl with
      | [] -> Result.Fail (`Lab (c, `Any, `Any))
      | [(d, ys1, ys2)] -> 
	  subtype_one_one hyps (c, x1, x2) (d, ys1, ys2) 
      |	ysl -> subtype_one_many hyps (c, x1, x2) ysl)
    xl

(* c(x1,x2) <: d(ys1,ys2) iff x1 <: ys1 and x2 <: ys2 
   It works because empty states have already been eliminated.  *)
(*VG ... eliminated by tidy_labels -- ??? *)

and subtype_one_one hyps (c, x1, x2) (d, ys1, ys2) =
  match subtype_main hyps [x2] ys2 with
  | Result.Fail w -> Result.Fail (`Lab (c, `Any, w))
  | Result.Succ -> 
      (match subtype_main hyps [x1] ys1 with
      |	Result.Fail w -> Result.Fail (`Lab (c, w, `Any))
      | Result.Succ -> Result.Succ)

and subtype_one_many hyps (c,x1, x2) ysl =
  if false then begin
    let desc s = match s.Ta.desc with Some x -> x | None -> SD.sty_epsilon in
    printf "===========@.";
    printf "- %a@." SD.pr_labty c;
    printf "-----------@.";
    printf "- %a@." SD.pr_seqty (desc x1);
    printf "-----------@.";
    printf "- %a@." SD.pr_seqty (desc x2);
    printf "~~~~~~~~~~~@.";
    let l0 = List.map (fun (a,_,_) -> a) ysl in
    let l1 = List.map (fun (_,a,_) -> a) ysl in
    let l2 = List.map (fun (_,_,a) -> a) ysl in
    Printing.print_listH (fun () -> ()) (fun x -> printf "- %a@." SD.pr_labty x)
      l0;
    printf "-----------@.";
    Printing.print_listH (fun () -> ()) (fun x -> printf "- %a@." SD.pr_seqty x)
      (List.map (fun l -> SD.alt (List.map (fun s -> desc s) l)) l1);
    printf "-----------@.";
    Printing.print_listH (fun () -> ()) (fun x -> printf "- %a@." SD.pr_seqty x)
      (List.map (fun l -> SD.alt (List.map (fun s -> desc s) l)) l2);
  end;
  subtype_one_many_iter hyps (c, x1, [], x2, []) ysl

(* 
   + c(x1\xs1, x2\xs2) <: d(ys1,ys2)|T
     iff (c\d)(x1\xs1, x2\xs2) <: T 
         and c(x1\xs1\ys1, x2\xs2) <: T
         and c(x1\xs1, x2\xs2\ys2) <: T
   + c(x1\xs1, x2\xs2) <: 0 iff x1 <: xs1 and x2 <: xs2
*)

and subtype_one_many_iter hyps (c, x1, xs1, x2, xs2) ysl =
  match ysl with
    (d, ys1, ys2) :: ysl' -> 
      if (*old SD.is_emptylab c || *)  (*VG always false for simple labels *)
         subtype_main hyps [x1] xs1 = Result.Succ ||
	 subtype_main hyps [x2] xs2 = Result.Succ then Result.Succ else
(*old VG  appears to be a "failfast optimization"
	(match
	  subtype_one_many_iter hyps 
	    (boot.sb_labdenot.ld_diff c d, x1, xs1, x2, xs2) 
	    ysl'
	with
	  Result.Fail w -> Result.Fail w
	| Result.Succ ->
*)
	    (match 
	      subtype_one_many_iter 
		hyps (c, x1, Ta.union_state_sets xs1 ys1, x2, xs2) ysl'
	    with
	      Result.Fail w -> Result.Fail w
	    | Result.Succ ->
		subtype_one_many_iter 
		  hyps (c, x1, xs1, x2, Ta.union_state_sets xs2 ys2) ysl')
  (*old    )  *)
  | [] -> 
      if (SD.is_emptylab c) then Result.Succ else
      (match subtype_main hyps [x2] xs2 with
      | Result.Succ -> Result.Succ
      | Result.Fail w2 ->
	  (match subtype_main hyps [x1] xs1 with
	  | Result.Succ -> Result.Succ
	  | Result.Fail w1 -> Result.Fail (`Lab (c, w1, w2))))
	

and sub_ta x y =
  let x = Ta.elim_empty x in
  let y = Ta.elim_empty y in
  let res = subtype_main Store.sub_hyps x y in
  res


(* 

+ There used be a rule A <: B|C and B <: C iff A <: C.  But this was
removed because this sometimes becomes harmful.  Also, what the
function did wasn't clear.

+ Sometimes we get a redundant label like (~ U ~): needs a fix in
Label.union?


Some ideas for further improvements:

+ cheap (incomplete) subtyping: maybe useful for speculative checking
in subtype_one_many_iter/isect/infer/...

+ randomly generate a sample tree to test failure of subtyping: maybe
useful for purposes similar to the cheap subtyping

*)


(*************************************************************************
                       Sequence types
**************************************************************************)
and buildTa x =  Buildta.build_ta x


and sub_seqty (sty1: SD.seqty) (sty2: SD.seqty) : Typath.t Result.t = 
  let x1 = buildTa sty1 in
  let x2 = buildTa sty2 in
  sub_ta [x1] [x2] 

and bsub_seqty sty1 sty2 = R.boolOf(sub_seqty sty1 sty2)


(*************************************************************************
                          Full types
**************************************************************************)

and sub_ty  (ty1: SD.ty) (ty2: SD.ty) =
  match (SD.ty' ty1, SD.ty' ty2) with 
  | SD.Thost ht1, SD.Thost ht2 -> sub_hostty ht1 ht2
  | SD.Tseq seq1, SD.Tseq seq2 -> sub_seqty seq1 seq2
  | SD.Tseq _, SD.Thost(ht2) -> 
      if SD.isHostGEQSeq ht2 then R.Succ else R.Fail(`Ty(ty1, ty2))
  | SD.Thost(ht1), SD.Tseq seq2 -> 
      if SD.isHostLEQSeq ht1 then sub_seqty !Store.any_seqty_def  seq2
      else R.Fail(`Ty(ty1, ty2))
  | SD.PTyBind(_, t1), _  ->  sub_ty t1 ty2
  |  _, SD.PTyBind(_, t2) ->  sub_ty ty1 t2


let bsub_ty t1 t2 = R.boolOf(sub_ty t1 t2)


let host_seq_partion typs = 
  let rec looper hs ss = function
    | [] -> (hs, ss)
    | ty::tys -> 
	match SD.ty' ty with
	| SD.PTyBind(_, t) -> looper hs ss (t::tys)
	| SD.Thost h -> looper (h::hs) ss tys
	| SD.Tseq  s -> looper hs (s::ss) tys
  in
  looper [] [] typs
    

let maximum_ty typs = 
  let (htyps, styps) = host_seq_partion typs in
  match htyps, styps with 
  | [],   []   -> Some SD.ty_Null
  | [],   _::_ ->
(*      let styp = List.fold_left (fun acc sty -> SD.alt[acc; sty]) 
	  SD.sty_empty styps
          *)
        let styp = SD.alt styps 
      in Some(SD.ty (SD.Tseq styp))
  | _::_, _    -> 
      let htyps1 = match styps with
      |	[]   -> List.map SD.whole_hostty htyps
      |	_::_ -> List.map SD.whole_hostty (SD.hty_Seq :: htyps)
      in  match maximum_hostty htyps1 with
      |	None -> None
      |	Some hty -> Some(SD.ty (SD.Thost hty))
      

