(***********************************************************************
       Syntactic definitions for the "data model" part of the language. 
***********************************************************************)

(* Uncomment one of the following variations of hashing policy: *)
(* 
include SynTypPublic.Make(SynTypHashCompute.M(
			  struct let meaningful_nodes = 1
			         let total_nodes = 2   end))
*)
(* 
include SynTypPublic.Make(SynTypHashCompute.M(
			  struct let meaningful_nodes = 3
			         let total_nodes = 5   end))
*)
(* this is the fastest: *) 
include SynTypPublic.Make(SynTypHashCarry.M)
(* 
 *)


(** Patterns --  defined mostly for documentation. *)
type hostpa = hostty
type pa = ty
type labpa = labty
type seqpa = seqty

(** Some convenience functions and constructors  *)

let basiclab ty =  labty (TLab ty)

let is_emptylab lty = false
    (** Always false: there is no empty label in the absence of label sets *)

let crack_labty (lty: labty) : hostty * (Ident.id list) = 
   let rec looper vars lt = 
   match labty' lt with
   | TLab hty -> (hty, List.sort compare vars)
   | PLabBind(v, l) -> looper (v::vars) l
   in looper [] lty

let slap_labty vars hty = 
   let rec looper lty = function
   | [] -> lty
   | v::vs -> 
      let newlty = labty(PLabBind(v, lty)) in
      looper newlty vs
   in
   looper (labty(TLab hty)) vars



let attsty_any = attsty TAttAny
let attsty_empty = attsty (PAtt([], None))

let tree lty aty sty = seqty (TTree(lty, aty, sty))
let leaf ty = seqty (TTree(labty (TLab ty), attsty_any, sty_epsilon))

let cat stys = seqty (TCat stys)
let alt stys = seqty (TAlt stys)
let rep sty = seqty (TRep sty)
let seqref id = seqty (TSeqRef id)
let seqbind  vid sty = seqty (PSeqBind(vid, sty))
let tailbind vid sty = seqty (PTailBind(vid, sty))



(****************   Printing ********************************************)
let fpr = Format.fprintf

(*******************************************************************)
                    (** Types printing *)

let pr_attpat ff = function 
  | (aname, id, _) -> fpr ff "@[%s=%a@]" aname Ident.print id

let pr_attsty ff x = 
  match attsty' x with 
  | TAttAny -> ()
  | PAtt(attpats, catchopt) ->
      Printing.print_list pr_attpat " " "" ff attpats;
      match catchopt with
      | None -> ()
      | Some(id, _) -> fpr ff "%s%a" "@a" Ident.print id

(* Host types *)
let rec pr_hostty ff x = 
  match hostty' x with 
  | TClasRef cid -> Ident.print ff cid
  | TIntfRef cid -> Ident.print ff cid
  | TObject -> fpr ff "Object"
  | TNull -> fpr ff "NULL"
  | TBool -> fpr ff "bool"
  | Tbool b -> Format.pp_print_bool ff b
  | TInt -> fpr ff "int"
  | Tint n -> Format.pp_print_int ff n
  | TString -> fpr ff "string"
  | Tstring s -> fpr ff "\"%a\"" Format.pp_print_string s
  | TChar -> fpr ff "char"
  | Tchar c -> Format.pp_print_char ff c
  | TTag -> fpr ff "Tag"
  | Ttag t -> fpr ff "Tag(%s)" t 
  | TSeq -> fpr ff "Seq"
  | TArray t -> fpr ff "%a[]" pr_ty t

and pr_ty ff x = 
  match ty' x with
  | Thost ht -> pr_hostty ff ht
  | Tseq seq -> fpr ff "@[{[%a]}@]" pr_seqty seq
  | PTyBind (id, ty) ->  fpr ff "@[%a@ %a,@]" pr_ty ty Ident.print id
                               
and pr_labty ff x = 
  match labty' x with  
  | TLab hty -> 
      fpr ff "(%a)" pr_hostty hty
  | PLabBind(vid, lty) -> 
      fpr ff "%a %a," pr_labty lty Ident.print vid

and pr_seqty ff x = 
  match seqty' x with 
  | TTree(lab, atts, seq) -> 
      fpr ff "@[%s%a%a%s%a%s@]" 
	"<(" pr_labty lab pr_attsty atts ")>" pr_seqty seq "</>"
  | TCat seqs -> Printing.print_list pr_seqty " "  "()" ff seqs
  | TAlt seqs -> Printing.print_list pr_seqty "|"  "[EMPTY-ALT]" ff seqs
  | TRep seq -> fpr ff "@[(%a)*@]" pr_seqty seq
  | TSeqRef id -> Ident.print ff id
  | PSeqBind(id, seqty) -> 
      fpr ff "@[%a@ as@ %a@]" pr_seqty seqty Ident.print id
  | PTailBind(id, seqty) -> 
      fpr ff "@[%a@ as@ %a@]" pr_seqty seqty Ident.print id

let pr_ty_option ff = function
  | None -> fpr ff "VOID"
  | Some ty -> pr_ty ff ty
      

(*******************************************************************)
                    (** Type stamp printing *)

let _module_name_ = ref ""

let set_module_name s =
  _module_name_ := s

let pr_ts_attpat ff = function 
  | (aname, id, _) -> fpr ff "(%s %a)" aname Ident.print id

let pr_ts_attsty ff x = 
  match attsty' x with 
    | TAttAny -> ()
    | PAtt(attpats, catchopt) ->
        fpr ff "(patt %a %s)"
          (Printing.print_list pr_ts_attpat " " "") attpats
          (match catchopt with
            | None -> "(none)"
            | Some(id, _) -> "(some " ^ Printing.toString Ident.print id ^ ")")

let rec pr_ts_hostty ff x = 
  match hostty' x with 
    | TClasRef cid -> fpr ff "(tclassref %a)" Ident.print cid
    | TIntfRef cid -> fpr ff "(tintfref %a)" Ident.print cid
    | TObject -> fpr ff "(tobject)"
    | TNull -> fpr ff "(tnull)"
    | TBool -> fpr ff "(tbool)"
    | Tbool b -> fpr ff "(tbool %a)" Format.pp_print_bool b
    | TInt -> fpr ff "(tint)"
    | Tint n -> fpr ff "(tint %a)" Format.pp_print_int n
    | TString -> fpr ff "(tstring)"
    | Tstring s -> fpr ff "(tstring %a)" Format.pp_print_string s
    | TChar -> fpr ff "(tchar)"
    | Tchar c -> fpr ff "(tchar %a)" Format.pp_print_char c
    | TTag -> fpr ff "(ttag)"
    | Ttag t -> fpr ff "(ttag %s)" t
    | TSeq -> fpr ff "(tseq)"
    | TArray t -> fpr ff "(tarray %a)" pr_ts_ty t

and pr_ts_ty ff x = 
  match ty' x with
    | Thost ht -> fpr ff "(thost %a)" pr_ts_hostty ht
    | Tseq seq -> fpr ff "(tseq %a)" pr_ts_seqty seq
    | PTyBind (id, ty) -> pr_ts_ty ff ty

and pr_ts_labty ff x = 
  match labty' x with  
    | TLab hty -> fpr ff "(tlab %a)" pr_ts_hostty hty
    | PLabBind(vid, lty) -> pr_ts_labty ff lty

and pr_ts_seqty ff x = 
  match seqty' x with 
  | TTree(lab, atts, seq) ->
      fpr ff "(ttree %a %a %a)" pr_ts_labty lab pr_ts_attsty atts pr_ts_seqty seq
  | TCat seqs ->
      fpr ff "(tcat %a)" (Printing.print_list pr_ts_seqty " "  "()") seqs
  | TAlt seqs ->
      fpr ff "(talt %a)" (Printing.print_list pr_ts_seqty " "  "()") seqs
  | TRep seq ->
      fpr ff "(trep %a)" pr_ts_seqty seq
  | TSeqRef id ->
      fpr ff "%s.%a" !_module_name_ Ident.print id
  | PSeqBind(_, seqty) | PTailBind(_, seqty) -> 
      pr_ts_seqty ff seqty


(*******************************************************************)
               (** Exported printing function printing *)

let toString_hostty hostty = 
  Printing.toString pr_hostty hostty

let toString_labty labty = 
  Printing.toString pr_labty labty

let toString_seqty ty = 
  Printing.toString pr_seqty ty

let toTypeStampString_seqty ty = 
  Printing.toString pr_ts_seqty ty

let toString_ty ty = 
  Printing.toString pr_ty ty

let toString_ty_option ty = 
  Printing.toString pr_ty_option ty

let toString_list_ty = function
  | [] -> ""
  | t :: r -> List.fold_left (fun acc ty -> acc^", "^(toString_ty ty))
    (toString_ty t) r


(******* More convenience definitions **************)

let hty_ClasRef id = hostty (TClasRef id)
let hty_IntfRef id = hostty (TIntfRef id)
let hty_Object = hostty TObject
let hty_Null = hostty TNull
let hty_Bool = hostty TBool
let hty_Int = hostty TInt
let hty_Char = hostty TChar
let hty_String = hostty TString
let hty_Tag = hostty TTag
let hty_Seq = hostty TSeq

let ty_ClasRef id = ty (Thost (hty_ClasRef id))
let ty_IntfRef id = ty (Thost (hty_IntfRef id))
let ty_Object = ty (Thost hty_Object)
let ty_Null = ty (Thost hty_Null)
let ty_Bool = ty (Thost hty_Bool)
let ty_Int = ty (Thost hty_Int)
let ty_Char = ty (Thost hty_Char)
let ty_String = ty (Thost hty_String)
let ty_Tag = ty (Thost hty_Tag)
let ty_Seq = ty (Thost hty_Seq)

let sty_anyone = seqref Kerids.anyone_id
let sty_any = seqref Kerids.any_id
let sty_xml = seqref Kerids.xml_id
let sty_pcdata = seqref (Kerids.pcdata_id)
let sty_attseq = seqref Kerids.attseq_id

let ty_empty = ty (Tseq sty_empty)
let ty_epsilon = ty (Tseq sty_epsilon)
let ty_anyone = ty (Tseq sty_anyone)
let ty_any = ty (Tseq sty_any)
let ty_xml = ty (Tseq sty_xml)
let ty_pcdata = ty (Tseq sty_pcdata)
let ty_attseq = ty (Tseq sty_attseq)



let whole_hostty hty = 
  match hostty' hty with
  | Tbool _   -> hty_Bool
  | Tint _    -> hty_Int
  | Tchar _   -> hty_Char
  | Tstring _ -> hty_String
  | Ttag _    -> hty_Tag
  | other -> hty

let rec whole_ty t = 
  match ty' t with
  | Thost hty -> ty (Thost(whole_hostty hty))
  | Tseq sty -> ty_Seq
  | PTyBind(_, tt) -> whole_ty tt


(****)

let clasOf_hostty t = 
  match hostty' t with 
  | TNull -> assert false  (* should not be called on these *)
  | TClasRef cid -> cid
  | TIntfRef cid -> cid
  | TObject -> Kerids.object_id

  | TBool  -> Kerids.bool_id
  | Tbool _ -> Kerids.bool_id
  | TInt  -> Kerids.int_id
  | Tint _ -> Kerids.int_id
  | TChar -> Kerids.char_id
  | Tchar _ -> Kerids.char_id
  | TString  -> Kerids.string_id
  | Tstring _ -> Kerids.string_id
  | TTag  -> Kerids.tag_id
  | Ttag _ -> Kerids.tag_id
  | TSeq -> Kerids.seq_id
  | TArray _ -> failwith "calling SD.clasOf_hostty on array"

let rec clasOf_ty t = 
  match ty' t with
  | Thost hty -> clasOf_hostty hty 
  | Tseq _ -> Kerids.seq_id
  | PTyBind(_, ty) -> clasOf_ty ty


let isHostGEQSeq hostty =  
 match hostty' hostty with
  | TSeq -> true
  | TObject -> true
  | _ -> false

let isHostLEQSeq hostty =  
  match hostty' hostty with
  | TSeq -> true
  | TNull -> true
  | _ -> false


(********** Determine whether a pattern contains binders  ********)

let binders_attsty atts = 
  match attsty' atts with
  | TAttAny -> false
  | PAtt([], None) -> false
  | PAtt _ -> true

let rec binders_ty t =
  match ty' t with
    | Thost _ -> false
    | Tseq seq -> binders_seqty seq
    | PTyBind (id, ty) -> true
                               
and binders_labty lty = 
  match labty' lty with
    | TLab tys -> false  (*NB possible binders in tys are meaningless *)
    | PLabBind _ -> true

and binders_seqty s =
  match seqty' s with
    | TTree(prs, atts, seq) -> 
	binders_labty prs || binders_attsty atts || binders_seqty seq
    | TCat seqs -> Utils.ormap binders_seqty seqs
    | TAlt seqs -> Utils.ormap binders_seqty seqs
    | TRep seq -> binders_seqty seq
    | TSeqRef id -> false
    | PSeqBind(id, seqty) -> true
    | PTailBind(id, seqty) -> true
