(** Syntax and printing for the source program constructs:
    types, patterns, expressions, classes.  *)

(*--VG## TODO: Consider using OO types for the AST, in order to avoid
  duplication (of defs here and of code in other modules) for XXX and
  xti_XXX features.  *)

module N = Naming

(**************** File info stuff *************************************)

type 'a info = 'a Finfo.finfo


(************** Names (for types and variables) ************************)
type name = string info
type longname = name list 

(*## These types have only mnemonic value now; perhaps can be in-lined... *)
type decname = name      (** for declaring a new name *)
type refname = longname  (** for referencing a type through its long name *)
type onename = name      (** for referencing a type through its short name *)

type qualname = name option * name     (* (prefix opt, local) *)

type uri = string info


(****************** Types *****************************************)

type attpat = attpat' info
and attpat' = name * decname   (* (attr_name, var) *)

type attsty = attsty' info
and attsty' = 
  | TAttAny
  | PAtt of attpat list * decname option  (* (individ attrs, catch-all var) *)

type hostty = hostty' info
and hostty' = 
  | TCSharp of refname   (* reference to class, interface, ... *)
  | TNull 
  | TBool
  | Tbool of bool
  | TInt
  | Tint of int
  | TString
  | Tstring of string
  | TChar
  | Tchar of char
  | TTag 
  | Ttag of string 
  | TSeq
  | TArray of ty

and labty = labty' info
and labty' = 
  | TLab of hostty
  | PLabBind of decname * labty

and seqty = seqty' info 
and seqty' = 
  | TTree of labty * attsty * seqty
  | TCat of seqty list
  | TAlt of seqty list
  | TRep of seqty
  | TSeqRef of qualname
  | PSeqBind of decname * seqty


and ty = ty' info 
and ty' = 
  | Thost of hostty
  | Tseq of seqty
  | PTyBind of decname * ty


(******************* Patterns ***************************************)
type pa = ty
type seqpa = seqty


(*************** Expressions *********************************)


type exp = exp' info
and exp' = 
  (** Expressions for literals *)
  | Ebool of bool
  | Eint of int
  | Estring of string 
  | Echar of char 
  | Etag of string 
  | Enull

  (** OO host expressions *)
  | EThis
  | ENamedLoc of refname   (** elaborates in Source into EFieldOrVar plus
			       field projections *)
  | ENew of refname * (exp list)
  | ENewArray of ty * exp * int   (*!!-> itemtype *)
  | EFieldAccess of exp * onename 
  | EFieldAccessSuper of onename
  | EMethInvkBase of exp * onename * (exp list)
  | EMethInvkLocal of onename * (exp list)  (* for meth of current class *)
  | EMethInvkSuper of onename * (exp list)  (* for meth of the superclass *)
  | EArrayAccess of exp * exp
  | EAssign of exp * exp   (* lhs := rhs *)
  | ECast of ty * exp * bool
  | ETypeStamp of seqty * exp
  | EUnary of SynOp.unaryop * exp
  | EBinary of exp * SynOp.binaryop * exp

  (** Sequence expressions *)
  | ESeq of exp list            (* {| ... |}  *)
  | ETree of exp * exp list * exp list     
      (*  (label, attributes, contents) *)
  | EPcdata of string
   (** Syntactically, attributes are a regular sequence, but they are
   treated in a special way: when evaluated to a value, the sequence
   is checked to be "attributive".  Currently this means to have
   distinct tags as labels and to have pcdata as values.  *)

  | EString2Pcdata of exp   (* a hack, used only for short-hand attributes *)

  (** Statements *)
type block = block' info
and block' = stmt list

and stmt = stmt' info
and stmt' = 
  | SBlock of block
  | SExp of exp 
  | SLocVars of ((decname * exp option) list * ty)
  | SReturn of exp' option info 
  | SIf of exp * block * block   
  | SWhile of exp * block        
(*??  | SSwitch of exp * (clause list)     (* for host matching *)  *)
  | SMatch  of exp * (clause list)      (* for XDuce matching *)
  | SForeachRE of (seqpa * exp * block)    (* XML's foreach *)
  | SEmpty 
  | SBreak

and clause  = clause' info 
and clause' = pa * block
  

(*** Classes ***)

(*--VG The types clas, tunit, cunit (and their subordinates) have been
  generalized to handle both xt and xti files by making some of the
  fields optional.  This compromises type safety somewhat, but reduces
  code duplication of in XxxCtx modules (no need to write a second set
  of 'extend' functions).  *)

type method_modifier = SynOp.method_modifier info
    (*## SynOp.method_modifier can be used directly if less precise error
      messages in CtxCls.check_meth would be tolerated... *)

type mangle = string 
     (** mangling name suffix, for entries with regtypes *)

  
type clas = clas' info
and  clas' = 
    {clas_name: decname;
     clas_extends: refname list;
     clas_sealed: bool;
     clas_constr: constr;
     clas_fields: field list;
     clas_meths:  meth list }

and field = field' info
and field' = 
    ( decname *           (* field name *)
      mangle option *     (* mangling suffix - Some -> xti, None -> xt *)
      exp option *        (* init expr - Some -> xt, None -> xti, xt *)
      SynOp.status *      (* Static or Instance *)
      ty )                (* field type *)

and meth = meth' info
and meth' = 
    ( decname *                      (* method name *)
      mangle option *                (* mangling suffix - Some -> xti *)
      SynOp.status *                 (* MStatic or MInstance *)
      method_modifier *              (* modifier *)
      ty' option info *              (* return type  -- None is for void!*)
      (decname option * ty) list *   (* parameters -- None -> xti *)
      block option)                  (* body -- None -> xti *)
    
and constr = constr' info
and constr' = 
    ( onename *                      (* constructor's class name *)
      (decname option * ty) list *   (* parameters *)
      invoke  *                      (* expl envok of another constructor *)
      block option )                 (* body -- None -> xti *)
	
and invoke = invoke' info
and invoke' = 
  | InvkNone
  | InvkThis of exp list  
  | InvkSuper of exp list
  
type intf = intf' info
and intf' =
   {intf_name: decname;
    intf_implements: refname list;
    intf_meths:  (decname * imeth) list }
and imeth = imeth' info
and imeth' = 
    (  ty' option info *              (* return type -- None -> void !*)
       (decname option * ty) list)    (* parameters -- None -> xti *)


(***  Compilation untits ***)

type usedir = usedir' info
and usedir' = 
  | CAlias of decname * refname
  | COpen of refname
  | XPrefix of decname * uri
  | XOpen of uri
  | XAlias of decname * qualname

type tunit = tunit' info     (* top-level unit *)
and tunit' = {
  tunit_usedirs: usedir list;
  tunit_decls: (decname * tmember) list; }

and tmember = 
  | TMemClas of clas
  | TMemIntf of intf
  | TMemSpace of cunit
  | TMemSeq of seqty
  | TMemXspace of xunit

and cunit = cunit' info     (* inner C# unit *)
and cunit' = {
  cunit_usedirs: usedir list;
  cunit_decls: (decname * cmember) list;  }

and cmember = 
  | CMemClas of clas
  | CMemIntf of intf
  | CMemSpace of cunit

and xunit = xunit' info    (* inner XML/Reg unit *)
and xunit' = {
  xunit_usedirs: usedir list;
  xunit_decls: (decname * xmember) list; }

and xmember = 
  | XMemSeq of seqty



(******************* Program (flattened) *************************)

type clstbl = (Naming.ctx * clas) Ident.tbl
type intftbl = (Naming.ctx * intf) Ident.tbl
type seqtbl = (Naming.ctx * seqty) Ident.tbl

type prog = 
    { 
      cls: clstbl;
      intfs: intftbl;
      seq: seqtbl; 
    } 


(*******************************************************************
                     xti file entries
*******************************************************************)
(*--VG??
type mangle = string option   
     (** mangling name suffix, for entries with regtypes *)

type xti_clas = xti_clas' info
and  xti_clas' = 
    {xti_clas_name: decname;
     xti_clas_extends: refname list;
     xti_clas_sealed: bool;
     xti_clas_constr: xti_constr;
     (* the bool in the fields is true if the field is static *)
     xti_clas_fields: ((decname * mangle) * SynOp.status * ty) list ;  
     xti_clas_meths:  ((decname * mangle) * meth) list }

and xti_meth = xti_meth' info
and xti_meth' = 
    ( SynOp.status *        (* MStatic or MInstance *)
      method_modifier *     (* modifier *)
      ty' option info *     (* return type *)
      ty list )             (* parameters *)
    
and xti_constr = xti_constr' info
and xti_constr' = 
    (onename *               (* constructor's class name--to be checked so *)
     ty list )               (* parameters *)
	
  
type xti_intf = intf' info
and xti_intf' =
   {xti_intf_name: decname;
    xti_intf_implements: refname list;
    xti_intf_methsdecls:  (decname * xti_methdecl) list }
and xti_methdecl = xti_methdecl' info
and xti_methdecl' = 
    (  ty' option info *        (* return type *)
       ty list)                 (* parameters *)


(***  Compilation untits ***)

(*--VG--        hoping to reuse  usedir...
type xti_usedir = usedir' info
and usedir' = 
  | CAlias of decname * refname
  | COpen of refname
  | XPrefix of decname * uri
  | XOpen of uri
  | XAlias of decname * qualname
*)

type xti_tunit = xti_tunit' info     (* top-level unit *)
and xti_tunit' = {
  xti_tunit_usedirs: usedir list;
  xti_tunit_decls: (decname * xti_tmember) list; }

and xti_tmember = 
  | XTI_TMemClas of xti_clas
  | XTI_TMemIntf of xti_intf
  | XTI_TMemSpace of xti_cunit
  | XTI_TMemSeq of seqty
  | XTI_TMemXspace of xunit

and xti_cunit = xti_cunit' info     (* inner C# unit *)
and xti_cunit' = {
  xti_cunit_usedirs: usedir list;
  xti_cunit_decls: (decname * xti_cmember) list;  }

and xti_cmember = 
  | XTI_CMemClas of xti_clas
  | XTI_CMemIntf of xti_intf
  | XTI_CMemSpace of xti_cunit

(*--VG--    hoping to reuse xunit... 
and xti_xunit = xunit' info    (* inner XML/Reg unit *)
and xunit' = {
  xunit_usedirs: usedir list;
  xunit_decls: (decname * xmember) list; }

and xmember = 
  | XMemSeq of seqty
*)

(* do we need type xti_prog ? *)

--*)


(*******************************************************************
                          Pretty printing 
*******************************************************************)
let fpr = Format.fprintf

let pr_opt pr ff x = 
  match x with 
    | None -> ()
    | Some x' -> pr ff x'

(***** Names ***********)

let pr_name ff n = 
  Format.pp_print_string ff (Finfo.drop n)

let pr_longname ff ln = 
  let rec prList ff lst = 
    match lst with
    | [n] -> pr_name ff n
    | n::ns -> fpr ff "%a.%a" pr_name n prList ns
    | [] -> raise (Failure "BUG: zero-length longname")
  in prList ff ln

let pr_decname ff n = pr_name ff n
let pr_refname ff n = pr_longname ff n
let pr_onename ff n = pr_name ff n

let pr_commalist pr_func ff names = 
  let rec pr fff = function 
    | [] -> assert false
    | [n] -> pr_func fff n
    | n::ns -> fpr fff "%a, %a" pr_func n pr ns
  in pr ff names

let pr_qualname ff qn = 
  match qn with 
    | (None, locname) -> pr_name ff locname
    | (Some pref, locname) -> fpr ff "%a:%a" pr_name pref pr_name locname

let pr_mangle ff mangle = 
  fpr ff "_%s" mangle

(***** Types ************)

let pr_attpat ff att =
  let (aname, dn) = Finfo.drop att in 
    fpr ff "%a=%a" pr_name aname pr_decname dn

let pr_attsty ff attsty = 
  match Finfo.drop attsty with 
  | TAttAny -> ()
  | PAtt([], None) -> fpr ff "%s()" "@"
  | PAtt(attpats, catchopt)  -> 
      Printing.print_list pr_attpat " " "" ff attpats;
      match catchopt with
      | None -> ()
      | Some dname -> fpr ff "%s%a" "@a" pr_decname dname
    

let rec pr_hostty ff hty = 
  match (Finfo.drop hty) with 
  | TCSharp refname -> pr_refname ff refname
  | TNull -> fpr ff "NULL"
  | TBool -> fpr ff "bool"
  | Tbool b -> Format.pp_print_bool ff b
  | TInt -> fpr ff "int"
  | Tint n -> Format.pp_print_int ff n
  | TString -> fpr ff "string"
  | Tstring s -> Format.pp_print_string ff s 
  | TChar -> fpr ff "char"
  | Tchar c -> Format.pp_print_char ff c
  | TTag -> fpr ff "tag"
  | Ttag t ->  fpr ff "%s" t 
  | TSeq -> fpr ff "seq"
  | TArray t -> fpr ff "%a[]" pr_ty t

and pr_labty ff labty = 
  match Finfo.drop labty with 
  | TLab hty -> 
      fpr ff "@[(%a)@]" pr_hostty hty
  | PLabBind(vname, lty) -> 
      fpr ff "@[%a@ %a,@]" pr_labty lty pr_decname vname
	
and pr_seqty ff seqty = 
  match (Finfo.drop seqty) with 
  | TTree (lty, atts, seqty) -> 
      fpr ff "@[%s(%a)%a%s%a%s@]" 
	"<" pr_labty lty pr_attsty atts ">" pr_seqty seqty "</>"
  | TCat seqtyps -> Printing.print_list pr_seqty " " "()" ff seqtyps 
  | TAlt seqtyps -> Printing.print_list pr_seqty "|" "[EMPTY-ALT]" ff seqtyps 
  | TRep seq -> fpr ff "@[(%a)*@]" pr_seqty seq
  | TSeqRef qualname -> pr_qualname ff qualname
  | PSeqBind (vname, spa) -> fpr ff "%a as %a" pr_seqty spa pr_decname vname

and pr_ty ff ty =
  match Finfo.drop ty with
  | Thost hty -> pr_hostty ff hty
  | Tseq seq -> fpr ff "@[{[%a]}@]" pr_seqty seq
  | PTyBind (vname, ty) -> fpr ff "@[%a %a@]" pr_ty ty pr_decname vname

let pr_ty_option ff = function
| (info, None) -> fpr ff "VOID"
| (info, Some ty) -> pr_ty ff (info,ty)

(******** Expressions **********)

let rec pr_exp ff expr = 
  match (Finfo.drop expr) with 
  | Ebool b -> Format.pp_print_bool ff b
  | Eint n -> Format.pp_print_int ff n
  | Estring s -> Format.pp_print_string ff s 
  | Echar c -> Format.pp_print_char ff c
  | Etag t -> fpr ff "%s" t 
  | Enull -> fpr ff "null"

  | ENamedLoc name -> pr_refname ff name
  | EThis ->  fpr ff "this"
  | ENew(tname, args) ->
      fpr ff "@[new %a%a@]" pr_refname tname pr_args args
  | ENewArray(ty, dim_exp, dims_impl) -> 
      let print_expl_dim exp = 
	fpr ff "@[[";  pr_exp ff exp;  fpr ff "]@]"
      in
      let rec pr_dims = function
        | 0 -> ()
        | n -> fpr ff "@[[]@]"; pr_dims (n-1)
      in
      fpr ff "@[new "; pr_ty ff ty;  
      print_expl_dim dim_exp;  
      pr_dims dims_impl;
      fpr ff "@]"
  | EFieldAccess(exp, name) ->
      fpr ff "%a.%a" pr_exp exp pr_onename name
  | EFieldAccessSuper(name) ->
      fpr ff "base.%a" pr_onename name
  | EMethInvkBase(exp, name, args) -> 
      fpr ff "%a.%a%a" pr_exp exp pr_onename name pr_args args
  | EMethInvkLocal(name, args) -> 
      fpr ff "%a%a" pr_onename name  pr_args args
  | EMethInvkSuper(name, args) -> 
      fpr ff "base.%a%a" pr_onename name  pr_args args
  | EArrayAccess(arr, ind) -> 
      pr_exp ff arr;  
      fpr ff "[";  pr_exp ff ind;  fpr ff "]"
  | ECast(ty, exp, _) -> 
      fpr ff "(%a)%a" pr_ty ty pr_exp exp
  | ETypeStamp(sty, exp) -> 
      fpr ff "(%a)%a" pr_seqty sty pr_exp exp
  | EAssign(lhs, rhs) -> 
      fpr ff "(";
      pr_exp ff lhs;  fpr ff " = ";  pr_exp ff rhs;
      fpr ff ")"
  | EUnary(uop, exp) -> 
      fpr ff "@[(";  SynOp.pr_unary ff uop;  pr_exp ff exp;  fpr ff ")@]"
  | EBinary (exp1, bop, exp2) ->
      fpr ff "(";
      pr_exp ff exp1;  SynOp.pr_binary ff bop;  pr_exp ff exp2;
      fpr ff ")"
  | ESeq exps -> 
      fpr ff "@[{|";  Printing.print_list pr_exp " " "" ff exps;   fpr ff "|}@]"
  | ETree(lbl, atts, conts) ->
      fpr ff "@[%s(%a) %a%s" 
	"<" pr_exp lbl (Printing.print_list pr_attexp " " "") atts ">";  
      List.iter (pr_exp ff) conts;  
      fpr ff "</>"
  | EPcdata(lbl) ->
      fpr ff "%s" lbl;
  | EString2Pcdata exp -> 
      fpr ff "@[str2pcdata(%a)@]" pr_exp exp

and pr_exp_option ff expr =
  match Finfo.drop expr with
  | None -> fpr ff "<VOID>"
  | Some e -> pr_exp ff (Finfo.get expr, e)
      
and pr_attexp ff aexp = 
  match Finfo.drop aexp with 
  | ETree(aname, acnt, []) -> 
      let str_aname = match (Finfo.drop aname) with 
      |	Estring s -> s
      |	_ -> "NON-STRING-NAME"   (* error case *)
      in
      fpr ff "%s%s=(%a)" "@" str_aname (Printing.print_list pr_exp " " "") acnt
  | _ -> 
      fpr ff "%s(%a)" "@WRONG" pr_exp aexp   (* error case *)


and pr_args ff exps = 
  let print_arg exp = 
    pr_exp ff exp; fpr ff ",@ "
  in
  fpr ff "@[("; List.iter print_arg exps; fpr ff ")@]"


let rec pr_block ff block = 
  fpr ff "@[{@[";
  (match Finfo.drop block with
  | block' -> Printing.print_list pr_stmt "\n" "" ff block'
  );
  fpr ff "@]}@]"

and pr_vardecl ff (dn,bindexp) = 
  pr_decname ff dn;
  match bindexp with
    | None -> ()
    | Some(e) -> fpr ff " = %a" pr_exp e

and pr_stmt ff stmt = 
  match Finfo.drop stmt with
  | SBlock block ->
      pr_block ff block
  | SExp exp -> 
      fpr ff "%a;@ " pr_exp exp
  | SLocVars (vnames, ty) ->
      fpr ff "@[%a@ %a;@ @]" pr_ty ty (pr_commalist pr_vardecl) vnames
  | SReturn exp ->
      fpr ff "@[return(%a);@ @]" pr_exp_option exp
  | SIf(cond, block1, block2) ->
      fpr ff "@[";
      fpr ff "if (%a)@\n" pr_exp cond;
      fpr ff "%a@\n" pr_block block1;
      fpr ff "else %a@\n" pr_block block2;
  | SWhile(cond, body) -> 
      fpr ff "@[while (%a)@ @[%a@] @\n@]" pr_exp cond pr_block body
  | SMatch (exp, clauses) ->
      fpr ff "@[match (";  pr_exp ff exp;  fpr ff ") {@\n";
      fpr ff "  @[";
      List.iter (pr_clause ff) clauses;
      fpr ff "}@]@\n@]"
  | SForeachRE (spa, exp, body) -> 
      fpr ff "@[foreach (%a in %a)@ @[%a@] @\n@]" 
	pr_seqty spa pr_exp exp pr_block body
(*??  | SSwitch (exp, clauses) ->
      fpr ff "@[switch (";  pr_exp ff exp;  fpr ff ") {@\n";
      fpr ff "  @[";
      List.iter (pr_clause ff) clauses;
      fpr ff "}@]@\n@]"
*)
  | SEmpty ->
      fpr ff ";@ "
  | SBreak ->
      fpr ff "break;@ "


and pr_clause ff clause = 
  let (pat, block) = Finfo.drop clause 
  in
  fpr ff "@[case %a:@]@\n" pr_ty pat;
  fpr ff "    @[%a@]@\n" pr_block block


(********** Classes ***********)

let pr_params ff params = 
  let pr_param (pname, ty) = 
    fpr ff "@[";   pr_opt pr_decname ff pname;  fpr ff " : ";  
    pr_ty ff ty;  
    fpr ff ",@ @]"
  in
  fpr ff "@[(";  List.iter pr_param params;  fpr ff ")@]"

let pr_constr ff constr =
  let (hostclas, params, inv_opt, body) = Finfo.drop constr in
  let print_invopt () = 
    match inv_opt with 
    | (inf, InvkThis args) -> 
	fpr ff "@[this";  pr_args ff args; fpr ff ";@\n@]"
    | (inf, InvkSuper args) -> 
	fpr ff "@[super"; pr_args ff args; fpr ff ";@\n@]"
    | (inf, InvkNone) -> ()
  in
  pr_params ff params;   fpr ff " =@\n";  
  fpr ff "  ";
  print_invopt ();  pr_opt pr_block ff body;
  fpr ff "@\n"

let pr_status ff kind = 
  match kind with 
    | SynOp.Static -> fpr ff "static "
    | SynOp.Instance -> ()

let pr_mod ff modifier =
    match (Finfo.drop modifier) with
    | SynOp.Virtual -> fpr ff "virtual "
    | SynOp.Override -> fpr ff "override "
    | SynOp.New -> fpr ff "new "
    | SynOp.Normal -> ()

let pr_meth ff (meth : meth) = 
  let (name, mangopt, status, modifier, rettyp, params, bodyopt) = 
    Finfo.drop meth in
  pr_status ff status;
  pr_mod ff modifier;
  pr_decname ff name;  pr_opt pr_mangle ff mangopt; fpr ff " : "; 
  pr_params ff params;  fpr ff " -> "; 
  pr_ty_option ff rettyp; fpr ff " =@\n";  
  fpr ff "  ";
  pr_opt pr_block ff bodyopt;
  fpr ff "@\n"

let pr_clas ff clas = 
  let print_field fld = 
    let (name, mangopt, expopt, status, ty) = Finfo.drop fld in
      fpr ff "@[FLD ";
      pr_status ff status;
      pr_decname ff name;
      pr_opt pr_mangle ff mangopt;
      fpr ff " : ";   pr_ty ff ty;  
      pr_opt (fun ff e -> fpr ff " = %a" pr_exp e) ff expopt;
      fpr ff "@]@\n"   in 
  let print_constr cns = 
    fpr ff "@[CONSTR : "; pr_constr ff cns;  fpr ff "@]@\n" in
  let print_meth  mth = 
    fpr ff "@[METH "; 
    pr_meth ff mth;  fpr ff "@]@\n" 
  in
  let clas' = Finfo.drop clas in
  fpr ff "extends "; List.iter (pr_refname ff) clas'.clas_extends;  fpr ff " ";
  fpr ff "{@\n";
  fpr ff "  @[";
  List.iter print_field clas'.clas_fields;
  fpr ff "@\n";
  print_constr clas'.clas_constr;
  List.iter print_meth clas'.clas_meths;
  fpr ff "@]}"


(****** Program *********)

let pr_prog ff prog = 
  let prdecl_clas id (nctx, clas) = 
    fpr ff "@[class %a %a@]@\n" Ident.print id pr_clas clas
  in
  let prdecl_seqty id (nctx, seqty) = 
    fpr ff "@[regtype %a %a@]@\n" Ident.print id pr_seqty seqty
  in
  Ident.iter prdecl_seqty prog.seq;
  Ident.iter prdecl_clas prog.cls



