(** Signatures for modules implementing type expressions that are
    later included in SynDenot. *)
 
(** The module architecture imposed by the signatures below may seem
    complicated, but it is motivated by the following goals: 
    (1) make it easy to vary internal implementations of hashing and
        memoization for complex type expressions, while
    (2) providing a stable interface to the rest of the program, and
    (3) localizing in one place the things that require changes when
        switching from one implementation to another.

    Therefore, 
    - PUBLIC is the ADT for the syntax of types that provides a
      fuctionally complete set of operations with the guarantees of
      all invariants.  Later, SynDenot extends it with a few
      convenience functions.
    - IMPLEM specifies a module that localizes code specific to a
      hashing approach. There can be different implementations of this
      module.  PUBLIC is expected to be implemented by a _functor_
      that takes an IMPLEM module and extends it.
    - CORE is just a common part of PUBLIC and IMPLEM, included in
      both.
 *)

(** Since Xtatic sequence types can be large hierarchical expressions,
    their naive structural comparizon would be expensive.  Therefore,
    this interface for creating type expressions is designed to
    guarantee that, for any two type expressions ever created, if they
    are structurally equal (up to some Xtatic syntactic equivalence),
    then they are physically equal (as OCaml values). This has to be
    ensured by keeping a cache of all expressions ever created.
    Consequently, the physical equaity (==) can be used in tests. 

    The setup is as follows.  (See CORE and PUBLIC for real defs.)
    - For each category XXX of interest, e.g. (sequence types, or host
      types, etc.), there are two OCaml types, xxx and xxx'.
    - xxx' is an OCaml datatype with public constructors whose values
      are "descriptions" of XXX entities.
    - xxx is an abstract type for XXX entities whose values are to be
      used in the rest of the program.
    - Structurally equal values of xxx' are not guaranteed to be
      physically unique, but the values of xxx are.
    - A value of xxx can be created only by applying the "constructor
      function" xxx : xxx' -> xxx to a previously created description
      of type xxx'.  The costructor function consults the cache of
      previously created xxx to ensure physical uniqueness of the result.
    - The structure of an xxx value cannot be explored directly, but
      only after applying the destructor function xxx' : xxx -> xxx'
      to obtain its description.

    For this setup, the clients of this module do not have to worry
    about maintaining the physical uniqueness invariants -- they are
    taken care of by the module itself.
*)



(** CORE: type definitions for xxx and xxx' types. *)
(* No module implements just CORE -- it is merely the common part of
   PUBLIC and IMPLEM. *)
module type CORE = 
sig

type hostty (* host type *)
type ty (* type *)
type labty (* label type *)
type seqty (* sequence type *)
type attsty (* attribute type *)

type hostty' = 
  | TClasRef of Ident.id 
  | TIntfRef of Ident.id 
  | TObject
  | TNull
  | TBool
  | Tbool of bool
  | TInt
  | Tint of int
  | TChar
  | Tchar of char
  | TString
  | Tstring of string
  | TTag 
  | Ttag of string 
  | TSeq
  | TArray of ty

type ty' = 
  | Thost of hostty
  | Tseq of seqty
  | PTyBind of Ident.id * ty

type labty' = 
  | TLab of hostty
  | PLabBind of Ident.id * labty
	(** ty's appearing in labty must have no var binders *)

type seqty' = 
  | TTree of labty * attsty * seqty  
      (*adhoc: attrs are ignored everywhere, except in pattern-matching *)
  | TCat of seqty list
  | TAlt of seqty list
  | TRep of seqty 
  | TSeqRef of Ident.id
  | PSeqBind of Ident.id * seqty
  | PTailBind of Ident.id * seqty

and attpat = string * Ident.id * seqty   (* (attr name, var, attr value type)*)

and attsty' = 
  | TAttAny
  | PAtt of attpat list * (Ident.id * seqty) option   
           (** individ attrs and catch-all *)
(**NB: seqty in the defs of attribute types is a hack needed for
   figuring out type of attribute variables during type inference in
   Infer.retrieve_lab_att_bindings.  seqty in attpat is always pcdata,
   and seqty in the catch-all is always attseq -- see src2denot. *)

(** Type destructors *)
val attsty': attsty -> attsty'
val hostty': hostty -> hostty'
val ty': ty -> ty'
val labty': labty -> labty'
val seqty': seqty -> seqty'

end (* CORE *)


(** PUBLIC: Signature of a module that ensures representation invariants. *)
(* There must be only one module (or functor) implementing this signature. *)


module type PUBLIC =  
sig
  include CORE
 
(**  Memoizing type constructors *)
(** Each type constructor ensures that if it ever returns two type
    expressions that are structurally equal, then in both cases the
    same physical Ocaml value was returned.  *)
val attsty: attsty' -> attsty
    (** Canonization: named attributes are ordered. *)
val hostty: hostty' -> hostty
    (** Canonization: classids for basic types (int, ...) are translated
        into corresponding native types (TInt, ...). *)
val ty: ty' -> ty
val labty: labty' -> labty
val seqty: seqty' -> seqty
    (** Canonization: As many as possible nested TAlt and TCat
    constructs gets collapsed, i.e. for any TCat[s1; s2; ...] occuring
    in the result, none of si is a TCat, similarly for TAlt.  Also,
    for TAlt only, duplicate members are removed and the remaining
    ones are sorted (by compare_seqty). *)

(** A couple of important pre-computed and pre-memoized types: *)
val sty_epsilon : seqty  (** TCat [] *)
val sty_empty   : seqty  (** TAlt [] *)

(** Comparison functions: for use in ordered storage datastructures. 
    (The order does not have semantic meaning for any other purpose.) *)
val compare_hostty: hostty -> hostty -> int
val compare_ty: ty -> ty -> int
val compare_labty: labty -> labty -> int
val compare_seqty: seqty -> seqty -> int
val compare_attsty: attsty -> attsty -> int

(** Hash tables **)
module SeqTbl : Hashtbl.S with type key = seqty

end (* PUBLIC *)


(** Implementation of a particular hashing mechanism.  *)
(* There can be alternative modules implementing just this signature. *)
module type IMPLEM = 
sig
  include CORE

(** NB The actual definitions of xxx types are supposed to be done in
    this module.  Depending on the particular choice, either cons_xxx
    or hash_xxx functions can be trivial.  Also, the actual
    computation of hashes can happen either in cons_xxx or in
    hash_xxx. *)

(** Pre-constructors (non-memoizing) *)
(** The pre-constructors assume that an xxx' value passed to them is 
    already canonized.  *)
val cons_attsty: attsty' -> attsty
val cons_hostty: hostty' -> hostty
val cons_ty: ty' -> ty
val cons_labty: labty' -> labty
val cons_seqty: seqty' -> seqty

(* Hash values *)
val hash_attsty: attsty -> int
val hash_hostty: hostty -> int
val hash_ty: ty -> int
val hash_labty: labty -> int
val hash_seqty: seqty -> int

val hash_attsty': attsty' -> int
val hash_hostty': hostty' -> int
val hash_ty': ty' -> int
val hash_labty': labty' -> int
val hash_seqty': seqty' -> int

end (* IMPLEM *)
