(** An implementation of hashing policy for type expressions: hash
    values are stored in each node of the expression, to avoid
    recomputation.
*)

module M : SynTyp.IMPLEM = 
struct

type hostty' = 
  | TClasRef of Ident.id 
  | TIntfRef of Ident.id
  | TObject
  | TNull
  | TBool
  | Tbool of bool
  | TInt
  | Tint of int
  | TChar
  | Tchar of char
  | TString
  | Tstring of string
  | TTag 
  | Ttag of string 
  | TSeq
  | TArray of ty
and hostty = 
    { hty_hash: int;  hty_def: hostty' }

and labty' = 
  | TLab of hostty
  | PLabBind of Ident.id * labty
	(** ty's appearing in labty must have no var binders *)
and labty = 
    { lty_hash: int; lty_def: labty' }

and seqty' = 
  | TTree of labty * attsty * seqty  
      (*adhoc: attrs are ignored everywhere, except in pattern-matching *)
  | TCat of seqty list
  | TAlt of seqty list     (* in seqty, the list is sorted *)
  | TRep of seqty 
  | TSeqRef of Ident.id
  | PSeqBind of Ident.id * seqty
  | PTailBind of Ident.id * seqty
and seqty = { sty_hash: int;  sty_def: seqty' }

and attpat = string * Ident.id * seqty   (* (attr name, var, attr value type)*)

and attsty' = 
  | TAttAny
  | PAtt of attpat list * (Ident.id * seqty) option   
            (* individ attrs and catch-all *)
and attsty = 
    { aty_hash: int;  aty_def: attsty' }

and ty' = 
  | Thost of hostty
  | Tseq of seqty
  | PTyBind of Ident.id * ty
and ty = 
    { ty_hash: int;  ty_def: ty' }


(************* Destructors *************)
let attsty' t = t.aty_def
let hostty' t = t.hty_def
let ty' t = t.ty_def
let labty' t = t.lty_def
let seqty' t = t.sty_def


(************* Hash readers *************)
let hash_attsty t = t.aty_hash
let hash_hostty t = t.hty_hash
let hash_labty t = t.lty_hash
let hash_seqty t = t.sty_hash
let hash_ty t = t.ty_hash



(********** Hash computation *******)

let hash_combine h h' = h + 17 * h'

let hash_attpat (aname, avar, atyp) = 
  hash_combine (Hashtbl.hash aname)
    (hash_combine (Ident.hash avar) (atyp.sty_hash))

let hash_attsty' = function
  | TAttAny -> 1
  | PAtt(aps, catchopt) -> 
      let h_aps = 
	List.fold_left (fun h p -> hash_combine h (hash_attpat p)) 1 aps in
      let h_catchopt = match catchopt with
      | None -> 1
      | Some (var, typ) -> hash_combine (Ident.hash var) (typ.sty_hash)
      in hash_combine h_aps h_catchopt
  

let hash_hostty' hty' = Hashtbl.hash hty'


let hash_labty' = function 
  | TLab hty -> hty.hty_hash
  | PLabBind(var, lty) -> hash_combine (Ident.hash var) (lty.lty_hash)
  

let hash_seqty' = function 
  | TTree (lty, aty, sty) -> 
      hash_combine (lty.lty_hash)
	(hash_combine (aty.aty_hash) (sty.sty_hash))
  | TCat lst -> 
      List.fold_left (fun h s -> hash_combine h (s.sty_hash)) 1 lst
  | TAlt lst -> 
      List.fold_left (fun h s -> hash_combine h (s.sty_hash)) 2 lst
  | TRep sty -> 
      hash_combine 3 sty.sty_hash
  | TSeqRef name -> Ident.hash name
  | PSeqBind (var, sty) -> 
      hash_combine 4 
	(hash_combine (Ident.hash var) (sty.sty_hash))
  | PTailBind (var, sty) -> 
      hash_combine 5
	(hash_combine (Ident.hash var) (sty.sty_hash))
  

let hash_ty' = function
  | Thost hty -> hty.hty_hash
  | Tseq sty -> sty.sty_hash
  | PTyBind(var, ty) -> hash_combine (Ident.hash var) (ty.ty_hash)


(************ Pre-constructors ********) 

let cons_attsty t' = 
  { aty_hash = hash_attsty' t';  
    aty_def = t' }

let cons_hostty t' = 
  { hty_hash = hash_hostty' t';  
    hty_def = t' }

let cons_labty t' = 
  { lty_hash = hash_labty' t';  
    lty_def = t' }

let cons_seqty t' = 
  { sty_hash = hash_seqty' t';  
    sty_def = t' }

let cons_ty t' = 
  { ty_hash = hash_ty' t';  
    ty_def = t' }


end
