(** An implementation of hashing policy: hashes of type expressions
    are not stored, but rather computed every time they are needed.
    However, the computation traverses the expressions only to
    pre-determined depth given by the functor parameters (see
    hash_xxx' functions for their use). 

    This saves on memory space (compared to hash-carry approach), and
    the depth-bounded traversal during hashing might be cheap enough. 
*)


module M (MaxNodes : 
	    sig val meaningful_nodes : int 
	        val total_nodes : int     end) 
  : SynTyp.IMPLEM  = 
struct

(* NB: in this implementation, xxx and xxx' are the same type; this
   fiddling is needed to comply to the signature designed to
   accomodate more sophisticated implementations. *)

type hostty' = 
  | TClasRef of Ident.id 
  | TIntfRef of Ident.id
  | TObject
  | TNull
  | TBool
  | Tbool of bool
  | TInt
  | Tint of int
  | TChar
  | Tchar of char
  | TString
  | Tstring of string
  | TTag 
  | Ttag of string 
  | TSeq
  | TArray of ty
and hostty = hostty'

and labty' = 
  | TLab of hostty
  | PLabBind of Ident.id * labty
	(** ty's appearing in labty must have no var binders *)
and labty = labty'

and seqty' = 
  | TTree of labty * attsty * seqty  
      (*adhoc: attrs are ignored everywhere, except in pattern-matching *)
  | TCat of seqty list
  | TAlt of seqty list     (* in seqty, the list is sorted *)
  | TRep of seqty 
  | TSeqRef of Ident.id
  | PSeqBind of Ident.id * seqty
  | PTailBind of Ident.id * seqty
and seqty = seqty'

and attpat = string * Ident.id * seqty   (* (attr name, var, attr value type)*)

and attsty' = 
  | TAttAny
  | PAtt of attpat list * (Ident.id * seqty) option   
            (* individ attrs and catch-all *)
and attsty = attsty' 
      (* but the list is sorted by attr names *)

and ty' = 
  | Thost of hostty
  | Tseq of seqty
  | PTyBind of Ident.id * ty
and ty = ty'


(**********************************************************************)
      (** Destructors  *)

let attsty' attsty = attsty
let hostty' hostty = hostty
let ty' ty = ty
let labty' labty = labty
let seqty' seqty = seqty

(********* Pre-constructors ****************)
let cons_attsty t' = t'
let cons_hostty t' = t'
let cons_labty t' = t'
let cons_seqty t' = t'
let cons_ty t' = t'


(************************************************************************)
                        (** Hash functions *****)

(*  Generic hash function instead of custom ones:
 (This makes the thing work, but could eliminate the efficiency gains that 
   are expected from hash-augmented seqtypes...)  *)
(*## Hey, using H.hash_param instead of H.hash or custom functions can be a 
  good idea not just for debugging.  Then there is no need to augment 
  seqtys with the hash values, what can significantly save on memory, 
  while reasonable parameters for hash_param might provide 
  collision-elimination benefit comparable to that of custom hash functions. *)

let hash_attsty' = Hashtbl.hash_param 3 5
let hash_hostty' = Hashtbl.hash_param 3 5
let hash_ty' =     Hashtbl.hash_param 3 5
let hash_labty' =  Hashtbl.hash_param 3 5
let hash_seqty' =  Hashtbl.hash_param 3 5

let hash_attsty = hash_attsty'
let hash_hostty = hash_hostty'
let hash_ty = hash_ty'
let hash_labty = hash_labty'
let hash_seqty = hash_seqty'

end
