module Make(SynTypImplem : SynTyp.IMPLEM) : SynTyp.PUBLIC = 
struct
  include SynTypImplem

(**********************************************************************)
      (** Comparison functions *)

let compare_hashable (hash: 'a -> int) t1 t2 = 
  if t1 == t2 then 0 else
  let h1, h2 = hash t1, hash t2 in
  if h1 < h2 then -1 else
  if h1 > h2 then 1 else
  Pervasives.compare t1 t2


let compare_attsty = compare_hashable hash_attsty
let compare_hostty = Pervasives.compare
let compare_labty = compare_hashable hash_labty
let compare_seqty = compare_hashable hash_seqty
let compare_ty = compare_hashable hash_ty


(***********************************************************************)
      (** Equality functions *)

(** These equality functions compute the same function that the
    OCaml's structural equality (=) would do (on "fully unfolded" xxx'
    descriptions, to be precise), but more efficiently, since they
    short-circuit via (==) below the first level of xxx' data
    constructors.  This is safe, since below the constructors equality
    (==) is guaranteed to coinside with the structural equality (this
    is ensured by the constructor functions.)  *)

let rec equal_attpats aps1 aps2 = 
    (match aps1, aps2 with
    | [], [] -> true
    | (an1, v1, s1)::aa1, (an2, v2, s2)::aa2 -> 
	an1 = an2 && v1 == v2 && equal_attpats aa1 aa2
	(*##ATT ignore s1, s2 -- both must be pcdata *)
    | _ -> false)
 
let rec equal_attsty' ats1' ats2' = 
  match ats1', ats2' with 
  | TAttAny, TAttAny -> true 
  | PAtt(aps1, catch1), PAtt(aps2, catch2) ->
      (match catch1, catch2 with 
      | None, None -> true
      | Some(id1,s1), Some(id2,s2) -> 
	  id1 == id2 && s1 == s2 
      | _ -> false)
      &&
      equal_attpats aps1 aps2
  | _ -> false
(* Assumes at1, at2 are sorted *)
(*NB neither same_list nor physical equality are appropriate for atts *)


let rec same_list l l' = 
  l == l' ||
  match l, l' with
    [], [] -> true
  | x :: r, x' :: r' -> x == x' && same_list r r'
  | _ -> false


let equal_hostty' hty1' hty2' = 
  (hty1' = hty2')
  (* This is efficient only for hostty; in the other types we 
     we use == to prevent structural traversal inside complex values. *)


let  equal_ty' ty1' ty2' = 
   match ty1', ty2' with
   | Thost ht1, Thost ht2 -> ht1 == ht2
   | Tseq seq1, Tseq seq2 -> seq1 == seq2
   | PTyBind(id1, t1), PTyBind(id2, t2) -> 
       id1 == id2 && t1 == t2
   | _, _ -> false


let equal_labty' lty1' lty2' = 
  match lty1', lty2' with 
  | TLab ty1, TLab ty2 -> ty1 == ty2
  | PLabBind(id1, l1), PLabBind(id2, l2) -> 
      id1 == id2 && l1 == l2
  | _, _ -> false


let equal_seqty' seqty1' seqty2' =
  match  seqty1',  seqty2' with
  | TTree(lab1, ats1, seq1), TTree(lab2, ats2, seq2) -> 
       lab1 == lab2 && ats1 == ats2 && seq1 == seq2
  | TCat seqs1, TCat seqs2 -> same_list seqs1 seqs2
  | TAlt seqs1, TAlt seqs2 -> same_list seqs1 seqs2
  | TRep seq1, TRep seq2 -> seq1 == seq2
  | TSeqRef id1, TSeqRef id2 -> 
      id1 == id2
  | PSeqBind(v1, seq1), PSeqBind(v2, seq2) -> 
      v1 == v2 && seq1 == seq2
  | PTailBind(v1, seq1), PTailBind(v2, seq2) -> 
      v1 == v2 && seq1 == seq2
  | _, _ -> false


                  (**  Hash tables *)

(** Tables for hash consing. 
    For a type expression t, a table maps it to itself (that is, the
    table contains identity pairs (t,t).  Since the equality function
    of the table checks for structural equality (see def of eq
    functions above), the table effectively maps a class of
    structurally equivalent type expressions to a single "canonical"
    element of the class.
*)

(** These tables are only for use within this module, for caching.
    Exported tables with xxx as keys are defined later. *)
module AttsTbl' = 
  Hashtbl.Make (struct type t = attsty' 
    let equal = equal_attsty' let hash = hash_attsty' end)

module HostTbl' = 
  Hashtbl.Make (struct type t = hostty'
    let equal = equal_hostty' let hash = hash_hostty' end)

module TyTbl' = 
  Hashtbl.Make (struct type t = ty'
    let equal = equal_ty' let hash = hash_ty' end)

module LabTbl' = 
  Hashtbl.Make (struct type t = labty'
    let equal = equal_labty' let hash = hash_labty' end)

module SeqTbl' = 
  Hashtbl.Make (struct type t = seqty'
    let equal = equal_seqty' let hash = hash_seqty' end)


(************  Canonization and memoization  ********************************)

(* A canonization function transforms a type expression into a
  "canonical" form; this is the form of expressions stored in the
  cache tables. *)

(* hostty *)
let canonize_hostty' (hty': hostty') : hostty' = 
  match hty' with
  | (TClasRef cid) as clsref -> 
      if (cid == Kerids.object_id) then TObject else
      if (cid == Kerids.int_id) then TInt else
      if (cid == Kerids.bool_id) then TBool else
      if (cid == Kerids.string_id) then TString else
      if (cid == Kerids.char_id) then TChar else
      if (cid == Kerids.tag_id) then TTag else
      if (cid == Kerids.seq_id) then TSeq else
      clsref
  | nonclassref -> nonclassref

let cacheHostty = HostTbl'.create 100

let hostty t' = 
  let t' = canonize_hostty' t' in
  try
    HostTbl'.find cacheHostty t'
  with Not_found ->
    let t = cons_hostty t' in
    HostTbl'.add cacheHostty t' t;
    t


(* ty *)
let cacheTy = TyTbl'.create 99

let ty t' = 
  try
     TyTbl'.find cacheTy t'
   with Not_found ->
     let t = cons_ty t' in
     TyTbl'.add cacheTy t' t;
     t


(* labty *)
let cacheLabty = LabTbl'.create 93

let labty t' = 
  try 
    LabTbl'.find cacheLabty t'
  with Not_found -> 
    let t = cons_labty t' in 
    LabTbl'.add cacheLabty t' t;
    t


(* seqty *)
let cacheSeqty = SeqTbl'.create 153

let sty_epsilon' = TCat[]
let sty_epsilon = cons_seqty sty_epsilon'
let _ = SeqTbl'.add cacheSeqty sty_epsilon' sty_epsilon
let sty_empty' = TAlt[]
let sty_empty   = cons_seqty sty_empty'
let _ = SeqTbl'.add cacheSeqty sty_empty' sty_empty

let canon_cat stys = 
  let canon sty result = 
    match seqty' sty with 
    | TAlt[] -> raise Exit (* The whole sequence is TAlt[] in this case *)
    | TCat[] -> result
    | TCat[t] -> assert false  (* by canonization, there are no singletons *)
    | TCat s -> s @ result
    | _ -> sty :: result
  in List.fold_right canon stys []

let canon_alt stys = 
  let alt_merge = Sortedlist.merge compare_seqty
  in
  let canon sty result = 
    match seqty' sty with
    | TAlt [] -> result
    | TAlt s -> alt_merge s result
    | _ -> alt_merge [sty] result
  in List.fold_right canon stys []
    
type canon_result = 
  | CanonSeqty of seqty     (* canonization resulted in a "full" type  *)
  | CanonSeqty' of seqty'   (* "description" type in canonical form *)

let canonize_seqty (sty': seqty') : canon_result = 
  match sty' with
  | TCat stys -> 
      (try 
	(match canon_cat stys with 
	| [t] -> CanonSeqty t
	| []  -> CanonSeqty sty_epsilon
	| lst -> CanonSeqty' (TCat lst)
	)
      with Exit -> CanonSeqty sty_empty)  
  | TAlt stys -> 
      (match canon_alt stys with
      | [t] -> CanonSeqty t
      | []  -> CanonSeqty sty_empty
      | lst -> CanonSeqty' (TAlt lst)
      )
  | _ -> CanonSeqty' sty'


let  seqty t' = 
  match canonize_seqty t' with
  | CanonSeqty t -> t
  | CanonSeqty' t' -> 
      try 
	SeqTbl'.find cacheSeqty t'
      with Not_found -> 
	let t = cons_seqty t' in
	SeqTbl'.add cacheSeqty t' t;
	t



(* attsty *)
let canonize_attsty = function 
  | TAttAny -> TAttAny
  | PAtt(attpats, catchopt) -> 
      let c_attpats = 
	(List.sort (fun (an1, _, _) (an2, _, _) -> compare an1 an2) attpats) 
      in PAtt(c_attpats, catchopt)

let cacheAttsty = AttsTbl'.create 67

let attsty t' = 
  let t' = canonize_attsty t' in
  try
    AttsTbl'.find cacheAttsty t'
  with Not_found ->
    let t = cons_attsty t' in
    AttsTbl'.add cacheAttsty t' t;
    t


           (** Exported hash tables *)

module SeqTbl = 
  Hashtbl.Make (struct type t = seqty
    let equal = (==)   let hash = hash_seqty end)


end
