module SD = SynDenot

type exp =
  { exp_desc: exp_desc;
    exp_type: SD.ty option; }

and exp_desc = 
  (** Expressions for literals *)
  | Ebool of bool
  | Eint of int
  | Echar of char 
  | Estring of string 
  | Etag of string 
  | Enull

  (** OO host expressions *)
  | EThis
  | EClass of Ident.id (* for static access *)
  | EVar of Ident.id
  | ENew of Ident.id * (exp list)
  | ENewArray of SynDenot.ty * exp * int 
  | EFieldAccess of exp * Ident.id
  | EMethInvk of exp * (Ident.id option) * Ident.id * string * (exp list)
  | EProcMethInvk of exp * (Ident.id option) * Ident.id * string * (exp list)
      (** (Some cid) imposes early binding: at run time, the method 
	 _must_ be taken from cid class, regardless whether the class
	 of base is a proper subtype of cid.
	 None indicates late binding: the run-time class of base should
	 be used.  *)
       (**
          EProcMethInvk does not return a result (its return type is None), EMethInvk does
     *)
       (** the string near the end is the mangled method name (identical to the
         * name if no mangling *)

  | EArrayAccess of exp * exp
  | EAssign of exp * exp   (* lhs := rhs *)
  | ECast of SynDenot.ty * exp * bool
  | ETypeStamp of SynDenot.seqty * exp  
  | EUnary of SynOp.unaryop * exp            (*  ! - +                       *)
  | EBinaryArith of exp * SynOp.binaryop * exp  (*  + - / % * < > <= >=    *)
  | EBinaryBool of exp * SynOp.binaryop * exp   (*  && ||                   *)
  | EConcatString of exp * exp            (*  +                           *)
  | EBinaryEq of exp * bool * exp   (*  == !=                       *) 


  (** Sequence expressions *)
  | ESeq of exp list            (* {| ... |}  *)
  | ETree of exp * exp list * exp list         
     (*  (label, attributes, contents) *)
  | EPcdata of string
  | EString2Pcdata of exp   (* a hack, used only for short-hand attributes *)


  (** Statements *)
and block = stmt list

and stmt = 
  | SBlock of block
  | SExp of exp
  | SLocVars of ((Ident.id * exp option) list * SynDenot.ty * bool)
      (* the last component is a flag which = true if the declared
         type is a pattern with binders, and = false if it is a proper type *)
  | SReturn of exp option
  | SIf of exp * block * block
  | SWhile of exp * block
(*??  | SSwitch of exp * clause list     (* for host matching *) *)
  | SForeachRE of SynDenot.seqpa * exp * block
  | SMatch of exp * clause list      (* for XDuce matching *)
  | SEmpty 
  | SBreak

and clause = SynDenot.pa * block 


type form_param = Ident.id * SynDenot.ty

type constr = invoke * (form_param list) * block
and invoke = 
  | InvkNone
  | InvkThis of exp list
  | InvkSuper of exp list

type meth = SynOp.method_modifier * SynOp.status * (form_param list) * SynDenot.ty option * block
type meth_decl = (form_param list) * SynDenot.ty option

type clas = {
    clas_id: Ident.id;
    clas_parent: Ident.id;
    clas_implements: Ident.id list;
    clas_fields: (Ident.id * (SynOp.status * SynDenot.ty * exp option)) list; 
    clas_constr: constr;
    clas_meths: (Ident.id * meth * string) list;
    (** last string is mangled name *)
  } 

type intf = {
    intf_id: Ident.id;
    intf_implements: Ident.id list;
    intf_meths_decls: (Ident.id * meth_decl * string) list;
    (** last string is mangled name *)
  } 

(* NB: clas does not have to carry info about fields and declared types of 
   constructors and classes, since it will be available from clsctx. *)
      
type prog =
  { classes : clas Ident.tbl;
    intfs : intf Ident.tbl
  }

let get_desc x = x.exp_desc
let get_type x = x.exp_type

let get_real_type x =
  match x.exp_type with
    | Some t -> t
    | None -> assert false

let make e t = {exp_desc = e; exp_type = Some t}
