module F =  Format
module SD = SynDenot
module AR = Array
module HT = Hashtbl
module ST = Stack
open List

(* Preactions *)

type act = 
  | Begin of Ident.id
  | End of Ident.id
  | Rest of Ident.id
  | Fixed of Ident.id * int
type pre = act list

type binder =
  | BNAttr of Ident.id
  | BNLabel of Ident.id * SD.hostty
  | BNSeq of act

let print_act ch = function
  | Begin x -> F.fprintf ch "B(%a)" Ident.print x
  | End x -> F.fprintf ch "E(%a)" Ident.print x
  | Rest x -> F.fprintf ch "R(%a)" Ident.print x
  | Fixed (x,i) -> F.fprintf ch "F%d(%a)" i Ident.print x

let print_pre = Printing.print_list print_act " " "" 

(* Tree automata *)

type state =
  { num : int;    (*VG was: state.id *)
    mutable final : bool;
    mutable finpre : pre;
    mutable next : trans list;
    mutable desc : SD.seqty option;
    mutable empty : bool option; 
    mutable binders : binder list option;
    mutable preclos : pre;
(*--VG not used...
    mutable nochk : bool;
    mutable exact : bool option;
    mutable always : bool;
*)
  }
and trans = (SD.labty * SD.attsty * state * state * pre) 

let compare_act x y = match (x,y) with
  | Begin x, Begin y -> Ident.compare x y
  | Begin _, _ -> (-1)
  | _, Begin _ -> 1
  | End x, End y -> Ident.compare x y
  | End _, _ -> (-1)
  | _, End _ -> 1
  | Rest x, Rest y -> Ident.compare x y
  | Rest _, _ -> (-1)
  | _, Rest _ -> 1
  | Fixed(x,k), Fixed(y,j) ->
      let x = Ident.compare x y in
      if x != 0 then x
      else compare k j

let compare_state s1 s2 = compare s1.num s2.num

let compare_pre b1 b2 =
  Sortedlist.compare compare_act (sort compare_act b1) (sort compare_act b2)

let compare_trans (c1,a1,s11,s12,b1) (c2,a2,s21,s22,b2) =
  let x = SD.compare_labty c1 c2 in
  if x != 0 then x
  else let x = SD.compare_attsty a1 a2 in
  if x != 0 then x
  else let x = compare (s11.num,s12.num) (s21.num,s22.num) in
  if x != 0 then x
  else compare_pre b1 b2
    

let last_num = ref 0

let number_of_states () : int = !last_num + 1

let bogus =
  {num = 0; final = false; finpre = []; next = []; desc = None;
   empty = None; binders = None; preclos = [] }

let (_states_ : state array ref) = ref (AR.make 1 bogus)

let ensure_size (i : int) : unit =
  let len = AR.length !_states_ in
  if len <= i then
    _states_ :=  AR.append !_states_ !_states_

let index_to_state (i : int) : state =
  !_states_.(i)

let create_state final next desc =
  incr last_num;
  ensure_size !last_num;
  let s = { num = !last_num; 
            final = final;
            finpre = [];
            next = next;
            empty = None;
            binders = None;
            desc = desc;
            preclos = [];
          (*-- not used...
            exact = None;
            nochk = false;
            always = false;
          *)
          } in
  !_states_.(!last_num) <- s;
  s

(* special states *)

let empty = create_state false [] None
let epsilon = create_state true [] None


(* Operations on state sets *)

let union_trans l1 l2 = l1 @ l2

let state_set_next x = 
  Listutil.union_many union_trans (map (fun s -> s.next) x)

let state_set_final s = exists (fun x -> x.final) s

(*?? let state_set_is_obviously_any x = exists (fun s -> s == !any()) x *)

let compare_state_sets l1 l2 = Sortedlist.compare compare_state l1 l2
let union_state_sets l1 l2 = Sortedlist.merge compare_state l1 l2
let diff_state_sets = Sortedlist.diff compare_state

module StSet = 
  Set.Make(struct
    type t = state
    let compare x y = compare x.num y.num
  end)

module StMap = 
  Map.Make(struct
    type t = state
    let compare x y = compare x.num y.num
  end)

(* Empty state elimination *)

(* just a helper, redefined below for toplevel *)
let rec is_empty env x1 =
  match x1.empty with
    Some b -> b
  | None ->
      not x1.final &&
      (StSet.mem x1 env ||
       is_empty_l (StSet.add x1 env) x1.next)

(* just a helper, redefined below for toplevel *)
and is_empty_l env xl =
  for_all (fun (c,a,x1,x2,_) ->
    SD.is_emptylab c ||
    is_empty env x1 || 
    is_empty env x2)
    xl

let rec elim_empty1 x1 =
  match x1.empty with
    Some b -> not b 
  | None ->
      let b = is_empty StSet.empty x1 in
      x1.empty <- Some b;
      x1.next <- elim_empty_l x1.next;
      not b

and elim_empty_l xl =
  filter (fun (c,a,y1,y2,_) ->
    let y1 = elim_empty1 y1 in
    let y2 = elim_empty1 y2 in
    not(SD.is_emptylab c) && y1 && y2)
    xl

let is_empty x = is_empty StSet.empty x
let is_empty_l l = is_empty_l StSet.empty l

let elim_empty x = 
  let res = filter elim_empty1 x in
  res

let elim_empty_l xl = 
  let res = elim_empty_l xl in
  res

(** get_binders utility *)

let comp x1 x2 = compare x1.num x2.num

let uns = Sortedlist.merge comp

let diffs = Sortedlist.diff comp

let substates (x : state) : state list =
  if x == empty then []
  else
    fold_right (fun (_,_,s1,s2,_) res -> uns [s1] (uns [s2] res))
      x.next
      []

let (sort : binder list -> binder list) = Sort.list (fun x y -> compare x y < 0)

let unb (b1:binder list) (b2:binder list) =
  let res = Sortedlist.merge compare b1 b2 in
  res

let lab2bns (l : SD.labty) : binder list =
  let h, ids = SD.crack_labty l in
  sort (map (fun id -> BNLabel(id, h)) ids)

let att2bns (atts : SD.attsty) : binder list = match SD.attsty' atts with
  | SD.TAttAny -> []
  | SD.PAtt(attpats, resto) ->
      fold_right (fun (_, id, _) res -> unb [BNAttr id] res) attpats
        (match resto with
          | None -> []
          | Some(id,_) -> [BNAttr id])

let pre2bns (p : pre) : binder list =
  sort (map (fun a -> BNSeq a) p)

let subbinders (x : state) : binder list =
  let combine_binders (l,atts,_,_,p) (res:binder list) =
    unb res (unb (lab2bns l) (unb (att2bns atts) (pre2bns p)))
  in fold_right combine_binders x.next (pre2bns x.finpre) 

let substates_binders (x : state) : state list * binder list =
  match x.binders with
    | None ->
        let sts = substates x in
        let bns = subbinders x in
        sts, bns
    | Some(bns) -> [], bns

let collect_binders (x : state) : binder list =
  let rec loop sts unprocessed = match unprocessed with
    | [] -> []
    | s::unprocessed ->
        let ss,bnds = substates_binders s in
        unb bnds (loop (uns ss sts) (uns unprocessed (diffs ss sts)))
  in loop [x] [x]

let get_binders (x : state) : binder list =
  match x.binders with
    | Some(bns) -> bns
    | None ->
        let bnds = collect_binders x in
        x.binders <- Some bnds;
        bnds

let has_binders (x : state) : bool = get_binders x != []



(***************** Map and Set modules ******************)

(* Hyp, HypSet, HypMap -- from Haruo's Sub *)
let rec hashl xs =
  match xs with
    [] -> 0
  | x :: xs' -> x.num + 239 * hashl xs'

module IntLsLs = 
  struct
    type t = int * state list * state list
    let compare (h1, xs1, xs2 : t) (h2, ys1, ys2 : t) =
      let c = compare h1 h2 in
      if c <> 0 then c else
      let c = compare_state_sets xs1 ys1 in
      if c <> 0 then c else
      let c = compare_state_sets xs2 ys2 in
      c
  end

module IntLsLsSet = Set.Make(IntLsLs)
module IntLsLsMap = Map.Make(IntLsLs)


(* From Haruo's Isect, Infer *)

module StSt =    (* pair of states as a key (with ordering) *)
  struct
    type t = state * state
    let compare (v1, v1' : t) (v2, v2' : t) =
      let c = compare_state v1 v2 in
      if c <> 0 then c else
      let c = compare_state v1' v2' in
      c
  end

module StStMap = Map.Make(StSt) (* map indexed by a pair of states *)
module StStSet = Set.Make(StSt) (* set .... *)


module StLs =     (* state and a list of states as a key *)
  struct
    type t = state * state list
    let compare (v1, v1' : t) (v2, v2' : t) =
      let c = compare_state v1 v2 in
      if c <> 0 then c else
      let c = compare_state_sets v1' v2' in
      c
  end

module StLsMap = Map.Make(StLs)  (* map indexed by (state, state list) *) 

let number_of_empty_states () : int =
  let res = ref 0 in
  for i = 0 to !last_num do
    if is_empty !_states_.(i) then incr res
  done;
  !res


(*** printer ***)

let _cache_ : (int, unit) HT.t = HT.create 97
let _states_to_print_ : state ST.t = ST.create ()

let rec print_rec ch =
  let x = ST.pop _states_to_print_ in
  try
    HT.find _cache_ x.num
  with Not_found ->
    HT.add _cache_ x.num ();
    if x.final && x.next = [] then
      F.fprintf ch "{%d}(%a)@." x.num print_pre x.finpre
    else if x.final then (
      HT.add _cache_ x.num ();
      F.fprintf ch "{%d}[(%a) | " x.num print_pre x.finpre;
      printl ch x.next;
      F.fprintf ch "]@."
    ) else (
      HT.add _cache_ x.num ();
      F.fprintf ch "{%d}[" x.num;
      printl ch x.next;
      F.fprintf ch "]@."
    )

and printl ch xl =
  if xl = [] then () else
  let rec loop ch = function
    | [] -> F.fprintf ch ""
    | [(c,_,x1,x2,p)] ->                (*##ATT not prnted *)
        SD.pr_labty ch c;
        F.fprintf ch "(%d, %d" x1.num x2.num;
        ST.push x1 _states_to_print_;
        ST.push x2 _states_to_print_;
        if p != [] then F.fprintf ch ", ";
        print_pre ch p;
        F.fprintf ch ")"
    | (c,_,x1,x2,p)::xs ->
        SD.pr_labty ch c;
        F.fprintf ch "(%d, %d" x1.num x2.num;
        ST.push x1 _states_to_print_;
        ST.push x2 _states_to_print_;
        if p != [] then F.fprintf ch ", ";
        print_pre ch p;
        F.fprintf ch ") | ";
        loop ch xs
  in loop ch xl

let print ch states =
  ST.clear _states_to_print_;
  iter (fun s -> ST.push s _states_to_print_) states;
  while not (ST.is_empty _states_to_print_) do print_rec ch done
