module SO = SynOp
module SS = SynSource
module SD = SynDenot
module ST = SynTyped
module C = Contexts 
module U = Utils
module R = Result
module Pr = Pretty

exception TyCheck of Finfo.t * string

let fpr = Format.fprintf
let outf = !Pref.output_formatter

type rettype =
  | NoRet
  | MaybeRet of SD.ty option
  | SureRet of SD.ty option

let warning inf s =
  Format.eprintf "@[Warning: @[%s@] : @[%s@]@." (Finfo.finfo_to_string inf) s

(* Convenience type constructors *)
let ty_hostty hty' = SD.ty (SD.Thost (SD.hostty hty'))

let ty_attributive = SD.ty (SD.Tseq (SD.seqref Kerids.attseq_id))


(******  Wrappers for class context  ******)

(** Finds a field description using field's string name *)
let find_field cctx cid fname = 
  List.find (fun (fid, _) -> fname = (Ident.name fid))
    (cctx.C.cc_fields cid)


(** Checks whether all typs are subtypes of bounds. In case of error,
    uses fileinfo from srcexps to report it.  All three list arguments must 
   have the same length.  *)
let check_argparam inf srcexps typs bounds = 
   let rec looper = function 
     | [], [], [] -> () (* success *)
     | e::es, t::ts, b::bs -> 
	 (match Semant.sub t b with
	 | R.Succ -> looper(es, ts, bs)
	 | R.Fail f -> raise
	       (TyCheck(Finfo.get e, "Argument not subtype of parameter. \n" ^
			"Argument type:  " ^ (Pr.toString_ty t) ^ "\n" ^
			"Parameter type: " ^ (Pr.toString_ty b) ^ "\n" ^
			"At path: " ^ (Typath.toString f) ^"\n"))
	 )

     | _, [], _::_ 
     | _, _::_, [] -> raise
	   (TyCheck(inf,
		    "Mismatch in the numbers of arguments and parameters. \n"))
     | _, _, _ -> 
	 assert false (* lists srcexps and typs must have equal length *)
   in looper (srcexps, typs, bounds)     

(** helper function to find the best method according to a name and type
    returns a class and an id
    may raise Not_found or AmbiguousMethod
*)

exception AmbiguousCall of ((SD.ty list) list) * string

let rec find_best_method cctx cid name tyargs =
  let rec check_subs = function
    | t::ts, b::bs ->
        if not (Semant.bsub t b) then false
        else check_subs (ts,bs)
    | [], [] -> true
    | _ -> assert false (* lists were filtered to have the same size *)
  in
  let rec find_best_match minl = function
    | [] ->
        begin
          match minl with
          | [m] -> m
          | l -> raise (AmbiguousCall ((List.map (fun (_,s,m,x,y) -> y) l), Ident.name
          cid))
        end
    | (_,is_s,mo,rty,argty) as elt :: r ->
        (* first case: we already have a minimum smaller than this *)
        if List.exists (fun (_,s,m,r,t) -> check_subs (t, argty)) minl
        then find_best_match minl r
        else 
          (* we only consider elements that are not supertypes of the new one *)
          let newminl = elt :: 
                        (List.filter (fun (_,s,m,r,t) -> not (check_subs (argty, t))) minl)
          in
          find_best_match newminl r 
  in
  let candidates = cctx.C.cc_meththistype cid name in
  let nbargs = List.length tyargs in
  let candidates' = List.filter (fun (_,s,mo,r,c) -> (List.length c) = nbargs) candidates in
  let candidates'' = List.filter (fun (_,s,mo,r,c) -> check_subs (tyargs,c)) candidates' in
  match candidates'' with
  | [] -> find_best_method cctx (Semant.parent cid) name tyargs
  | [m] -> m
  | h :: r -> find_best_match [h] r

let fm_timer = Timer.create "fm_timer"

let check_if_seq ty inf msg = 
  if not (Semant.bsub ty SD.ty_any) then raise (TyCheck(inf, msg))

(** Main expression-typing routine  *)
let rec typ_exp nctx cctx env expr : ST.exp = 
  let check_array_access ex =
    let ({ST.exp_desc=e;ST.exp_type=te} as pr) = typ_exp nctx cctx env ex in
    match te with
    | None -> 
        raise (TyCheck(Finfo.get ex, "Array size cannot have type VOID, must have type int.\n"))
    | Some t ->
        if SD.ty_Int <> (SD.whole_ty t) then
          raise (TyCheck(Finfo.get ex, "Array size must have type int.\n"))
        else
          pr
  in    
  let (inf_expr, def_expr) = expr in
  match def_expr with 
  | SS.Ebool b ->   {ST.exp_desc=ST.Ebool b; ST.exp_type=Some (ty_hostty(SD.Tbool b))}
  | SS.Eint n ->    {ST.exp_desc=ST.Eint n; ST.exp_type=Some (ty_hostty(SD.Tint n))}
  | SS.Echar c ->   {ST.exp_desc=ST.Echar c; ST.exp_type=Some (ty_hostty(SD.Tchar c))}
  | SS.Estring s ->
      {ST.exp_desc=ST.Estring s; ST.exp_type=Some (ty_hostty(SD.Tstring s))}

  | SS.ENewArray (t, ex, d) ->
      let tbase = Src2denot.ty nctx t in
      let rec mktype = function
        | 0 -> tbase
        | n -> SD.ty (SD.Thost (SD.hostty (SD.TArray (mktype (n-1)))))
      in
      {ST.exp_desc=ST.ENewArray(tbase, check_array_access ex, d);
       ST.exp_type=Some ( mktype (1 + d))}
  
  | SS.Etag t -> 
	{ST.exp_desc=ST.Etag t; ST.exp_type=Some (ty_hostty(SD.Ttag t))}

  | SS.Enull -> {ST.exp_desc=ST.Enull; ST.exp_type=Some SD.ty_Null}

  | SS.EThis -> {ST.exp_desc=ST.EThis;
                 ST.exp_type=Some (ty_hostty (SD.TClasRef (TyEnv.get_this env)))}
  | SS.ENamedLoc rname  -> 
      typ_location nctx cctx env rname
  | SS.ENew(rname, args) -> 
      let cid = Naming.resolve_class nctx rname in 
      let (t_args, typ_args) = typ_explist nctx cctx env args in
      let typ_params = try cctx.C.cc_constrtype cid
	  with Not_found -> raise
	      (Bug.Bug("Tycheck.typ_exp!ENew", 
		       "Class not found: " ^ (Ident.toString cid)))
      in check_argparam inf_expr args typ_args typ_params;
      let t_exp = ST.ENew(cid, t_args) in 
      let ty_exp = SD.ty (SD.Thost(SD.hostty (SD.TClasRef cid))) in 
      {ST.exp_desc=t_exp; ST.exp_type=Some ty_exp}

  | SS.EFieldAccess((base:SS.exp), oname) -> 
      let ((t_base:ST.exp), cid_base) =
        match typ_is_class nctx cctx env base with
        | Some cid ->
            let e = {ST.exp_desc=ST.EClass cid;ST.exp_type=None} in
            (e, cid)
        | None -> typ_invkbase nctx cctx env base 
      in typ_fieldaccess nctx cctx env (t_base, cid_base, oname)

  | SS.EFieldAccessSuper(oname) -> 
      let base_desc = ST.EThis in
      let base_type = ty_hostty (SD.TClasRef (TyEnv.get_this env)) in
      let base = {ST.exp_desc=base_desc;ST.exp_type=Some base_type} in
      let cid_base = TyEnv.get_super env in 
      typ_fieldaccess nctx cctx env (base, cid_base, oname)

  | SS.EArrayAccess(earray, eindex) ->
      let t_ea = typ_exp nctx cctx env earray in
      let t =
        match(
          match ST.get_type t_ea with
          | Some te ->
              (match SD.ty' te with
              | SD.Thost ht ->
                  (match SD.hostty' ht with
                  | SD.TArray ta -> Some ta
                  | _ -> None
                  )
              | _ -> None
              )
          | None -> None)
        with
        | Some ta -> ta
        | None ->
            raise (TyCheck(Finfo.get earray, "Must have an array type.\n"))
      in
      let t_ei = check_array_access eindex in
      {ST.exp_desc=ST.EArrayAccess(t_ea, t_ei); ST.exp_type=Some t}
      
  | SS.EMethInvkBase(base, oname, args) ->
      let (t_base, cid_base) =
        match typ_is_class nctx cctx env base with
        | Some cid ->
            let e_desc = ST.EClass cid in
            let e_type = ty_hostty (SD.TClasRef (TyEnv.get_this env)) in
            let e = {ST.exp_desc=e_desc;ST.exp_type=Some e_type} in
            (e, cid)
        | None -> typ_invkbase nctx cctx env base 
      in
      typ_methinvk nctx cctx env inf_expr
	(Some t_base, cid_base, oname, args)

  | SS.EMethInvkLocal(oname, args) -> 
      typ_methinvk nctx cctx env inf_expr
        (None, TyEnv.get_this env, oname, args)

  | SS.EMethInvkSuper(oname, args) ->
      let e_desc = ST.EThis in
      let e_type = ty_hostty (SD.TClasRef (TyEnv.get_super env)) in
      let e = {ST.exp_desc=e_desc;ST.exp_type=Some e_type} in
      typ_methinvk nctx cctx env inf_expr
	(Some e, TyEnv.get_super env, oname, args)

  | SS.EAssign(lhs, rhs) -> 
      let t_lhs = typ_exp nctx cctx env lhs in
      let t_rhs = typ_exp nctx cctx env rhs in
      let typ_lhs = ST.get_type t_lhs in
      let typ_rhs = ST.get_type t_rhs in
      (match ST.get_desc t_lhs with 
      |	ST.EVar _ |  ST.EFieldAccess _ | ST.EArrayAccess _  -> () (* ok, a location *)
      |	_ -> raise 
	    (TyCheck(Finfo.get lhs, 
		     "Assignment's left-hand side must be a location.\n")) );
      let leftty =
      match typ_lhs with 
      |	Some t  -> t (* ok, a type *)
      |	_ -> assert false
      in
      let retty =
      match typ_rhs with 
      |	Some t  -> t (* ok, a type *)
      |	_ -> raise 
	    (TyCheck(Finfo.get lhs, 
		     "Assignment's right-hand side type must not be VOID.\n")) 
      in
      (match Semant.sub retty leftty with
      | R.Fail f -> raise 
	  (TyCheck(inf_expr, "Assignment's rhs is not subtype of lhs \n" ^
		   "RHS type:  " ^ (Pr.toString_ty_option typ_rhs) ^ "\n" ^
                   "LHS type: " ^ (Pr.toString_ty_option typ_lhs) ^ "\n" ^ 
                   "Path for RHS <: LHS: " ^ (Typath.toString f) ^ "\n") )
      | _ -> ());
      {ST.exp_desc=ST.EAssign(t_lhs, t_rhs); ST.exp_type=typ_lhs}

  | SS.ECast(s_ty, exp, issafe) -> 
      let ty = Src2denot.ty nctx s_ty in
      let ({ST.exp_desc=t_exp; ST.exp_type=ty_exp} as e) = typ_exp nctx cctx env exp in
      let retty = match ty_exp with
      | Some t -> t
      | None -> raise (TyCheck(Finfo.get exp, "Cannot cast something whose type is VOID"))
      in
      (match Semant.sub retty ty with 
      | R.Succ -> {ST.exp_desc=t_exp; ST.exp_type=Some ty}          (* upcast *)
      | R.Fail uf -> 
	  (match Semant.sub ty retty with 
          | R.Succ -> {ST.exp_desc=ST.ECast(ty, e, issafe);
                       ST.exp_type=Some ty}                         (* downcast *)
	  | R.Fail df -> raise                          (* sidecast *)
		(TyCheck(inf_expr,  
			 "Statically incorrect cast. \n" ^
			 "Expr type: "^(Pr.toString_ty_option ty_exp)^"\n" ^
			 "Target type: " ^ (Pr.toString_ty ty) ^ "\n" ^
			 "Path for exp <: target: "^(Typath.toString uf)^"\n" ^
 			 "Path for target <: exp: "^(Typath.toString df)^"\n" 
			))
	  ))

  | SS.ETypeStamp(s_ty, exp) -> 
      let sty = Src2denot.seqty nctx s_ty in
      let e = typ_exp nctx cctx env exp in
      let t = match ST.get_type e with
        | Some t -> t
        | None ->
            raise (TyCheck(Finfo.get exp, "Cannot stamp something whose type is VOID"))
      in check_if_seq t (Finfo.get exp) "This is not a sequence expression";
      let t2 = SD.ty (SD.Tseq sty) in
      if not (Semant.bsub t t2) then
	raise (TyCheck(inf_expr,
                       "The actual type " ^
                         (Pr.toString_ty t) ^
                         " is not a subtype of the type stamp " ^ (Pr.toString_ty t2)))
      else
        {ST.exp_desc=ST.ETypeStamp(sty,e); ST.exp_type=Some SD.ty_Object}
  | SS.EUnary(uop, exp)  -> 
      let ({ST.exp_desc=_; ST.exp_type=ty_exp} as t_exp) = typ_exp nctx cctx env exp in
      let retty = match ty_exp with
      | Some t -> t
      | None -> raise 
	    (TyCheck(Finfo.get exp, "Argument cannot have type VOID.\n"))
      in
      (match uop with 
      |	SO.UnNot -> 
	  if not (Semant.bsub retty SD.ty_Bool) then raise
	      (TyCheck(inf_expr, "Logical NOT argument must be bool.\n" ^
		       "Actual type: " ^ (Pr.toString_ty_option ty_exp) ^"\n"))
	  else {ST.exp_desc=ST.EUnary(uop, t_exp); ST.exp_type=Some SD.ty_Bool}
      |	SO.UnPlus | SO.UnMinus ->
	  if not (Semant.bsub retty SD.ty_Int) then raise
	      (TyCheck(inf_expr, "Unary PLUS/MINUS argument must be int.\n" ^
		       "Actual type: " ^ (Pr.toString_ty_option ty_exp)^"\n"))
	  else {ST.exp_desc=ST.EUnary(uop, t_exp); ST.exp_type=Some SD.ty_Int}
      )

  | SS.EBinary(exp1, bop, exp2) ->
      let t_exp1 = typ_exp nctx cctx env exp1 in
      let t_exp2 = typ_exp nctx cctx env exp2 in
      let {ST.exp_desc=_; ST.exp_type=ty_exp1} = t_exp1 in
      let {ST.exp_desc=_; ST.exp_type=ty_exp2} = t_exp2 in
      let retty1 = match ty_exp1 with
      | Some t -> t
      | None -> raise 
	    (TyCheck(Finfo.get exp1, "Argument cannot have type VOID.\n"))
      in
      let retty2 = match ty_exp2 with
      | Some t -> t
      | None -> raise 
	    (TyCheck(Finfo.get exp2, "Argument cannot have type VOID.\n"))
      in
      let wty1 = SD.whole_ty retty1 in
      let wty2 = SD.whole_ty retty2 in
      (* auxiliary handlers for different situations *)
      let handle_binop_arith () = 
	if not(SD.ty_Int == wty1  &&  SD.ty_Int == wty2)
	then raise (TyCheck(inf_expr, "An argument of an arithmetic binary " ^
			    "operation is not an integer.\n"))
	else 
	  let ty_res = match bop with 
	  | SO.BiPlus | SO.BiMinus | SO.BiMult | SO.BiDiv | SO.BiMod -> 
	      SD.ty_Int
	  | SO.BiLt | SO.BiGt | SO.BiLeq | SO.BiGeq | SO.BiEq | SO.BiNeq -> 
	      SD.ty_Bool
	  | _ -> assert false
	  in {ST.exp_desc=ST.EBinaryArith(t_exp1, bop, t_exp2); ST.exp_type=Some ty_res}
      in
      let handle_binop_bool () = 
	if not(SD.ty_Bool == wty1  &&  SD.ty_Bool == wty2)
	then raise (TyCheck(inf_expr, "An argument of a boolean binary " ^
			    "operation is not a boolean.\n"))
	else 
	  let ty_res = match bop with 
	  | SO.BiAnd | SO.BiOr -> SD.ty_Bool
	  | _ -> assert false
	  in {ST.exp_desc=ST.EBinaryBool(t_exp1, bop, t_exp2); ST.exp_type=Some ty_res}
      in
      let handle_binop_equality eqflag = 
	if (SD.ty_Bool == wty1  &&  SD.ty_Bool == wty2) ||
           (SD.ty_Char == wty1  &&  SD.ty_Char == wty2) ||
           (SD.ty_Int == wty1  &&  SD.ty_Int == wty2) ||
	    (Semant.bsub retty1 retty2 || Semant.bsub retty2 retty1)
	then {ST.exp_desc=ST.EBinaryEq(t_exp1, eqflag, t_exp2);
              ST.exp_type=Some SD.ty_Bool}
	else raise 
	  (TyCheck(inf_expr, "Arguments to a comparison operator " ^
		   "have incomparable types.\n"))
      in
      (* now analyse the expression and use the appropriate handler *)
      (match bop with
      |	SO.BiPlus -> 
	  if (SD.ty_String == wty1) || (SD.ty_String == wty2)
	  then {ST.exp_desc=ST.EConcatString(t_exp1, t_exp2);
                ST.exp_type=Some SD.ty_String}
	  else handle_binop_arith ()
      |	SO.BiMinus | SO.BiMult | SO.BiDiv | SO.BiMod 
      | SO.BiLt | SO.BiGt | SO.BiLeq | SO.BiGeq ->
	  handle_binop_arith ()
      |	SO.BiAnd | SO.BiOr ->
	  handle_binop_bool ()
      |	SO.BiEq  ->
	  handle_binop_equality true
      | SO.BiNeq ->
	  handle_binop_equality false
      )

  | SS.ESeq exps ->
      let (t_exps, sty_exp) = styp_explist nctx cctx env exps in
      {ST.exp_desc=ST.ESeq t_exps; ST.exp_type=Some (SD.ty (SD.Tseq sty_exp))}

  | SS.ETree(lab, atts, exps) ->
      let t_lab = typ_exp nctx cctx env lab in 
      let {ST.exp_desc=_; ST.exp_type=ty_lab} = t_lab in
      let (t_exps, sty_exp) = styp_explist nctx cctx env exps in
      let (t_atts, ty_atts) = typ_atts nctx cctx env atts in
      let t_expr = ST.ETree(t_lab, t_atts, t_exps) in 
      let rettylab = match ty_lab with
      | Some t -> t
      | None -> 
	  raise (TyCheck(Finfo.get lab, "A label cannot have type VOID.\n"))
      in
      let ty_lab1 = match SD.ty' rettylab with
      |	SD.Thost hty -> hty
      |	SD.Tseq sty -> SD.hostty SD.TSeq
      |	SD.PTyBind _ -> assert false (* pattern cannot be a typing result *)
      in let lty_lab = SD.labty(SD.TLab ty_lab1) in
      let ty_expr = SD.ty (SD.Tseq (SD.tree lty_lab ty_atts sty_exp)) in
      {ST.exp_desc=t_expr; ST.exp_type=Some ty_expr}
      (**NB: This clause effectively permits to have a tree expression on 
	 top level, unprotected by sequence quotes (they get added 
	 automatically in this clause).  However, the concrete syntax
	 does not allow this (yet?).  *)

  | SS.EPcdata str ->
      let n = String.length str in
      let inf = Finfo.bogus() in
      let exps = ref [] in
      for i=0 to n-1 do
        let c = String.get str i in
        exps := (inf, SS.ETree((inf, SS.Echar c), [], [])) :: !exps
      done;
      let (_, t) = styp_explist nctx cctx env (List.rev !exps) in
      {ST.exp_desc=ST.EPcdata str; ST.exp_type=Some (SD.ty (SD.Tseq t))}

  | SS.EString2Pcdata exp -> 
      let t_exp = typ_exp nctx cctx env exp in
      let {ST.exp_desc=_; ST.exp_type=ty_exp} = t_exp in
      let retty = match ty_exp with
      | Some t -> t
      | None -> raise (TyCheck(Finfo.get exp, "cannot have type VOID"))
      in
      if SD.ty_String == (SD.whole_ty retty)
      then {ST.exp_desc=ST.EString2Pcdata t_exp; ST.exp_type=Some SD.ty_pcdata}
      else raise(TyCheck(inf_expr, "A string expression was expected " ^
			 "for coersion to PCDATA, but found type \n" ^
			 (Pr.toString_ty_option ty_exp) ^ "\n" ))


(** Resolves a long name as either a local variable, or a field, or a
   static field in a class -- either possibly followed by a sequence
   of field projections.  *)
and typ_location nctx cctx env longname : ST.exp = 
 let (base_exp, base_typ, projections) = 
    match longname with
    | [] -> assert false  (* never produced by parsing *)
    | name::rest -> 
      (* First, see if longname starts with a field or var *)
	let (name_inf, name_str) = name in
	(try  
	  (match TyEnv.find env name_str with
	  | TyEnv.Var, vid, typ -> (ST.EVar vid, typ, rest)
	  | TyEnv.Fld, fid, typ ->
              let this_ty = ty_hostty (SD.TClasRef (TyEnv.get_this env)) in
              let this_t = {ST.exp_desc=ST.EThis; ST.exp_type=Some this_ty} in
              (ST.EFieldAccess(this_t, fid), typ, rest)
	  ) with Not_found ->
	(* Otherwise, see if longname is a static field in a class *)
	    match Naming.chop_leaf nctx longname with 
	      | Some ((cid, Cnaming.LClas), []) -> 
		  raise (TyCheck(Naming.info_of_longname longname, 
				 "Class name cannot be used as a location.\n"))
	      | Some ((cid, Cnaming.LClas), statfld::_) -> 
		  raise (Bug.NotYetFI(Naming.info_of_longname longname, 
				      "Tycheck.typ_location",
				      "Access to class's static fields"))
	      | Some ((_, Cnaming.LIntf), _) -> 
		  raise (TyCheck(Naming.info_of_longname longname, 
				 "Location cannot be inside an interface"))
	      | None ->
                (try 
                  let cid_base = TyEnv.get_this env in
                  let (id_fld, (status, typ_fld)) = 
                    find_field cctx cid_base name_str
                  in
		    (match status with
		    | SO.Static ->
                        let t_clas = ST.EClass cid_base in
                        let t_inst = {ST.exp_desc=t_clas; ST.exp_type=None} in
                        (ST.EFieldAccess(t_inst,id_fld),typ_fld, rest)
                    | SO.Instance ->  raise Not_found  )
                 with
                | Not_found -> raise (TyCheck(Naming.info_of_longname longname,
     	         "This longname does not start with a location or a static field: " ^ 
	         (Printing.toString SS.pr_longname longname) ^"\n" )))
	)
 in
 let e = {ST.exp_desc=base_exp;ST.exp_type=Some base_typ} in
 typ_fieldaccess_chain nctx cctx env (e, projections)

(** this code is very similar to typ_location, except that it returns
    Some(cid) if the name is the name of a class, and None otherwise. *)
and typ_is_class nctx cctx env (longname:SS.exp) = 
  match Finfo.drop longname with
  | SS.ENamedLoc [] -> assert false  (* never produced by parsing *)
  | SS.ENamedLoc ((name::_) as ln) -> 
    (* First, see if longname starts with a field or var *)
    let (name_inf, name_str) = name in
      (try  
        (match TyEnv.find env name_str with
        | _ -> None
        ) 
      with Not_found ->
      (* Otherwise, see if longname is a static field in a class *)
          match Naming.chop_leaf nctx ln with 
          | Some ((cid, Cnaming.LClas), []) -> Some cid
          | _ -> None
      )
  | _ -> None

(** Checks whether a source expression is a suitable base for
   method or field invokation (not null or void), and returns its typed version 
   and class id.  *)
and typ_invkbase nctx cctx env exp : ST.exp * Ident.id = 
   let t_base = typ_exp nctx cctx env exp in
   let {ST.exp_desc=_; ST.exp_type=typ_base} = t_base in
   let retbase = match typ_base with
   | Some t -> t
   | None -> raise (TyCheck(Finfo.get exp, 
            "Receiver of method invocation is VOID.\n"))
   in
   match SD.ty' retbase with 
   | SD.Thost hty -> 
     (match SD.hostty' hty with 
       | SD.TNull -> raise (TyCheck(Finfo.get exp, 
            "Receiver of method invocation is verifiably NULL.\n"))
       | _ -> (t_base, SD.clasOf_ty retbase)
      )
   | _ -> (t_base, SD.clasOf_ty retbase)

(** Applies typ_exp_noreturns to a list of expressions and returns 
   a list of corresponding typed expresions and a list of their types. *)
and typ_explist nctx cctx env exps : ST.exp list * SD.ty list = 
   let rec looper rexps rtyps = function 
   | [] -> (List.rev rexps, List.rev rtyps)
   | e::es -> let ({ST.exp_desc=_; ST.exp_type=ty} as ex) = typ_exp nctx cctx env e
      in
      let rty = match ty with
      | Some t -> t
      | None -> raise 
	    (TyCheck(Finfo.get e, "Expression cannot have type VOID.\n"))
      in
      looper (ex::rexps) (rty::rtyps) es
   in looper [] [] exps


and typ_fieldaccess nctx cctx env ((t_base:ST.exp), cid_base, oname) = 
  let (inf_fname, fname) = oname in
  let (id_fld, (status, typ_fld)) = 
    try 
      find_field cctx cid_base fname
    with 
    | Not_found -> raise 
        (TyCheck(inf_fname, "Field " ^ fname ^ 
	       " not found in class " ^ (Ident.toString cid_base) ^"\n"))
  in 
  {ST.exp_desc=ST.EFieldAccess(t_base, id_fld); ST.exp_type=Some typ_fld}

and typ_fieldaccess_chain nctx cctx env (e, longname) : ST.exp = 
  match longname with
  | [] -> e
  | fld::flds ->
      let {ST.exp_desc=_;ST.exp_type=ty_base_opt} = e in
      match ty_base_opt with
        | Some ty_base ->
            let cid_base = SD.clasOf_ty ty_base in
            let e = typ_fieldaccess nctx cctx env (e, cid_base, fld) in
            typ_fieldaccess_chain nctx cctx env (e, flds)
        | None ->
            assert false

and typ_methinvk nctx cctx env inf 
    ((t_base_opt : ST.exp option), cid_base, oname, s_args) = 
  let (inf_mname, mname) = oname in
  let (t_args, typ_args) = typ_explist nctx cctx env s_args in
  let (mid, status, mods, ret_ty, arg_ty) = 
    try
      Timer.wrap fm_timer (fun () -> find_best_method cctx cid_base mname typ_args)
    with
    | Not_found -> 
        raise (TyCheck(inf_mname, 
		       "Method "^ mname ^" with arguments of types \n"^
                       Pr.toString_list_ty typ_args ^ "\n" ^ 
		       " not found for class " ^ (Ident.name cid_base)))
    | AmbiguousCall (tl, classname) ->
        raise (TyCheck(inf_mname, 
		       "Method "^ mname ^" with arguments of types\n"^
                       (Pr.toString_list_ty typ_args) ^ "\n" ^
		       "can correspond to the " ^
                       "following signatures from class "^classname^":" ^ 
		       (List.fold_left (fun acc elt ->
			 acc ^ "\n" ^(Pr.toString_list_ty elt)) "" tl)^"\n" ^ 
		       "problem: there is no best signature"
                     )) 
  in
  (* What is the declared method type?:
  let typ_ret = 
    try cctx.C.cc_methrettype cid_base (MethIdent.getid mname typ_args) with
      Not_found -> raise 
	  (TyCheck(inf_mname, "Method " ^ mname ^ 
		   " not found for class " ^ (Ident.name cid_base)))*)
 (* Do arguments conform to parameters?: *)
  (* to change
  check_argparam inf s_args typ_args typ_params; *)
 (* Costruct the typed expression: *)
  let base_clas, t_base =
    match status with 
    | SO.Static -> 
	(match t_base_opt with
	   | None ->
               Some cid_base,
               {ST.exp_desc=ST.EClass cid_base; ST.exp_type=None}
	   | Some {ST.exp_desc=ST.EClass c;ST.exp_type=_} ->
               Some cid_base,
               {ST.exp_desc=ST.EClass c; ST.exp_type=None}
	   | _ -> raise 
              (TyCheck(inf_mname, "non static invocation of a static method")))
    | SO.Instance -> 
	(match mods with
	   | SynOp.New | SynOp.Normal -> Some cid_base
	   | SynOp.Virtual | SynOp.Override -> None) ,
	(match t_base_opt with
	   | None ->
               let e_type = ty_hostty (SD.TClasRef (TyEnv.get_this env)) in
               {ST.exp_desc=ST.EThis; ST.exp_type=Some e_type}
	   | Some {ST.exp_desc=ST.EClass c;ST.exp_type=_} ->
               raise (TyCheck(inf_mname, "static invocation of a non static method"))
	   | Some e -> e)
  in
  let mangled_name = Ident.find_same cctx.C.cc_mangled_name mid in
  let t_exp = match ret_ty with
  | None -> ST.EProcMethInvk(t_base, base_clas, mid, mangled_name, t_args)
  | Some t -> 
      (*
      print_string ("return type for "^mname^" with input type ");
      print_string (Pr.toString_list_ty arg_ty);
      print_string (": ");
      print_endline (Pr.toString_ty t);
      flush stdout;
      *)
      ST.EMethInvk(t_base, base_clas, mid, mangled_name, t_args)
  in {ST.exp_desc=t_exp; ST.exp_type=ret_ty}
 

and typ_atts nctx cctx env (atts: SS.exp list) : ST.exp list * SD.attsty = 
  let typ_attexp aexp : ST.exp list = 
    let (t_aexps, sty_aexp) = styp_explist nctx cctx env [aexp] in
    if not (Semant.bsub (SD.ty (SD.Tseq sty_aexp)) ty_attributive) 
    then raise
	(TyCheck(Finfo.get aexp, 
		 "This attribute expression is not attributive. " ^ 
		 "Computed type: \n" ^
		 (Pr.toString_seqty sty_aexp) ^ "\n" ))
    else t_aexps
  in
  let t_atts = List.concat (List.map typ_attexp atts) in
  let ty_atts = SD.attsty_empty in  (*##ATT attsty will be ignored *)
  (t_atts, ty_atts)


(********** styp_xxx  *****************)

and styp_explist nctx cctx env exps : ST.exp list * SD.seqty = 
  let rec looper rexps rstys = function
    | [] -> (List.rev rexps, SD.cat (List.rev rstys))
    | ex::rest ->
        let e = typ_exp nctx cctx env ex in
	let {ST.exp_desc=_; ST.exp_type=ty_ex} = e in
        let retty = match ty_ex with
        | Some t -> t
        | None -> raise 
	      (TyCheck(Finfo.get ex, "Expression cannot have type VOID.\n"))
        in
	let sty_ex = 
	  (match SD.ty' retty with
	  | SD.PTyBind _ -> assert false (* cannot be result of typing *)
	  | SD.Tseq sty -> sty
	  | SD.Thost hty -> 
              (match SD.hostty' hty with
	       | SD.TSeq -> SD.sty_any 
	       | SD.TClasRef cid when cid == Kerids.seq_id ->
		             raise (Bug.Bug("Tycheck.styp_explist", 
			                    "TClasRef(seq_id) that must be Seq "))
	       | _ -> raise 
		     (TyCheck(Finfo.get ex, 
			      "Expression must have a sequence type, \n" ^
			      "but has type " ^ (Pr.toString_ty retty) ^"\n"))
	      )
	  )
	in looper (e::rexps) (sty_ex::rstys) rest
  in looper [] [] exps


  
  
(** Some auxiliary checking / error reporting functions. *)

let ensure_seqty ty inf msg : SD.seqty = 
  match (SD.ty' ty) with
  | SD.PTyBind _ -> assert false (* should not be called on patterns *)
  | SD.Tseq t -> t
  | SD.Thost hty -> 
      if SD.isHostLEQSeq hty then SD.sty_any
      else raise (TyCheck(inf, msg))
	  

(******** Blocks and statements  ********)

(** Typechecks the list of block's statements either till its end, or
    till the first return statement is encountered.  In the latter
    case, checks whether the return is the last statement in the list.
    Each successive statement is checked in the environment returned
    by the previous statement.  The result is a typed block expression
    and the type of the return, if any.  *)

(*## This attempted to check for return after a return, but turned out 
     too restrictive, since relied on rty in deciding whether the previous 
     statement was a return.  Problematic case: when the prev statement is 
     conditional, but returns only in one of the branches, and the following
     statement is intended to return for sure.
let rec typ_block nctx cctx env block = 
  let rec looper ev revstmts = function 
    | [] -> 
	(List.rev revstmts,  None)
    | stm::rest -> 
	let (t_stm, retty_stm, ev1) = 
	  typ_stmt nctx cctx ev stm in
	match retty_stm with
	| None -> looper ev1 (t_stm::revstmts) rest
	| Some retty -> 
	    (match rest with
	    | [] -> (List.rev (t_stm::revstmts), Some retty)
	    | r::_ -> raise
		  (TyCheck(Finfo.get r, "Statement after a return.\n"))
	    )
  in looper env [] (Finfo.drop block)
*)

(**## Note: this typ_block allows one return statement to immediately
 follow another (). It still isn't capable of checking that all paths
 in branching code do return.  *)

let check_ret_or_error inf_stm ty1 ty2 sinit s1 s2 =
  match ty1, ty2 with
  | None, None -> None
  | Some t1, Some t2 ->
    (match Semant.maximum [t1; t2] with
	| Some ty -> Some ty
	| None -> raise 
	      (TyCheck(inf_stm, 
                   sinit ^
	    	   s1 ^ (Pr.toString_ty t1) ^ "\n" ^
	    	   s2 ^ (Pr.toString_ty t2)  ))
	)
  | _ -> raise
	      (TyCheck(inf_stm, 
                 sinit ^
	    	 s1 ^
                 (match ty1 with None -> "void" | Some t1 -> (Pr.toString_ty t1)) 
                 ^ "\n" ^
                 s2 ^
                 (match ty2 with None -> "void" | Some t2 -> (Pr.toString_ty t2)) 
                 ))



let rec typ_block nctx cctx env block =
  let check_sub stm tya tys =
    check_ret_or_error (Finfo.get stm) tya tys 
      ("The type of return in this statement, and "^
       "returns from the earlier statements are " ^ 
       "incompatible.\n")
      "Earlier: "
      "Now:     "
  in
  let rec looper ev revstmts retty_acc = function 
    | [] -> 
	(List.rev revstmts,  retty_acc)
    | stm::rest -> 
	let (t_stm, retty_stm, ev1) = 
	  typ_stmt nctx cctx ev stm in
	match retty_acc, retty_stm with
	| NoRet,      _ -> looper ev1 (t_stm::revstmts) retty_stm rest
	| (MaybeRet _ | SureRet _), NoRet -> looper ev1 (t_stm::revstmts) retty_acc rest
        | MaybeRet ty_acc, MaybeRet ty_stm -> looper ev1 (t_stm::revstmts)
                                                         (MaybeRet (check_sub stm ty_acc ty_stm))
                                                         rest
        | (MaybeRet ty_acc | SureRet ty_acc), 
          (MaybeRet ty_stm | SureRet ty_stm) -> looper ev1 (t_stm::revstmts)
                                                           (SureRet (check_sub stm ty_acc ty_stm))
                                                           rest
  in looper env [] NoRet (Finfo.drop block) 
   
   

and check_subtypes exps typs ty =
  List.iter2
    (fun e t ->
       match (e,t) with
         | None,None -> ()
         | (Some e,Some t) ->
	     (match Semant.sub t ty with
	     | R.Succ -> ()
	     | R.Fail f -> raise
		   (TyCheck(Finfo.get e,
                            "Type of binding expression is not a subtype " ^
                            "of the declared type. \n" ^
                            "Expr type: \n" ^ Pr.toString_ty t ^ "\n" ^
                            "Decl type: \n" ^ Pr.toString_ty ty ^ "\n" ^
			    "At path: " ^ (Typath.toString f) ^"\n"))  )
         | _ -> assert false)
    exps typs


and check_decls exps typs ty env vids fi flag =
  check_subtypes exps typs ty;
  if flag then
    match exps,typs,vids with
      | [Some(e)],[Some(t)],[x] ->
          (match Semant.tyMatch t ty env with
            | None -> raise
               (Bug.Bug(Finfo.finfo_to_string(Finfo.get e) ^
			" [in Tycheck.check_decls] ",
                        "Type of binding expression is incomparable " ^
                        "to the declared type; " ^
                        "BT = " ^ Pr.toString_ty t ^ "; " ^
                        "DT = " ^ Pr.toString_ty ty))
	    | Some(isec_ty, env1) -> 
		(assert (Semant.bsub isec_ty ty);
		TyEnv.addVar env1 (x,ty))
	  )
      | _ -> raise
            (TyCheck(fi,
                     "Exactly one binding expression must be present " ^
                     "when the declared type contains variables.\n"))
  else 
    let tys = List.map (fun _ -> ty) typs in
    List.fold_left2 (fun ev x t -> TyEnv.addVar ev (x,t)) env vids tys


and typ_decls nctx cctx env = 
  let rec loop exps typs = function
    | [] -> List.rev exps, List.rev typs
    | (x,e)::rest ->
        match e with
          | None -> loop (None::exps) (None::typs) rest
          | Some exp ->
              let e = typ_exp nctx cctx env exp in
              let {ST.exp_desc=_; ST.exp_type=t} = e in
              let rett = match t with
              | Some t -> t
              | None -> raise 
		    (TyCheck(Finfo.get exp, 
			     "Expression cannot have type VOID.\n"))
              in
              loop ((Some e)::exps) ((Some rett)::typs) rest
  in loop [] []

and typ_stmt nctx cctx env stmt = 
  let (inf_stmt, def_stmt) = stmt in
  match def_stmt with 
  | SS.SBlock block -> 
      let (t_block, retty_block) = typ_block nctx cctx env block
      in (ST.SBlock t_block, retty_block, env)

  | SS.SExp exp -> 
      let t_exp = typ_exp nctx cctx env exp in
      (ST.SExp t_exp, NoRet, env)

  (*## Check with C# on local var scoping rules... *)
  | SS.SLocVars (decls, s_ty) -> 
      let ty = Src2denot.ty nctx s_ty in 
      let vids = List.map 
	  (fun decl -> Ident.create (Finfo.drop (fst decl))) decls in 
      let srcexps = List.map snd decls in 
      let (exps,typs) = typ_decls nctx cctx env decls in
      let flag = SD.binders_ty ty in
      let env1 = check_decls srcexps typs ty env vids inf_stmt flag in
      let prs = List.map2 (fun x y -> x,y) vids exps in
      (ST.SLocVars(prs, ty, flag), NoRet, env1)

  | SS.SReturn ex -> 
      let (inf_ex, def_ex) = ex in
      begin
      match def_ex with
      | None -> (ST.SReturn None, SureRet None, env)
      | Some exp ->
          let e = typ_exp nctx cctx env (inf_ex, exp) in
          let {ST.exp_desc=_; ST.exp_type=typ_ex} = e in
          if typ_ex = None then 
          raise (TyCheck(Finfo.get ex, "Expression cannot have type VOID.\n"))
        else
        (ST.SReturn (Some e), SureRet typ_ex, env)
      end

  | SS.SIf(cond, bthen, belse) ->
      let t_cond = typ_exp nctx cctx env cond in
      let {ST.exp_desc=_; ST.exp_type=ty_cond} = t_cond in
      let retcond = match ty_cond with
      | Some t -> t
      | None -> raise (TyCheck(Finfo.get cond, "Expression cannot have type VOID.\n"))
      in
      if not (Semant.bsub retcond SD.ty_Bool) then raise
	  (TyCheck(Finfo.get cond, "IF condition must be bool, but it is " ^
		   (Pr.toString_ty retcond))); 
      let (t_bthen, rty_bthen) = typ_block nctx cctx env bthen in
      let (t_belse, rty_belse) = typ_block nctx cctx env belse in
      let rty_stmt = 
        match rty_bthen, rty_belse with
        | NoRet, NoRet -> NoRet
        | (NoRet, (MaybeRet t | SureRet t)) | ((MaybeRet t | SureRet t), NoRet)  -> MaybeRet t
        | SureRet t1, SureRet t2 -> 
            SureRet (check_ret_or_error inf_stmt t1 t2 
                               ("Types of returns from IF branches cannot be reconciled.\n")
	                        "Return from THEN: "
	                        "Return from ELSE: ")
        | (MaybeRet t1 | SureRet t1), (MaybeRet t2 | SureRet t2) -> 
            MaybeRet (check_ret_or_error inf_stmt t1 t2 
                               ("Types of returns from IF branches cannot be reconciled.\n")
	                        "Return from THEN: "
	                        "Return from ELSE: ")
      in
      (ST.SIf(t_cond, t_bthen, t_belse), rty_stmt, env)

  | SS.SWhile (cond, body) ->
      let t_cond = typ_exp nctx cctx env cond in
      let {ST.exp_desc=_; ST.exp_type=ty_cond} = t_cond in
      let retcond = match ty_cond with
      | Some t -> t
      | None -> raise (TyCheck(Finfo.get cond, "Expression cannot have type VOID"))
      in
      if not (Semant.bsub retcond SD.ty_Bool) then raise
	  (TyCheck(Finfo.get cond, "WHILE condition must be bool, but it is " ^
		   (Pr.toString_ty retcond))); 
      let (t_body, rty_body) = typ_block nctx cctx env body in
      (ST.SWhile (t_cond, t_body), rty_body, env)

  | SS.SForeachRE (s_pa, arg, body) -> 
      (* Prepare the input argument *)
      let t_arg = typ_exp nctx cctx env arg in
      let {ST.exp_desc=_; ST.exp_type=typeof_arg} = t_arg in
      let sty_arg = match typeof_arg with
      | Some t -> ensure_seqty t (Finfo.get arg) 
	    "FOREACH-RE input must be a sequence expression"
      | None -> raise (TyCheck(Finfo.get arg, 
			       "FOREACH-RE input cannot have type VOID"))
      in
      (* Prepare the pattern *)
      let sty_pat = Src2denot.seqty nctx s_pa 
      in
      (* Check matchability and extend the environment *)

(*        fpr outf "--- Foreach TA at %s:\n" (Finfo.finfo_to_string (Finfo.get arg)); *)

      let body_env = match Semant.styChop sty_arg sty_pat env with
      | None -> raise 
	    (TyCheck(Finfo.get s_pa, 
		     "A portion of the FOREACH-RE input " ^ 
		     "cannot be covered by its pattern."))
      | Some ev -> ev 
      in
      (* Typecheck the body *)
      let (t_body, rty_body) = typ_block nctx cctx body_env body in
      (ST.SForeachRE(sty_pat, t_arg, t_body), rty_body, env)

  | SS.SMatch(arg, clauses) ->
      let t_arg = typ_exp nctx cctx env arg in
      let {ST.exp_desc=_; ST.exp_type=ty_arg} = t_arg in
      let retarg = match ty_arg with
      | Some t -> t
      | None -> raise (TyCheck(Finfo.get arg, "Expression cannot have type VOID"))
      in
      check_if_seq retarg (Finfo.get arg) "This is not a sequence expression";
      (* Process the clauses *)
      let rec looper revclauses revrtybodies patsunion = function 
	| [] -> (List.rev revclauses, List.rev revrtybodies, patsunion)
	| cl::rest -> 
	    let (inf_cl, (s_pa, body)) = cl in 
            Debug.trace ("Typechecking clause at "^(Finfo.finfo_to_string
            inf_cl));
	    let pa = Src2denot.ty nctx s_pa in 
	    check_if_seq  pa (Finfo.get s_pa) "This is not a sequence pattern";
	    if (Semant.bsub pa patsunion) then (*raise 
		(TyCheck(Finfo.get s_pa, "This pattern is redundant"));*)
   	      Format.eprintf "@[%s: @[%s@]@]@." (Finfo.finfo_to_string (Finfo.get s_pa)) 
	       ("Warning from Tycheck: This pattern is redundant");
	    (* Infer types of pattern variables and typecheck the body *)
	    let (t_body, rty_body) = 
	      try 

(*                  fpr outf "-----@."; *)
(*                  fpr outf "retarg = %a@." Pretty.pr_ty retarg; *)
(*                  fpr outf "pa = %a@." Pretty.pr_ty pa; *)
(*                  fpr outf "--- Input TA at %s:\n" (Finfo.finfo_to_string inf_cl); *)

		(match Semant.tyMatch retarg pa env with
		| None -> raise
		      (TyCheck(Finfo.get s_pa, 
			       "This pattern does not capture input. \n" ^
			       "Input type: " ^ (Pr.toString_ty retarg)^"\n" ))
		| Some(_, env1) -> 
		    typ_block nctx cctx env1 body
		) with Infer.NoLabsMaximum(lvar, htys) -> raise
		    (TyCheck(Finfo.get s_pa, 
			     "The label variable " ^ (Ident.name lvar) ^
			     " gets bound to types none of which is an " ^ 
			     "upper bound of the others: \n" ^ 
			     (List.fold_right
				(fun h s -> 
				  "Type: " ^ (Pr.toString_hostty h)^ "\n" ^ s)
				htys "")))
	    in
	    (* Prepare recursive call *)
	    let t_cl = (pa, t_body) in 
	    let punion = match SD.ty' pa,  SD.ty' patsunion with
	    | SD.Tseq spa, SD.Tseq spats -> 
		SD.ty (SD.Tseq (SD.alt [spa; spats]))
	    | _ -> assert false   (* known to be sequence types *)
	    in
	    looper (t_cl::revclauses) (rty_body::revrtybodies) punion rest
      in 
      let (t_clauses, rty_bodies, ty_pats) = looper [] [] SD.ty_empty clauses 
      in
      (* Check exhaustiveness *)
      (match Semant.sub retarg ty_pats with
      |	R.Succ -> ()
      |	R.Fail tpath -> raise 
	    (TyCheck(inf_stmt, 
		     "Patterns of the MATCH are not exhaustive." ^
		     "Subtyping of the input type into the union of " ^ 
		     "patterns failed on \n" ^ (Typath.toString tpath)))
      );
      (* Check consistency of returns across the clause bodies *)
      let (sureret_bodtys, mayberet_bodtys, nonret_bodtys) = 
        let rec loop sb mb nb = function
          | [] -> sb, mb, nb
          | b :: rest ->
              (match b with
              | NoRet -> loop sb mb (b :: nb) rest
              | MaybeRet t -> loop sb (t :: mb) nb rest
              | SureRet t -> loop (t :: sb) mb nb rest)
        in
        loop [] [] [] rty_bodies in
      let rty_stmt = 
        let check tl =
          let rsome, rnone = List.partition (function Some _ -> true | None -> false) tl in 
          (match rsome, rnone with
          | [], _ -> None
          | _, [] ->
              (match Semant.maximum (List.map (function Some rty -> rty | None -> assert false) rsome)  with
              | None -> raise
                        (TyCheck(inf_stmt, "Among return types of the MATCH bodies, " ^
			 "there is no one that bounds all the others."))
              | Some rty -> Some rty)
          | _, _ -> raise
                        (TyCheck(inf_stmt, "Among return types of the MATCH bodies, " ^
			 "there are void and non-void return types."))
          )
        in
        match (sureret_bodtys, mayberet_bodtys, nonret_bodtys) with
        | [], [], _ -> NoRet
        | _::_, [], [] -> SureRet (check sureret_bodtys)
        | _, _, _ -> MaybeRet (check (sureret_bodtys @ mayberet_bodtys))
      in 
      (ST.SMatch(t_arg, t_clauses), rty_stmt, env)

  | SS.SEmpty -> 
      (ST.SEmpty, NoRet, env)

  | SS.SBreak -> 
      (ST.SBreak, NoRet, env)



