(*********************

This module is responsible for printing Xtatic source programs in
a "scheme-y" Xduce syntax. This function is turned on by the
outputxduce command line flag.

Related to this is the capability of outputting parenthesized XML
documents. See the associated function xdpr_seqva in xmlIO.mli

*********************)

open List
open SynSource
open Format

let fpr = Format.fprintf

(***** Names ***********)
(*--VG why is this here?  The same fns come from the opened SynSource!
let pr_name ff n = 
  Format.pp_print_string ff (Finfo.drop n)

let pr_longname ff ln = 
  let rec prList ff lst = 
    match lst with
    | [n] -> pr_name ff n
    | n::ns -> fpr ff "%a.%a" pr_name n prList ns
    | [] -> raise (Failure "BUG: zero-length longname")
  in prList ff ln
*)

let pr_prefix_longname ff (pre,ln) = 
  let rec prList ff lst = 
    match lst with
    | [n] -> pr_name ff n
    | n::ns -> fpr ff "%a.%a" pr_name n prList ns
    | [] -> raise (Failure "BUG: zero-length longname")
  in fpr ff "%s_%a" pre prList ln


(***** Types ************)

let rec pr_hostty ff hty = 
  match (Finfo.drop hty) with 
    | TCSharp refname ->
        (match refname with
          | [_,"string"] -> fpr ff "string"
          | _ -> fpr ff "class_%a" pr_longname refname)
    | TNull -> raise (Failure "TNull unsupported")
    | TBool -> fpr ff "bool"
    | Tbool b -> raise (Failure "Tbool unsupported")
    | TInt -> fpr ff "int"
    | Tint n -> raise (Failure "Tint unsupported")
    | TString -> fpr ff "string"
    | Tstring s -> raise (Failure "Tstring unsupported")
    | TChar -> raise (Failure "TChar unsupported")
    | Tchar c -> raise (Failure "Tchar unsupported")
    | TTag -> raise (Failure "TTag unsupported")
    | Ttag t -> raise (Failure "Ttag unsupported")
    | TSeq -> raise (Failure "TSeq unsupported")
    | TArray t -> raise (Failure "TArray unsupported")

let rec pr_attpat ff atts =
  let att ff a =
    let (aname,vname) = Finfo.drop a in
    fpr ff "(lab %s (tvar %s))" (Finfo.drop aname) (Finfo.drop aname)
  in 
  match (Finfo.drop atts) with
    | TAttAny -> fpr ff ""
    | PAtt(atts,rest) ->
        fpr ff "(lab @@ (cat %a))" (Printing.print_list att " " "") atts

let rec pr_seqty ff seqty = 
  match (Finfo.drop seqty) with 
    | TTree (lty, atts, seqty) ->
        (match (Finfo.drop lty) with
          | TLab hty -> 
              (match (Finfo.drop hty) with 
                | TCSharp refname ->
                    (match refname with
                      | [_,"char"] ->
                          fpr ff "char" (* ignore seqty *)
                      | [_,"tag"] ->
                          fpr ff "(lab ~ (cat %a %a))" pr_attpat atts
                            pr_seqty seqty
                      | _ -> raise (Failure "non-char class label"))
                | Ttag t -> fpr ff "(lab %s (cat %a %a))" t pr_attpat atts
                    pr_seqty seqty
                | Tchar ch -> fpr ff "#\\%c " ch
                | _ -> raise (Failure "wrong type of label"))
          | PLabBind(vname, lty) -> raise (Failure "PLabBind unsupported"))
    | TCat seqtyps -> fpr ff "(cat %a)" (Printing.print_list pr_seqty " " "") seqtyps 
    | TAlt seqtyps -> fpr ff "(alt %a)" (Printing.print_list pr_seqty " " "") seqtyps 
    | TRep seq -> fpr ff "(star %a)" pr_seqty seq
    | TSeqRef refname -> pr_qualname ff refname
    | PSeqBind (vname, spa) -> fpr ff "(tvar %a %a)" pr_decname vname pr_seqty spa

let rec pr_ty ff ty =
  match Finfo.drop ty with
    | Thost hty -> pr_hostty ff hty
    | Tseq seq -> pr_seqty ff seq
    | PTyBind (vname, ty) -> raise (Failure "PTyBind unsupported")

let pr_ty_option ff = function
  | (info, None) -> fpr ff "()"
  | (info, Some ty) -> pr_ty ff (info,ty)

(******** Expressions **********)

let pr_lhs ff expr =
  match (Finfo.drop expr) with 
    | ENamedLoc name -> pr_longname ff name
    | _ -> raise (Failure "only long id is accepted as an lhs")

let pr_lbl ff expr =
  match (Finfo.drop expr) with 
    | Etag t -> fpr ff "%s" t
    | _ -> raise (Failure "only Etag is accepted as a label")

let rec pr_atts_exp ff = function
  | [],[] -> fpr ff "()"
  | [],exps -> fpr ff "(cat %a)" (Printing.print_list pr_exp " " "") exps
  | atts,[] -> fpr ff "(lab @@ (cat %a))" (Printing.print_list pr_exp " " "") atts
  | atts,exps ->
      fpr ff "(cat (lab @@ (cat %a)) %a)" (Printing.print_list pr_exp " " "") atts
        (Printing.print_list pr_exp " " "") exps

and pr_exp ff expr = 
  match (Finfo.drop expr) with 
  | Ebool b -> fpr ff "%s" (if b then "#t" else "#f")
  | Eint n -> fpr ff "%d" n
  | Estring s -> fpr ff "\"%s\"" s
  | Echar c -> raise (Failure "Echar is unimplemented")
  | Etag t -> raise (Failure "Etag is unimplemented")
  | Enull -> raise (Failure "Enull is unimplemented")
  | ENamedLoc name -> pr_longname ff name
  | EThis ->  raise (Failure "EThis is unimplemented")
  | ENew(tname, args) -> fpr ff "(new_%a %a)" pr_longname tname
        (Printing.print_list pr_exp " " "") args
  | ENewArray _ ->  raise (Failure "ENewArray is unimplemented")
  | EFieldAccess(_, name) -> pr_name ff name
  | EFieldAccessSuper(name) -> raise (Failure "EFieldAccessSuper is unimplemented")
  | EArrayAccess _ ->  raise (Failure "EArrayAccess is unimplemented")
  | EMethInvkBase(exp, name, args) ->
      fpr ff "(%a %a %a)" pr_name name pr_exp exp
        (Printing.print_list pr_exp " " "") args
  | EMethInvkLocal(name, args) -> fpr ff "(%a %a)" pr_name name
      (Printing.print_list pr_exp " " "") args
  | EMethInvkSuper(name, args) -> raise (Failure "EMethInvkSuper is unimplemented")
  | ECast(ty, exp, _) -> fpr ff "(validate %a %a)" pr_exp exp pr_ty ty
  | ETypeStamp(t,e) -> fpr ff "(typestamp %a %a)" pr_seqty t pr_exp e
  | EAssign(lhs, rhs) -> fpr ff "(set! %a %a)" pr_lhs lhs pr_exp rhs
  | EUnary(uop, exp) -> fpr ff "(%a %a)" SynOp.pr_unary uop pr_exp exp
  | EBinary (exp1, bop, exp2) -> fpr ff "(%a %a %a)" SynOp.pr_binary bop pr_exp exp1
      pr_exp exp2
  | ESeq exps -> fpr ff "(cat %a)" (Printing.print_list pr_exp " " "") exps
  | ETree(lbl, atts, conts) ->
      (match (Finfo.drop lbl) with 
        | Etag t -> fpr ff "(lab %s %a)" t pr_atts_exp (atts,conts)
        | Echar ch -> fpr ff "#\\%c " ch
        | _ -> raise (Failure "only Etag or Tchar are accepted as a label"))
  | EPcdata(lbl) -> assert false
  | EString2Pcdata exp -> pr_exp ff exp

and pr_exp_option ff expr =
  match Finfo.drop expr with
    | None -> fpr ff "()"
    | Some e -> pr_exp ff (Finfo.get expr, e)

let rec pr_block ff = function
  | [] -> fpr ff "()"
  | [s] -> pr_stmt ff s
  | s::rest ->
      (match (Finfo.drop s) with
        | SLocVars (vnames, ty) ->
            let pr_field ff = function
              | (_,None) -> raise (Failure ("let var must be initialized "))
              | (name,Some(e)) ->
                  if snd name = Ident.bogus_name then fpr ff "(%a %a)" pr_ty ty pr_exp e
                  else fpr ff "((tvar %a %a) %a)" pr_name name pr_ty ty pr_exp e
            in 
            fpr ff "(let (%a) %a)" (Printing.print_list pr_field " " "") vnames
              pr_block rest
        | _ -> fpr ff "(seq %a %a)" pr_stmt s pr_block rest)

and pr_stmt ff (s:stmt) = 
  match (Finfo.drop s) with
  | SBlock block -> pr_block ff (Finfo.drop block)
  | SExp exp -> pr_exp ff exp
  | SLocVars (vnames, ty) -> raise (Failure "unexpected SLocVars")
  | SReturn exp -> pr_exp_option ff exp
  | SIf(cond, b1, b2) ->
      fpr ff "(if %a %a %a)" pr_exp cond pr_block (Finfo.drop b1) pr_block
        (Finfo.drop b2)
  | SWhile(cond, body) -> 
      fpr ff "(while %a %a)" pr_exp cond pr_block (Finfo.drop body)
  | SMatch (exp, clauses) ->
      fpr ff "(match %a %a)" pr_exp exp pr_clauses clauses
  | SForeachRE (spa, exp, body) -> 
      fpr ff "(foreach %a %a %a)\n" pr_seqty spa pr_exp exp pr_block (Finfo.drop body)
  | SEmpty -> fpr ff "()"
  | SBreak -> failwith "Implement SBreak case for pr_stmt in xdprint.ml"

and pr_clauses ff =
  let pr_cl clause =
    let (pat, block) = Finfo.drop clause in
    fpr ff "(%a %a)" pr_ty pat pr_block (Finfo.drop block)
  in iter pr_cl


(********** Classes ***********)

let pr_params ff =
  iter (fun (name_opt,ty) -> 
	  let name = 
	    Bug.from_opt "Xdprint.pr_params" "missing param name" name_opt in
	    fpr ff "(tvar %a %a) " pr_name name pr_ty ty)

let pr_clas ff (_,clas) = 
  let print_field fld = 
    let(name,_,exp,_,ty) = Finfo.drop fld in
    match exp with
      | None -> raise (Failure "initializer must be present")
      | Some(e) ->
          fpr ff "(let (tvar %a %a) %a)\n" pr_name name pr_ty ty pr_exp e in 
  let print_meth meth =
    let (name, _, _, _, rettyp, params, body_opt) = Finfo.drop meth in
    let body = Bug.from_opt "Xdprint.pr_clas" "missing meth bosy" body_opt in
    fpr ff "(fun (%a %a) %a %a)\n" pr_name name pr_params params pr_ty_option rettyp
      pr_block (Finfo.drop body) in
  let clas' = Finfo.drop clas in
  iter print_field clas'.clas_fields;
  iter print_meth clas'.clas_meths


(****** Program *********)

let pr_prog ff prog = 
  let prdecl_seqty id (nctx, seqty) = 
    fpr ff "(type %s %a)\n" (Ident.name id) pr_seqty seqty
  in
  Ident.iter prdecl_seqty prog.seq;
  Ident.iter (fun _ -> (pr_clas ff)) prog.cls;
  fpr ff "@."
