type mid = Ident.id   (* member id *)
type locname = string
type pref = NoPref | Pref of string
type qname = pref * locname
type uri = NoUri | Uri of string
type expname = uri * locname

exception AmbiguousName of string
exception NameAlreadyUsed of string 

  
(*** Collection of types (or other members) defined in XML namespaces ******)

module MemberMap = Map.Make(struct 
				type t = expname 
				let compare = Pervasives.compare  end)

type spaces = mid MemberMap.t


let string_of_expname = function 
  | (NoUri, ln) -> ln
  | (Uri u, ln) -> "{" ^ u ^ "}:" ^ ln


let spacesEmpty = MemberMap.empty

let spacesAddMember spaces expname mid = 
  if MemberMap.mem expname spaces 
  then raise (NameAlreadyUsed (string_of_expname expname))
  else MemberMap.add expname mid spaces

let spacesFindMember spaces expname = 
  MemberMap.find expname spaces


(**********  Imports  **************)
type import =
  | BindPrefix of pref * uri      (* declaring a prefix for a namespace *)
  | UnbindPrefix of pref          (* undeclaring a namespace prefix *)
  | BindDefault of uri            (* declaring the default namespace *)
  | UnbindDefault                 (* undeclaring the default namespace *)
  | AliasMember of locname * mid  (* an alias for a namespace member *)


(*******  Contexts ******)

module BindsMap = Map.Make (struct 
			      type t = pref  
			      let compare = Pervasives.compare end)

module AliasesMap = Map.Make (struct
				type t = locname
				let compare = Pervasives.compare end)
type ctx = 
    { ctx_spaces: spaces;
      ctx_binds: uri BindsMap.t; 
      ctx_aliases: mid AliasesMap.t; }

let ctxInit spaces = 
  { ctx_spaces = spaces;
    ctx_binds = BindsMap.empty; 
    ctx_aliases = AliasesMap.empty; }

let ctxApplyImport ctx = function 
  | BindPrefix(pref, uri) -> 
      { ctx with ctx_binds = BindsMap.add pref uri ctx.ctx_binds }
  | UnbindPrefix pref -> 
      { ctx with ctx_binds = BindsMap.remove pref ctx.ctx_binds }
  | BindDefault uri ->
      { ctx with ctx_binds = BindsMap.add NoPref uri ctx.ctx_binds }
  | UnbindDefault ->
      { ctx with ctx_binds = BindsMap.remove NoPref ctx.ctx_binds }
  | AliasMember(locname, mid) -> 
      { ctx with ctx_aliases = AliasesMap.add locname mid ctx.ctx_aliases }

let ctxApplyImports ctx imports = 
  List.fold_left ctxApplyImport ctx imports


let ctxFindUri ctx pref = 
  try 
    BindsMap.find pref ctx.ctx_binds
  with Not_found -> 
    if pref = NoPref then NoUri
    else raise Not_found

let ctxLocateMember ctx expname = 
  spacesFindMember ctx.ctx_spaces expname

let ctxResolveQname ctx qname = 
  let (pref, locname) = qname in 
  let uri = ctxFindUri ctx pref in
  let deref_opt = 
    (try 
       Some (spacesFindMember ctx.ctx_spaces (uri, locname))
     with Not_found -> None) in
  let alias_opt = 
    (try 
       if pref = NoPref then Some(AliasesMap.find locname ctx.ctx_aliases)
       else None
     with Not_found -> None) in
    match deref_opt, alias_opt with 
      | None, None -> raise Not_found
      | None, Some mid
      | Some mid, None -> mid
      | Some _, Some _ -> raise (AmbiguousName locname)
