using System;
using System.IO;
using System.Net;
using System.Xml;
using System.Xml.Schema;
using System.Collections;
using System.Collections.Specialized;
using System.Text;

namespace Xtatic {

  public class Attributes : HybridDictionary {}
  
  public class SeqBuilder {
    private Seq fst = null;
    public Seq last = null;
    
    public SeqBuilder() {}
    
    public void Append(Seq s) {
      if (s == Utils.the_empty_seq) return;
      Seq s2 = s.Clone();
      if (fst == null) fst = last = s2;
      else { 
        last.next = s2;
        last = s2;
      }
    }
    
    // assumes that SeqBuilder.Append* is not called after
    // SeqBuilder.AppendNoClone is invoked. Only SeqBuilder.ToSeq
    // remains to be called at this point
    public void AppendNoClone(Seq s) {
      if (fst == null) fst = s;
      else last.next = s;
    }
    
    public Seq ToSeq() {
      return fst;
    }
  }

  public abstract class Seq {
    public enum Kind {Empty, Element, Object, Bool, Int, Char,
                      String, Substring, Append, Wrap, Typestamp};
    public Kind kind;
    public Attributes attributes;
    public Seq contents;
    public Seq next;
    public override String ToString() { return Utils.ToString(this); }
    public abstract Seq Clone();
  }
  
  public class SeqEmpty : Seq {
    public SeqEmpty() { 
      kind = Seq.Kind.Empty;
      next = null; // [XXX] bogus and arbitrary
    }
    public override Seq Clone() { return Utils.the_empty_seq; }
  }
  
  public class SeqTypestamp : Seq {
    public object typestamp;
    public SeqTypestamp(object typestamp, Seq contents) { 
      kind = Seq.Kind.Typestamp;
      this.typestamp = typestamp;
      next = contents;
    }
    public override Seq Clone() { 
      throw new SystemException("SeqTypestamp.Clone: should not be cloned");
    }
  }
  
  public class SeqElement : Seq {
    public Tag tag;
    public SeqElement(Tag tag, Seq contents, Seq next) { 
      kind = Kind.Element; 
      this.tag = tag;
      this.contents = contents;
      this.next = next;
    }
    public SeqElement(Tag tag, Seq contents) { 
      kind = Kind.Element; 
      this.tag = tag;
      this.contents = contents;
      this.next = Utils.the_empty_seq;
    }
    public SeqElement(string tag, Seq contents, Seq next) { 
      kind = Kind.Element; 
      this.tag = Tag.Intern(tag); 
      this.contents = contents; 
      this.next = next; 
    }
    public override Seq Clone() { 
      Seq res = new SeqElement(tag,contents,next);
      res.attributes = attributes;
      return res;
    }
  }

  public class SeqObject : Seq {
    public object tag;
    public SeqObject(object tag, Seq contents, Seq next) { 
      kind = Kind.Object; 
      this.tag = tag; 
      this.contents = contents; 
      this.next = next; 
    }
    public override Seq Clone() { 
      Seq res = new SeqObject(tag,contents,next);
      res.attributes = attributes;
      return res;
    }
  }

  public class SeqBool : Seq {
    public bool tag;
    public SeqBool(bool tag, Seq contents, Seq next) { 
      kind = Kind.Bool; 
      this.tag = tag;
      this.contents = contents;
      this.next = next;
    }
    public override Seq Clone() { 
      Seq res = new SeqBool(tag,contents,next);
      res.attributes = attributes;
      return res;
    }
  }

  public class SeqInt : Seq {
    public int tag;
    public SeqInt(int tag, Seq contents, Seq next) { 
      kind = Kind.Int; 
      this.tag = tag;
      this.contents = contents;
      this.next = next;
    }
    public override Seq Clone() { 
      Seq res = new SeqInt(tag,contents,next);
      res.attributes = attributes;
      return res;
    }
  }

  public class SeqChar : Seq {
    public char tag;
    public SeqChar(char tag, Seq contents, Seq next) { 
      kind = Kind.Char; 
      this.tag = tag;
      this.contents = contents;
      this.next = next;
    }
    public SeqChar(char tag, Seq next) : this(tag, Utils.the_empty_seq,next) { }
    public override Seq Clone() { 
      Seq res = new SeqChar(tag,contents,next);
      res.attributes = attributes;
      return res;
    }
  }

  public class SeqString : Seq {
    new public string contents;
    public SeqString(string contents, Seq next) { 
      kind = Kind.String;
      this.contents = contents;
      this.next = next;
    }
    public override Seq Clone() { 
      Seq res = new SeqString(contents,next);
      return res;
    }
  }

  public class SeqSubstring : Seq {
    new public string contents;
    public int offset;
    public int length;
    public SeqSubstring(string contents, int offset, int length, Seq next) { 
      kind = Kind.Substring;
      this.contents = contents;
      this.offset = offset;
      this.length = length;
      this.next = next;
    }
    public SeqSubstring(string cont, Seq next) : this(cont, 0, cont.Length, next) { }
    public override Seq Clone() { 
      Seq res = new SeqSubstring(contents,offset,length,next);
      return res;
    }
  }

  public class SeqAppend : Seq {
    public Seq fst;
    public Seq snd;
    public SeqAppend(Seq fst, Seq snd) { 
      kind = Kind.Append;
      this.fst = fst;
      this.snd = snd;
    }
    public override Seq Clone() { 
      Seq res = new SeqAppend(fst,snd);
      return res;
    }
  }

  internal abstract class SeqWrap : Seq {
    protected Seq seq;
    public abstract Seq Normalize();
  }

  internal class SeqWrapDom : SeqWrap {
    private XmlElement dom;

    public SeqWrapDom(XmlElement elt) {
      kind = Seq.Kind.Wrap;
      seq = null;
      dom = elt;
    }

    private SeqWrapDom(Seq s, XmlElement e) { //to be used in cloning
      kind = Seq.Kind.Wrap;
      seq = s;
      dom = e;
    }

    public override Seq Clone() {
      return new SeqWrapDom(seq, dom);
    }

    private Seq SeqFromXmlNode(XmlNode n) {
      if (n == null) return Utils.the_empty_seq;
      switch (n.NodeType) {
      case XmlNodeType.Element: return new SeqWrapDom((XmlElement)n);
      case XmlNodeType.Text: {
        Seq next =  SeqFromXmlNode(n.NextSibling); 
        Utils.SetSubstringPcdataHandling();
        Seq s = Utils.convText(((XmlText)n).Value, next); 
        Utils.ResetSubstringPcdataHandling();
        return s;
      } default: {
        throw new Exception("In SeqWrapDom: encountered an untreated DOM node");
      }
      }
    }

    public override Seq Normalize() {
      if (seq != null) {
        return seq;
      } else if (Utils.appendpolicy == Utils.AppendPolicy.Lazynormalize) {
        Seq child = SeqFromXmlNode(dom.FirstChild);
        Seq next = SeqFromXmlNode(dom.NextSibling);
        string tag = dom.LocalName;
        return new SeqElement(tag, child, next);
      } else if (Utils.appendpolicy == Utils.AppendPolicy.Eagernormalize) {
        Utils.SetSubstringPcdataHandling();
        Seq child = Utils.convNode(dom.FirstChild, Utils.the_empty_seq);
        Seq next = Utils.convNode(dom.NextSibling, Utils.the_empty_seq);
        Utils.ResetSubstringPcdataHandling();
        string tag = dom.LocalName;
        return new SeqElement(tag, child, next);
      } else {
        throw new SystemException("SeqDomWrap.Normalize: should never be reached");
      }
    }
  }


  public class Tag {
    static Hashtable tbl = new Hashtable();
    private string name;
    public Tag(string x) { name = x; } /*VG??: shouldn't be private? */
    public string Name { get { return name; } }
    public static Tag Intern(string s) {
      if (tbl.Contains(s)) {
        return (Tag) tbl[s];
      }
      Tag tag = new Tag(s);
      tbl[s] = tag;
      return tag;
    }
    public static Tag Create(string s) { return Intern(s);  }    

    public override string ToString() { return name; }

    //Equals and GetHashCode must be overriden together 
    public override bool Equals (object x) { return (this == x);  }
    public override int GetHashCode() { return base.GetHashCode(); }

  }

  public class Typestamp {
    static Hashtable tbl = new Hashtable();
    public static object Intern(string s) {
      if (tbl.Contains(s)) return tbl[s];
      tbl[s] = s;
      return s;
    }
  }

  public class Utils {
    internal enum PcdataEncoding { Char, String, Substring, Hybrid }
    internal static PcdataEncoding pcdataencoding;

    internal enum AppendPolicy { Eager, Lazynormalize, Eagernormalize }
    internal static AppendPolicy appendpolicy;

    /// <summary>
    /// HTML-encodes a string and returns the encoded string.
    /// </summary>
    /// <param name="s">The text string to encode. </param>
    /// <returns>The HTML-encoded text.</returns>
    public static string HtmlEncode (string s) 
    {
      if (s == null)
        return null;
      
      StringBuilder output = new StringBuilder ();
      
      foreach (char c in s) 
        switch (c) {
        case '&' :
          output.Append ("&amp;");
          break;
        case '>' : 
          output.Append ("&gt;");
          break;
        case '<' :
          output.Append ("&lt;");
          break;
        case '"' :
          output.Append ("&quot;");
          break;
        default:
          // MS starts encoding with &# from 160 and stops at 255.
          // We don't do that. One reason is the 65308/65310 unicode
          // characters that look like '<' and '>'.
          if (c > 159) {
            output.Append ("&#");
            output.Append (((int) c).ToString (System.Globalization.CultureInfo.InvariantCulture));
            output.Append (";");
          } else {
            output.Append (c);
          }
          break;
        }
      return output.ToString ();
    }

    public static void SetPcdataEncoding(string what) {
      if (what == "char") pcdataencoding = PcdataEncoding.Char;
      else if (what == "string") pcdataencoding = PcdataEncoding.String;
      else if (what == "substring") pcdataencoding = PcdataEncoding.Substring;
      else if (what == "hybrid") pcdataencoding = PcdataEncoding.Hybrid;
      else throw new SystemException("SetPcdataEncoding: unknown encoding " + what);
    }

    public static void SetAppendPolicy(string what) {
      if (what == "eager") appendpolicy = AppendPolicy.Eager;
      else if (what == "lazynormalize") appendpolicy = AppendPolicy.Lazynormalize;
      else if (what == "eagernormalize") appendpolicy = AppendPolicy.Eagernormalize;
      else throw new SystemException("SetAppendPolicy: unknown encoding " + what);
    }

    public static Seq the_empty_seq = new SeqEmpty();

    public static bool switchA(Attributes attr, Tag [] tags, Seq [] values) {
      int i;

      if (attr == null && tags.Length > 0) return false;

      for (i = 0; i < tags.Length; i++) {
	Seq value = (Seq) attr[tags[i]];
	if (value == null) return false;
	values.SetValue(value, i);
      }

      if (attr == null) {
        values[i] = the_empty_seq;
        return true;
      } else {
        IDictionaryEnumerator e = attr.GetEnumerator();
        Seq rest = the_empty_seq;
        while (e.MoveNext()) {
          Tag tag = (Tag)e.Key;
          if (Array.IndexOf(tags, tag) >= 0) continue;
          rest = new SeqElement(tag, (Seq)e.Value, rest);
        }
        values[i] = rest;
        return true;
      }
    }

    public static Seq AttrsToSeq(Attributes attr) {
      if (attr == null) {
        return the_empty_seq;
      } else {
        IDictionaryEnumerator e = attr.GetEnumerator();
        Seq rest = the_empty_seq;
        while (e.MoveNext()) {
          Tag tag = (Tag)e.Key;
          rest = new SeqElement(tag, (Seq)e.Value, rest);
        }
        return rest;
      }
    }

    private static Seq convText1(string txt, Seq next) {
      int n = txt.Length;
      for (int i = n-1; i >= 0; i--) next = new SeqChar(txt[i], the_empty_seq, next);
      return next;
    }

    private static Seq convText2(string txt, Seq next) {
      return new SeqString(txt,next);
    }

    private static StringBuilder b = null;
    private static Stack q = null;
    private static string dummy = "";

    internal static void SetSubstringPcdataHandling() {
      b = new StringBuilder();
      q = new Stack();
      dummy = "";
    }

    internal static void ResetSubstringPcdataHandling() {
      string tmp = b.ToString();
      while (q.Count > 0) {
        SeqSubstring x = (SeqSubstring) q.Pop();
        x.contents = tmp;
      }
      b = null;
      q = null;
    }

    private static Seq convText3(string txt, Seq next) {
      int offset = b.Length;
      int len = txt.Length;
      b.Append(txt);
      Seq res = new SeqSubstring(dummy, offset, len, next);
      q.Push(res);
      return res;
    }

    internal static Seq convText(string txt, Seq next) {
      switch (pcdataencoding) {
      case PcdataEncoding.Char: 
        return convText1(txt, next);
      case PcdataEncoding.String: 
        return convText2(txt, next);
      case PcdataEncoding.Substring: 
      case PcdataEncoding.Hybrid: 
        return convText3(txt, next);
      default:
        throw new SystemException("convText: should never be reached");
      }
    }

    public static Seq convSeq(XmlNode n) {
      Seq res = the_empty_seq;
      Stack q = new Stack();
      for (; n != null; n = n.NextSibling) q.Push(n);
      while (q.Count > 0) {
        XmlNode n2 = (XmlNode) q.Pop();
        res = convNode(n2, res);
      }
      return res;
    }

    internal static Seq convNode(XmlNode n, Seq next) {
      switch (n.NodeType) {
      case XmlNodeType.Element:
        SeqElement el = new SeqElement(n.Name, convSeq(n.FirstChild), next);
        XmlAttributeCollection atts = n.Attributes;
        if (atts.Count != 0) {
          Attributes d = new Attributes();
          foreach (XmlAttribute a in atts) {
            d[Tag.Intern(a.Name)] = convText(a.Value, the_empty_seq);
          }
          el.attributes = d;
        }
        return el;
      case XmlNodeType.Text: return convText(n.InnerText, next);
      case XmlNodeType.Comment: return next;
      case XmlNodeType.EntityReference: 
        throw new SystemException("convNode: got an unresolved entity reference from the parser " + n.Name);
      case XmlNodeType.CDATA:
	return convText(n.Value, next);

      default:
        throw new SystemException("convNode: unexpected node type" + n);
      }
    }


    /** Loads an XML document from file without validation.
        In particular, does not resolve non-standard entities.    */
    public static Seq LoadXml(string fname) {
      StreamReader s = null;
      try {
        XmlDocument doc = new XmlDocument();

        //TODO: Michael, please document why this dancing with a StreamReader
        //is necessary, and a doc.Load(fname) cannot be used instead.
        //(Or at least do doc.Load(s) rather than doc.LoadXml(str))

        //      doc.Load(fname);
        s = 
          new StreamReader(fname,System.Text.Encoding.GetEncoding("ISO-8859-1"));
        string str = s.ReadToEnd();
        s.Close();
      
        //Timer t = new Timer();
        //t.StartTimer();

        doc.LoadXml(str);

        //Console.WriteLine("loading time: " + t.GetTimer());
        //GC.Collect();
        //System.Threading.Thread.Sleep(10000);

        //t.StartTimer();

        SetSubstringPcdataHandling();
        Seq res = convNode(doc.DocumentElement,the_empty_seq); 
        ResetSubstringPcdataHandling();

        //Console.WriteLine("conversion time: " + t.GetTimer());
        //GC.Collect();
        //System.Threading.Thread.Sleep(10000);

        return res;
      }
      catch (XmlException xe){
        System.Console.WriteLine("\r\nException : There was a load or parse error in the XML");
        System.Diagnostics.Debug.Fail("\r\nException : There was a load or parse error in the XML");
        throw xe;
      }
      catch (Exception e) {
        System.Console.WriteLine("\r\nException : The file could not be read");
        System.Console.WriteLine(e);
        System.Diagnostics.Debug.Fail("\r\nException : The file could not be read");
        throw e;;
      }
      finally {
        if (s != null) {
          s.Close();
        }
      }
    }

    /**  Loads an XML document and attempts to validate it according
     * to the DTD specified inside.  Actual DTD location is done by
     * standard .NET means (XmlUriResolver).        */
    public static Seq LoadXmlValidated (string fname) {
      StreamReader s= null;
      try{
        s = 
          new StreamReader(fname,System.Text.Encoding.GetEncoding("ISO-8859-1"));

        XmlValidatingReader reader = 
          new XmlValidatingReader(new XmlTextReader(s));
        reader.ValidationType = ValidationType.DTD;
        //TODO for simplicity, we do not yet instantiate
        //reader.ValidationEventHandelr.  This silently skips validation errors.

        XmlDocument doc = new XmlDocument();
        doc.Load(reader);

        SetSubstringPcdataHandling();
        Seq res = convNode(doc.DocumentElement,the_empty_seq); 
        ResetSubstringPcdataHandling();

        return res;
      }
      catch (XmlException xe){
        System.Console.WriteLine("\r\nException : There was a load or parse error in the XML");
        System.Diagnostics.Debug.Fail("\r\nException : There was a load or parse error in the XML");
        throw xe;
      }
      catch (Exception e) {
        System.Console.WriteLine("\r\nException : The file could not be read");
        System.Diagnostics.Debug.Fail("\r\nException : The file could not be read");
        throw e;
      }
      finally {
        if (s != null) {
          s.Close();
        }
      }
    }

    /**
     * <summary> This implementation outputs the Seq to a file by use of a
     * LIFO Stack.  Uses an XmlTextWriter to ensure text is indented,
     * correctly formatted, and escapes XML Special characters. </summary>
     *
     * <remark> By default, ToString escapes special characters. However,
     * if you want to build part of your xml by hand, call ToString with
     * the raw flag activated to directly copy the content of your Xml
     * </remark>
     *
     */

    public static string ToString(Seq s, bool raw) {
      StringBuilder res = new StringBuilder();
      XmlTextWriter w = new XmlTextWriter (new StringWriter(res));
      w.Formatting = Formatting.Indented;
      Stack q = new Stack();
      q.Push(s);
      while (q.Count > 0) {
        object o = q.Pop();
        if (o is string) {
          w.WriteEndElement();
          continue;
        };
        s = (Seq)o;
        if (s.kind != Seq.Kind.Append && 
            s.kind != Seq.Kind.Empty && 
            s.kind != Seq.Kind.Wrap)
          q.Push(s.next);
        string towrite = null;
        switch (s.kind) {
        case Seq.Kind.Empty:
          break;
        case Seq.Kind.Element:          
        case Seq.Kind.Object: {
          string t = null;
          if (s.kind == Seq.Kind.Object) {
            t=((SeqObject) s).tag.ToString();
          }
          else {
            t =((SeqElement) s).tag.ToString();
          }
          w.WriteStartElement(t);
          if (s.attributes != null)
            foreach (DictionaryEntry de in s.attributes) {
              if (raw) {w.WriteAttributeString(((Tag)de.Key).Name,
                                               de.Value.ToString());
              }
              else {
                w.WriteAttributeString(((Tag)de.Key).Name,
                                       HtmlEncode(de.Value.ToString()));
              }
            };
          if (s.contents == Utils.the_empty_seq) {
            w.WriteEndElement();
          }
          else {
            q.Push(t);
            q.Push(s.contents);
          }
          break;
        } 
        case Seq.Kind.Append: {
          SeqAppend sc = (SeqAppend) s;
          q.Push(sc.snd);
          if (sc.fst != null) q.Push(sc.fst);
          break;
        }
        case Seq.Kind.Wrap: {
          SeqWrap sw = (SeqWrap)s;
          Seq tmp = sw.Normalize();
          q.Push(tmp);
          break;
        }
        case Seq.Kind.Bool: 
          towrite = ((SeqBool) s).tag.ToString();
          goto default;
        case Seq.Kind.Int: 
          towrite = ((SeqInt) s).tag.ToString();
          goto default;
        case Seq.Kind.Char: 
          towrite = ((SeqChar) s).tag.ToString();
          goto default;
        case Seq.Kind.String: 
          towrite = ((SeqString) s).contents;
          goto default;
        case Seq.Kind.Substring:
          SeqSubstring sc = ((SeqSubstring) s);
          towrite = sc.contents.Substring(sc.offset,
                                          sc.length);
          continue;
        default: {          
          if (towrite != null) {
            if (raw) {
              w.WriteRaw(towrite);
            }
            else {
              w.WriteString(towrite);
            }
          }
          if (s.next.kind != Seq.Kind.Empty 
              && (s.kind == Seq.Kind.Bool 
                  || s.kind ==Seq.Kind.Int)) 
            w.WriteWhitespace(" ");
          break;
        }
        }
      }
      w.Flush();
      w.Close();
      return res.ToString();
    }

    /** 
     * <remark> In order to preserve backwards-compatibility, ToString
     * escapes special characters by default </remark>
     */

    public static string ToString (Seq s) {
      return ToString(s,false);
    }

    private static Seq AppendE(Seq v1, Seq v2) {
      if (v1.kind == Seq.Kind.Empty) return v2;
      else {
        Seq tmp = AppendE(v1.next, v2);
        switch (v1.kind) {
        case Seq.Kind.Element: {
          SeqElement x = (SeqElement) v1;
          SeqElement res = new SeqElement(x.tag, x.contents, tmp);
          res.attributes = x.attributes;
          return res;
        } case Seq.Kind.Object: {
          SeqObject x = (SeqObject) v1;
          SeqObject res = new SeqObject(x.tag, x.contents, tmp);
          res.attributes = x.attributes;
          return res;
        } case Seq.Kind.Bool: {
          SeqBool x = (SeqBool) v1;
          return new SeqBool(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.Int: {
          SeqInt x = (SeqInt) v1;
          return new SeqInt(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.Char: {
          SeqChar x = (SeqChar) v1;
          return new SeqChar(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.String: {
          SeqString x = (SeqString) v1;
          return new SeqString(x.contents, tmp);
        } case Seq.Kind.Substring: {
          SeqSubstring x = (SeqSubstring) v1;
          return new SeqSubstring(x.contents, x.offset, x.length, tmp);
        } default:
            throw new SystemException("AppendE: should never be reached");
        }
      }
    }

    private static Seq AppendL(Seq v1, Seq v2) {
      if (v1.kind == Seq.Kind.Empty) return v2;
      else if (v2.kind == Seq.Kind.Empty) return v1;
      else if (v1.kind != Seq.Kind.Append && 
               v1.kind != Seq.Kind.Wrap &&
               v1.next.kind == Seq.Kind.Empty) {
        Seq res = v1.Clone();
        res.next = v2;
        return res;
      } else {
        return new SeqAppend(v1,v2);
      }
    }

    public static Seq Append(Seq v1, Seq v2) {
      if (appendpolicy == AppendPolicy.Eager) return AppendE(v1,v2);
      else return AppendL(v1,v2);
    }    

    public static Seq NormalizeE(Seq s) {
      if (s.kind != Seq.Kind.Append && s.kind != Seq.Kind.Wrap) return s;
      if (s.kind == Seq.Kind.Append) {
        SeqAppend top = (SeqAppend) s;
        if (top.fst == null) return top.snd;
        SeqBuilder res = new SeqBuilder();
        Stack q = new Stack();
        Stack tomodify = new Stack();
        q.Push(s);
        while (q.Count > 0) {
          s = (Seq) q.Pop();
          switch (s.kind) {
          case Seq.Kind.Empty: {
            continue;
          } case Seq.Kind.Object: 
          case Seq.Kind.Bool: 
          case Seq.Kind.Int: 
          case Seq.Kind.Char: 
          case Seq.Kind.String: 
          case Seq.Kind.Substring:
          case Seq.Kind.Element: {
            if (q.Count == 0) {
              res.AppendNoClone(s);
              break;
            } else {
              res.Append(s);
              q.Push(s.next);
            }
            break;
          } case Seq.Kind.Append: {
            if (q.Count == 0 && res.last != null) {
              tomodify.Push(s);
              tomodify.Push(res.last);
            }
            SeqAppend tmp = (SeqAppend) s;
            q.Push(tmp.snd);
            if (tmp.fst != null) q.Push(tmp.fst);
            break;
          } case Seq.Kind.Wrap: {
            throw new Exception
              ("NormalizeE: Eager norm not implemented on SeqWrap");
          } default: {
            throw new SystemException("NormalizeE: should never be reached");
          }
          }
        }
        while (tomodify.Count > 0) {
          Seq next = (Seq) tomodify.Pop();
          SeqAppend s2 = (SeqAppend) tomodify.Pop();
          s2.fst = null;
          s2.snd = next;
        }
        Seq res2 = res.ToSeq();
        top.fst = null;
        top.snd = res2;
        return res2;
      } else if (s.kind == Seq.Kind.Wrap) {
        throw new Exception("NormalizeE: Eager normalization not implemented on SeqWrap");
      } else {
        throw new Exception("NormalizeE: unexpected Seq kind " + s.kind);
      }
    }

    public static Seq NormalizeL(Seq s) {
      if (s.kind != Seq.Kind.Append && s.kind != Seq.Kind.Wrap) return s;
      if (s.kind == Seq.Kind.Append) {
        SeqAppend top = (SeqAppend) s;
        if (top.fst == null) return top.snd;
        Queue q = new Queue();
        while (s.kind == Seq.Kind.Append) {
          SeqAppend tmp = (SeqAppend) s;
          if (tmp.fst == null) {
            s = tmp.snd;
          } else {
            q.Enqueue(tmp.snd);
            s = tmp.fst;
          }
        }
        Seq res;
        if (s.kind == Seq.Kind.Wrap) res = ((SeqWrap)s).Normalize();
        else res = s.Clone();
        Seq next = (Seq) q.Dequeue();
        while (q.Count > 0) {
          s = (Seq) q.Dequeue();
          if (s.kind != Seq.Kind.Append && 
              s.kind != Seq.Kind.Wrap && 
              s.next.kind == Seq.Kind.Empty) {
            Seq tmp = s.Clone();
            tmp.next = next;
            next = tmp;
          } else {
            next = new SeqAppend(s,next);
          }
        }
        res.next = Utils.Append(res.next,next);
        top.fst = null;
        top.snd = res;
        return res;
      } else if (s.kind == Seq.Kind.Wrap) {
        return ((SeqWrap)s).Normalize(); 
      } else {
        throw new Exception("NormalizeL: unexpected Seq kind " + s.kind);
      }
    }

    public static Seq Normalize(Seq s) {
      if (appendpolicy == AppendPolicy.Lazynormalize) return NormalizeL(s);
      else if (appendpolicy == AppendPolicy.Eagernormalize) return NormalizeE(s);
      else throw new SystemException("Normalize: should never be reached");
    }
 
    /** Traverses v1 until v2 is met.  Prepends items met along the way 
        to rest and returns.
    */
    /* --VG??  implemented by recursion on v1; is it memory-efficient? 
     */
    public static Seq Prefix(Seq v1, Seq v2, Seq rest) {
      if (v1.kind == Seq.Kind.Append) {
        if (((SeqAppend)v1).fst == null) v1 = ((SeqAppend)v1).snd;
        else v1 = Normalize(v1);
      } else if (v1.kind == Seq.Kind.Wrap) {
        v1 = ((SeqWrap)v1).Normalize();
      }
      if ((object)v1 == (object)v2) 
        return rest;
      if (v1.kind == Seq.Kind.Empty) {
        throw new SystemException("Prefix: unexpected empty sequence: ");
      } else {
        Seq tmp = Prefix(v1.next, v2, rest);
        switch (v1.kind) {
        case Seq.Kind.Element: {
          SeqElement x = (SeqElement) v1;
          SeqElement res = new SeqElement(x.tag, x.contents, tmp);
          res.attributes = x.attributes;
          return res;
        } case Seq.Kind.Object: {
          SeqObject x = (SeqObject) v1;
          SeqObject res = new SeqObject(x.tag, x.contents, tmp);
          res.attributes = x.attributes;
          return res;
        } case Seq.Kind.Bool: {
          SeqBool x = (SeqBool) v1;
          return new SeqBool(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.Int: {
          SeqInt x = (SeqInt) v1;
          return new SeqInt(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.Char: {
          SeqChar x = (SeqChar) v1;
          return new SeqChar(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.String: {
          SeqString x = (SeqString) v1;
          return new SeqString(x.contents, tmp);
        } case Seq.Kind.Substring: {
          SeqSubstring x = (SeqSubstring) v1;
          return new SeqSubstring(x.contents, x.offset, x.length, tmp);
        } default:
            throw new SystemException("Prefix: should never be reached");
        }
      }
    }

    public static Seq FixedPrefix(Seq v1, int cnt) {
      if (v1.kind == Seq.Kind.Append) {
        if (((SeqAppend)v1).fst == null) v1 = ((SeqAppend)v1).snd;
        else v1 = Normalize(v1);
      } else if (v1.kind == Seq.Kind.Wrap) {
        v1 = ((SeqWrap)v1).Normalize();
      }
      if (cnt == 0) return the_empty_seq;
      else if (v1.kind == Seq.Kind.Empty) 
        throw new SystemException("FixedPrefix: unexpected empty sequence: " + cnt);
      else if (v1.kind == Seq.Kind.String) {
        SeqString x = (SeqString) v1;
        string cont = x.contents;
        int len = cont.Length;
        if (len<=cnt) {
          Seq tmp = FixedPrefix(v1.next, cnt-len);
          return new SeqString(cont, tmp);
        } else if (cnt == 1 && pcdataencoding == PcdataEncoding.Hybrid) {
          return new SeqChar(cont[0], the_empty_seq);
        } else {
          return new SeqString(cont.Substring(0,cnt), the_empty_seq);
        }
      } else if (v1.kind == Seq.Kind.Substring) {
        SeqSubstring x = (SeqSubstring) v1;
        string cont = x.contents;
        int pos = x.offset;
        int len = x.length;
        if (len<=cnt) {
          Seq tmp = FixedPrefix(v1.next, cnt-len);
          return new SeqSubstring(cont, pos, len, tmp);
        } else if (cnt == 1 && pcdataencoding == PcdataEncoding.Hybrid) {
          return new SeqChar(cont[pos], the_empty_seq);
        } else {
          return new SeqSubstring(cont, pos, cnt, the_empty_seq);
        }
      } else {
        Seq tmp = FixedPrefix(v1.next, cnt-1);
        switch (v1.kind) {
        case Seq.Kind.Element: {
          SeqElement x = (SeqElement) v1;
          SeqElement res = new SeqElement(x.tag, x.contents, tmp);
          res.attributes = x.attributes;
          return res;
        } case Seq.Kind.Object: {
          SeqObject x = (SeqObject) v1;
          SeqObject res = new SeqObject(x.tag, x.contents, tmp);
          res.attributes = x.attributes;
          return res;
        } case Seq.Kind.Bool: {
          SeqBool x = (SeqBool) v1;
          return new SeqBool(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.Int: {
          SeqInt x = (SeqInt) v1;
          return new SeqInt(x.tag, the_empty_seq, tmp);
        } case Seq.Kind.Char: {
          SeqChar x = (SeqChar) v1;
          return new SeqChar(x.tag, the_empty_seq, tmp);
        } default:
            throw new SystemException("FixedPrefix: should never be reached");
        }
      }
    }

    public static Attributes CreateAttributes(object key, object value) {
      Attributes res = new Attributes();
      res.Add(key,value);
      return res;
    }

    public static Attributes ProcessAtts(Seq atts) {
      if (atts == the_empty_seq) return null;
      else {
        Attributes res = new Attributes();
        while (atts != the_empty_seq) {
          switch (atts.kind) {
          case Seq.Kind.Element: {
            SeqElement s = (SeqElement) atts;
            res.Add(s.tag, s.contents);
            atts = s.next;
            break;
          } default: {
            throw new SystemException("ProcessAtts: only element nodes allowed in attributes");
          }
          }
        }
        return res;
      }
    }

    public static Seq CreateTree(Tag tag, Seq atts, Seq contents) {
      SeqElement el = new SeqElement(tag, contents, the_empty_seq);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq CreateTree(object tag, Seq atts, Seq contents) {
      SeqObject el = new SeqObject(tag, contents, the_empty_seq);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq CreateTree(bool tag, Seq atts, Seq contents) {
      SeqBool el = new SeqBool(tag, contents, the_empty_seq);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq CreateTree(int tag, Seq atts, Seq contents) {
      SeqInt el = new SeqInt(tag, contents, the_empty_seq);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq CreateTree(char tag, Seq atts, Seq contents) {
      SeqChar el = new SeqChar(tag, contents, the_empty_seq);
      el.attributes = ProcessAtts(atts);
      return el;
    }

    public static Seq StringToPcdata(string s) {
      SetSubstringPcdataHandling();
      Seq res = convText(s, the_empty_seq);
      ResetSubstringPcdataHandling();
      return res;
    }

    public static Seq StringOfIntToPcdata (int i){
      string alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
      int offset = (i % 26);
      return StringToPcdata(alphabet.Substring(offset,1));
    }
    
    public static int StringToInt(string x) {
      return Convert.ToInt32(x);
    }

    public static int PcdataToInt(Seq x) {
      return StringToInt(ToString(x));
    }

    public static double StringToDouble(string x) {
      return Convert.ToDouble(x);
    }

    public static double PcdataToDouble(Seq x) {
      return StringToDouble(ToString(x));
    }

    public static Seq IntToPcdata(int n) {
      string s = n.ToString();
      return StringToPcdata(s);
    }

    public static Seq PcdataToUpper(Seq x) {
      string y = ToString(x);
      return StringToPcdata(y.Substring(0,1).ToUpper() + y.Substring(1,y.Length-1));
    }

    public static Seq PccharUpperCase(Seq c) {
      string y = ToString(c);
      return StringToPcdata(y.Substring(0,1).ToUpper());
    }

    public static Seq PccharLowerCase(Seq c) {
      string y = ToString(c);
      return StringToPcdata(y.Substring(0,1).ToLower());
    }

    public static string Coalesce(Seq x, ref Seq rest) {
      if (x.kind != Seq.Kind.Char && 
          x.kind != Seq.Kind.String && 
          x.kind != Seq.Kind.Substring && 
          x.kind != Seq.Kind.Append) {
        rest = x;
        return "";
      }
      if (x.kind == Seq.Kind.String && 
          x.next.kind != Seq.Kind.Char &&
          x.next.kind != Seq.Kind.String &&
          x.next.kind != Seq.Kind.Substring &&
          x.next.kind != Seq.Kind.Append) {
        rest = x.next;
        return ((SeqString)x).contents;
      }
      if (x.kind == Seq.Kind.Substring && 
          x.next.kind != Seq.Kind.Char &&
          x.next.kind != Seq.Kind.String &&
          x.next.kind != Seq.Kind.Substring &&
          x.next.kind != Seq.Kind.Append) {
        rest = x.next;
        SeqSubstring y = (SeqSubstring) x;
        return y.contents.Substring(y.offset, y.length);
      }
      StringBuilder res = new StringBuilder();
      while (true) {
        if (x.kind == Seq.Kind.Append) x = Normalize(x);
        switch (x.kind) {
        case Seq.Kind.Char: {
          SeqChar y = (SeqChar) x;
          res.Append(y.tag);
          x = x.next;
          break;
        } case Seq.Kind.String: {
          SeqString y = (SeqString) x;
          res.Append(y.contents);
          x = x.next;
          break;
        } case Seq.Kind.Substring: {
          SeqSubstring y = (SeqSubstring) x;
          res.Append(y.contents, y.offset, y.length);
          x = x.next;
          break;
        } default: {
          rest = x;
          return res.ToString();
        }
        }
      }
    }

    public static string Coalesce(Seq x, ref int offset, ref int length, ref Seq rest) {
      if (x.kind != Seq.Kind.Char && 
          x.kind != Seq.Kind.String && 
          x.kind != Seq.Kind.Substring && 
          x.kind != Seq.Kind.Append) {
        rest = x;
        offset = 0;
        length = 0;
        return "";
      }
      if (x.kind == Seq.Kind.String && 
          x.next.kind != Seq.Kind.Char &&
          x.next.kind != Seq.Kind.String &&
          x.next.kind != Seq.Kind.Substring &&
          x.next.kind != Seq.Kind.Append) {
        rest = x.next;
        string str = ((SeqString)x).contents;
        offset = 0;
        length = str.Length;
        return str;
      }
      if (x.kind == Seq.Kind.Substring && 
          x.next.kind != Seq.Kind.String &&
          x.next.kind != Seq.Kind.Char &&
          x.next.kind != Seq.Kind.Substring &&
          x.next.kind != Seq.Kind.Append) {
        rest = x.next;
        SeqSubstring y = (SeqSubstring) x;
        offset = y.offset;
        length = y.length;
        return y.contents;
      }
      StringBuilder res = new StringBuilder();
      while (true) {
        if (x.kind == Seq.Kind.Append) x = Normalize(x);
        switch (x.kind) {
        case Seq.Kind.Char: {
          SeqChar y = (SeqChar) x;
          res.Append(y.tag);
          x = x.next;
          break;
        } case Seq.Kind.String: {
          SeqString y = (SeqString) x;
          res.Append(y.contents);
          x = x.next;
          break;
        } case Seq.Kind.Substring: {
          SeqSubstring y = (SeqSubstring) x;
          res.Append(y.contents, y.offset, y.length);
          x = x.next;
          break;
        } default: {
          rest = x;
          string str2 = res.ToString();
          offset = 0;
          length = str2.Length;
          return str2;
        }
        }
      }
    }

    //--VG This private method is not used, and mcs annoyingly warns 
    //  about the fact.  
    /*
      private static bool EqualPcdataString(Seq x, Seq y) {
      if (x.kind==Seq.Kind.Empty && y.kind==Seq.Kind.Empty) return true;
      else if (x.kind==Seq.Kind.Empty || y.kind==Seq.Kind.Empty) return false;
      else {
      Seq bogus = null;
      string x1 = Coalesce(x,ref bogus);
      string y1 = Coalesce(y,ref bogus);
      return (x1==y1);
      }
      }
    */
    public static bool EqualPcdata(Seq x, Seq y) {
      Seq bogus = null;
      string str1 = Coalesce(x, ref bogus);
      string str2 = Coalesce(y, ref bogus);
      return str1 == str2;
    }

    public static bool PrefixPcdata(Seq x, Seq y) {
      Seq bogus = null;
      string str1 = Coalesce(x,ref bogus);
      string str2 = Coalesce(y,ref bogus);
      if (str1.Length > str2.Length) return false;
      else {
        for (int i = 0; i<str1.Length; i++) 
          if (str1[i]!=str2[i]) {
            return false;
          };
        return true;
      }
    }

    public static bool PrefixString(string fst, int pos, int len, string snd, int len2) {
      if (len2 > len) return false;
      for(int i = 0; i < len2; i++) if (fst[pos+i] != snd[i]) return false;
      return true;
    }


    /**
     * <summary> Basic implementation that makes use of an XmlTextWriter
     * to add a correct XML header that complies with the standard's
     * requirement and MSFT's use of UTF-16 to represent Xml
     * data. </summary>
     * <param name ="x"> the sequence to Pretty-Print</param>
     * <param name ="writeStartDocument"> Wether to add a xml 1.0 header 
     *                                    to the file</param>
     * <param name="destinationEncoding"> Transform the Xml from MSFT's UTF16
     *                                    representation to another encoding.
     *                                    </param>
     * <param name="useRaw"> the Raw parameter in ToString, controlling the
     *                       escaping of special characters.</param>
     */

    public static string ToPretty(Seq x, 
                                  bool writeStartDocument, 
                                  bool useRaw,
                                  string destinationEncoding) {

      // <remark>If at some point strict encoding choices need to be
      // applied to the files, this is the right place.</remark>
      // <remark>here's an url worth consulting on the matter :
      // http://weblogs.asp.net/rmclaws/archive/2003/07/31/22080.aspx
      // </remark>
      try {
        string initialString = ToString (x,useRaw);
        string encodedString = null;
        if (writeStartDocument 
            || (Encoding.GetEncoding(destinationEncoding) != Encoding.Unicode)) {
          XmlTextReader r = new XmlTextReader(new StringReader(initialString));
          MemoryStream ms = new MemoryStream();
          XmlTextWriter w = new XmlTextWriter(ms, 
                                              Encoding.GetEncoding(destinationEncoding));
          w.Formatting = Formatting.Indented;
          w.Indentation = 3;
          if (writeStartDocument) 
            w.WriteStartDocument();
          w.WriteNode(r, true);
          w.Flush();
          ms.Position = 0;
          StreamReader sr = new StreamReader(ms);
          encodedString = sr.ReadToEnd();
          w.Close();
          return encodedString;
        }
        else {
          return initialString;
        }
      }
      //Happens on an empty document
      catch (System.Xml.XmlException) {
        return "";
      }
    }

    /**
     * <remark> Overloading interface, considering useRaw = false is the
     * most common case of all. Beware, to sepcify useRaw = true, you have
     * to specify all arguments (which makes sense, since this pretty much
     * means you're building some of your XML by hand).</remark>
     */

    public static string ToPretty (Seq x) {
      return ToPretty(x,true,false,"utf-16");
    }

    public static string ToPretty (Seq x, bool y) {
      return ToPretty(x,y,false,"utf-16");
    }

    public static string ToPretty (Seq x,bool y,bool z) {
      return ToPretty(x,y,z,"utf-16");
    }

    public static void WriteToFile(string f, string contents) {
      StreamWriter w = 
        new StreamWriter(f,false,System.Text.Encoding.GetEncoding("ISO-8859-1"));
      w.WriteLine(contents);
      w.Close();
    }

    public static void Exit(int code) {
      throw new SystemException("unexpected exit");
    }

    //      public static float GetGCHeapSize(string prog) {
    //        PerformanceCounter cnt = 
    //          new PerformanceCounter(".NET CLR Memory", "# Bytes in all Heaps", prog);
    //        return cnt.NextValue();
    //      }

    public static Seq LoadDom(string filename) {
      XmlDocument doc = new XmlDocument();
      doc.Load(filename);
      return new SeqWrapDom(doc.DocumentElement);
    }


    /**
     * <summary> This method prints the content of a Seq without any
     * escaping if itself and all its followers (linked by the 'next' field)
     * are of the type of the current pcdataencoding (i.e. Char <-> Char,
     * String <-> String, SubString <-> SubString | Hybrid). </summary>
     * <param name="s">The Pcdata-based Sequence</param>
     */

    private static StringBuilder sb = null;

    public static string PcdataToString(Seq s){
      sb = new StringBuilder();
      convToString(s);
      return sb.ToString();
    }
    
    private static void convToString(Seq s) {
      switch (s.kind) {
      case Seq.Kind.Empty:
        break;
      case Seq.Kind.Element:          
      case Seq.Kind.Object: 
      case Seq.Kind.Bool:
      case Seq.Kind.Int:
          //We know that the currently treated String is not Pcdata
        throw new SystemException("convToString : trying to convert a sequence that is not pcdata");
      case Seq.Kind.Append: {
        SeqAppend sc = (SeqAppend) s;
        if (sc.fst != null) convToString(sc.fst);
        convToString(sc.snd);
        break;
      }
      case Seq.Kind.Wrap: {
        SeqWrap sw = (SeqWrap)s;
        Seq tmp = sw.Normalize();
        convToString(tmp);
        break;
      }
      case Seq.Kind.Char: 
        sb.Append(((SeqChar) s).tag.ToString());
        convToString(s.contents);
        convToString(s.next);
        break;
      case Seq.Kind.String: 
          sb.Append(((SeqString) s).contents);
          convToString(s.next);
          break;
      case Seq.Kind.Substring:
        SeqSubstring sc = ((SeqSubstring) s);
        sb.Append(sc.contents.Substring(sc.offset,
                                        sc.length));
        convToString(s.next);
        break;
      default: 
        throw new SystemException("convToString : should never be reached");
      }
    }

    /**
     * <summary>Writes a string to a file.</summary>
     * <param name="s"> The String to be written</param>
     * <param name="filename"> The complete file path to write to.</param>
     * <param name="encoding"> The name of the encoding to use</param>
     */
    public static void WriteStringToFile (string s, string filename, 
                                          string encoding) {
      WriteStringToFile(s, filename, Encoding.GetEncoding(encoding));
    }

    /** An oveload that uses the default system's encoding 
    */
    public static void WriteStringToFile (string s, string filename) {
      WriteStringToFile(s,filename, Encoding.Default);
    }
  
    /* An overload that actually does the work 
    */
    private static void WriteStringToFile (string s, string filename, 
                                          Encoding encoding) {
      StreamWriter sw = new StreamWriter (filename, false, encoding);
      sw.Write(s);
      sw.Close();
    }
  }

  public class Timer {
    private double timer;
    private bool ticking = false;
    private DateTime current_time;

    public Timer () {
      timer = 0;
      ticking = false;
    }
    
    public void StartTimer() {
      current_time = DateTime.Now;
      ticking = true;
    }

    public void StopTimer() {
      if (ticking) {
        DateTime t_end = DateTime.Now;
        TimeSpan t = t_end - current_time;
        timer += t.TotalMilliseconds;
        ticking = false;
      };
    }

    public double GetTimer() {
      StopTimer();
      double res = timer;
      timer = 0;
      return res;
    }

  }
}


/**
 *   New I/O interface
 *
 */

namespace Xtatic {


  public class XmlIO {

    /** Loads an XML document from a URI (can be a file name) into a Seq.
        The document's character encoding (via a byte order mark or an 
        <?xml?> declaration is honored by the underlying parser.  
        If the doc contains or refers to a DTD, the DTD's entity declarations
        are used for interpreting entity references, but the DTD's 
        constraints on document structure are ignored (use Xtatic's
        retypes to check those).
    */
    //TODO: Bypass DOM (XmlDocument and friends) 
    //      and load a Seq directly from an XmlReader

    public static Seq Load(string uri) {

      XmlValidatingReader reader = null;
      ValidationEventHandler eventHandler = new 
        ValidationEventHandler(Xtatic.XmlIO.ValidationCallback);

      try {
        //We use a _validating_ reader only for its automatic entity expansion.
        //The doc structure is not checked (ValidationType == None).
        reader = new XmlValidatingReader(new XmlTextReader(uri));
        reader.ValidationType = ValidationType.None; 
        reader.ValidationEventHandler += eventHandler;

        XmlDocument doc = new XmlDocument();
        doc.Load(reader);

        Utils.SetSubstringPcdataHandling();
        Seq res = Utils.convNode(doc.DocumentElement,Utils.the_empty_seq);
        Utils.ResetSubstringPcdataHandling();
        return res;
      }
      catch(Exception e) {
        string msg = 
          "\r\nXtatic error (In XmlIO.Load):";
        System.Console.WriteLine(msg);
        System.Diagnostics.Debug.Fail(msg);
        throw e;
      }
      finally {
        if ( reader != null )
          reader.Close();
      }
    }

    static void ValidationCallback(object sender, ValidationEventArgs args ) {
      Console.WriteLine("Validation error loading: {0}", args.Message);
    }


    public static Seq LoadLazyDom(string filename) {
      XmlDocument doc = new XmlDocument();
      doc.Load(filename);
      return new SeqWrapDom(doc.DocumentElement);
    }




    /**
     * <summary> This implementation outputs the Seq to a file by use of a
     * LIFO Stack.  Uses an XmlWriter to ensure text is indented,
     * correctly formatted, and escapes XML Special characters. </summary>
     *
     * <remark> This method forces escaping of special characters, and
     * therefore breaks if you try to craft entities by hand.  </remark>
     * <remark> This method will assume no special characters others than
     * simple and double quotes are in the attributes of your Seq object
     * </remark>
     * <param name = "s">The Xtatic Sequence to be written</param>
     * <param name = "w">The XmlWriter to which the sequence is to be written</param>
     */

    static void Write(Seq s, XmlWriter w) {
      Stack q = new Stack();
      q.Push(s);
      w.WriteStartDocument();
      while (q.Count > 0) {
        object o = q.Pop();
        if (o is string) {
          w.WriteEndElement();
          continue;
        };
        s = (Seq)o;
        if (s.kind != Seq.Kind.Append && 
            s.kind != Seq.Kind.Empty && 
            s.kind != Seq.Kind.Wrap)
          q.Push(s.next);
        string towrite = null;
        switch (s.kind) {
        case Seq.Kind.Empty:
          break;
        case Seq.Kind.Element:          
        case Seq.Kind.Object: {
          string t = null;
          if (s.kind == Seq.Kind.Object) {
            t=((SeqObject) s).tag.ToString();
          }
          else {
            t =((SeqElement) s).tag.ToString();
          }
          w.WriteStartElement(t);
          if (s.attributes != null)
            foreach (DictionaryEntry de in s.attributes) {
              Seq value = (Seq)(de.Value);
              w.WriteAttributeString(((Tag)de.Key).Name, Utils.PcdataToString(value));
            };
          if (s.contents == Utils.the_empty_seq) {
            w.WriteEndElement();
          }
          else {
            q.Push(t);
            q.Push(s.contents);
          }
          break;
        } 
        case Seq.Kind.Append: {
          SeqAppend sc = (SeqAppend) s;
          q.Push(sc.snd);
          if (sc.fst != null) q.Push(sc.fst);
          break;
        }
        case Seq.Kind.Wrap: {
          SeqWrap sw = (SeqWrap)s;
          Seq tmp = sw.Normalize();
          q.Push(tmp);
          break;
        }
        case Seq.Kind.Bool: 
          towrite = ((SeqBool) s).tag.ToString();
          goto default;
        case Seq.Kind.Int: 
          towrite = ((SeqInt) s).tag.ToString();
          goto default;
        case Seq.Kind.Char: 
          towrite = ((SeqChar) s).tag.ToString();
          goto default;
        case Seq.Kind.String: 
          towrite = ((SeqString) s).contents;
          goto default;
        case Seq.Kind.Substring:
          SeqSubstring sc = ((SeqSubstring) s);
          towrite = sc.contents.Substring(sc.offset,
                                          sc.length);
          continue;
        default: {          
          if (towrite != null) {
            w.WriteString(towrite);
          }
          if (s.next.kind != Seq.Kind.Empty 
              && (s.kind == Seq.Kind.Bool 
                  || s.kind ==Seq.Kind.Int)) 
            w.WriteWhitespace(" ");
          break;
        }
        }
      }
      w.Flush();
    }

    /**
     * <summary> This method outputs an Xtatic Sequence to a File, by use of
     * the previous, hidden write method</summary>
     * <param name="s"> The Xtatic Sequence to be written</param>
     * <param name = "filename"> The name of the file to which the sequence
     * will be written</param>
     * <param name="destinationEncoding">The optional encoding in which the
     * Sequence is to be translated </param>
     */

    public static void WriteToFile (Seq s, string filename, 
                                    string destinationEncoding){
	Encoding e = Encoding.GetEncoding(destinationEncoding);
        WriteToFile(s, filename,e);
    }
 
    /** The same, but using the default system encoding. 
    */
    public static void WriteToFile (Seq s, string filename) {
      WriteToFile (s, filename, Encoding.Default);
    }

    private static void WriteToFile (Seq s, string filename, Encoding enc){
       try {
        XmlTextWriter w = new XmlTextWriter (filename, enc);
        w.Formatting = Formatting.Indented;
        Write(s, w);
        w.Close();
      }
      catch (Exception e){
        Console.WriteLine("Exception writing to file : {0}",e.Message);
      }
    }


   /** Writes xml to string -- mostly to be used in error messages. 
       The string receives the same contents as the file in WriteToFile.
       (Since .NET strings are only UTF-16, there is no way to specify an 
       encoding.) 
   */
   public static string WriteToString (Seq seq) {
     try { 
     System.IO.StringWriter sw = new System.IO.StringWriter();
      XmlTextWriter xw = new XmlTextWriter(sw);
      xw.Formatting = Formatting.Indented;
      Write(seq, xw);
      xw.Close();
      sw.Close();
      return (sw.ToString());
     }
     catch (Exception e){
       Console.WriteLine("Exception writing to string : {0}", e.Message);
       return "";
     }
    }
}
}


namespace Xtatic { 
  public class CharEntities { 
    public CharEntities () {}

    public Seq amp  = Utils.StringToPcdata("\u0026");
    public Seq lt   = Utils.StringToPcdata("\u003C");
    public Seq gt   = Utils.StringToPcdata("\u003E");
    public Seq nbsp = Utils.StringToPcdata("\u00A0");
    public Seq copy = Utils.StringToPcdata("\u00A9");
  }
}
